#!/bin/bash
# make debian packages from source tarball
# copyright 2011 jamie twycross, jpt AT cs.nott.ac.uk
# released under GNU GPL version 3

# run this script as root in the top-level source directory
# if successful, the debian packages will be found in the debs/ subdirectory
# a log file debian.log is created in the top-level source directory
# you should view this log file after the script has finished
# (especially the lintian output for each package at the end of the log)

# you will also need to create and edit the file "control" in the top-level
# source directory to provide package specific options
# also the files README, AUTHORS, VERSION, NAME, ChangeLog need to be present
# and properly formatted in the top-level source directory

######################################################
# should only need to perhaps change these variables #
######################################################

export DEBFULLNAME="Jamie Twycross"
export DEBEMAIL="jpt@cs.nott.ac.uk"
SRCURL="http://www.infobiotics.org/"

LIB_EXTRA_FILES=""
DEV_EXTRA_FILES=""
BIN_EXTRA_FILES="usr/bin/* usr/share/doc/mcss/README usr/share/doc/mcss/AUTHORS usr/share/doc/mcss/VERSION usr/share/doc/mcss/mcss.html usr/share/doc/mcss/mcss-postprocess.html"
DOC_EXTRA_FILES="AUTHORS VERSION changelog"

MAJOR_RELEASE="1" # the number the lib package is appended with e.g. libecsb1

######################################################
# end of user changes                                #
######################################################

export DEB_BUILD_OPTIONS="nocheck"

PKGNAME=$(cat NAME)
PKGVER=$(cat VERSION)
BUILDBASE=${PWD}/debian.tmp
SRCDIR=${PWD}
PKGDIR=${PWD}/debs

PKGFULL=${PKGNAME}-${PKGVER}
BUILDDIR=${BUILDBASE}/${PKGFULL}
TMPFILE=${PKGFULL}.tmp
BUILDLOG=${PWD}/debian.log

# redirect stderr to build log
rm -f ${BUILDLOG}
exec 2>> ${BUILDLOG}

echo "building ${PKGFULL} debian packages..."

# prepare build directory
echo "preparing build directory..."
rm -rf ${BUILDDIR} ${PKGDIR} &&
mkdir -p ${BUILDDIR} ${PKGDIR}
if [ $? != 0 ]
then
	echo "error: build directory preparation failed (see ${BUILDLOG} for details)"
	exit 1
fi

# make source tarball
echo "making source tarball..."
./configure &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: configure failed (see ${BUILDLOG} for details)"
	exit 1
fi
make dist &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: make dist failed (see ${BUILDLOG} for details)"
	exit 1
fi

# extract source tarball
echo "extracting source tarball..."
pushd ${BUILDDIR} &> /dev/null &&
tar -xzvf ${SRCDIR}/${PKGFULL}.tar.gz 1> /dev/null 2>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: tar failed (see ${BUILDLOG} for details)"
	exit 1
fi

# configure debian package
echo "configuring debian packages..."
pushd ${PKGFULL} &> /dev/null &&
echo | dh_make --library -p lib${PKGNAME} -c gpl3 --createorig &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: dh_make failed (see ${BUILDLOG} for details)"
	exit 1
fi
# fix original source archive name
mv ../lib${PKGNAME}_${PKGVER}.orig.tar.gz ../${PKGNAME}_${PKGVER}.orig.tar.gz
if [ $? != 0 ]
then
	echo "error: dh_make failed (see ${BUILDLOG} for details)"
	exit 1
fi

# edit default package configuration
echo "editing package configuration..."
# copy debian control file
cp control debian/ &&
# edit copyright
head -n 2 debian/copyright > ${TMPFILE} &&
printf "Source: ${SRCURL}\n\n" >> ${TMPFILE} &&
printf "Files: *\n" >> ${TMPFILE} &&
printf "Copyright: %s ${DEBFULLNAME} <${DEBEMAIL}>\n" $(date +%Y) >> ${TMPFILE} &&
tail -n +8 debian/copyright | head -n -11 >> ${TMPFILE} &&
mv ${TMPFILE} debian/copyright &&
# update changelog
if [ -e /usr/share/doc/${PKGNAME}${MAJOR_RELEASE}/changelog.Debian.gz ]
then 
	zcat /usr/share/doc/${PKGNAME}${MAJOR_RELEASE}/changelog.Debian.gz > ${TMPFILE}
else
	rm -f ${TMPFILE}
fi
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
LINE=$(egrep -n "^version ${PKGVER}" ChangeLog | cut -f 1 -d ':') &&
if [ -e ${TMPFILE} ]
then
	echo >> ${TMPFILE}
fi
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
head -n 2 debian/changelog | sed "s/^lib${PKGNAME}/${PKGNAME}/g" >> ${TMPFILE} &&
if [ -n "${LINE}" ]
then
	tail -n +${LINE} ChangeLog | egrep '^\*' |
	while read comment
	do
		echo "  ${comment}" >> ${TMPFILE}
	done
else
	echo "  * New upstream release." >> ${TMPFILE}
fi
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
echo >> ${TMPFILE} &&
tail -n 1 debian/changelog >> ${TMPFILE} &&
mv ${TMPFILE} debian/changelog &&
# adjust debian installed files list
# remove usr/share/pkgconfig entries
egrep -v '^usr/share/pkgconfig/' debian/lib${PKGNAME}-dev.install > ${TMPFILE} &&
mv ${TMPFILE} debian/lib${PKGNAME}-dev.install &&
# add extra package files
for FILE in ${LIB_EXTRA_FILES}
do
	echo "${FILE}" >> debian/lib${PKGNAME}${MAJOR_RELEASE}.install || exit 1
done
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
for FILE in ${DEV_EXTRA_FILES}
do
	echo "${FILE}" >> debian/lib${PKGNAME}-dev.install || exit 1
done
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
for FILE in ${BIN_EXTRA_FILES}
do
	echo "${FILE}" >> debian/${PKGNAME}.install || exit 1
done
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
for FILE in ${DOC_EXTRA_FILES}
do
	echo "${FILE}" >> debian/docs || exit 1
done
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi
# remove debian readmes
rm -f debian/README.{Debian,source} &&
# copy ChangeLog to debian compliant name
cp ChangeLog changelog
if [ $? != 0 ]
then
	echo "error: editing package configuration failed (see ${BUILDLOG} for details)"
	exit 1
fi

# build packages
echo "building debian packages..."
dpkg-buildpackage -uc -us &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: dpkg-buildpackage failed (see ${BUILDLOG} for details)"
	exit 1
fi

# copy built packages
echo "copying packages..."
cp ${BUILDDIR}/${PKGNAME}*.{deb,gz,dsc,changes} ${BUILDDIR}/lib${PKGNAME}*.deb ${PKGDIR} &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: copy packages failed (see ${BUILDLOG} for details)"
	exit 1
fi

# run lintian on built packages
echo "running lintian on built packages..."
for PACKAGE in ${PKGDIR}/*.deb
do
	echo "*** lintian output for $(basename ${PACKAGE}) ***" >> ${BUILDLOG}
	lintian ${PACKAGE} >> ${BUILDLOG}
done

# tidy up
echo "tidying up..."
popd &> /dev/null &&
popd &> /dev/null &&
rm -rf ${BUILDBASE} &>> ${BUILDLOG}
if [ $? != 0 ]
then
	echo "error: tidying up failed (see ${BUILDLOG} for details)"
	exit 1
fi

echo "successfully built ${PKGFULL} debian packages."
