// alphabet class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Alphabet.h"

#include <iostream>
#include <ecsb/ecsb.h>
#include <stdlib.h>

using namespace Simulation;
using namespace ECSB;

// constructor
Alphabet::Alphabet(unsigned long n)
{
	unsigned long int i;

	alphabet_size = n;

	// create storage for names
	if(!(name = (std::string **) calloc(alphabet_size, sizeof(std::string *)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < alphabet_size; i++)
		name[i] = new string();

	// allocate memory for distribution-based species
	if(!(constant = (bool *) calloc(alphabet_size, sizeof(bool)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution = (bool *) calloc(alphabet_size, sizeof(bool)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(dividing = (bool *) calloc(alphabet_size, sizeof(bool)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < alphabet_size; i++) {
		constant[i] = false;
		distribution[i] = false;
		dividing[i] = true;
	}
	if(!(distribution_type = (unsigned int *) calloc(alphabet_size, \
				sizeof(unsigned int)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution_mean = (double *) calloc(alphabet_size, \
				sizeof(double)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution_sd = (double *) calloc(alphabet_size, \
				sizeof(double)))) {
		cerr << "error: unable to allocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
}

// destructor
Alphabet::~Alphabet()
{
	unsigned long i;

	// free memory
	if(name) {
		for(i = 0; i < alphabet_size; i++)
			if(name[i])
				delete name[i];
		free(name);
	}
	if(constant)
		free(constant);
	if(distribution)
		free(distribution);
	if(dividing)
		free(dividing);
	if(distribution_type)
		free(distribution_type);
	if(distribution_mean)
		free(distribution_mean);
	if(distribution_sd)
		free(distribution_sd);
}

// general access methods
unsigned long Alphabet::size(void) { return alphabet_size; }
std::string Alphabet::getName(unsigned long i) { return *name[i]; }
bool Alphabet::isDistribution(unsigned long i) { return distribution[i]; }
void Alphabet::setConstant(unsigned long i) { constant[i] = true; }
bool Alphabet::isConstant(unsigned long i) { return constant[i]; }
void Alphabet::setNondividing(unsigned long i) { dividing[i] = false; }
bool Alphabet::isNondividing(unsigned long i) { return !dividing[i]; }
double Alphabet::getDistributionMean(unsigned long i) \
		{ return distribution_mean[i]; }

// get index of species given species name
unsigned long Alphabet::getSpeciesIndex(string s)
{
	register unsigned long i;

	for(i = 0; i < alphabet_size; i++)
		if(*name[i] == s)
			return(i);
	cerr << "index of species " << s << " not found\n";
	exit(EXIT_FAILURE);
}

// test if species name is degradation species
bool Alphabet::isDegradation(string s)
{
	if(s == "s0" || s.find("_degraded") != string::npos)
		return true;
	return false;
}

// set species name + attributes
void Alphabet::setName(unsigned long ind, std::string s)
{
	unsigned long count;
	string::size_type index;
	string name2, sc, scc;
	char *endptr;

	// check if distribution-based species
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}
	if(!count) {
		// plain species
		*name[ind] = s;
		distribution[ind] = false;
	} else {
		// distribution-based species
		*name[ind] = s;
		distribution[ind] = true;
		name2 = s.substr(s.find(':') + 1);
		// set distribution
		sc = name2.substr(0, name2.find(':'));
		if(sc == "gaussian")
			distribution_type[ind] = GAUSSIAN_DISTRIBUTION;

		// set distribution mean
		sc = name2.substr(name2.find(':') + 1);
		distribution_mean[ind] = strtod(sc.c_str(), &endptr);

		// set distribution standard deviation
		scc = sc.substr(sc.find(':') + 1);
		distribution_sd[ind] = strtod(scc.c_str(), &endptr);

		// set as nondividing
		dividing[ind] = false;
	}
}

// return distribution-based species level
unsigned long Alphabet::getLevel(unsigned long ind)
{
	switch(distribution_type[ind]) {
// 		case GAUSSIAN_DISTRIBUTION:
// 			return (unsigned long) gaussianRandomNumberNN( \
// 					distribution_mean[ind], distribution_sd[ind]);
// 			break;

		default:
			std::cerr << "error: unknown distribution function\n";
			return 0;
			break;
	}

}

// add species to alphabet
void Alphabet::addSpecies(string n)
{
	unsigned long i;

	// check species with same name doesn't already exist
	for(i = 0; i < alphabet_size; i++)
		if(*name[i] == n) {
			cerr << "error: unable to add species " << n << \
					" as species with that name already exists\n";
			exit(EXIT_FAILURE);
		}

	// increase size of alphabet
	alphabet_size++;

	// realloc memory
	if(!(name = (std::string **) realloc(name, \
				alphabet_size * sizeof(std::string *)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory for distribution-based species
	if(!(constant = (bool *) realloc(constant, \
				alphabet_size * sizeof(bool)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution = (bool *) realloc(distribution, \
				alphabet_size * sizeof(bool)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(dividing = (bool *) realloc(dividing, \
				alphabet_size * sizeof(bool)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution_type = (unsigned int *) realloc(distribution_type, \
				alphabet_size * sizeof(unsigned int)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution_mean = (double *) realloc(distribution_mean, \
				alphabet_size * sizeof(double)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}
	if(!(distribution_sd = (double *) realloc(distribution_sd, \
				alphabet_size * sizeof(double)))) {
		cerr << "error: unable to reallocate memory for alphabet\n";
		exit(EXIT_FAILURE);
	}

	// set variables
	name[alphabet_size - 1] = new string(n);
	constant[alphabet_size - 1] = false;
	distribution[alphabet_size - 1] = false;
	dividing[alphabet_size - 1] = true;
}
