// compartment class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Compartment.h"

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <string>
#include <ecsb/Misc.h>

using namespace Simulation;

// constructor
Simulation::Compartment::Compartment(std::string n)
{
	// set default initialisation values
	setInitialisationValues();

	// set compartment id
	id = n;
}

// destructor
Simulation::Compartment::~Compartment()
{
	unsigned long i;

	// free memory
	freeValidRules();
	if(valid_rules)
		free(valid_rules);
	if(ruleset)
		free(ruleset);
	if(ruleset_index)
		free(ruleset_index);
	if(num_rules)
		free(num_rules);
	if(num_updates)
		free(num_updates);
	if(updates)
		free(updates);
	if(rc)
		free(rc);
	if(propensity)
		free(propensity);
	if(num_dist_species)
		free(dist_species);
	if(level)
		free(level);
	if(constant)
		free(constant);
	if(!volume_constant && growth_type == FUNCTION_GROWTH) {
		if(growth) {
			free_stochastic_comparisons_variables(growth_function);
			if(num_growth_species) {
				free(growth_species);
				free(growth_species_values);
			}
		}
		if(division) {
			free_stochastic_comparisons_variables(division_function);
			if(num_division_species) {
				free(division_species);
				free(division_species_values);
			}
		}
	}
	if(neighbours)
		free(neighbours);
	if(event) {
		for(i = 0; i < num_events; i++)
				delete event[i];
		free(event);
	}
	if(child_index)
		free(child_index);
}

// copy constructor
Simulation::Compartment::Compartment(const Compartment &comp)
{
	register unsigned long i, j;
	std::string varname;
	mu::varmap_type variables;
	mu::varmap_type::const_iterator iter;

	// set default initialisation values
	setInitialisationValues();

	// do a deep copy...
	id = comp.id;  // compartment id
	name = comp.name;  // compartment name

	tau = comp.tau;  // time to next reaction
	mu = comp.mu;  // index of next reaction
	rs = comp.rs;  // index of ruleset of next reaction

	// species levels
	initialiseLevels(comp.num_species);
	memcpy(level, comp.level, num_species * sizeof(unsigned long));
	memcpy(constant, comp.constant, num_species * sizeof(bool));

	// distribution species
	initialiseDistributionSpecies(comp.num_dist_species);
	memcpy(dist_species, comp.dist_species, num_dist_species * \
			sizeof(unsigned long));

	// rulesets
	setNumRuleSets(comp.num_rulesets);
	memcpy(ruleset, comp.ruleset, num_rulesets * \
			sizeof(RuleSet *));
	memcpy(ruleset_index, comp.ruleset_index, num_rulesets * \
			sizeof(unsigned long));
	for(i = 0; i < num_rulesets; i++) {
		initialiseValidRules(i, comp.num_rules[i]);
		memcpy(valid_rules[i], comp.valid_rules[i], comp.num_rules[i] * \
			sizeof(unsigned long));
		memcpy(num_updates[i], comp.num_updates[i], comp.num_rules[i] * \
			sizeof(unsigned long));
		memcpy(rc[i], comp.rc[i], comp.num_rules[i] * \
			sizeof(double));
		memcpy(propensity[i], comp.propensity[i], comp.num_rules[i] * \
			sizeof(double));
		for(j = 0; j < num_rules[i]; j++)
			memcpy(updates[i][j], comp.updates[i][j], sizeof(unsigned long));
	}

	memcpy(position, comp.position, 3 * sizeof(unsigned long));  // position

	creation_time = comp.creation_time;  // creation time
	destruction_time = comp.destruction_time;  // destruction time

	volume = comp.volume;  // current volume of compartment
	initial_volume = comp.initial_volume;  // initial volume of compartment
	volume_unit = comp.volume_unit;  // flag indicating if volume is one
	volume_constant = comp.volume_constant;  // flag indicating volume constant
	generation_time = comp.generation_time;  // generation time for compartment
	volume_gradient = comp.volume_gradient;  // volume gradient
	division_volume = comp.division_volume;  // volume compartment divides
	growth_type = comp.growth_type;  // volume growth type
	volume_time = comp.volume_time;  // time copy
	volume_tau0 = comp.volume_tau0;  // time increment copy
	dividing = comp.dividing;  // dividing flag
	division = comp.division;  // division flag
	growth = comp.growth;  // growth flag
	enclosed_compartment = comp.enclosed_compartment;  // enclosed compartment

	// copy parent compartments
	parent_index = comp.parent_index;

	// copy child compartments
	for(i = 0; i < comp.num_child_compartments; i++)
		addChildCompartment(comp.child_index[i]);

	// copy growth species
	if(growth) {
		num_growth_species = comp.num_growth_species;
		if(num_growth_species) {
			if(!(growth_species = (unsigned long *) \
					calloc(num_growth_species, sizeof(unsigned long)))) {
				std::cerr << "error: unable to allocate memory for growth species\n";
				exit(EXIT_FAILURE);
			}
			if(!(growth_species_values = (double *) \
					calloc(num_growth_species, sizeof(double)))) {
				std::cerr << "error: unable to allocate memory for growth species\n";
				exit(EXIT_FAILURE);
			}
			memcpy(growth_species, comp.growth_species, num_growth_species * \
					sizeof(unsigned long));
			memcpy(growth_species_values, comp.growth_species_values, \
					num_growth_species * sizeof(double));
		}
		if(growth_type == FUNCTION_GROWTH) {
			growth_function = comp.growth_function;
			// reassign variables
			variables = growth_function.GetUsedVar();
			if(variables.size()) {
				for(iter = variables.begin(); iter != variables.end(); ++iter) {
					varname = iter->first;
					if(strcmp(varname.c_str(), "V") == 0) {
						// volume
						growth_function.DefineVar(varname, &volume);
						continue;
					}
					if(strcmp(varname.c_str(), "VI") == 0) {
						// initial volume
						growth_function.DefineVar(varname, &initial_volume);
						continue;
					}
					if(strcmp(varname.c_str(), "t") == 0) {
						// current simulation time
						growth_function.DefineVar(varname, &volume_time);
						continue;
					}
					if(strcmp(varname.c_str(), "dt") == 0) {
						// simulation time increment
						growth_function.DefineVar(varname, &volume_tau0);
						continue;
					}
					// compartment object
					for(i = 0; i < num_growth_species; i++)
						if(iter->second == &comp.growth_species_values[i]) {
							growth_function.DefineVar(varname, \
									&growth_species_values[i]);
							break;
						}
				}
			}
			// update stochastic comparison variables
			update_stochastic_comparisons_variables(growth_function);
		}
	}

	// copy division species
	if(division) {
		num_division_species = comp.num_division_species;
		if(num_division_species) {
			if(!(division_species = (unsigned long *) \
					calloc(num_division_species, sizeof(unsigned long)))) {
				std::cerr << "error: unable to allocate memory for division " \
						"species\n";
				exit(EXIT_FAILURE);
			}
			if(!(division_species_values = (double *) \
					calloc(num_division_species, sizeof(double)))) {
				std::cerr << "error: unable to allocate memory for division " \
						"species\n";
				exit(EXIT_FAILURE);
			}
			memcpy(division_species, comp.division_species, \
					num_division_species * sizeof(unsigned long));
			memcpy(division_species_values, comp.division_species_values, \
					num_division_species * sizeof(double));
		}
		if(growth_type == FUNCTION_GROWTH) {
			division_function = comp.division_function;
			// reassign variables
			variables = division_function.GetUsedVar();
			if(variables.size()) {
				for(iter = variables.begin(); iter != variables.end(); ++iter) {
					varname = iter->first;
					if(strcmp(varname.c_str(), "V") == 0) {
						// volume
						division_function.DefineVar(varname, &volume);
						continue;
					}
					if(strcmp(varname.c_str(), "VI") == 0) {
						// initial volume
						growth_function.DefineVar(varname, &initial_volume);
						continue;
					}
					if(strcmp(varname.c_str(), "t") == 0) {
						// current simulation time
						division_function.DefineVar(varname, &volume_time);
						continue;
					}
					if(strcmp(varname.c_str(), "dt") == 0) {
						// simulation time increment
						division_function.DefineVar(varname, &volume_tau0);
						continue;
					}
					// compartment object
					for(i = 0; i < num_division_species; i++)
						if(iter->second == &comp.division_species_values[i]) {
							division_function.DefineVar(varname, \
									&division_species_values[i]);
							break;
						}
				}
			}
			// update stochastic comparison variables
			update_stochastic_comparisons_variables(division_function);
		}
	}

	// copy neighbours
	initialiseNeighbours(comp.num_neighbours);
	memcpy(neighbours, comp.neighbours, num_neighbours * sizeof(unsigned long));

	// copy events
	num_events = comp.num_events;
	if(!(event = (Simulation::Event **) calloc(num_events, \
			sizeof(Simulation::Event *)))) {
		std::cerr << "error: unable to allocate memory for events\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_events; i++) {
		event[i] = new Simulation::Event(*comp.event[i]);
	}
}

// set default values for initialised class
void Simulation::Compartment::setInitialisationValues(void)
{
	tau = 0.0;
	mu = 0;
	volume = initial_volume = 1.0;
	volume_unit = volume_constant = true;
	volume_gradient = division_volume = 0.0;
	growth_type = NO_GROWTH;
	num_growth_species = num_division_species = 0;
	num_dist_species = 0;
	generation_time = 0.0;
	creation_time = destruction_time = 0;
	dividing = false;
	division = false;
	growth = false;
	level = NULL;
	dist_species = NULL;
	num_rulesets = 0;
	ruleset = NULL;
	ruleset_index = NULL;
	num_rules = NULL;
	valid_rules = NULL;
	num_updates = NULL;
	updates = NULL;
	rc = NULL;
	propensity = NULL;
	enclosed_compartment = false;
	num_child_compartments = 0;
	child_index = NULL;

	num_neighbours = 0;
	neighbours = NULL;

	num_events = 0;
	event = NULL;
}

// general access methods
void Simulation::Compartment::setId(std::string n) { id = n; }
std::string Simulation::Compartment::getId(void) { return id; }
void Simulation::Compartment::setName(std::string n) { name = n; }
std::string Simulation::Compartment::getName(void) { return name; }
unsigned long Simulation::Compartment::getNumRuleSets(void) \
		{ return num_rulesets; }
unsigned long *Simulation::Compartment::getRuleSets(void) \
		{ return ruleset_index; }
//void Simulation::Compartment::setLabel(std::string n) { label = n; }
//std::string Simulation::Compartment::getLabel(void) { return label; }
void Simulation::Compartment::setRuleSetIndex(unsigned long rs, \
		unsigned long ind) { ruleset_index[rs] = ind; }
unsigned long Simulation::Compartment::getRuleSetIndex(unsigned long rs) \
		{ return ruleset_index[rs]; }
void Simulation::Compartment::setTimeToNextReaction(double t) { tau = t; }
double Simulation::Compartment::getTimeToNextReaction(void) { return tau; }
void Simulation::Compartment::setNextReaction(unsigned long i) { mu = i; }
unsigned long Simulation::Compartment::getNextReaction(void) \
		{ return mu; }
void Simulation::Compartment::setNextRuleSet(unsigned long i) { rs = i; }
unsigned long Simulation::Compartment::getNextRuleSet(void) \
		{ return rs; }
void Simulation::Compartment::setNumRules(unsigned long rs, unsigned long i) \
		{ num_rules[rs] = i; }
unsigned long Simulation::Compartment::getNumRules(unsigned long rs) \
		{ return num_rules[rs]; }
void Simulation::Compartment::setRuleSet(unsigned long rs, RuleSet *r) \
		{ ruleset[rs] = r; }
class RuleSet *Simulation::Compartment::getRuleSet(unsigned long rs) \
		{ return ruleset[rs]; }
Simulation::Rule *Simulation::Compartment::getRule(unsigned long rs, \
		unsigned long r) { return ruleset[rs]->getRule(valid_rules[rs][r]); }
unsigned long Simulation::Compartment::getRuleIndex(unsigned long rs, \
		unsigned long r) { return valid_rules[rs][r]; }
void Simulation::Compartment::setXPosition(unsigned long i) { position[0] = i; }
unsigned long Simulation::Compartment::getXPosition(void) \
		{ return position[0]; }
void Simulation::Compartment::setYPosition(unsigned long i) { position[1] = i; }
unsigned long Simulation::Compartment::getYPosition(void) \
		{ return position[1]; }
void Simulation::Compartment::setZPosition(unsigned long i) { position[2] = i; }
unsigned long Simulation::Compartment::getZPosition(void) \
		{ return position[2]; }
void Simulation::Compartment::setPosition(unsigned long *pos) \
		{ position[0] = pos[0]; position[1] = pos[1]; position[2] = pos[2]; }
const unsigned long *Simulation::Compartment::getPosition(void) \
		{ return position; }
void Simulation::Compartment::getPosition(unsigned long *pos) \
		{ pos[0] = position[0]; pos[1] = position[1]; pos[2] = position[2]; }
unsigned long Simulation::Compartment::getLevel(unsigned long i) \
		{ return(level[i]); }
void Simulation::Compartment::setLevel(unsigned long i, \
		unsigned long c) { level[i] = c; }
void Simulation::Compartment::setLevelConstant(unsigned long i) \
		{ constant[i] = true; }
bool Simulation::Compartment::isLevelConstant(unsigned long i) \
		{ return constant[i]; }
unsigned long Simulation::Compartment::decrementLevel(unsigned long i) \
		{ return --level[i]; }
unsigned long Simulation::Compartment::incrementLevel(unsigned long i) \
		{ return ++level[i]; }
unsigned long Simulation::Compartment::getReactantIndex(unsigned long rs, \
		unsigned long ind, unsigned long i) \
		{ return ruleset[rs]->getReactantIndex(ind, valid_rules[rs][i]); }
unsigned long Simulation::Compartment::getProductIndex(unsigned long rs, \
		unsigned long ind, unsigned long i) \
		{ return ruleset[rs]->getProductIndex(ind, valid_rules[rs][i]); }
int Simulation::Compartment::isChannel(unsigned long rs, unsigned long r) \
		{ return ruleset[rs]->isChannel(valid_rules[rs][r]); }
const char *Simulation::Compartment::getRuleName(unsigned long rs, \
		unsigned long i) \
		{ return ruleset[rs]->getRuleName(valid_rules[rs][i]); }
unsigned long Simulation::Compartment::getNumReactants(unsigned long rs, \
		unsigned long i) \
		{ return ruleset[rs]->getNumReactants(valid_rules[rs][i]); }
unsigned long Simulation::Compartment::getNumProducts(unsigned long rs, \
		unsigned long i) \
		{ return ruleset[rs]->getNumProducts(valid_rules[rs][i]); }
unsigned long Simulation::Compartment::getValidRule(unsigned long rs, \
		unsigned long r) { return valid_rules[rs][r]; }
long int Simulation::Compartment::getRuleXVector(unsigned long rs, \
		unsigned long r) \
		{ return ruleset[rs]->getRuleXVector(valid_rules[rs][r]); }
long int Simulation::Compartment::getRuleYVector(unsigned long rs, \
		unsigned long r) \
		{ return ruleset[rs]->getRuleYVector(valid_rules[rs][r]); }
long int Simulation::Compartment::getRuleZVector(unsigned long rs, \
		unsigned long r) \
		{ return ruleset[rs]->getRuleZVector(valid_rules[rs][r]); }
void Simulation::Compartment::setVolume(double v) { volume = v; }
void Simulation::Compartment::increaseVolume(double v) { volume += v; }
void Simulation::Compartment::decreaseVolume(double v) { volume -= v; }
double Simulation::Compartment::getVolume(void) { return volume; }
void Simulation::Compartment::setInitialVolume(double v) { initial_volume = v; }
double Simulation::Compartment::getInitialVolume(void) \
		{ return initial_volume; }
void Simulation::Compartment::setVolumeUnit(bool b) { volume_unit = b; }
bool Simulation::Compartment::isVolumeUnit(void) { return volume_unit; }
void Simulation::Compartment::setVolumeConstant(bool b) { volume_constant = b; }
bool Simulation::Compartment::isVolumeConstant(void) { return volume_constant; }
void Simulation::Compartment::setVolumeGradient(double d) \
		{ volume_gradient = d; }
double Simulation::Compartment::getVolumeGradient(void) \
		{ return volume_gradient; }
void Simulation::Compartment::setDivisionVolume(double d) \
		{ division_volume = d; }
double Simulation::Compartment::getDivisionVolume(void) \
		{ return division_volume; }
void Simulation::Compartment::setGrowthType(int i) \
		{ growth_type = i; }
int Simulation::Compartment::getGrowthType(void) \
		{ return growth_type; }
bool Simulation::Compartment::isDividing(void) { return dividing; }
void Simulation::Compartment::setDividing(bool b) { dividing = b; }
unsigned long Simulation::Compartment::getDistributionSpecies(unsigned long i) \
		{ return dist_species[i]; }
void Simulation::Compartment::setDistributionSpecies(unsigned long i, \
		unsigned long s) { dist_species[i] = s; }
unsigned long Simulation::Compartment::getNumDistributionSpecies(void) \
		{ return num_dist_species; }
void Simulation::Compartment::setGenerationTime(double v) \
		{ generation_time = v; }
double Simulation::Compartment::getGenerationTime(void) \
		{ return generation_time; }
void Simulation::Compartment::setCreationTime(double v) \
		{ creation_time = v; }
double Simulation::Compartment::getCreationTime(void) \
		{ return creation_time; }
void Simulation::Compartment::setDestructionTime(double v) \
		{ destruction_time = v; }
double Simulation::Compartment::getDestructionTime(void) \
		{ return destruction_time; }
bool Simulation::Compartment::isEnclosedCompartment(void) \
		{ return enclosed_compartment; }
void Simulation::Compartment::setEnclosedCompartment(bool b) \
		{ enclosed_compartment = b; }
unsigned long Simulation::Compartment::getParentCompartmentIndex(void) \
		{ return parent_index; }
void Simulation::Compartment::setParentCompartmentIndex(unsigned long v) \
		{ parent_index = v; }
void Simulation::Compartment::setNumEvents(unsigned long v) { num_events = v; }
unsigned long Simulation::Compartment::getNumEvents(void) \
		{ return num_events; }
void Simulation::Compartment::setChildCompartmentIndex(unsigned long i, \
		unsigned long v) { child_index[i] = v; }
unsigned long Simulation::Compartment::getChildCompartmentIndex( \
		unsigned long i) { return child_index[i]; }
unsigned long Simulation::Compartment::getNumChildCompartments(void) \
		{ return num_child_compartments; }
void Simulation::Compartment::setNumChildCompartments(unsigned long v) \
		{ num_child_compartments = v; }
Simulation::Event *Simulation::Compartment::getEvent(unsigned long i) \
		{ return event[i]; }

void Simulation::Compartment::initialiseLevels(unsigned long a)
{
	unsigned long i;

	num_species = a;
	if(!(level = (unsigned long *) calloc(a, sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for levels\n";
		exit(EXIT_FAILURE);
	}
	if(!(constant = (bool *) calloc(a, sizeof(bool)))) {
		std::cerr << "error: couldn't allocate memory for constant levels\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < a; i++) {
		level[i] = 0;
		constant[i] = false;
	}
}

// add storage space for additional species
void Simulation::Compartment::addSpecies(void)
{
	// check if first species added
	if(!num_species) {
		// allocate memory
		initialiseLevels(1);
	} else {
		num_species++;
		if(!(level = (unsigned long *) realloc(level, \
					num_species * sizeof(unsigned long)))) {
			std::cerr << "error: couldn't reallocate memory for levels\n";
			exit(EXIT_FAILURE);
		}
		if(!(constant = (bool *) realloc(constant, \
					num_species * sizeof(bool)))) {
			std::cerr << "error: couldn't reallocate memory for constant levels\n";
			exit(EXIT_FAILURE);
		}
		level[num_species - 1] = 0;
		constant[num_species - 1] = false;
	}
}

void Simulation::Compartment::show(void)
{
	unsigned long i, nr;

	nr = 0;
	for(i = 0; i < num_rulesets; i++)
		nr += num_rules[i];
	std::cout << "id=" << id << " name=" << name \
			<< " volume=" << volume \
			<< " tau=" << tau << " mu=" << mu \
			<< " num_rules=" << nr << " ruleset_id=";
	for(i = 0; i < num_rulesets; i++) {
		if(i)
			std::cout << ",";
		std::cout << ruleset_index[i];
	}
	std::cout << "\n";
}

// initialise list of valid rules
void Simulation::Compartment::initialiseValidRules(unsigned long rs, \
		unsigned long nr)
{
	unsigned long i;

	num_rules[rs] = nr;
	if(!nr)
		return;
	if(!(valid_rules[rs] = (unsigned long *) calloc(nr, \
			sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for valid rules\n";
		exit(EXIT_FAILURE);
	}
	if(!(num_updates[rs] = (unsigned long *) calloc(nr, \
		sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	if(!(updates[rs] = (unsigned long **) calloc(nr, \
			sizeof(unsigned long *)))) {
		std::cerr << "error: couldn't allocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < nr; i++) {
		if(!(updates[rs][i] = (unsigned long *) calloc(1, \
				sizeof(unsigned long)))) {
			std::cerr << "error: couldn't allocate memory for updates\n";
			exit(EXIT_FAILURE);
		}
		num_updates[rs][i] = 0;
	}
	if(!(rc[rs] = (double *) calloc(nr, \
			sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for reaction constants\n";
		exit(EXIT_FAILURE);
	}
	if(!(propensity[rs] = (double *) calloc(nr, \
			sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for propensities\n";
		exit(EXIT_FAILURE);
	}
}

// free list of valid rules
void Simulation::Compartment::freeValidRules(void)
{
	register unsigned long i, j;

	// free memory
	if(valid_rules) {
		for(i = 0; i < num_rulesets; i++) {
			if(num_rules[i]) {
				free(valid_rules[i]);
				free(num_updates[i]);
				for(j = 0; j < num_rules[i]; j++)
					free(updates[i][j]);
				free(updates[i]);
				free(rc[i]);
				free(propensity[i]);
			}
		}
	}
}

void Simulation::Compartment::setValidRule(unsigned long rs, unsigned long r, \
		unsigned long v)
{
	valid_rules[rs][r] = v;
	rc[rs][r] = ruleset[rs]->getRule(v)->getReactionConstant();
}

// add valid rule to list of valid rules
void Simulation::Compartment::addValidRule(unsigned long rs, \
		unsigned long r)
{
	unsigned long i;

	num_rules[rs]++;
	if(!(valid_rules[rs] = (unsigned long *) realloc(valid_rules[rs], \
			num_rules[rs] * sizeof(unsigned long)))) {
		std::cerr << "error: couldn't reallocate memory for valid rules\n";
		exit(EXIT_FAILURE);
	}
	if(!(num_updates[rs] = (unsigned long *) realloc(num_updates[rs], \
			num_rules[rs] * sizeof(unsigned long)))) {
		std::cerr << "error: couldn't reallocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	if(!(updates[rs] = (unsigned long **) realloc(updates[rs], \
			num_rules[rs] * sizeof(unsigned long *)))) {
		std::cerr << "error: couldn't allocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	if(!(updates[rs][num_rules[rs] - 1] = (unsigned long *) calloc(1, \
			sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	if(!(rc[rs] = (double *) realloc(rc[rs], \
			num_rules[rs] * sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for reaction constants\n";
		exit(EXIT_FAILURE);
	}
	if(!(propensity[rs] = (double *) realloc(propensity[rs], \
			num_rules[rs] * sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for propensities\n";
		exit(EXIT_FAILURE);
	}
	valid_rules[rs][num_rules[rs] - 1] = r;
	rc[rs][num_rules[rs] - 1] = ruleset[rs]->getRule(r)->getReactionConstant();
	propensity[rs][num_rules[rs] - 1] = 0.0;
	num_updates[rs][num_rules[rs] - 1] = 0;
	updates[rs][num_rules[rs] - 1][0] = 0;
}

// initialise list of rulesets
void Simulation::Compartment::setNumRuleSets(unsigned long nrs)
{
	num_rulesets = nrs;
	if(valid_rules) {
		if(!(valid_rules = (unsigned long **) realloc(valid_rules, nrs * \
				sizeof(unsigned long *)))) {
			std::cerr << "error: couldn't reallocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(valid_rules = (unsigned long **) calloc(nrs, \
				sizeof(unsigned long *)))) {
			std::cerr << "error: couldn't allocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	}
	if(ruleset) {
		if(!(ruleset = (RuleSet **) realloc(ruleset, nrs * \
					sizeof(RuleSet *)))) {
			std::cerr << "error: couldn't reallocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(ruleset = (RuleSet **) calloc(nrs, sizeof(RuleSet *)))) {
			std::cerr << "error: couldn't allocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	}
	if(ruleset_index) {
		if(!(ruleset_index = (unsigned long *) realloc(ruleset_index, nrs * \
				sizeof(unsigned long)))) {
			std::cerr << "error: couldn't reallocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(ruleset_index = (unsigned long *) calloc(nrs, \
				sizeof(unsigned long)))) {
			std::cerr << "error: couldn't allocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	}
	if(num_rules) {
		if(!(num_rules = (unsigned long *) realloc(num_rules, nrs * \
				sizeof(unsigned long)))) {
			std::cerr << "error: couldn't reallocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(num_rules = (unsigned long *) calloc(nrs, \
				sizeof(unsigned long)))) {
			std::cerr << "error: couldn't allocate memory for compartment " \
					"rulesets\n";
			exit(EXIT_FAILURE);
		}
	}
	if(num_updates) {
		if(!(num_updates = (unsigned long **) realloc(num_updates, nrs * \
				sizeof(unsigned long *)))) {
			std::cerr << "error: couldn't reallocate memory for updates\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(num_updates = (unsigned long **) calloc(nrs, \
				sizeof(unsigned long *)))) {
			std::cerr << "error: couldn't allocate memory for updates\n";
			exit(EXIT_FAILURE);
		}
	}
	if(updates) {
		if(!(updates = (unsigned long ***) realloc(updates, nrs * \
				sizeof(unsigned long **)))) {
			std::cerr << "error: couldn't reallocate memory for updates\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(updates = (unsigned long ***) calloc(nrs, \
				sizeof(unsigned long **)))) {
			std::cerr << "error: couldn't allocate memory for updates\n";
			exit(EXIT_FAILURE);
		}
	}
	if(rc) {
		if(!(rc = (double **) realloc(rc, nrs * \
				sizeof(double *)))) {
			std::cerr << "error: couldn't reallocate memory for reaction " \
					"constants\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(rc = (double **) calloc(nrs, \
				sizeof(double *)))) {
			std::cerr << "error: couldn't allocate memory for reaction " \
					"constants\n";
			exit(EXIT_FAILURE);
		}
	}
	if(propensity) {
		if(!(propensity = (double **) realloc(propensity, nrs * \
				sizeof(double *)))) {
			std::cerr << "error: couldn't reallocate memory for propensities\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(propensity = (double **) calloc(nrs, \
				sizeof(double *)))) {
			std::cerr << "error: couldn't reallocate memory for propensities\n";
			exit(EXIT_FAILURE);
		}
	}
}

// initialise storage for update compartments for rule in ruleset
void Simulation::Compartment::initialiseUpdates(unsigned long rs, \
		unsigned long r)
{
	// only one update compartment at the moment (for translocation rule)
/*	if(!(updates[rs][r] = (unsigned long *) calloc(1, \
			sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for updates\n";
		exit(EXIT_FAILURE);
	}
	num_updates[rs][r] = 1;*/
}

// set update compartments for rule in ruleset
void Simulation::Compartment::setUpdates(unsigned long rs, unsigned long r, \
		unsigned long *up, unsigned long nup)
{
	register unsigned long i;

	num_updates[rs][r] = nup;
	for(i = 0; i < num_updates[rs][r]; i++)
		updates[rs][r][i] = up[i];
}

// get target compartment of rule in ruleset
unsigned long Simulation::Compartment::getRuleTarget(unsigned long rs, \
		unsigned long r)
{
	return updates[rs][r][0];
}

// return propensity of reaction in compartment
double Simulation::Compartment::getPropensity(unsigned long rs, \
		unsigned long r) \
{
	return propensity[rs][r];
}

// set propensity of reaction in compartment
void Simulation::Compartment::setPropensity(unsigned long rs, \
		unsigned long r, double v) \
{
	propensity[rs][r] = v;
}

// initialise list of distribution-based species for compartment
void Simulation::Compartment::initialiseDistributionSpecies(unsigned long a)
{
	num_dist_species = a;
	if(!(dist_species = (unsigned long *) calloc(a, \
			sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for distribution species\n";
		exit(EXIT_FAILURE);
	}
}

// reset all levels to zero
void Simulation::Compartment::resetLevels(void)
{
	memset(level, 0, sizeof(unsigned long) * num_species);
}

// calculate volume of compartment
double Simulation::Compartment::calculateVolume(double time, double tau0)
{
	register unsigned long i;

	switch(growth_type) {
		case NO_GROWTH:
			return(volume);
			break;

		case LINEAR_GROWTH:
			return(volume + volume_gradient * tau0);
			break;

		case EXPONENTIAL_GROWTH:
			return(volume * exp(volume_gradient * tau0));
			break;

		case FUNCTION_GROWTH:
			if(!growth)
				return(volume);
			volume_time = time;
			volume_tau0 = tau0;
			for(i = 0; i < num_growth_species; i++)
				growth_species_values[i] = (double) level[growth_species[i]];
			return(growth_function.Eval());
			break;

		default:
			break;
	}
}

// get time to next division time of compartment
double Simulation::Compartment::getTimeToNextDivision(double time, double tau0)
{
	register unsigned long i;

	// compartment doesn't divide
	if(!division)
		return (tau0 + 1.0);

	switch(growth_type) {
		case NO_GROWTH:
			return (tau0 + 1.0);
			break;

		case LINEAR_GROWTH:
			return ((division_volume - volume) / volume_gradient);
			break;

		case EXPONENTIAL_GROWTH:
			return (log(division_volume / volume) / volume_gradient);
			break;

		case FUNCTION_GROWTH:
			volume_time = time;
			volume_tau0 = tau0;
			for(i = 0; i < num_division_species; i++)
				division_species_values[i] = (double) \
						level[division_species[i]];
			if(division_function.Eval())
				return (tau0 / 2.0);
			else
				return (tau0 + 1.0);
			break;

		default:
			break;
	}
}

// calculate if comopartment has divided
bool Simulation::Compartment::evaluateDivisionFunction(double time, double tau0)
{
	register unsigned long i;

	// return if compartment doesn't divide
	if(!division)
		return false;

	switch(growth_type) {
		case NO_GROWTH:
			return false;

		case LINEAR_GROWTH:
		case EXPONENTIAL_GROWTH:
			if(volume > division_volume)
				return true;
			return false;
			break;

		case FUNCTION_GROWTH:
			volume_time = time;
			volume_tau0 = tau0;
			for(i = 0; i < num_division_species; i++)
				division_species_values[i] = (double) \
						level[division_species[i]];
			return division_function.Eval();
			break;

		default:
			break;
	}
	return false;
}

// increase volume based on a given time increment
void Simulation::Compartment::advanceVolume(double time, double tau0)
{
	register unsigned long i;

	switch(growth_type) {
		case NO_GROWTH:
			break;

		case LINEAR_GROWTH:
			volume += (volume_gradient * tau0);
			break;

		case EXPONENTIAL_GROWTH:
			volume *= exp(volume_gradient * tau0);
			break;

		case FUNCTION_GROWTH:
			volume_time = time;
			volume_tau0 = tau0;
			for(i = 0; i < num_growth_species; i++)
				growth_species_values[i] = (double) level[growth_species[i]];
			volume = growth_function.Eval();

			break;

		default:
			break;
	}
}

// get growth function from compartment name
std::string Simulation::Compartment::getGrowthFunction(std::string s)
{
	unsigned int count;
	std::string::size_type index;
	char *endptr;

	count = index = 0;
	while((index = s.find(':', index)) != std::string::npos) {
		index++;
		count++;
	}
	if(count != 5)
		return "";
	count = 4;
	index = 0;
	while(count--)
		index = s.find(':', index) + 1;

	return s.substr(index, s.rfind(':') - index);
}

// get division function from compartment name
std::string Simulation::Compartment::getDivisionFunction(std::string s)
{
	unsigned int count;
	std::string::size_type index;

	count = index = 0;
	while((index = s.find(':', index)) != std::string::npos) {
		index++;
		count++;
	}
	if(count != 5)
		return "";

	return s.substr(s.rfind(':') + 1);
}

// initialise function based growth and division
void Simulation::Compartment::initialiseGrowthDivision(std::string n, \
		Alphabet *alphabet)
{
	unsigned long i;
	std::string varname;
	mu::varmap_type variables;
	mu::varmap_type::const_iterator iter;

	// check if growth required
	if(getGrowthFunction(n) == "") {
		growth = false;  // no growth
		volume_constant = true;
	}
	else
		growth = true;  // no growth

	if(growth) {
		// parse growth function
		try {
			// set growth function from compartment name
			growth_function.SetExpr(getGrowthFunction(n));
		}
		catch(mu::Parser::exception_type &e) {
			std::cerr << "error: unable to parse growth function " << \
					getGrowthFunction(n) << "\n";
			exit(EXIT_FAILURE);
		}

		// set pointers to variables
		variables = growth_function.GetUsedVar();
		if(variables.size()) {
			num_growth_species = 0;
			for(iter = variables.begin(); iter != variables.end(); ++iter) {
				varname = iter->first;
				if(strcmp(varname.c_str(), "V") == 0) {
					// volume
					growth_function.DefineVar(varname, &volume);
					continue;
				}
				if(strcmp(varname.c_str(), "VI") == 0) {
					// volume
					growth_function.DefineVar(varname, &initial_volume);
					continue;
				}
				if(strcmp(varname.c_str(), "t") == 0) {
					// current simulation time
					growth_function.DefineVar(varname, &volume_time);
					continue;
				}
				if(strcmp(varname.c_str(), "dt") == 0) {
					// simulation time increment
					growth_function.DefineVar(varname, &volume_tau0);
					continue;
				}
				// compartment object
				for(i = 0; i < num_species; i++)
					if(alphabet->getName(i) == varname) {
						num_growth_species++;
						break;
					}
				if(i == num_species) {
					std::cerr << "error: unknown variable " << varname << \
							" in growth function " << getGrowthFunction(name) \
							<< "\n";
					exit(EXIT_FAILURE);
				}
			}
			if(num_growth_species) {
				if(!(growth_species = (unsigned long *) calloc(num_growth_species, \
						sizeof(unsigned long)))) {
					std::cerr << "error: unable to allocate memory for growth species\n";
					exit(EXIT_FAILURE);
				}
				if(!(growth_species_values = (double *) calloc(num_growth_species, \
						sizeof(double)))) {
					std::cerr << "error: unable to allocate memory for growth species\n";
					exit(EXIT_FAILURE);
				}
				num_growth_species = 0;
				for(iter = variables.begin(); iter != variables.end(); ++iter) {
					varname = iter->first;
					for(i = 0; i < num_species; i++)
						if(alphabet->getName(i) == varname) {
							growth_species[num_growth_species] = i;
							growth_function.DefineVar(varname, \
									&growth_species_values[num_growth_species]);
							num_growth_species++;
							break;
						}
				}
			}
		}
		// enable stochastic comparisons
		initialise_stochastic_comparisons(growth_function);
	}

	// check if division required
	if(getDivisionFunction(n) == "")
		division = false;  // no division
	else
		division = true;  // no division

	// parse division function
	if(division) {
		try {
			// set division function from compartment name
			division_function.SetExpr(getDivisionFunction(n));
		}
		catch(mu::Parser::exception_type &e) {
			std::cerr << "error: unable to parse division function " << \
					getDivisionFunction(n) << "\n";
			exit(EXIT_FAILURE);
		}

		// set pointers to variables
		variables = division_function.GetUsedVar();
		if(variables.size()) {
			num_division_species = 0;
			for(iter = variables.begin(); iter != variables.end(); ++iter) {
				varname = iter->first;
				if(strcmp(varname.c_str(), "V") == 0) {
					// volume
					division_function.DefineVar(varname, &volume);
					continue;
				}
				if(strcmp(varname.c_str(), "VI") == 0) {
					// volume
					division_function.DefineVar(varname, &initial_volume);
					continue;
				}
				if(strcmp(varname.c_str(), "t") == 0) {
					// current simulation time
					division_function.DefineVar(varname, &volume_time);
					continue;
				}
				if(strcmp(varname.c_str(), "dt") == 0) {
					// simulation time increment
					division_function.DefineVar(varname, &volume_tau0);
					continue;
				}
				// compartment object
				for(i = 0; i < num_species; i++)
					if(alphabet->getName(i) == varname) {
						num_division_species++;
						break;
					}
				if(i == num_species) {
					std::cerr << "error: unknown variable " << varname << \
							" in division function " << \
							getGrowthFunction(name) << "\n";
					exit(EXIT_FAILURE);
				}
			}
			if(num_division_species) {
				if(!(division_species = (unsigned long *) \
						calloc(num_division_species, sizeof(unsigned long)))) {
					std::cerr << "error: unable to allocate memory for division " \
							"species\n";
					exit(EXIT_FAILURE);
				}
				if(!(division_species_values = (double *) \
						calloc(num_division_species, sizeof(double)))) {
					std::cerr << "error: unable to allocate memory for division " \
							"species\n";
					exit(EXIT_FAILURE);
				}
				num_division_species = 0;
				for(iter = variables.begin(); iter != variables.end(); ++iter) {
					varname = iter->first;
					for(i = 0; i < num_species; i++)
						if(alphabet->getName(i) == varname) {
							division_species[num_division_species] = i;
							division_function.DefineVar(varname, \
									&division_species_values[ \
									num_division_species]);
							num_division_species++;
							break;
						}
				}
			}
		}
		// enable stochastic comparisons
		initialise_stochastic_comparisons(division_function);
	}
}

// reset compartment volume
void Simulation::Compartment::resetVolume(void)
{
	if(growth_type == FUNCTION_GROWTH) {
		volume /= 2.0;
		initial_volume = volume;
	}
	else
		volume = initial_volume;
}

// add ruleset to compartment
void Simulation::Compartment::addRuleSet(unsigned long rsid)
{
	// increase storage space for rulesets
	setNumRuleSets(num_rulesets + 1);

	// set new ruleset index
	ruleset_index[num_rulesets - 1] = rsid;
}

// add event to compartment
void Simulation::Compartment::addEvent(RuleSet *rs, Alphabet *alphabet)
{
	// allocate memory for new event
	if(!(event = (Simulation::Event **) realloc(event, (num_events + 1) * \
			sizeof(Simulation::Event *)))) {
		std::cerr << "error: unable to allocate memory for event " << \
				rs->getName() << " in compartment " << name << "\n";
		exit(EXIT_FAILURE);
	}

	// create event
	event[num_events] = new Simulation::Event(rs->getId());
	event[num_events]->setTriggerFunction(rs->getEventTrigger(), alphabet);
	event[num_events]->setAssignmentFunction( \
			rs->getEventAssignmentVariable(), \
			rs->getEventAssignmentExpression(), alphabet);

	num_events++;
}

// handle events in compartment
bool Simulation::Compartment::handleEvents(double time)
{
	register unsigned long i, j, res;
	register bool ret;
	register Simulation::Event *ev;

	ret = false;
	for(i = 0; i < num_events; i++) {
		ev = event[i];
		// set variables
		ev->setTime(time);
		ev->setVolume(volume);
		ev->setInitialVolume(initial_volume);
		for(j = 0; j < ev->getNumTriggerSpecies(); j++)
			ev->setTriggerSpeciesValue(j, level[ev->getTriggerSpecies(j)]);

		// evaluate trigger
		if(ev->evaluateTrigger()) {
			// set variables
			for(j = 0; j < ev->getNumAssignmentSpecies(); j++)
				ev->setAssignmentSpeciesValue(j, \
						level[ev->getAssignmentSpecies(j)]);

			// evaluate assignment functions
			res = ev->evaluateAssignment();

			// set compartment species
			level[ev->getAssignmentVariableIndex()] = res;

			ret = true;
		}
	}

	return ret;
}

// get local reaction constant
double Simulation::Compartment::getReactionConstant(unsigned long rs, \
		unsigned long r)
{
	if(getRule(rs, r)->isDistribution())
		return getRule(rs, r)->getReactionConstant();

	return rc[rs][r];
}

// set local reaction constant
void Simulation::Compartment::setReactionConstant(unsigned long rs, \
		unsigned long r, double val)
{
	rc[rs][r] = val;
}

// initialise list of compartment neighbours
void Simulation::Compartment::initialiseNeighbours(unsigned long nn)
{
	num_neighbours = nn;
	if(!nn)
		return;
	if(!(neighbours = (unsigned long *) calloc(nn, sizeof(unsigned long)))) {
		std::cerr << "error: unable to allocate memory for neighbours\n";
		exit(EXIT_FAILURE);
	}
}

// set compartment neighbour
void Simulation::Compartment::setNeighbour(unsigned long n, unsigned long cid)
{
	neighbours[n] = cid;
}

// get compartment neighbour
unsigned long Simulation::Compartment::getNeighbour(unsigned long n)
{
	return neighbours[n];
}

// get number of compartment neighbours
unsigned long Simulation::Compartment::getNumNeighbours(void)
{
	return num_neighbours;
}

// add child compartment
void Simulation::Compartment::addChildCompartment(unsigned long v)
{
	num_child_compartments++;
	if(!(child_index = (unsigned long *) realloc(child_index, \
				num_child_compartments * sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for child compartment\n";
		exit(EXIT_FAILURE);
	}
	child_index[num_child_compartments - 1] = v;
}
