// compartment class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#ifndef COMPARTMENT_H
#define COMPARTMENT_H

#include <string>
#include <vector>
#include <muParser.h>

#include "Alphabet.h"
#include "RuleSet.h"
#include "Event.h"

// volume growth types
#define NO_GROWTH			0
#define LINEAR_GROWTH		1
#define EXPONENTIAL_GROWTH	2
#define FUNCTION_GROWTH		3

namespace Simulation {

// compartment class
class Compartment {
		std::string id;
		std::string name;

		double tau;  // time to next reaction
		unsigned long mu;  // index of next reaction
		unsigned long rs;  // index of ruleset of next reaction

		unsigned long num_species;  // number of species
		unsigned long *level;  // amounts of species
		bool *constant;  // flag indicating if species amount constant

		unsigned long num_dist_species; // number of distribution-based species
		unsigned long *dist_species;  // indices of distribution-based species

		unsigned long num_rulesets;  // number of rulesets of compartment
		class RuleSet **ruleset;  // pointer to rulesets of compartment
		unsigned long *ruleset_index;  // indices of rulesets of compartment
		unsigned long *num_rules;  // number of valid rules in each ruleset
		unsigned long **valid_rules;  // indices of valid rules in rulesets
		unsigned long **num_updates;  // number of update compartments for rule
		unsigned long ***updates;  // compartments to update for each rule
		double **rc;  // local reaction constant values
		double **propensity;  // local reaction propensities

		unsigned long num_events;  // number of events in compartment
		class Simulation::Event **event;  // compartment events

		unsigned long position[3];  // position of compartment
		double denisty;  // density of compartment

		double creation_time;  // creation time
		double destruction_time;  // destruction time

		double volume;  // current volume of compartment
		double initial_volume;  // initial volume of compartment
		bool volume_unit;  // flag indicating if volume is one
		bool volume_constant;  // flag indicating if volume is constant
		double generation_time;  // generation time for compartment
		double volume_gradient;  // volume gradient
		double division_volume;  // volume at which compartment divides
		int growth_type;  // volume growth type
		mu::Parser growth_function;  // growth function
		mu::Parser division_function;  // division function
		double volume_time, volume_tau0;  // time and time increment copies
		unsigned long num_growth_species;  // num of growth species copies
		unsigned long *growth_species;  // copies of growth species indices
		double *growth_species_values;  // copies of growth species levels
		unsigned long num_division_species;  // num of division species copies
		unsigned long *division_species;  // copies of division species indices
		double *division_species_values;  // copies of division species levels
		bool dividing;  // flag indicating if compartment is dividing
		bool division;  // flag indicating if compartment able to divide
		bool growth;  // flag indicating if compartment able to grow
		bool enclosed_compartment;  // flag indicating if enclosed compartment
		unsigned long parent_index;  // index of parent compartment
		unsigned long num_child_compartments;  // number of enclosed compartments
		unsigned long *child_index;  // indices of enclosed compartments

		unsigned long num_neighbours;  // number of neighbouring compartments
		unsigned long *neighbours;  // indices of neighbouring compartments

		// set default values for initialised class
		void setInitialisationValues(void);

	public:
		// constructor
		Compartment(std::string);

		// destructor
		~Compartment();

		// copy constructor
		Compartment(const Compartment &);

		// general access methods
		void setId(std::string);
		std::string getId(void);
		void setName(std::string);
		std::string getName(void);
//		void setLabel(std::string);
//		std::string getLabel(void);
		void setNumRuleSets(unsigned long);
		unsigned long getNumRuleSets(void);
		unsigned long *getRuleSets(void);
		void setRuleSetIndex(unsigned long, unsigned long);
		unsigned long getRuleSetIndex(unsigned long);
		void setTimeToNextReaction(double);
		double getTimeToNextReaction(void);
		void setNextReaction(unsigned long);
		unsigned long getNextReaction(void);
		void setNextRuleSet(unsigned long);
		unsigned long getNextRuleSet(void);
		void setNumRules(unsigned long, unsigned long);
		unsigned long getNumRules(unsigned long);
		void setRuleSet(unsigned long, RuleSet *);
		RuleSet *getRuleSet(unsigned long);
		Simulation::Rule *getRule(unsigned long, unsigned long);
		unsigned long getRuleIndex(unsigned long, unsigned long);
		void show(void);
		void initialiseLevels(unsigned long);
		unsigned long getLevel(unsigned long);
		void setLevel(unsigned long, unsigned long);
		void resetLevels(void);
		void setLevelConstant(unsigned long);
		bool isLevelConstant(unsigned long);
		void setXPosition(unsigned long);
		unsigned long getXPosition(void);
		void setYPosition(unsigned long);
		unsigned long getYPosition(void);
		void setZPosition(unsigned long);
		unsigned long getZPosition(void);
		void setPosition(unsigned long *);
		const unsigned long *getPosition(void);
		void getPosition(unsigned long *);
		unsigned long decrementLevel(unsigned long);
		unsigned long incrementLevel(unsigned long);
		unsigned long getReactantIndex(unsigned long, unsigned long, \
				unsigned long);
		unsigned long getProductIndex(unsigned long, unsigned long, \
				unsigned long);
		int isChannel(unsigned long, unsigned long);
		const char *getRuleName(unsigned long, unsigned long);
		unsigned long getNumReactants(unsigned long, unsigned long);
		unsigned long getNumProducts(unsigned long, unsigned long);
		void initialiseValidRules(unsigned long, unsigned long);
		void freeValidRules(void);
		void setValidRule(unsigned long, unsigned long, unsigned long);
		unsigned long getValidRule(unsigned long, unsigned long);
		long int getRuleXVector(unsigned long, unsigned long);
		long int getRuleYVector(unsigned long, unsigned long);
		long int getRuleZVector(unsigned long, unsigned long);
		double getPropensity(unsigned long, unsigned long);
		void setPropensity(unsigned long, unsigned long, double);
		void initialiseUpdates(unsigned long, unsigned long);
		void setUpdates(unsigned long, unsigned long, unsigned long *, \
				unsigned long);
		unsigned long getRuleTarget(unsigned long, unsigned long);
		void setVolume(double);
		void increaseVolume(double);
		void decreaseVolume(double);
		void resetVolume(void);
		double getVolume(void);
		void setInitialVolume(double);
		double getInitialVolume(void);
		void setVolumeUnit(bool);
		bool isVolumeUnit(void);
		void setVolumeConstant(bool);
		bool isVolumeConstant(void);
		void setVolumeGradient(double);
		double getVolumeGradient(void);
		void setDivisionVolume(double);
		double getDivisionVolume(void);
		double calculateVolume(double, double);
		double getTimeToNextDivision(double, double);
		void advanceVolume(double, double);
		void setGrowthType(int);
		int getGrowthType(void);
		bool isDividing(void);
		void setDividing(bool);
		void initialiseDistributionSpecies(unsigned long);
		unsigned long getDistributionSpecies(unsigned long);
		void setDistributionSpecies(unsigned long, unsigned long);
		unsigned long getNumDistributionSpecies(void);
		double getGenerationTime(void);
		void setGenerationTime(double);
		double getCreationTime(void);
		void setCreationTime(double);
		double getDestructionTime(void);
		void setDestructionTime(double);
		std::string getGrowthFunction(std::string);
		std::string getDivisionFunction(std::string);
		void initialiseGrowthDivision(std::string, class Alphabet *);
		bool evaluateDivisionFunction(double, double);
		bool isEnclosedCompartment(void);
		void setEnclosedCompartment(bool b);
		unsigned long getParentCompartmentIndex(void);
		void setParentCompartmentIndex(unsigned long v);
		void setNumEvents(unsigned long);
		unsigned long getNumEvents(void);
		void setChildCompartmentIndex(unsigned long i, unsigned long v);
		unsigned long getChildCompartmentIndex(unsigned long i);
		unsigned long getNumChildCompartments(void);
		void setNumChildCompartments(unsigned long v);
		Simulation::Event *getEvent(unsigned long i);

		// add storage space for additional species
		void addSpecies(void);

		// add ruleset to compartment
		void addRuleSet(unsigned long rsid);

		// add event to compartment
		void addEvent(RuleSet *rs, Alphabet *alphabet);

		// add valid rule to list of valid rules
		void addValidRule(unsigned long rs, unsigned long r);

		// get local reaction constant
		double getReactionConstant(unsigned long rs, unsigned long r);

		// set local reaction constant
		void setReactionConstant(unsigned long rs, unsigned long r, double val);

		// initialise list of compartment neighbours
		void initialiseNeighbours(unsigned long nn);

		// set compartment neighbour
		void setNeighbour(unsigned long n, unsigned long cid);

		// get compartment neighbour
		unsigned long getNeighbour(unsigned long n);

		// get number of compartment neighbours
		unsigned long getNumNeighbours(void);

		// handle events in compartment
		bool handleEvents(double time);

		// add child compartment
		void addChildCompartment(unsigned long v);
};

}

#endif
