// deterministic ODE solver class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "DeterministicODE.h"

#include <ecsb/ecsb.h>
#include <list>
#include <gsl/gsl_matrix.h>

using namespace Simulation;
using namespace ECSB;

// constructor
DeterministicODE::DeterministicODE(const unsigned long seed) \
		: SimulationAlgorithm("Deterministic ODE solver", seed)
{
	num_variables = 0;
	ode1 = ode2 = NULL;
	variable_lookup = NULL;
	ode_state = NULL;
	time_offset = 0.0;

	ode_evolution = NULL;
	ode_control = NULL;
	ode_stepper = NULL;
}

// destructor
DeterministicODE::~DeterministicODE()
{
	register unsigned long i;

	// free memory
	if(num_variables) {
		for(i = 0; i < num_variables; i++) {
			delete ode1[i];
			delete ode2[i];
		}
		free(ode1);
		free(ode2);
		free(variable_lookup);
		free(ode_state);
	}

	if(ode_evolution)
		gsl_odeiv_evolve_free(ode_evolution);
	if(ode_control)
		gsl_odeiv_control_free(ode_control);
	if(ode_stepper)
		gsl_odeiv_step_free(ode_stepper);
}

// general access methods
Psystem *DeterministicODE::getPsystem(void) \
		{ return psystem; }
unsigned long DeterministicODE::getNumVariables(void) \
		{ return num_variables; }
vector<ODEVariableEntry> **DeterministicODE::getODE1(void) \
		{ return ode1; }
vector<ODEVariableEntry> **DeterministicODE::getODE2(void) \
		{ return ode2; }
ODEVariableLookupEntry *DeterministicODE::getVariableLookup(void) \
		{ return variable_lookup; }

// return derivatives evaluated at given point
static int func(double t, const double y[], double f[], void *params)
{
	register unsigned long i;
	register long double val, total;
	vector<ODEVariableEntry>::iterator term;

	// set variables
	DeterministicODE *de = (DeterministicODE *) params;
	Psystem *psystem = de->getPsystem();
	unsigned long num_variables = de->getNumVariables();
	vector<ODEVariableEntry> **ode1 = de->getODE1();
	vector<ODEVariableEntry> **ode2 = de->getODE2();
	ODEVariableLookupEntry *variable_lookup = de->getVariableLookup();

	// evaluate odes at specified point
	for(i = 0; i < num_variables; i++) {
		total = 0.0;
		for(term = ode1[i]->begin(); term != ode1[i]->end(); term++) {
			val = ((long double) term->factor * psystem->getReactionConstant( \
					term->compartment, term->ruleset, term->rule) / \
					psystem->getCompartmentVolume(term->compartment));
			if(term->variable[0] < num_variables)
				val *= ((long double) y[term->variable[0]]);
			else
				val *= ((long double) psystem->getSpeciesLevel( \
						term->species[0], term->compartment));
			total += (double) val;
		}
		for(term = ode2[i]->begin(); term != ode2[i]->end(); term++) {
			val = ((long double) term->factor * psystem->getReactionConstant( \
					term->compartment, term->ruleset, term->rule) / \
					psystem->getCompartmentVolume(term->compartment));
			if(term->species[0] == term->species[1])
				val *= 0.5;
			if(term->variable[0] < num_variables)
				val *= ((long double) y[term->variable[0]]);
			else
				val *= ((long double) psystem->getSpeciesLevel( \
						term->species[0], term->compartment));
			if(term->variable[1] < num_variables)
				val *= ((long double) y[term->variable[1]]);
			else
				val *= ((long double) psystem->getSpeciesLevel( \
						term->species[1], term->compartment));
			total += (double) val;
		}
		f[i] = (double) ((long double) total * psystem->getCompartmentVolume( \
				variable_lookup[i].compartment));
	}

	return GSL_SUCCESS;
}

// return jacobian evaluated at a given point
static int jac(double t, const double y[], double *dfdy, double dfdt[], \
		void *params)
{
	register unsigned long i, j;
	register long double val;
	vector<ODEVariableEntry>::iterator term;

	// set variables
	DeterministicODE *de = (DeterministicODE *) params;
	Psystem *psystem = de->getPsystem();
	unsigned long num_variables = de->getNumVariables();
	vector<ODEVariableEntry> **ode1 = de->getODE1();
	vector<ODEVariableEntry> **ode2 = de->getODE2();
	ODEVariableLookupEntry *variable_lookup = de->getVariableLookup();

	gsl_matrix_view dfdy_mat = gsl_matrix_view_array(dfdy, num_variables, \
			num_variables);
	gsl_matrix *m = &dfdy_mat.matrix;

	// evaluate jacobian of odes at specified point
	for(i = 0; i < num_variables; i++) {
		for(j = 0; j < num_variables; j++) {
			val = 0.0;
			for(term = ode1[i]->begin(); term != ode1[i]->end(); term++) {
				if(term->variable[0] == j)
					val += ((long double) term->factor * \
							psystem->getReactionConstant( \
							term->compartment, term->ruleset, term->rule) / \
							psystem->getCompartmentVolume(term->compartment));
			}
			for(term = ode2[i]->begin(); term != ode2[i]->end(); term++) {
				if(term->variable[0] == j && term->variable[1] == j) {
					val += ((long double) 0.5 * term->factor * \
							psystem->getReactionConstant( \
							term->compartment, term->ruleset, term->rule) * \
							y[j]);
					continue;
				}
				if(term->variable[0] == j) {
					if(term->variable[1] < num_variables)
						val += ((long double) term->factor * \
								psystem->getReactionConstant( \
								term->compartment, term->ruleset, term->rule) \
								* y[term->variable[1]]);
					else
						val += ((long double) term->factor * \
								psystem->getReactionConstant( \
								term->compartment, term->ruleset, term->rule) \
								* psystem->getSpeciesLevel( \
								term->species[1], term->compartment));
					continue;
				}
				if(term->variable[1] == j) {
					if(term->variable[0] < num_variables)
						val += ((long double) term->factor * \
								psystem->getReactionConstant( \
								term->compartment, term->ruleset, term->rule) \
								* y[term->variable[0]]);
					else
						val += ((long double) term->factor * \
								psystem->getReactionConstant( \
								term->compartment, term->ruleset, term->rule) \
								* psystem->getSpeciesLevel( \
								term->species[0], term->compartment));
				}
			}
			gsl_matrix_set(m, i, j, (double) ((long double) val * \
					psystem->getCompartmentVolume( \
					variable_lookup[i].compartment)));
		}
		dfdt[i] = 0.0;
	}

	return GSL_SUCCESS;
}

// initialise algorithm
int DeterministicODE::preprocess(void)
{
	register unsigned long i, j, rs, r, ns;
	vector<ODEVariableEntry> ***ve1, ***ve2;
	vector<ODEVariableEntry>::iterator term;
	ODEVariableEntry entry;
	Compartment *compartment;
	register double t0;  // start time
	register double t1;  // end time
	register int status;

	// create odes
	// allocate memory for construction list of ode variables
	if(!(ve1 = (vector<ODEVariableEntry> ***) calloc( \
				num_compartments, sizeof(vector<ODEVariableEntry> **)))) {
		cerr << "error: couldn't allocate memory for ode variables\n";
		exit(EXIT_FAILURE);
	}
	if(!(ve2 = (vector<ODEVariableEntry> ***) calloc( \
				num_compartments, sizeof(vector<ODEVariableEntry> **)))) {
		cerr << "error: couldn't allocate memory for ode variables\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_compartments; i++) {
		if(!(ve1[i] = (vector<ODEVariableEntry> **) calloc( \
					num_species, sizeof(vector<ODEVariableEntry> *)))) {
			cerr << "error: couldn't allocate memory for ode variables\n";
			exit(EXIT_FAILURE);
		}
		if(!(ve2[i] = (vector<ODEVariableEntry> **) calloc( \
					num_species, sizeof(vector<ODEVariableEntry> *)))) {
			cerr << "error: couldn't allocate memory for ode variables\n";
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < num_species; j++) {
			ve1[i][j] = new vector<ODEVariableEntry>;
			ve2[i][j] = new vector<ODEVariableEntry>;
		}
	}

	// populate construction list of ode variables
	for(i = 0; i < num_compartments; i++) {
		compartment = psystem->getCompartment(i);
		for(rs = 0; rs < compartment->getNumRuleSets(); rs++)
			for(r = 0; r < compartment->getNumRules(rs); r++) {
				// reactants
				switch(compartment->getNumReactants(rs, r)) {
					// first order reaction
					case 1:
						if(compartment->isLevelConstant( \
								compartment->getReactantIndex(rs, 0, r)))
							break;
						entry.compartment = i;
						entry.ruleset = rs;
						entry.rule = r;
						entry.species[0] = \
								compartment->getReactantIndex(rs, 0, r);
						entry.factor = -1.0;
						ve1[i][compartment->getReactantIndex(rs, 0, r)]-> \
								push_back(entry);
						break;

					// second order reaction
					case 2:
						entry.compartment = i;
						entry.ruleset = rs;
						entry.rule = r;
						entry.species[0] = \
								compartment->getReactantIndex(rs, 0, r);
						entry.species[1] = \
								compartment->getReactantIndex(rs, 1, r);
						if(entry.species[0] == entry.species[1]) {
							entry.factor = -2.0;
							if(!compartment->isLevelConstant( \
									compartment->getReactantIndex(rs, 0, r)))
								ve2[i][compartment->getReactantIndex(rs, 0, \
										r)]->push_back(entry);
						} else {
							entry.factor = -1.0;
							if(!compartment->isLevelConstant( \
									compartment->getReactantIndex(rs, 0, r)))
								ve2[i][compartment->getReactantIndex(rs, 0, \
										r)]->push_back(entry);
							if(!compartment->isLevelConstant( \
									compartment->getReactantIndex(rs, 1, r)))
								ve2[i][compartment->getReactantIndex(rs, 1, \
										r)]->push_back(entry);
						}
						break;

					default:
						cerr << "error: invalid number of reactants\n";
						exit(EXIT_FAILURE);
						break;
				}
				// products
				for(j = 0; j < compartment->getNumProducts(rs, r); j++) {
					if(compartment->isChannel(rs, r)) {
						if(psystem->getCompartment( \
								compartment->getRuleTarget(rs, r)) \
								->isLevelConstant(compartment-> \
								getProductIndex(rs, j, r)))
							continue;
					} else
						if(compartment->isLevelConstant( \
								compartment->getProductIndex( \
								rs, j, r)))
							continue;
					switch(compartment->getNumReactants(rs, r)) {
						// first order reaction
						case 1:
							entry.compartment = i;
							entry.ruleset = rs;
							entry.rule = r;
							entry.species[0] = \
									compartment->getReactantIndex(rs, 0, r);
							entry.factor = 1.0;
							if(compartment->isChannel(rs, r))
								ve1[compartment->getRuleTarget(rs, r)][ \
										compartment->getProductIndex( \
										rs, j, r)]->push_back(entry);
							else
								ve1[i][compartment-> \
										getProductIndex(rs, j, r)]-> \
										push_back(entry);
							break;

						// second order reaction
						case 2:
							entry.compartment = i;
							entry.ruleset = rs;
							entry.rule = r;
							entry.species[0] = compartment-> \
									getReactantIndex(rs, 0, r);
							entry.species[1] = compartment-> \
									getReactantIndex(rs, 1, r);
							if(entry.species[0] == entry.species[1])
								entry.factor = 1.0;
							else
								entry.factor = 1.0;
							if(compartment->isChannel(rs, r))
								ve2[compartment->getRuleTarget(rs, r)][ \
										compartment->getProductIndex( \
										rs, j, r)]->push_back(entry);
							else
								ve2[i][compartment->getProductIndex( \
										rs, j, r)]->push_back(entry);
							break;

						default:
							cerr << "error: invalid number of reactants\n";
							exit(EXIT_FAILURE);
							break;
					}
				}
			}
	}

	// find number of nonzero variables
	num_variables = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < num_species; j++)
			if(!(ve1[i][j]->empty() && ve2[i][j]->empty()))
				num_variables++;

	// allocate memory for pruned odes
	if(!(ode1 = (vector<ODEVariableEntry> **) calloc( \
				num_variables, sizeof(vector<ODEVariableEntry> *)))) {
		cerr << "error: couldn't allocate memory for ode variables\n";
		exit(EXIT_FAILURE);
	}
	if(!(ode2 = (vector<ODEVariableEntry> **) calloc( \
				num_variables, sizeof(vector<ODEVariableEntry> *)))) {
		cerr << "error: couldn't allocate memory for ode variables\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory for lookup table of ode variables to compartment species
	if(!(variable_lookup = (ODEVariableLookupEntry *) calloc( \
				num_variables, sizeof(ODEVariableLookupEntry)))) {
		cerr << "error: couldn't allocate memory for ode lookup variables\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory for ode state
	if(!(ode_state = (double *) calloc( \
				num_variables, sizeof(double)))) {
		cerr << "error: couldn't allocate memory for ode initial state\n";
		exit(EXIT_FAILURE);
	}

	// prune odes to remove zero entries
	num_variables = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < num_species; j++)
			if(!(ve1[i][j]->empty() && ve2[i][j]->empty())) {
				ode1[num_variables] = ve1[i][j];
				ode2[num_variables] = ve2[i][j];
				variable_lookup[num_variables].compartment = i;
				variable_lookup[num_variables].species = j;
				ode_state[num_variables] = (double) \
						psystem->getSpeciesLevel(j, i);
				num_variables++;
			} else {
				delete ve1[i][j];
				delete ve2[i][j];
			}

	// set which variables are set by solver or constant
	for(i = 0; i < num_variables; i++) {
		for(term = ode1[i]->begin(); term != ode1[i]->end(); term++) {
			for(j = 0; j < num_variables; j++)
				if(variable_lookup[j].compartment == term->compartment && \
						variable_lookup[j].species == term->species[0])
					break;
			term->variable[0] = j;
		}
		for(term = ode2[i]->begin(); term != ode2[i]->end(); term++) {
			for(j = 0; j < num_variables; j++)
				if(variable_lookup[j].compartment == term->compartment && \
						variable_lookup[j].species == term->species[0])
					break;
			term->variable[0] = j;
		}
		for(term = ode2[i]->begin(); term != ode2[i]->end(); term++) {
			for(j = 0; j < num_variables; j++)
				if(variable_lookup[j].compartment == term->compartment && \
						variable_lookup[j].species == term->species[1])
					break;
			term->variable[1] = j;
		}
	}

	// free memory for construction list of ode variables
	for(i = 0; i < num_compartments; i++) {
		free(ve1[i]);
		free(ve2[i]);
	}
	free(ve1);
	free(ve2);

	// initialise gsl ode solver
	ode_stepper = gsl_odeiv_step_alloc(ode_algorithm, num_variables);
	ode_control = gsl_odeiv_control_standard_new(ode_eps_abs, ode_eps_rel, \
			ode_yscaling, ode_dyscaling);
	ode_evolution = gsl_odeiv_evolve_alloc(num_variables);
	ode_system.function = func;
	ode_system.jacobian = jac;
	ode_system.dimension = num_variables;
	ode_system.params = (void *) this;
	ode_step_size = ode_initial_step_size;

	// perform initial iteration of solver
	// iterate ode solver
	t0 = time;
	t1 = t0 + log_interval;
	while(t0 < t1) {
		// execute solver
		status = gsl_odeiv_evolve_apply(ode_evolution, ode_control, \
				ode_stepper, &ode_system, &t0, t1, &ode_step_size, \
				ode_state);

		// failed to solve
		if(status != GSL_SUCCESS)
			return 0;
	}

	// update global simulation time
	tau0global = t1;
	tau0 = tau0global - time;

	// all ok
	return 1;
}

// perform one iteration of algorithm for all compartments
int DeterministicODE::execute(void)
{
	register double t0;  // start time
	register double t1;  // end time
	register int status;
	register unsigned long i, j, it, si;
	register Compartment *compartment;
	register bool flag;

	// update simulation time
	setTime(tau0global);

	// set species levels
	for(i = 0; i < num_variables; i++) {
		compartment = psystem->getCompartment(variable_lookup[i].compartment);
		compartment->setLevel(variable_lookup[i].species, (unsigned long) \
				ode_state[i]);
	}

	// handle events
	if(events_defined) {
		flag = false;
		for(it = 0; it < num_event_compartments; it++) {
			compartment = psystem->getCompartment(event_compartment[it]);
			if(compartment->handleEvents(tau0global)) {
				// event executed
				// update ode_state
				j = compartment->getNumEvents();
				while(j--) {
					si = compartment->getEvent(j)->getAssignmentVariableIndex();
					for(i = 0; i < num_variables; i++)
						if(variable_lookup[i].compartment == \
								event_compartment[it] && \
								variable_lookup[i].species == si) {
							ode_state[i] = (double) compartment->getLevel(si);
							flag = true;
							break;
						}
				}
			}
		}
		if(flag) {
			// reset solver
			// free old solver
			gsl_odeiv_evolve_free(ode_evolution);
			gsl_odeiv_control_free(ode_control);
			gsl_odeiv_step_free(ode_stepper);
			// reinitialise solver
			ode_stepper = gsl_odeiv_step_alloc(ode_algorithm, num_variables);
			ode_control = gsl_odeiv_control_standard_new(ode_eps_abs, \
					ode_eps_rel, ode_yscaling, ode_dyscaling);
			ode_evolution = gsl_odeiv_evolve_alloc(num_variables);
			ode_step_size = ode_initial_step_size;
			time_offset = tau0global;
		}
	}

	// iterate ode solver
	t0 = tau0global - time_offset;
	t1 = t0 + log_interval;
	while(t0 < t1) {
		// execute solver
		status = gsl_odeiv_evolve_apply(ode_evolution, ode_control, \
				ode_stepper, &ode_system, &t0, t1, &ode_step_size, \
				ode_state);

		// failed to solve
		if(status != GSL_SUCCESS)
			return 0;
	}

	// update global simulation time
	tau0global = t1 + time_offset;
	tau0 = tau0global - time;


	// all ok
	return 1;
}

// perform one iteration of algorithm for one compartment
void DeterministicODE::execute(unsigned long c)
{
}
