// event class definitions
// copyright 2008, 2009, 2010 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Event.h"

#include <string.h>

using namespace Simulation;

// muparser functions
double mpfunction_leq(double x, double y) { return (double) (x <= y); }
double mpfunction_geq(double x, double y) { return (double) (x >= y); }
double mpfunction_lt(double x, double y) { return (double) (x < y); }
double mpfunction_gt(double x, double y) { return (double) (x > y); }
double mpfunction_eq(double x, double y) { return (double) (x == y); }
double mpfunction_neq(double x, double y) { return (double) (x != y); }

// constructor
Simulation::Event::Event(std::string n)
{
	// set event id
	id = n;
	name = "";
	index = 0;
	fired = false;
	time = volume = initial_volume = 0.0;
	num_trigger_species = 0;
	trigger_species = NULL;
	trigger_species_value = NULL;
	assignment_variable = "";
	assignment_variable_index = 0;
	num_assignment_species = 0;
	assignment_species = NULL;
	assignment_species_value = NULL;
}

// destructor
Simulation::Event::~Event()
{
	if(trigger_species)
		free(trigger_species);
	if(trigger_species_value)
		free(trigger_species_value);
	if(assignment_species)
		free(assignment_species);
	if(assignment_species_value)
		free(assignment_species_value);
}

// copy constructor
Simulation::Event::Event(const Event &ev)
{
	unsigned long i;
	std::string varname;
	mu::varmap_type variables;
	mu::varmap_type::const_iterator iter;

	id = ev.id;
	name = ev.name;
	index = ev.index;
	fired = ev.fired;
	assignment_variable = ev.assignment_variable;
	assignment_variable_index = ev.assignment_variable_index;

	time = ev.time;
	volume = ev.volume;
	initial_volume = ev.initial_volume;
	num_trigger_species = ev.num_trigger_species;
	trigger_species = NULL;
	trigger_species_value = NULL;
	num_assignment_species = ev.num_assignment_species;
	assignment_species = NULL;
	assignment_species_value = NULL;

	// copy trigger function
	trigger_function = ev.trigger_function;

	// copy trigger species
	if(num_trigger_species) {
		if(!(trigger_species = (unsigned long *) calloc(num_trigger_species, \
				sizeof(unsigned long)))) {
			std::cerr << "error: unable to allocate memory for trigger " <<
					"species\n";
			exit(EXIT_FAILURE);
		}
		memcpy(trigger_species, ev.trigger_species, num_trigger_species * \
				sizeof(unsigned long));
		if(!(trigger_species_value  = (double *) calloc(num_trigger_species, \
				sizeof(double)))) {
			std::cerr << "error: unable to allocate memory for trigger " <<
					"species values\n";
			exit(EXIT_FAILURE);
		}
		memcpy(trigger_species_value, ev.trigger_species_value, \
				num_trigger_species * sizeof(double));
	}

	// reassign trigger function variables
	try {
		variables = trigger_function.GetUsedVar();
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		exit(EXIT_FAILURE);
	}
	if(variables.size()) {
		num_trigger_species = 0;
		for(iter = variables.begin(); iter != variables.end(); ++iter) {
			varname = iter->first;
			if(strcmp(varname.c_str(), "V") == 0) {
				// volume
				trigger_function.DefineVar(varname, &volume);
				continue;
			}
			if(strcmp(varname.c_str(), "VI") == 0) {
				// initial volume
				trigger_function.DefineVar(varname, &initial_volume);
				continue;
			}
			if(strcmp(varname.c_str(), "t") == 0) {
				// current simulation time
				trigger_function.DefineVar(varname, &time);
				continue;
			}
			// compartment object
			for(i = 0; i < num_trigger_species; i++)
				if(iter->second == &ev.trigger_species_value[i]) {
					trigger_function.DefineVar(varname, \
							&trigger_species_value[i]);
					break;
				}
		}
	}

	// copy assignment function
	assignment_function = ev.assignment_function;

	// copy assignment species
	if(num_assignment_species) {
		if(!(assignment_species = (unsigned long *) \
				calloc(num_assignment_species, sizeof(unsigned long)))) {
			std::cerr << "error: unable to allocate memory for assignment " <<
					"species\n";
			exit(EXIT_FAILURE);
		}
		memcpy(assignment_species, ev.assignment_species, \
				num_assignment_species * sizeof(unsigned long));
		if(!(assignment_species_value  = (double *) \
				calloc(num_assignment_species, sizeof(double)))) {
			std::cerr << "error: unable to allocate memory for assignment " <<
					"species values\n";
			exit(EXIT_FAILURE);
		}
		memcpy(assignment_species_value, ev.assignment_species_value, \
				num_assignment_species * sizeof(double));
	}

	// reassign assignment function variables
	try {
		variables = assignment_function.GetUsedVar();
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		exit(EXIT_FAILURE);
	}
	if(variables.size()) {
		num_assignment_species = 0;
		for(iter = variables.begin(); iter != variables.end(); ++iter) {
			varname = iter->first;
			if(strcmp(varname.c_str(), "V") == 0) {
				// volume
				assignment_function.DefineVar(varname, &volume);
				continue;
			}
			if(strcmp(varname.c_str(), "VI") == 0) {
				// initial volume
				assignment_function.DefineVar(varname, &initial_volume);
				continue;
			}
			if(strcmp(varname.c_str(), "t") == 0) {
				// current simulation time
				assignment_function.DefineVar(varname, &time);
				continue;
			}
			// compartment object
			for(i = 0; i < num_assignment_species; i++)
				if(iter->second == &ev.assignment_species_value[i]) {
					assignment_function.DefineVar(varname, \
							&assignment_species_value[i]);
					break;
				}
		}
	}
}

// general access methods
std::string Simulation::Event::getId(void) { return id; }
void Simulation::Event::setId(std::string s) { id = s; }
std::string Simulation::Event::getName(void) { return name; }
void Simulation::Event::setName(std::string s) { name = s; }
void Simulation::Event::setIndex(const unsigned long i) { index = i; }
unsigned long Simulation::Event::getIndex(void) { return index; }
void Simulation::Event::setTime(double v) { time = v; }
double Simulation::Event::getTime(void) { return time; }
void Simulation::Event::setVolume(double v) { volume = v; }
double Simulation::Event::getVolume(void) { return volume; }
void Simulation::Event::setInitialVolume(double v) { initial_volume = v; }
double Simulation::Event::getInitialVolume(void) { return initial_volume; }
void Simulation::Event::setNumTriggerSpecies(unsigned long v) \
		{ num_trigger_species = v; }
unsigned long Simulation::Event::getNumTriggerSpecies(void) \
		{ return num_trigger_species; }
void Simulation::Event::setTriggerSpecies(unsigned long i, unsigned long v) \
		{ trigger_species[i] = v; }
unsigned long Simulation::Event::getTriggerSpecies(unsigned long i) \
		{ return trigger_species[i]; }
void Simulation::Event::setTriggerSpeciesValue(unsigned long i, \
		unsigned long v) { trigger_species_value[i] = (double) v; }
unsigned long Simulation::Event::getTriggerSpeciesValue(unsigned long i) \
		{ return (unsigned long) trigger_species_value[i]; }
void Simulation::Event::setNumAssignmentSpecies(unsigned long v) \
		{ num_assignment_species = v; }
unsigned long Simulation::Event::getNumAssignmentSpecies(void) \
		{ return num_assignment_species; }
void Simulation::Event::setAssignmentSpecies(unsigned long i, unsigned long v) \
		{ assignment_species[i] = v; }
unsigned long Simulation::Event::getAssignmentSpecies(unsigned long i) \
		{ return assignment_species[i]; }
void Simulation::Event::setAssignmentSpeciesValue(unsigned long i, \
		unsigned long v) { assignment_species_value[i] = (double) v; }
unsigned long Simulation::Event::getAssignmentSpeciesValue(unsigned long i) \
		{ return (unsigned long) assignment_species_value[i]; }
unsigned long Simulation::Event::getAssignmentVariableIndex(void) \
		{ return assignment_variable_index; }

// set trigger function
void Simulation::Event::setTriggerFunction(std::string n, Alphabet *alphabet)
{
	unsigned long i;
	std::string varname;
	mu::varmap_type variables;
	mu::varmap_type::const_iterator iter;

	try {
		trigger_function.SetExpr(n);
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to parse event trigger " << n << "\n";
		exit(EXIT_FAILURE);
	}

	// define functions corresponding to SBML functions
	try {
		trigger_function.DefineFun("leq", mpfunction_leq, false);
		trigger_function.DefineFun("geq", mpfunction_geq, false);
		trigger_function.DefineFun("lt", mpfunction_lt, false);
		trigger_function.DefineFun("gt", mpfunction_gt, false);
		trigger_function.DefineFun("eq", mpfunction_eq, false);
		trigger_function.DefineFun("neq", mpfunction_neq, false);
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to add function for trigger " \
				<< n << "\n";
		exit(EXIT_FAILURE);
	}

	// set pointers to variables
	try {
		variables = trigger_function.GetUsedVar();
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to parse event variable for trigger " \
				<< n << "\n";
		exit(EXIT_FAILURE);
	}
	if(variables.size()) {
		num_trigger_species = 0;
		for(iter = variables.begin(); iter != variables.end(); ++iter) {
			varname = iter->first;
			if(strcmp(varname.c_str(), "V") == 0) {
				// volume
				trigger_function.DefineVar(varname, &volume);
				continue;
			}
			if(strcmp(varname.c_str(), "VI") == 0) {
				// initial volume
				trigger_function.DefineVar(varname, &initial_volume);
				continue;
			}
			if(strcmp(varname.c_str(), "t") == 0) {
				// current simulation time
				trigger_function.DefineVar(varname, &time);
				continue;
			}
			// compartment object
			for(i = 0; i < alphabet->size(); i++)
				if(alphabet->getName(i) == varname) {
					num_trigger_species++;
					break;
				}
			if(i == alphabet->size()) {
				std::cerr << "error: unknown variable " << varname << \
						" in trigger function " << n << "\n";
				exit(EXIT_FAILURE);
			}
		}
		if(num_trigger_species) {
			if(!(trigger_species = (unsigned long *) calloc(num_trigger_species, \
					sizeof(unsigned long)))) {
				std::cerr << "error: unable to allocate memory for trigger species\n";
				exit(EXIT_FAILURE);
			}
			if(!(trigger_species_value = (double *) calloc(num_trigger_species, \
					sizeof(double)))) {
				std::cerr << "error: unable to allocate memory for trigger species\n";
				exit(EXIT_FAILURE);
			}
			num_trigger_species = 0;
			for(iter = variables.begin(); iter != variables.end(); ++iter) {
				varname = iter->first;
				for(i = 0; i < alphabet->size(); i++)
					if(alphabet->getName(i) == varname) {
						trigger_species[num_trigger_species] = i;
						trigger_function.DefineVar(varname, \
								&trigger_species_value[num_trigger_species]);
						num_trigger_species++;
						break;
					}
			}
		}
	}
}

// set assignment function
void Simulation::Event::setAssignmentFunction(std::string v, std::string n, \
				Alphabet *alphabet)
{
	unsigned long i;
	std::string varname;
	mu::varmap_type variables;
	mu::varmap_type::const_iterator iter;

	// check assignment variable is valid
	for(i = 0; i < alphabet->size(); i++)
		if(alphabet->getName(i) == v)
			break;
	if(i == alphabet->size()) {
		std::cerr << "error: unknown assignment variable " << v << \
				" in event " << name << "\n";
		exit(EXIT_FAILURE);
	}
	// set assignment variable
	assignment_variable = v;
	assignment_variable_index = i;

	// set assignment function
	try {
		assignment_function.SetExpr(n);
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to parse assignment function " << n << "\n";
		exit(EXIT_FAILURE);
	}

	// define functions corresponding to SBML functions
	try {
		assignment_function.DefineFun("leq", mpfunction_leq, false);
		assignment_function.DefineFun("geq", mpfunction_geq, false);
		assignment_function.DefineFun("lt", mpfunction_lt, false);
		assignment_function.DefineFun("gt", mpfunction_gt, false);
		assignment_function.DefineFun("eq", mpfunction_eq, false);
		assignment_function.DefineFun("neq", mpfunction_neq, false);
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to add function for assignment function " \
				<< n << "\n";
		exit(EXIT_FAILURE);
	}

	// set pointers to variables
	try {
		variables = assignment_function.GetUsedVar();
	}
	catch(mu::Parser::exception_type &e) {
		std::cerr << e.GetMsg() << "\n";
		std::cerr << "error: unable to parse event variable for assignment " \
				"function " << n << "\n";
		exit(EXIT_FAILURE);
	}
	if(variables.size()) {
		num_assignment_species = 0;
		for(iter = variables.begin(); iter != variables.end(); ++iter) {
			varname = iter->first;
			if(strcmp(varname.c_str(), "V") == 0) {
				// volume
				assignment_function.DefineVar(varname, &volume);
				continue;
			}
			if(strcmp(varname.c_str(), "VI") == 0) {
				// initial volume
				assignment_function.DefineVar(varname, &initial_volume);
				continue;
			}
			if(strcmp(varname.c_str(), "t") == 0) {
				// current simulation time
				assignment_function.DefineVar(varname, &time);
				continue;
			}
			// compartment object
			for(i = 0; i < alphabet->size(); i++)
				if(alphabet->getName(i) == varname) {
					num_assignment_species++;
					break;
				}
			if(i == alphabet->size()) {
				std::cerr << "error: unknown variable " << varname << \
						" in assignment function " << n << "\n";
				exit(EXIT_FAILURE);
			}
		}
		if(num_assignment_species) {
			if(!(assignment_species = (unsigned long *) \
					calloc(num_assignment_species, sizeof(unsigned long)))) {
				std::cerr << "error: unable to allocate memory for " \
						"assignment species\n";
				exit(EXIT_FAILURE);
			}
			if(!(assignment_species_value = (double *) \
					calloc(num_assignment_species, sizeof(double)))) {
				std::cerr << "error: unable to allocate memory for " \
						"assignment species\n";
				exit(EXIT_FAILURE);
			}
			num_assignment_species = 0;
			for(iter = variables.begin(); iter != variables.end(); ++iter) {
				varname = iter->first;
				for(i = 0; i < alphabet->size(); i++)
					if(alphabet->getName(i) == varname) {
						assignment_species[num_assignment_species] = i;
						assignment_function.DefineVar(varname, \
								&assignment_species_value[num_assignment_species]);
						num_assignment_species++;
						break;
					}
			}
		}
	}
}

// evaluate trigger function
bool Simulation::Event::evaluateTrigger(void)
{
	bool res;

	// use same semantics as sbml event
	// don't fire at time t=0
	if(time == 0.0)
		return false;

	res = (bool) trigger_function.Eval();
	// fire only when trigger moves from false to true
	if(fired == false && res == true) {
		fired = res;
		return true;
	}
	fired = res;
	return false;
}

// evaluate assignment functions
unsigned long Simulation::Event::evaluateAssignment(void)
{
	unsigned long res;

	res = (unsigned long) assignment_function.Eval();

	return res;
}
