// lattice class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Lattice.h"

#include <iostream>
#include <ecsb/ecsb.h>
#include <stdlib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

using namespace Simulation;
using namespace ECSB;

// initialise a lattice from an xml file
Lattice::Lattice(char *filename, unsigned long seed)
{
	xmlDoc *document;
	xmlNode *node, *child;
	char *str;
	int i, j, k;
	unsigned int x, y;

	// set default values
	name = "";
	scale = 1.0;
	temperature = 1.0;

	// open lattice file
	if(!(document = xmlReadFile(filename, NULL, 0))) {
		cerr << "error: unable to open lattice file " << filename << "\n";
		exit(EXIT_FAILURE);
	}

	// extract data from xml file
	if(!(node = xmlDocGetRootElement(document))) {
		cerr << "error: unable to parse lattice file " << filename << "\n";
		exit(EXIT_FAILURE);
	}

	// extract data from root node
    if(node->type == XML_ELEMENT_NODE && strcmp((char *) node->name, \
			"lattice") == 0) {
		// required attributes
		str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "dimension");
		if(str) {
			dimension = atoi(str);
			free(str);
			if(dimension != 2) {
				cerr << "error: invalid lattice dimension " << dimension << \
						"\n";
				exit(EXIT_FAILURE);
			}
		} else {
			cerr << "error: lattice dimension not specified\n";
			exit(EXIT_FAILURE);
		}
		str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "order");
		if(str) {
			order = atoi(str);
			free(str);
			if(order != 6) {
				cerr << "error: invalid lattice order " << order << "\n";
				exit(EXIT_FAILURE);
			}
		} else {
			cerr << "error: lattice order not specified\n";
			exit(EXIT_FAILURE);
		}
		// optional attributes
		str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "name");
		if(str) {
			name = str;
			free(str);
		}
		str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "scale");
		if(str) {
			scale = atof(str);
			free(str);
		}
	} else {
		cerr << "error: root lattice node not found\n";
		exit(EXIT_FAILURE);
	}

	// extract data from child nodes
	node = node->children;
	xdim = 0;
	ydim = 0;
	max_spin = 0;
	// determine lattice dimensions
	while(node) {
        if(node->type == XML_ELEMENT_NODE && strcmp((char *) node->name, \
				"site") == 0) {
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "x");
			if(str) {
				x = atoi(str);
				free(str);
			} else {
				cerr << "error: invalid site entry (no x attribute)\n";
				exit(EXIT_FAILURE);
			}
			if(x > xdim)
				xdim = x;
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "y");
			if(str) {
				y = atoi(str);
				free(str);
			} else {
				cerr << "error: invalid site entry (no y attribute)\n";
				exit(EXIT_FAILURE);
			}
			if(y > ydim)
				ydim = y;
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "spin");
			if(str) {
				x = atoi(str);
				free(str);
			} else {
				cerr << "error: invalid site entry (no spin attribute)\n";
				exit(EXIT_FAILURE);
			}
			if(x > max_spin)
				max_spin = x;
		}
		node = node->next;
	}
	xdim++;
	ydim++;
	zdim = 1;

	// allocate memory for lattice
	if(!(lattice = (LatticeSite ****) calloc(xdim, sizeof(LatticeSite ***)))) {
		cerr << "error: couldn't allocate lattice\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < xdim; i++) {
		if(!(lattice[i] = (LatticeSite ***) calloc(ydim, \
				sizeof(LatticeSite **)))) {
			cerr << "error: couldn't allocate lattice\n";
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < ydim; j++) {
			if(!(lattice[i][j] = (LatticeSite **) calloc(zdim, \
					sizeof(LatticeSite *)))) {
				cerr << "error: couldn't allocate lattice\n";
				exit(EXIT_FAILURE);
			}
			for(k = 0; k < zdim; k++) {
				if(!(lattice[i][j][k] = (LatticeSite *) malloc( \
						sizeof(LatticeSite)))) {
					cerr << "error: couldn't allocate lattice\n";
					exit(EXIT_FAILURE);
				}
				lattice[i][j][k]->x = i;
				lattice[i][j][k]->y = j;
				lattice[i][j][k]->z = k;
				lattice[i][j][k]->spin = 0;
				lattice[i][j][k]->type = 0;
				lattice[i][j][k]->name = NULL;
			}
		}
	}

	// populate lattice
	node = xmlDocGetRootElement(document)->children;
	while(node) {
        if(node->type == XML_ELEMENT_NODE && strcmp((char *) node->name, \
				"site") == 0) {
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "x");
			x = atoi(str);
			free(str);
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "y");
			y = atoi(str);
			free(str);
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "spin");
			lattice[x][y][0]->spin = atoi(str);
			free(str);
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "type");
			if(str) {
				lattice[x][y][0]->type = atoi(str);
				free(str);
			} else {
				cerr << "error: invalid site entry (no type attribute)\n";
				exit(EXIT_FAILURE);
			}
			str = (char *) xmlGetNoNsProp(node, (const xmlChar *) "name");
			if(str) {
				lattice[x][y][0]->name = str;
			} else {
				cerr << "error: invalid site entry (no name attribute)\n";
				exit(EXIT_FAILURE);
			}
		}
		node = node->next;
	}

	// set number of neighbours
	neighbourhood_size = order;

	// set site volume
	site_volume = order * scale * scale * sin((double) 2.0 * M_PI / order) \
			/ 2.0;

	// allocate memory for vectors to calculate neighbours
	if(!(nxvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood xvectors\n";
		exit(EXIT_FAILURE);
	}
	if(!(nyvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood yvectors\n";
		exit(EXIT_FAILURE);
	}
	if(!(nzvector = (int *) calloc(neighbourhood_size, \
			sizeof(int)))) {
		cerr << "error: couldn't allocate neighbourhood zvectors\n";
		exit(EXIT_FAILURE);
	}

	// set neighbourhood vectors TODO: generalise for any lattice
	nxvector[0] = 1;
	nyvector[0] = 0;
	nzvector[0] = 0;
	nxvector[1] = -1;
	nyvector[1] = 1;
	nzvector[1] = 0;
	nxvector[2] = 0;
	nyvector[2] = -1;
	nzvector[2] = 0;
	nxvector[3] = -1;
	nyvector[3] = 0;
	nzvector[3] = 0;
	nxvector[4] = -1;
	nyvector[4] = 1;
	nzvector[4] = 0;
	nxvector[5] = 0;
	nyvector[5] = 1;
	nzvector[5] = 0;

	// allocate memory for temporary storage for site neighbours
	num_neighbours = 0;
	if(!(neighbours = (LatticeSite **) calloc(neighbourhood_size, \
			sizeof(LatticeSite *)))) {
		cerr << "error: couldn't allocate lattice neighbours\n";
		exit(EXIT_FAILURE);
	}

	// initialise random number generator
	rng = new RNG(seed);

	// close xml tissue file
	xmlFreeDoc(document);
}

// destructor
Lattice::~Lattice()
{
	register unsigned long i, j, k;

	// free memory
	delete rng;
	free(neighbours);
	free(nzvector);
	free(nyvector);
	free(nxvector);
	for(i = 0; i < xdim; i++) {
		for(j = 0; j < ydim; j++) {
			for(k = 0; k < zdim; k++)
				free(lattice[i][j][k]);
			free(lattice[i][j]);
		}
		free(lattice[i]);
	}
	free(lattice);
}

// get size of lattice neighbourhood
unsigned int Lattice::getNeighbourhoodSize(void) { return neighbourhood_size; }

// get volume of a single lattice site
double Lattice::getSiteVolume(void) { return site_volume; }

// set lattice temperature
void Lattice::setTemperature(double t) { temperature = t; }

// get lattice temperature
double Lattice::getTemperature(void) { return temperature; }

// get random site on lattice
LatticeSite *Lattice::getSite(unsigned int x, unsigned int y, unsigned int z)
{
		return lattice[x][y][z];
}

// get random site on lattice
LatticeSite *Lattice::getRandomSite(void)
{
	if(zdim == 1)
		// 2d lattice
		return lattice[rng->randomNonnegativeInteger(xdim)] \
				[rng->randomNonnegativeInteger(ydim)][0];
	else
		// 3d lattice
		return lattice[rng->randomNonnegativeInteger(xdim)] \
				[rng->randomNonnegativeInteger(ydim)] \
				[rng->randomNonnegativeInteger(zdim)];
}

// calculate list of alien neighbours i.e. with different spin
unsigned int Lattice::calculateAlienNeighbours(LatticeSite *site)
{
	register unsigned int i;
	register unsigned long site_spin;
	register LatticeSite *neighbour;

	site_spin = site->spin;
	num_neighbours = 0;
	if(zdim == 1) {
		// 2d lattice
		for(i = 0; i < neighbourhood_size; i++)
			if(site->x + nxvector[i] < xdim && site->y + nyvector[i] < \
						ydim) {
				neighbour = lattice[site->x + nxvector[i]][site->y + \
						nyvector[i]][0];
				if(neighbour->spin != site_spin)
					neighbours[num_neighbours++] = neighbour;
			}
	} else
		// 3d lattice
		for(i = 0; i < neighbourhood_size; i++)
			if(site->x + nxvector[i] < xdim && site->y + nyvector[i] < \
						ydim && site->z + nzvector[i] < zdim) {
				neighbour = lattice[site->x + nxvector[i]][site->y + \
						nyvector[i]][site->z + nzvector[i]];
				if(neighbour->spin != site_spin)
					neighbours[num_neighbours++] = neighbour;
			}

	return num_neighbours;
}

// select random alien neighbour (only use after calculateAlienNeighbours)
LatticeSite *Lattice::selectAlienNeighbour(void)
{
	return neighbours[rng->randomNonnegativeInteger(num_neighbours)];
}

// get alien neighbour spin (only use after calculateAlienNeighbours)
unsigned long Lattice::getAlienNeighbourSpin(unsigned int n)
{
	return neighbours[n]->spin;
}

// flip site from one spin to another
void Lattice::flipSite(LatticeSite *from, LatticeSite *to)
{
	from->spin = to->spin;
	from->type = to->type;
}

// calculate energy to flip to one spin from another
double Lattice::calculateFlipEnergy(LatticeSite *from, LatticeSite *to)
{
	register unsigned int i;
	register unsigned long from_spin, to_spin;
	register LatticeSite *neighbour;
	register double from_energy, to_energy;

	// calculate current energy and flipped energy of lattice site
	from_spin = from->spin;
	to_spin = to->spin;
	from_energy = 0.0;
	to_energy = 0.0;
	if(zdim == 1) {
		// 2d lattice
		for(i = 0; i < neighbourhood_size; i++)
			if(from->x + nxvector[i] < xdim && from->y + nyvector[i] < ydim) {
				neighbour = lattice[from->x + nxvector[i]][from->y + \
						nyvector[i]][0];
				if(neighbour->spin != from_spin)
					from_energy += 1.0;  // current energy
				if(neighbour->spin != to_spin)
					to_energy += 1.0;  // flipped energy
			}
	} else
		// 3d lattice
		for(i = 0; i < neighbourhood_size; i++)
			if(from->x + nxvector[i] < xdim && from->y + nyvector[i] < \
						ydim && from->z + nzvector[i] < zdim) {
				neighbour = lattice[from->x + nxvector[i]][from->y + \
						nyvector[i]][from->z + nzvector[i]];
				if(neighbour->spin != from_spin)
					from_energy += 1.0;  // current energy
				if(neighbour->spin != to_spin)
					to_energy += 1.0;  // flipped energy
			}

	return (to_energy - from_energy);
}

// get lattice dimensions
unsigned long Lattice::getXDim(void) { return xdim; }
unsigned long Lattice::getYDim(void) { return ydim; }
unsigned long Lattice::getZDim(void) { return zdim; }

// get maximum spin index
unsigned long Lattice::getMaxSpin(void) { return max_spin; }

// get spin of lattice site
unsigned long Lattice::getSpin(unsigned int x, unsigned int y, unsigned int z)
{
	return lattice[x][y][z]->spin;
}

// get type of lattice site
unsigned long Lattice::getType(unsigned int x, unsigned int y, unsigned int z)
{
	return lattice[x][y][z]->type;
}

// get name of lattice site
char *Lattice::getSiteName(unsigned int x, unsigned int y, unsigned int z)
{
	return lattice[x][y][z]->name;
}
