// lattice class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#ifndef LATTICE_H
#define LATTICE_H

#include <iostream>
#include <ecsb/ecsb.h>

using namespace ECSB;

namespace Simulation {

// lattice site structure
typedef struct LatticeSite {
	unsigned long x;
	unsigned long y;
	unsigned long z;
	unsigned long spin;
	unsigned long type;
	char *name;
} LatticeSite;

// lattice class
class Lattice {
	std::string name;
	unsigned long xdim, ydim, zdim;  // lattice dimensions
	LatticeSite ****lattice;  // lattice
	unsigned int neighbourhood_size;  // number of site neighbours
	int *nxvector, *nyvector, *nzvector;  // vectors to calculate neighbours
	double site_volume;  // volume of single lattice site
	double scale;  // scale of lattice
	double temperature;  // lattice temperature
	unsigned int dimension;  // lattice dimension (2D/3D)
	unsigned int order;  // lattice order
	unsigned long max_spin;  // maximum spin label
	RNG *rng;  // random number generator

	// temporary storage for site neighbours
	unsigned int num_neighbours;
	LatticeSite **neighbours;

	public:
		// constructor
		Lattice(char *filename, unsigned long seed = 0);

		// destructor
		~Lattice();

		// get size of lattice neighbourhood
		unsigned int getNeighbourhoodSize(void);

		// get volume of a single lattice site
		double getSiteVolume();

		// set lattice temperature
		void setTemperature(double);

		// get lattice temperature
		double getTemperature(void);

		// get site on lattice
		LatticeSite *getSite(unsigned int x, unsigned int y, \
				unsigned int z = 0);

		// get spin of lattice site
		unsigned long getSpin(unsigned int x, unsigned int y, \
				unsigned int z = 0);

		// get type of lattice site
		unsigned long getType(unsigned int x, unsigned int y, \
				unsigned int z = 0);

		// get random site on lattice
		LatticeSite *getRandomSite(void);

		// calculate list of alien neighbours i.e. with different spin
		unsigned int calculateAlienNeighbours(LatticeSite *);

		// select random alien neighbour (use after calculateAlienNeighbours)
		LatticeSite *selectAlienNeighbour(void);

		// get alien neighbour spin (use after calculateAlienNeighbours)
		unsigned long getAlienNeighbourSpin(unsigned int);

		// calculate energy to flip to one spin from another
		double calculateFlipEnergy(LatticeSite *from, LatticeSite *to);

		// flip site from one spin to another
		void flipSite(LatticeSite *from, LatticeSite *to);

		// get lattice dimensions
		unsigned long getXDim(void);
		unsigned long getYDim(void);
		unsigned long getZDim(void);

		// get maximum spin index
		unsigned long getMaxSpin(void);

		// get name of lattice site
		char *getSiteName(unsigned int x, unsigned int y, unsigned int z);
};

}

#endif
