// distributed mcss and cma-es parameter optimisation
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "McssCmaes.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "mcss/mcss.h"

using namespace Simulation;

static McssCmaesParameters cmaesparameters;
static SimulationParameters mcssparameters;
static double **target_data;
static int target_data_rows, target_data_columns;
static double *target_data_timepoints;

// return mcss-cmaes parameters
McssCmaesParameters *Simulation::cmaesParameters(void)
{
	return &cmaesparameters;
}

// return mcss parameters
SimulationParameters *Simulation::mcssParameters(void)
{
	return &mcssparameters;
}

// read in target data
void Simulation::readTargetData(void)
{
	FILE *fp;
	char buffer[1024], valstr[1024], *str, *str2, *endptr;
	int i, rows, columns, count;
	bool flag;
	double val;

	if(!(fp = fopen(cmaesparameters.target_data_file, "r"))) {
		fprintf(stderr, "error: unable to open target data file %s\n", \
				cmaesparameters.target_data_file);
		exit(EXIT_FAILURE);
	}

	// count number of rows and columns
	rows = columns = 0;
	while(freadline(buffer, 1024, fp)) {
		// ignore blank lines
		flag = true;
		for(i = 0; i < strlen(buffer); i++)
			if(!isspace(buffer[i])) {
				flag = false;
				break;
			}
		if(flag)
			continue;

		// count columns and sanity check
		str = buffer;
		count = 1;
		while((str = strchr(str, cmaesparameters.column_separator))) {
			str++;
			count++;
		}
		if(rows) {
			if(count != columns) {
				fprintf(stderr, "error: invalid number of columns in " \
						"target data file column \"%s\"\n", buffer);
				exit(EXIT_FAILURE);
			}
		} else
			columns = count;

		rows++;
	}
	if(!rows) {
		fprintf(stderr, "error: zero rows in target data\n");
		exit(EXIT_FAILURE);
	}

	// allocate memory
	target_data_rows = rows;
	target_data_columns = columns;
	if(!(target_data = (double **) calloc(rows, sizeof(double *)))) {
		fprintf(stderr, "error: unable to allocate memory for target data " \
				"rows\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < rows; i++)
		if(!(target_data[i] = (double *) calloc(columns, sizeof(double)))) {
			fprintf(stderr, "error: unable to allocate memory for target " \
					"data columns\n");
			exit(EXIT_FAILURE);
		}
	if(!(target_data_timepoints = (double *) calloc(rows, sizeof(double)))) {
		fprintf(stderr, "error: unable to allocate memory for target data " \
				"times\n");
		exit(EXIT_FAILURE);
	}

	// read in target data
	rewind(fp);
	rows = 0;
	while(freadline(buffer, 1024, fp)) {
		// ignore blank lines
		flag = true;
		for(i = 0; i < strlen(buffer); i++)
			if(!isspace(buffer[i])) {
				flag = false;
				break;
			}
		if(flag)
			continue;
		str = buffer;
		for(columns = 0; columns < target_data_columns; columns++) {
			str2 = strchr(str, cmaesparameters.column_separator);
			if(!str2)
				str2 = buffer + strlen(buffer);
			strncpy(valstr, str, str2 - str);
			valstr[str2 - str] = '\0';
			errno = 0;
			val = strtod(valstr, &endptr);
			if(errno != 0 || endptr == valstr || *endptr != '\0') {
				fprintf(stderr, "error: invalid entry in column %d of target " \
						"data column \"%s\"\n", target_data_columns - columns, \
						buffer);
				exit(EXIT_FAILURE);
			}
			target_data[rows][columns] = val;
			if(columns == 0)
				target_data_timepoints[rows] = val;
			str = str2 + 1;
		}

		rows++;
	}

	// tidy up
	fclose(fp);
}

// return number of rows in target data
int Simulation::targetDataRows(void)
{
	return target_data_rows;
}

// return number of columns in target data
int Simulation::targetDataColumns(void)
{
	return target_data_columns;
}

// return array of target data timepoints
double *Simulation::targetDataTimepoints(void)
{
	return target_data_timepoints;
}

// return entry in target data
double Simulation::targetData(int row, int column)
{
	if(row < 1 || row > target_data_rows) {
		fprintf(stderr, "error: invalid row %d in target data\n", row);
		exit(EXIT_FAILURE);
	}
	if(column < 1 || column > target_data_columns) {
		fprintf(stderr, "error: invalid column %d in target data\n", column);
		exit(EXIT_FAILURE);
	}
	return target_data[row - 1][column - 1];
}

// return target data array
double **Simulation::targetData(void)
{
	return target_data;
}

// run mcss from parameter file
void Simulation::run_mcss(const char *param_file, const char *params)
{
	int argc, i;
	char **argv, *params2, *ptr;

	// create parameters
	if(!(params2 = (char *) calloc(strlen(params) + 1, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameters\n");
		exit(EXIT_FAILURE);
	}
	strcpy(params2, params);
	argc = 0;
	if(!(argv = (char **) calloc(2, sizeof(char *)))) {
		fprintf(stderr, "error: unable to allocate memory for parameters\n");
		exit(EXIT_FAILURE);
	}
	if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameter\n");
		exit(EXIT_FAILURE);
	}
	strncpy(argv[argc++], "mcss", 128);
	if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameter\n");
		exit(EXIT_FAILURE);
	}
	strncpy(argv[argc++], param_file, 128);
	while((ptr = strsep(&params2, " "))) {
		if(strcmp(ptr, "") == 0)
			continue;
		if(!(argv = (char **) realloc(argv, sizeof(char *) * (argc + 1)))) {
			fprintf(stderr, \
					"error: unable to reallocate memory for parameters\n");
			exit(EXIT_FAILURE);
		}
		if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
			fprintf(stderr, "error: unable to allocate memory for parameter\n");
			exit(EXIT_FAILURE);
		}
		strncpy(argv[argc++], ptr, 128);
	}

	// read in parameters
	SimulationParameters *parameters = new SimulationParameters();
	parameters->read(argc, argv);

	// free memory
	for(i = 0; i < argc; i++)
		free(argv[i]);
	free(argv);
	free(params2);

	#ifdef HAVE_MPI
	// set mpi to operate in serial mode
	parameters->mpi_serial = true;
	#endif

	// create p-system from parameters specified in parameter file
	Psystem *psystem = new Psystem(*parameters);

	// execute simulation
	psystem->execute(parameters->max_time);

	// tidy up
	delete parameters;
	delete psystem;
}
