// p-system class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Psystem.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <map>
#include <libxml++/libxml++.h>
#include <strings.h>
#include <string.h>
#include <libgen.h>
#include <string>
#include <math.h>
#include <ctype.h>
#include <sstream>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_const_num.h>
#include <vector>

using namespace xmlpp;
using namespace Simulation;

// default constructor
Psystem::Psystem()
{
	// set initial values
	setInitialisationValues();
}

// constructor - create from parameter structure
Psystem::Psystem(SimulationParameters &parameters)
{
	// ignore signal handler while creating psystem
//	SignalHandler::ignore_handler(SIGINT);

	// set initial values
	setInitialisationValues();

	if(!parameters.just_psystem) {
		// set simulation algorithm
		if(strcmp(parameters.simulation_algorithm, "dm") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespie(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "ldm") == 0)
			setSimulationAlgorithm(new \
					LogarithmicDirectMethod(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmgd") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieGD(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmq2gd") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieQ2GD(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmq") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieQ(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmq2") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieQ2(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmqg") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieQG(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmq2g") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieQ2G(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "dmcp") == 0)
			setSimulationAlgorithm(new \
					MulticompartmentGillespieCellularPotts(parameters.seed));
		if(strcmp(parameters.simulation_algorithm, "ode1") == 0)
			setSimulationAlgorithm(new \
					DeterministicODE(parameters.seed));
	}

	// set lattice boundary conditions
	periodic_x = parameters.periodic_x;
	periodic_y = parameters.periodic_y;
	periodic_z = parameters.periodic_z;
	
	// set lattice neighbourhood (von neumann = 4, moore = 8)
	neighbourhood = parameters.neighbourhood;

	// duplicate initial amounts of template compartment species
	duplicate_initial_amounts = parameters.duplicate_initial_amounts;

	// keep levels of degraded species
	keep_degraded = parameters.log_degraded;

	// log volume data
	log_volumes = parameters.log_volumes;

	// set volume growth type
	if(strcmp(parameters.growth_type, "none") == 0)
		growth_type = NO_GROWTH;
	if(strcmp(parameters.growth_type, "linear") == 0)
		growth_type = LINEAR_GROWTH;
	if(strcmp(parameters.growth_type, "exponential") == 0)
		growth_type = EXPONENTIAL_GROWTH;
	if(strcmp(parameters.growth_type, "function") == 0)
		growth_type = FUNCTION_GROWTH;

	// create p system
	if(strcmp(parameters.model_format, "sbml") == 0)
		createFromSBML(parameters);
	if(strcmp(parameters.model_format, "xml") == 0 || \
				strcmp(parameters.model_format, "lpp") == 0) {
		#ifdef HAVE_BNF
		createFromXML(parameters);
		#else
		cerr << "error: xml models not supported.\n";
		exit(EXIT_FAILURE);
		#endif
	}

	// create + populate lattice if using gillespie cellular potts algorithm
	if(!parameters.just_psystem) {
		if(strcmp(parameters.simulation_algorithm, "dmcp") == 0)
			createLattice(parameters);
		else
			lattice = NULL;
	}

	// save list of initial compartments
	saveInitialCompartments();

	if(!parameters.just_psystem) {
		// set simulation algorithm parameters
		simalg->setPsystem(this);
		simalg->setNumCompartments(num_compartments);
		simalg->setNumSpecies(alphabet->size());
		simalg->setNumRuns(parameters.runs);
		simalg->setKeepDivisions(parameters.keep_divisions);
		simalg->setDivisionDirection(parameters.division_direction);
		simalg->setCompress(parameters.compress);
		simalg->setCompressionLevel(parameters.compression_level);
		simalg->setScaleTranslocation(parameters.scale_translocation);
		simalg->setStaticLattice(parameters.static_lattice);
		simalg->setMpiSerial(parameters.mpi_serial);

		// get random number generator from simulation algorithm
		rng = simalg->getRNG();

		// set potts rate if using gillespie cellular potts algorithm
		if(strcmp(parameters.simulation_algorithm, "dmcp") == 0)
			simalg->setPottsRate(parameters.potts_rate);

		// set volume rate if using gillespie queue growth algorithm
		if((strcmp(parameters.simulation_algorithm, "dmqg") == 0) || \
					(strcmp(parameters.simulation_algorithm, "dmq2g") == 0) || \
					(strcmp(parameters.simulation_algorithm, "dmq2gd") == 0))
			simalg->setVolumeRate(parameters.volume_rate);

		// set division rate if using gillespie queue growth division algorithm
		if((strcmp(parameters.simulation_algorithm, "dmq2gd") == 0))
			simalg->setDivisionRate(parameters.division_rate);

		// set ode solver algorithm if using deterministic ode algorithm
		// see gsl documentation for more information of each algorithm
		if((strcmp(parameters.simulation_algorithm, "ode1") == 0)) {
			if((strcmp(parameters.ode_solver, "rk2") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rk2);
			if((strcmp(parameters.ode_solver, "rk4") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rk4);
			if((strcmp(parameters.ode_solver, "rkf45") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rkf45);
			if((strcmp(parameters.ode_solver, "rkck") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rkck);
			if((strcmp(parameters.ode_solver, "rk8pd") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rk8pd);
			if((strcmp(parameters.ode_solver, "rk2imp") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rk2imp);
			if((strcmp(parameters.ode_solver, "rk4imp") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_rk4imp);
			if((strcmp(parameters.ode_solver, "bsimp") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_bsimp);
			if((strcmp(parameters.ode_solver, "gear1") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_gear1);
			if((strcmp(parameters.ode_solver, "gear2") == 0))
				simalg->setOdeSolver(gsl_odeiv_step_gear2);
			simalg->setAbsoluteError(parameters.ode_absolute_error);
			simalg->setRelativeError(parameters.ode_relative_error);
			simalg->setInitialStepSize(parameters.ode_initial_step_size);
			simalg->setYScaling(parameters.ode_yscaling);
			simalg->setDYScaling(parameters.ode_dyscaling);
		}

		// create file to store output data
		if(!parameters.dump)
			simalg->createOutputDataset(parameters);

		// setup event handling
		simalg->initialiseEvents();
	}

	// resume signal handler
//	SignalHandler::unignore_handler(SIGINT);
}

// constructor - create from sbml file - no parameters
Psystem::Psystem(char *model_file)
{
	// set initial values
	setInitialisationValues();

	// create from sbml
	createFromSBML(model_file);
}

// set default values for initialised class
void Psystem::setInitialisationValues(void)
{
	// set initial values
	id = "";
	simalg = NULL;
	num_compartments = 0;
	compartment = NULL;
	num_species = 0;
	alphabet = NULL;
	num_rulesets = 0;
	ruleset = NULL;
	num_initial_compartments = 0;
	initial_compartment = NULL;
	xdim = ydim = zdim = 0;
	lattice = NULL;
	initial_xdim = initial_ydim = initial_zdim = 0;
	periodic_x = periodic_y = periodic_z = false;
	sink_x = sink_y = sink_z = false;
	log_reactions = false;
	log_volumes = false;
	duplicate_initial_amounts = false;
	keep_degraded = false;
	growth_type = NO_GROWTH;
	volume_scale = 1.0;
	neighbourhood = 4;
	num_species_flat = 0;
	num_rules_flat = 0;
	flat_species_species_lookup = NULL;
	flat_species_compartment_lookup = NULL;
	flat_rule_compartment_lookup = NULL;
	flat_rule_ruleset_lookup = NULL;
	flat_rule_rule_lookup = NULL;
}

// destructor
Psystem::~Psystem()
{
	unsigned long i;

	// free memory
	if(simalg)
		delete simalg;
	if(alphabet)
		delete alphabet;
	if(ruleset) {
		for(i = 0; i < num_rulesets; i++)
			if(ruleset[i])
				delete ruleset[i];
		free(ruleset);
	}
	if(compartment) {
		for(i = 0; i < num_compartments; i++)
			if(compartment[i])
				delete compartment[i];
		free(compartment);
	}
	if(initial_compartment) {
		for(i = 0; i < num_initial_compartments; i++)
			if(initial_compartment[i])
				delete initial_compartment[i];
		free(initial_compartment);
	}
	if(lattice)
		delete lattice;

	if(flat_species_species_lookup)
		free(flat_species_species_lookup);
	if(flat_species_compartment_lookup)
		free(flat_species_compartment_lookup);
	if(flat_rule_compartment_lookup)
		free(flat_rule_compartment_lookup);
	if(flat_rule_ruleset_lookup)
		free(flat_rule_ruleset_lookup);
	if(flat_rule_rule_lookup)
		free(flat_rule_rule_lookup);
}

// general access methods
void Psystem::setId(std::string n) { id = n; }
std::string Psystem::getId(void) { return id; }
unsigned long Psystem::getNumCompartments(void) { return num_compartments; }
Alphabet *Psystem::getAlphabet(void) { return alphabet; }
SimulationAlgorithm *Psystem::getSimulationAlgorithm(void) { return simalg; }
void Psystem::setSimulationAlgorithm(SimulationAlgorithm *sa) { simalg = sa; }
Simulation::Compartment *Psystem::getCompartment(unsigned long i) \
		{ return compartment[i]; }
Lattice *Psystem::getLattice(void) { return lattice; }
unsigned long Psystem::getLatticeX(void) { return xdim; }
unsigned long Psystem::getLatticeY(void) { return ydim; }
unsigned long Psystem::getLatticeZ(void) { return zdim; }
Simulation::Rule *Psystem::getRule(unsigned long rs, unsigned long r, \
		unsigned long c) { return compartment[c]->getRule(rs, r); }
unsigned long Psystem::getNumSpecies(void) { return alphabet->size(); }
unsigned long Psystem::getNumRuleSets(void) { return num_rulesets; }
unsigned long Psystem::getNumRuleSets(unsigned long i) \
		{ return compartment[i]->getNumRuleSets(); }
Simulation::RuleSet *Psystem::getRuleSet(unsigned long i) \
		{ return ruleset[i]; }
unsigned long Psystem::getRuleSetIndex(unsigned long c, unsigned long rs) \
		{ return compartment[c]->getRuleSetIndex(rs); }
unsigned long Psystem::getNumRules(unsigned long c, unsigned long rs) \
		{ return compartment[c]->getNumRules(rs); }
void Psystem::logReactions(bool f) { log_reactions = f; }
unsigned long Psystem::getSpeciesIndex(string s) \
		{ return alphabet->getSpeciesIndex(s); }
bool Psystem::isDegradation(string s) { return alphabet->isDegradation(s); }
unsigned long Psystem::getNumTimepoints(void) \
		{  return simalg->getNumTimepoints(); }
unsigned int ***Psystem::getAmountsDataset(void) \
		{ return simalg->getAmountsDataset(); }
double Psystem::getCompartmentVolume(unsigned long c) \
		{ return compartment[c]->getVolume(); }
void Psystem::setCompartmentVolume(unsigned long c, double v) \
		{ compartment[c]->setVolume(v); }
double Psystem::getReactionConstant(unsigned long c, unsigned long rs, \
		unsigned long r) \
		{ return compartment[c]->getReactionConstant(rs, r); }
void Psystem::resetSimulation(void) { simalg->resetSimulation(); }
bool Psystem::isChannel(unsigned long c, unsigned long rs, unsigned long r) \
		{ return getRule(rs, r, c)->isChannel(); }
string Psystem::getTargetName(unsigned long c, unsigned long rs, unsigned long r) \
		{ return getRule(rs, r, c)->getTargetName(); }
string Psystem::getTargetType(unsigned long c, unsigned long rs, unsigned long r) \
		{ return getRule(rs, r, c)->getTargetType(); }
unsigned long Psystem::getTargetCompartment(unsigned long c, unsigned long rs, unsigned long r) \
		{ return compartment[c]->getRuleTarget(rs, r); }
double Psystem::getLogInterval(void) { return simalg->getLogInterval(); }
string Psystem::getCompartmentName(unsigned long c) \
		{ return compartment[c]->getName(); }
double Psystem::getSimulationTime(void) { return simalg->getTime(); }

// create p system from sbml model
void Psystem::createFromSBML(SimulationParameters &parameters)
{
	createFromSBML(parameters.model_file);
}

// create p system from sbml model
void Psystem::createFromSBML(char *model_file)
{
	SBMLDocument *sbml_document;

	sbml_document = readSBML(model_file);
	if(sbml_document->getNumErrors() && \
			sbml_document->getError(0)->getErrorId() == XMLFileUnreadable) {
		// try to find model file in same directory as parameter file
		char pfilename[1024];
		char filename[1024];
		strncpy(pfilename, model_file, 1024);
		strncpy(filename, dirname_r(pfilename), 1024);
		strncat(filename, "/", 1024);
		strncat(filename, model_file, 1023);
		sbml_document = readSBML(filename);
	}
	if(sbml_document->getNumErrors()) {
		unsigned int i;
		for(i = 0; i < sbml_document->getNumErrors(); i++) {
			// ignore mathml errors
			if(sbml_document->getError(i)->getErrorId() != BadMathMLNodeType)
				break;
		}
		if(i != sbml_document->getNumErrors()) {
			cout << "error reading sbml input file\n";
			sbml_document->printErrors(cerr);
			exit(EXIT_FAILURE);
		}
	}

	Model *sbml_model = sbml_document->getModel();

	// set model units
	setUnits(sbml_model);

	// set model name
	setId(sbml_model->getId());

	// extract alphabet information
	createAlphabet(sbml_model);

	// extract compartment information
	createCompartments(sbml_model);

	// free memory
	delete sbml_document;
}

#ifdef HAVE_BNF
// create p system from xml model
void Psystem::createFromXML(SimulationParameters &parameters)
{

  unsigned long i, j, ind;

  //Parse the LPPsystem model
  ECSB::LPPsystem model(parameters.model_file);

  //Name of the model
  this->id = model.getName();


  std::map<std::string,ECSB::PSystem> pSystems = model.getPsystems();

  //Convert P systems rules to translocation rules when necessary
  for(std::map<std::string,ECSB::PSystem>::iterator pSyst = pSystems.begin();
    pSyst != pSystems.end();
    ++pSyst)
  {
      (pSyst->second).convertTranslocation();
  }

  model.setPsystems(pSystems);
  //model.show();

  //Create alphabet from LPPsystem model
  std::vector<std::string> combinedAlphabets = model.getCombinedAlphabets();
  createAlphabet(combinedAlphabets);

  //Create rule sets from LPPsystem model. The rule sets of all the compartments of each
  //P system are stored in a map where the key is the concatenation of the P system name
  //with the rule set label (which corresponds with the compartment name).
  std::map<std::string,ECSB::LPPRuleSet> combinedRuleSets = model.getCombinedRuleSets();
  std::map<std::string,unsigned long> labelsId = createLabelsId(combinedRuleSets);

  createRuleSets(combinedRuleSets,labelsId);

  //Create initial multisets from LPPsystem model. The initial multiset of all the compartments
  //from each P system are stored in a map where the key is the concatenation of the P system
  //name and the initial multiset label (which corresponds with the compartment name).
  std::map<std::string,ECSB::InitialMultiset>
    combinedInitialMultisets = model.getCombinedInitialMultisets();

  //Create compartments from CAPsystem model.
  createCompartments(model,combinedInitialMultisets);
}

// create alphabet from xml model
void Psystem::createAlphabet(std::vector<std::string> combinedAlphabets)
{
  // create storage for alphabet
  num_species = combinedAlphabets.size();
  alphabet = new Simulation::Alphabet(num_species);

  // set object names
  int ind = 0;
  for(std::vector<std::string>::iterator obj = combinedAlphabets.begin();
      obj != combinedAlphabets.end();
      ++obj)
    {
      // set name
      alphabet->setName(ind,*obj);
      ind++;
    }
}

// create label ids from xml model
std::map<std::string,unsigned long> Psystem::createLabelsId( \
		std::map<std::string, ECSB::LPPRuleSet> combinedRuleSets)
{
  std::map<std::string,unsigned long> labelsId;
  unsigned long ind = 0;

  for(std::map<std::string,ECSB::LPPRuleSet>::iterator l = \
  		combinedRuleSets.begin();
      l != combinedRuleSets.end();
      ++l)
    {
      std::string labName = l->first;
      labelsId.insert(pair<string,unsigned long> (labName, ind));
      ind++;
    }

  return labelsId;
}

// create rulesets from xml model
void Psystem::createRuleSets(std::map<std::string,ECSB::LPPRuleSet> \
		combinedRuleSets, std::map<std::string, unsigned long> labelsId)
{
  //Initialise the number of rule sets
  this->num_rulesets = labelsId.size();
  //Allocate memory for ruleset
  this->ruleset = new Simulation::RuleSet * [num_rulesets];

  //Rule set index
  unsigned long ind = 0;

  for(std::map<std::string,ECSB::LPPRuleSet>::iterator combRuleSet = combinedRuleSets.begin();
      combRuleSet != combinedRuleSets.end();
      ++combRuleSet)
    {
      std::string rulesLabel = combRuleSet->first;
      ECSB::LPPRuleSet rules = combRuleSet->second;
      // allocate memory
      ruleset[ind] = new Simulation::RuleSet(rules.getNumberRules());

      // set basic information for each ruleset
      ruleset[ind]->setIndex(labelsId[rulesLabel]);
      ruleset[ind]->setId(rulesLabel);
      ruleset[ind]->setName(rulesLabel);

      // set the rules of each ruleset
      int ruleIndex = 0;
      std::vector<ECSB::LPPRule> LPPRules = rules.getRules();

      for(vector<ECSB::LPPRule>::iterator ru = LPPRules.begin();
	  ru != LPPRules.end();
	  ++ru)
	{
	  // get name and id of each rule
	  std::string ruleId = ru->getName();
	  ruleset[ind]->setRule(ruleIndex, new Simulation::Rule(ruleId));
	  (ruleset[ind]->getRule(ruleIndex))->setId(ruleId);
	  (ruleset[ind]->getRule(ruleIndex))->setName(ruleId);

	  // get reactants, only the objects inside or translocated are taken into account
	  std::vector<std::string> reacts;


	  if(ru->getType() == "translocation")
	    {
	      (ruleset[ind]->getRule(ruleIndex))->setChannel();
	      reacts = ru->getLhsTranslocation(); //translocation rule
	    }
	  else
	    reacts = ru->getLhsInside();        //boundary rule

	  // set number of reactants
	  (ruleset[ind]->getRule(ruleIndex))->setNumReactants(reacts.size());

	  // set products, only the objects inside or translocated are taken into account
	  std::vector<std::string> prods;
	  if(ru->getType() == "translocation")
	    prods = ru->getRhsTranslocation(); //translocation rule
	  else
	    prods = ru->getRhsInside();        //boundary rule

	  // set number of products
	  (ruleset[ind]->getRule(ruleIndex))->setNumProducts(prods.size());

	  // set stochastic reaction constant
	  (ruleset[ind]->getRule(ruleIndex))->setReactionConstant(ru->getConstant());

	  // set reactants
	  unsigned long i = 0;
	  for(std::vector<std::string>::iterator react = reacts.begin();
	      react != reacts.end();
	      ++react)
	    {
	      (ruleset[ind]->getRule(ruleIndex))->
		setReactantIndex(i, alphabet->getSpeciesIndex(*react));
	      i++;
	    }

	  // set products
	  i = 0;
	  for(std::vector<std::string>::iterator prod = prods.begin();
	      prod != prods.end();
	      ++prod)
	    {
	      (ruleset[ind]->getRule(ruleIndex))->
		setProductIndex(i, alphabet->getSpeciesIndex(*prod));
	      i++;
	    }
	  // set rule vector
	  if(ru->getType() == "translocation")
	    {
	      std::vector<int> ruVector = ru->getVector();
	      ruleset[ind]->getRule(ruleIndex)->setXVector(ruVector[0]);
	      ruleset[ind]->getRule(ruleIndex)->setYVector(ruVector[1]);
	      ruleset[ind]->getRule(ruleIndex)->setZVector(ruVector[2]);
	    }
	  ruleIndex++;
	}
      ind++;
    }
}

// create compartments from xml model
void Psystem::createCompartments(ECSB::LPPsystem model, \
		std::map<std::string,ECSB::InitialMultiset> combinedInitialMultisets)
{
  register unsigned long updates[1];

  //the compartments are created following the positions from the LPP model
  //std::map<vector<long>,std::string> positions = model.getPositions().getPositions();

  //std::vector<ECSB::automaton> cellAutomata = (model.getCellularAutomata()).getAutomata();

  //allocate memory for compartments
  this->num_compartments = model.getPositionsPtr()->getNumberOfCompartments();

  this->compartment = new Simulation::Compartment * [this->num_compartments];

  //populate compartments
  unsigned long ind = 0; //index of the current compartment being created
  this->xdim = 0;
  this->ydim = 0;
  this->zdim = 0;

  std::map<std::string,ECSB::PSystem> pSystems = model.getPsystems();

	for(std::map<vector<long>,std::string>::iterator position = \
		model.getPositionsPtr()->getPositionsPtr()->begin(); \
		position != model.getPositionsPtr()->getPositionsPtr()->end(); \
		++position)
    {
      unsigned long pos1[3];
      std::vector<long> automatonPosition1 = position->first;

      ECSB::PSystem pSystem = pSystems[position->second];

      std::map<std::string,ECSB::LPPCompartment> pSystemCompartments = pSystem.getMembraneStructure();
      for(std::map<std::string,ECSB::LPPCompartment>::iterator comp = pSystemCompartments.begin();
	  comp != pSystemCompartments.end();
	  ++comp)
	{
	  this->compartment[ind] = new Simulation::Compartment(this->id);

	  //The id of each compartment is obtained by concateneting the name of the
	  //corresponding P system and the name of the compartment in the LPP model
	  std::string compIdentification = pSystem.getName() + (comp->second).getName();
	  this->compartment[ind]->setId(compIdentification);
	  this->compartment[ind]->setName(compIdentification);

	  //Assigning the position of each compartment
      	  unsigned long pos[3];
	  std::vector<long> automatonPosition = position->first;
	  pos[0] = automatonPosition[0];
	  pos[1] = automatonPosition[1];
	  pos[2] = (comp->second).getNumberId();

	  this->compartment[ind]->setPosition(pos);

	  // determine size of lattice
	  if(this->compartment[ind]->getXPosition() > this->xdim)
	    this->xdim = this->compartment[ind]->getXPosition();
	  if(this->compartment[ind]->getYPosition() > this->ydim)
	    this->ydim = this->compartment[ind]->getYPosition();
	  if(this->compartment[ind]->getZPosition() > this->zdim)
	    this->zdim = this->compartment[ind]->getZPosition();

	  //Set the initial multiplicity of the objects
	  this->compartment[ind]->initialiseLevels(this->alphabet->size());

	  std::map<std::string, unsigned long> multiplicities =
	    combinedInitialMultisets[compIdentification].getMultiset();


	  for(int i = 0; i < this->alphabet->size(); i++)
          {
	    if(multiplicities.find(this->alphabet->getName(i)) !=	\
	       multiplicities.end())
            {
		this->compartment[ind]->setLevel(i, multiplicities[alphabet->getName(i)]);
	    }
	    else
	    {
		this->compartment[ind]->setLevel(i, 0);
	    }
          }

	  ind++; //next compartment to create

	}

    }

  for(ind = 0; ind < num_compartments; ++ind)
    {
      //There is only one single set of rules per compartment
      this->compartment[ind]->setNumRuleSets(1);

      //The set of rules associated with each compartment is determined by its
      //name
      for(int i = 0; i < this->num_rulesets; i++)
	{
	  if(this->ruleset[i]->getName() == this->compartment[ind]->getName())
	    {

	      this->compartment[ind]->setRuleSetIndex(0, i);
	      this->compartment[ind]->setRuleSet(0, this->ruleset[i]);

	      // all rules are valid in the xml specification
/*	      this->compartment[ind]->initialiseValidRules(0,this->ruleset[i]->getNumRules());
		  
	      for(int j = 0; j < ruleset[i]->getNumRules(); j++)
		{
		  this->compartment[ind]->setValidRule(0, j, j);

		  if(ruleset[i]->isChannel(j))
		    {
				// set update compartments
				if((updates[0] = calculateRuleTarget(0, j, ind)) != num_compartments)
					compartment[ind]->setUpdates(0, j, updates, 1);
		    }
		}*/
	    }
	}
	    setValidRules(ind);
    }
}
#endif

// set model units from sbml file
void Psystem::setUnits(Model *sbml_model)
{
	unsigned int i, j;
	ListOfUnitDefinitions *sbml_unit_definitions;
	UnitDefinition *sbml_unit_definition;
	ListOfUnits *sbml_units;
	Unit *sbml_unit;
	UnitKind_t sbml_unit_type;

	sbml_unit_definitions = sbml_model->getListOfUnitDefinitions();
	for(i = 0; i < sbml_unit_definitions->size(); i++) {
		sbml_unit_definition = sbml_unit_definitions->get(i);
		if(sbml_unit_definition->getId() == "volume") {
			sbml_units = sbml_unit_definition->getListOfUnits();
			for(j = 0; j < sbml_units->size(); j++) {
				sbml_unit = sbml_units->get(j);
				if(!sbml_unit->isSetKind()) {
					cerr << "error: unknown volume unit type\n";
					exit(EXIT_FAILURE);
				}
				sbml_unit_type = sbml_unit->getKind();
				switch(sbml_unit_type) {
					case UNIT_KIND_METRE:
					case UNIT_KIND_METER:
						if(sbml_unit->getExponent() != -3)
							continue;  // error
						volume_scale = 1.0E3 * pow((double) 10.0, \
								(double) sbml_unit->getScale()) * \
								sbml_unit->getMultiplier();
						break;

					case UNIT_KIND_LITRE:
					case UNIT_KIND_LITER:
						if(sbml_unit->getExponent() != 1)
							continue;  // error
						volume_scale = 1.0 * pow((double) 10.0, \
								(double) sbml_unit->getScale()) * \
								sbml_unit->getMultiplier();
						break;

					default:
						cerr << "error: unknown volume unit type\n";
						exit(EXIT_FAILURE);
						break;
				}
			}
		}
	}
}

// create alphabet from sbml model
void Psystem::createAlphabet(Model *sbml_model)
{
	std::string name;
	unsigned long int i, j;

	// create list of unique species names
	std::list<string> names = std::list<string> ();
	for(i = 0; i < sbml_model->getNumSpecies(); i++) {
		name = sbml_model->getSpecies(i)->getName();
		if(!isDegradation(name) || (isDegradation(name) && keep_degraded)) {
			validateSpeciesName(name);
			names.push_front(name);
		}
	}
	names.sort();
	names.unique();

	// initialise alphabet
	num_species = names.size();
	alphabet = new Simulation::Alphabet(num_species);

	// set alphabet names + attributes
	list<string>::iterator it = names.begin();
	for(i = 0; i < num_species; i++) {
		name = *it++;
		// set name
		alphabet->setName(i, name);
		// set flags
		for(j = 0; j < sbml_model->getNumSpecies(); j++)
			if(name == sbml_model->getSpecies(j)->getName()) {
				if(sbml_model->getSpecies(j)->getBoundaryCondition())
					alphabet->setNondividing(i);
				if(sbml_model->getSpecies(j)->getConstant()) {
					alphabet->setConstant(i);
					alphabet->setNondividing(i);
				}
			}
	}
}

// create compartments from sbml model
void Psystem::createCompartments(Model *sbml_model)
{
	register unsigned long i, j, k, rule_index, count, updates[1];
	register unsigned long ind, ind2, ind3, ind4;
	register unsigned long num_reactants, num_products, *rstmp;
	register int offset;
	long unsigned int pos[3];
	string name, id, name1, name2;
	RuleSet *rs;
	Rule *rule;
	Parameter *parameter;
	KineticLaw *kinetic_law;
	::Event *ev;

	// handle single compartment models
	if(sbml_model->getNumCompartments() == 1 && \
			sbml_model->getCompartment(0)->getId() == "default") {
		::Compartment *comp;
		// set compartment name
		name = "c1";
		sbml_model->getCompartment(0)->setId(name);
		name = "module:1:";
		sbml_model->getCompartment(0)->setName(name);
		// add instantiated compartment
		comp = sbml_model->createCompartment();
		name = "c2";
		comp->setId(name);
		name = "compartment::1:0,0";
		comp->setName(name);
		// adjust compartment names of species
		for(i = 0; i < sbml_model->getNumSpecies(); i++) {
			name = sbml_model->getSpecies(i)->getCompartment();
			if(name == "default")
				sbml_model->getSpecies(i)->setCompartment("c1");
		}
	}

	// count number of compartments and rule templates
	num_rulesets = 0;
	num_compartments = 0;
	for(i = 0; i < sbml_model->getNumCompartments(); i++) {
		// ignore default compartment
		if(sbml_model->getCompartment(i)->getId() == "default")
			continue;
		// validate compartment name compartment
		name = sbml_model->getCompartment(i)->getName();
		validateCompartmentName(name);
		if(isTemplateCompartment(name))
			num_rulesets++;  // it's a template compartment
		if(getCompartmentPositionFromSBML(name, pos))
			num_compartments++;  // it's on the lattice
	}

	// count number of event templates
	for(i = 0; i < sbml_model->getNumEvents(); i++) {
		name = sbml_model->getEvent(i)->getName();
		validateEventName(name);
		num_rulesets++;
	}

	// create compartments
	if(num_compartments) {
		if(!(compartment = (Simulation::Compartment **) calloc( \
					num_compartments, sizeof(Simulation::Compartment *)))) {
			cerr << "error: couldn't allocate memory for compartments\n";
			exit(EXIT_FAILURE);
		}
		ind = xdim = ydim = zdim = 0;
		for(i = 0; i < sbml_model->getNumCompartments(); i++) {
			// ignore default compartment
			if(sbml_model->getCompartment(i)->getId() == "default")
				continue;

			// parse compartment
			name = sbml_model->getCompartment(i)->getName();
			// if on lattice then add to list of compartments
			if(getCompartmentPositionFromSBML(name, pos)) {
				id = sbml_model->getCompartment(i)->getId();
				compartment[ind] = new Simulation::Compartment(id);
				compartment[ind]->setName(name);
				compartment[ind]->setPosition( \
						getCompartmentPositionFromSBML(name, pos));
				compartment[ind]->initialiseLevels(alphabet->size());
				// determine size of lattice
				if(compartment[ind]->getXPosition() > xdim)
					xdim = compartment[ind]->getXPosition();
				if(compartment[ind]->getYPosition() > ydim)
					ydim = compartment[ind]->getYPosition();
				// set compartment volume
				if(sbml_model->getCompartment(i)->isSetVolume()) {
					compartment[ind]->setVolume(volume_scale * \
							sbml_model->getCompartment(i)->getVolume());
					compartment[ind]->setInitialVolume(volume_scale * \
							sbml_model-> getCompartment(i)->getVolume());
				}
				compartment[ind]->setVolumeConstant(sbml_model-> \
						getCompartment(i)->getConstant());
				if(compartment[ind]->isVolumeConstant() && \
						compartment[ind]->getVolume() != 1.0)
					compartment[ind]->setVolumeUnit(false);
				// handle variable volumes
				if(!compartment[ind]->isVolumeConstant()) {
					compartment[ind]->setVolumeUnit(false);
					compartment[ind]->setGrowthType(growth_type);
					if(growth_type == LINEAR_GROWTH || \
							growth_type == EXPONENTIAL_GROWTH) {
						// linear or exponential growth and division
						compartment[ind]->setGenerationTime( \
								getGenerationTimeFromSBML(name));
						compartment[ind]->setDivisionVolume(2.0 * \
								compartment[ind]->getInitialVolume());
						if(growth_type == LINEAR_GROWTH)
							compartment[ind]->setVolumeGradient((double) \
									compartment[ind]->getInitialVolume() / \
									compartment[ind]->getGenerationTime());
						else
							compartment[ind]->setVolumeGradient((double) \
									log(2.0) / \
									compartment[ind]->getGenerationTime());
					} else {
						// function based growth and division
						compartment[ind]->initialiseGrowthDivision(name, \
								alphabet);
					}
				}
				ind++;
			}
		}
		initial_xdim = ++xdim;
		initial_ydim = ++ydim;
	}

	// validate rule name
	for(i = 0; i < sbml_model->getNumReactions(); i++) {
		name = sbml_model->getReaction(i)->getName();
		validateRuleName(name);
		validateRuleStructure(sbml_model->getReaction(i));
	}

	// initialise rulesets
	if(!(ruleset = (RuleSet **) calloc(num_rulesets, sizeof(RuleSet *)))) {
		cerr << "error: couldn't allocate memory for rulesets\n";
		exit(EXIT_FAILURE);
	}
	ListOfSpeciesReferences *reactants, *products;
	SpeciesReference *sr;
	ind = 0;
	for(i = 0; i < sbml_model->getNumCompartments(); i++) {
		// ignore default compartment
		if(sbml_model->getCompartment(i)->getId() == "default")
			continue;

		name = sbml_model->getCompartment(i)->getName();
		if(isTemplateCompartment(name)) {  // template compartment
			// count number of rules belonging to this compartment
			count = 0;
			id = sbml_model->getCompartment(i)->getId();
			for(j = 0; j < sbml_model->getNumReactions(); j++) {
				reactants = sbml_model->getReaction(j)->getListOfReactants();
				for(k = 0; k < reactants->size(); k++) {
					sr = (SpeciesReference *) reactants->get(k);
					// ignore rules with reaction constants set to zero
					if(sbml_model->getSpecies(sr->getSpecies())-> \
							getCompartment() == id) {
						count++;
						break;
					}
				}
			}
			ruleset[ind] = new Simulation::RuleSet(count);
			ruleset[ind]->setId(id);
 			ruleset[ind]->setIndex(getSBMLIndex(name));
			ruleset[ind]->setName(name);
			if(!count) {
				if(!sbml_model->getCompartment(i)->getConstant()) {
					// set as growth ruleset
					ruleset[ind]->setGrowthRuleSet(true);
					if(sbml_model->getCompartment(i)->isSetVolume())
						ruleset[ind]->setVolume(volume_scale * \
								sbml_model->getCompartment(i)->getVolume());
				} else
					// set as species density ruleset
					ruleset[ind]->setDensityRuleSet(true);
			}
			// set as internal compartment if required
			if(!getParentCompartmentFromSBML(name).empty()) {
				ruleset[ind]->setParentCompartmentName( \
						getParentCompartmentFromSBML(name));
				ruleset[ind]->setEnclosedCompartment(true);
			}
			ind++;
		}
	}

	// add events to rulesets
	for(i = 0; i < sbml_model->getNumEvents(); i++) {
		ev = sbml_model->getEvent(i);
		ruleset[ind] = new Simulation::RuleSet(0);
		ruleset[ind]->setEventRuleSet(true);
		ruleset[ind]->setId(ev->getId());
		ruleset[ind]->setName(ev->getName());
 		ruleset[ind]->setIndex(getEventIndexFromSBML(ev->getName()));
		if(!ev->isSetTrigger()) {
			cerr << "error: trigger not set for event " << ev->getName() << \
					"\n";
			exit(EXIT_FAILURE);
		}
		if(ev->isSetDelay()) {
			cerr << "error: event delays not supported (event " << \
					ev->getName() << ")\n";
			exit(EXIT_FAILURE);
		}
		ruleset[ind]->setEventTrigger(SBML_formulaToString(ev->getTrigger()-> \
				getMath()));
		if(ev->getNumEventAssignments() != 1) {
			cerr << "error: only one assignment per event supported (event " \
					<< ev->getName() << ")\n";
			exit(EXIT_FAILURE);
		}
		ruleset[ind]->setEventAssignmentVariable( \
				ev->getEventAssignment(0)->getVariable());
		ruleset[ind]->setEventAssignmentExpression( \
				SBML_formulaToString(ev->getEventAssignment(0)->getMath()));
		ind++;
	}

	// populate ruleset with rules
	for(i = 0; i < num_rulesets; i++) {
		// set child templates
		name = ruleset[i]->getName();
		if(getNumCompartmentTemplatesFromSBML(name)) {
			ruleset[i]->setNumChildRuleSets(\
					getNumCompartmentTemplatesFromSBML(name));
			getCompartmentTemplatesFromSBML(name, \
					ruleset[i]->getChildRuleSets());
			// adjust template ids to indices
			for(j = 0; j < ruleset[i]->getNumChildRuleSets(); j++) {
				for(k = 0; k < num_rulesets; k++)
					if(getSBMLIndex(ruleset[k]->getName()) == \
								ruleset[i]->getChildRuleSet(j))
						break;
				if(k == num_rulesets) {
					cerr << "error: unknown template " << \
						ruleset[i]->getChildRuleSet(j) << \
						" in ruleset " << name << "\n";
					exit(EXIT_FAILURE);
				}
				ruleset[i]->setChildRuleSet(j, k);
			}
		}

		// add rules to ruleset
		rule_index = 0;
		for(j = 0; j < sbml_model->getNumReactions(); j++) {
			num_reactants = sbml_model->getReaction(j)->getNumReactants();
			num_products = sbml_model->getReaction(j)->getNumProducts();

			reactants = sbml_model->getReaction(j)->getListOfReactants();
			if(num_products)
				products = sbml_model->getReaction(j)->getListOfProducts();
			else
				products = NULL;
			kinetic_law = sbml_model->getReaction(j)->getKineticLaw();

			// validate reactants
			if(num_reactants < 1 || num_reactants > 2) {
				cerr << sbml_model->getReaction(j)->getName() << \
						": invalid number of reactants\n";
				exit(EXIT_FAILURE);
			}
			sr = (SpeciesReference *) reactants->get(0);

			// continue if reaction not related to current compartment
			// or reaction constant zero + reaction constant constant
			if(sbml_model->getSpecies(sr->getSpecies())-> \
					getCompartment() != ruleset[i]->getId())
				continue;

			// validate products
			if(num_products < 0 || num_products > 3) {
				cerr << sbml_model->getReaction(j)->getName() << \
						": invalid number of products\n";
				exit(EXIT_FAILURE);
			}

			// initialise rule
			ruleset[i]->setRule(rule_index, new Rule \
					(sbml_model->getReaction(j)->getId()));
			ruleset[i]->getRule(rule_index)-> \
					setName(sbml_model->getReaction(j)->getName());
			ruleset[i]->getRule(rule_index)->setNumReactants(num_reactants);
			ruleset[i]->getRule(rule_index)->setNumProducts(num_products);

			// get constants
			switch(kinetic_law->getNumParameters())
			{
				case 1:  // non-diffusion rule
					parameter = kinetic_law->getParameter(0);
					// the stochastic reaction constant
					if(parameter->getConstant()) {
						// a constant constant!
						ruleset[i]->getRule(rule_index)-> \
								setReactionConstant(parameter->getValue());
					} else {
						// a distribution-based constant
						validateReactionDistribution(parameter->getName());
						ruleset[i]->getRule(rule_index)->setDistribution( \
								parameter->getName());
					}
					ruleset[i]->getRule(rule_index)->setReactionConstantName( \
							parameter->getName());
					break;

				case 3:  // diffusion rule
				case 4:
					if(num_products != 1)
						continue;  // only allow one product with diffusion
					// get diffusion constant
					parameter = kinetic_law->getParameter("D");
					if(!parameter) {
						cerr << sbml_model->getReaction(j)-> \
								getName() << \
								": diffusion parameter D not found\n";
						exit(EXIT_FAILURE);
					}
					ruleset[i]->getRule(rule_index)-> \
							setDiffusionConstant(parameter-> \
							getValue());
					// get subvolume length
					parameter = kinetic_law->getParameter("h");
					if(!parameter) {
						cerr << sbml_model->getReaction(j)-> \
							getName() << ": diffusion parameter h not found\n";
						exit(EXIT_FAILURE);
					}
					ruleset[i]->getRule(rule_index)-> \
							setDiffusionLength(parameter->getValue());
					// get number of subvolumes
					parameter = kinetic_law->getParameter("N");
					if(!parameter) {
						cerr << sbml_model->getReaction(j)-> \
							getName() << ": diffusion parameter N not found\n";
						exit(EXIT_FAILURE);
					}
					ruleset[i]->getRule(rule_index)-> \
							setDiffusionNumber((unsigned int) \
							parameter->getValue());
					// get diffusion start time
					parameter = sbml_model->getReaction(j)-> \
							getKineticLaw()->getParameter("t");
					if(parameter) {
						ruleset[i]->getRule(rule_index)-> \
								setDiffusionStart(parameter->getValue());
					}
					// set rule as diffusion rule
					ruleset[i]->getRule(rule_index)->setDiffusion();
					break;

				default:
					cerr << sbml_model->getReaction(j)-> \
							getName() << ": unknown reaction parameters\n";
					exit(EXIT_FAILURE);
					break;
			}

			// set reactant indices
			for(k = 0; k < num_reactants; k++) {
				sr = (SpeciesReference *) reactants->get(k);
				ruleset[i]->getRule(rule_index)-> \
						setReactantIndex(k, \
						getSpeciesIndex(sbml_model-> \
						getSpecies(sr->getSpecies())-> \
						getName()));
			}

			// set product indices
			if(!num_products) {
				// degradation rule
				ruleset[i]->getRule(rule_index)->setNumProducts(0);
				ruleset[i]->getRule(rule_index)->setXVector(0);
				ruleset[i]->getRule(rule_index)->setYVector(0);
				ruleset[i]->getRule(rule_index)->setZVector(0);
				rule_index++;
				continue;
			}

			sr = (SpeciesReference *) products->get(0);
			if(num_products == 1 && !keep_degraded && \
					isDegradation(sbml_model->getSpecies( \
					sr->getSpecies())->getName())) {
				// degradation rule
				ruleset[i]->getRule(rule_index)->setNumProducts(0);
			} else {
				// non-degradation rule
				for(k = 0; k < num_products; k++) {
					sr = (SpeciesReference *) products->get(k);
					ruleset[i]->getRule(rule_index)-> \
							setProductIndex(k, \
							getSpeciesIndex(sbml_model-> \
							getSpecies(sr->getSpecies())-> \
							getName()));
				}
			}

			// set rule target
			sr = (SpeciesReference *) products->get(0);
			if(num_products == 1 && isDegradation(sbml_model->getSpecies( \
						sr->getSpecies())->getName())) {
					// degradation rule
				ruleset[i]->getRule(rule_index)->setXVector(0);
				ruleset[i]->getRule(rule_index)->setYVector(0);
				ruleset[i]->getRule(rule_index)->setZVector(0);
			} else {
				if(sbml_model->getReaction(j)-> \
						getName().find(":") == string::npos) {
					// no target specified, so transformation rule
					ruleset[i]->getRule(rule_index)-> \
							setXVector(0);
					ruleset[i]->getRule(rule_index)-> \
							setYVector(0);
				} else {
					// target specified with rule
					long int target[2];
					if(getRuleVectorFromSBML(sbml_model->getReaction(j)-> \
							getName(), target)) {
						// vector specified with rule
						// set target vector
						ruleset[i]->getRule(rule_index)->setXVector(target[0]);
						if(target[0])
							ruleset[i]->getRule(rule_index)->setChannel();
						ruleset[i]->getRule(rule_index)->setYVector(target[1]);
						if(target[1])
							ruleset[i]->getRule(rule_index)->setChannel();
					} else {
						// compartment name specified
						// set target name
						ruleset[i]->getRule(rule_index)-> \
								setTargetName(getRuleTargetFromSBML( \
								sbml_model->getReaction(j)->getName()));
						ruleset[i]->getRule(rule_index)-> \
								setTargetType(getRuleTargetTypeFromSBML( \
								sbml_model->getReaction(j)->getName()));
						ruleset[i]->getRule(rule_index)->setChannel();
					}
				}
				ruleset[i]->getRule(rule_index)->setZVector(0);
			}
			rule_index++;
		}
	}

	// set initial levels of species stored by rules
	for(i = 0; i < num_rulesets; i++) {
		name1 = ruleset[i]->getName();
		for(j = 0; j < sbml_model->getNumSpecies(); j++) {
			name2 = sbml_model->getCompartment(sbml_model-> \
					getSpecies(j)->getCompartment())->getName();
			if(name1 == name2 && !isDegradation( \
					sbml_model->getSpecies(j)->getName())) {
				ind3 = getSpeciesIndex(sbml_model->getSpecies(j)->getName());
				if((ind2 = (unsigned long) sbml_model->getSpecies(j)-> \
						getInitialAmount()) != 0) {
					for(k = 0; k < ruleset[i]->getNumRules(); k++) {
						for(ind = 0; ind < ruleset[i]->getNumReactants(k); \
									ind++)
							if(ruleset[i]->getReactantIndex(ind, k) == ind3)
								ruleset[i]->getRule(k)->setInitialReactant(ind, ind2);
						for(ind = 0; ind < ruleset[i]->getNumProducts(k); ind++)
							if(ruleset[i]->getProductIndex(ind, k) == ind3)
								ruleset[i]->getRule(k)->setInitialProduct(ind, ind2);
					}
				}
			}
		}
	}

	// add rulesets to compartments
	for(i = 0; i < num_compartments; i++) {
		// get specified rulesets
		name = compartment[i]->getName();
		ind = getNumCompartmentTemplatesFromSBML(name);
		if(!ind)
			continue;  // no rulesets
		// allocate memory to temporarily store ruleset indices
		if(!(rstmp = (unsigned long *) calloc(ind, sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for temporary rulesets\n";
			exit(EXIT_FAILURE);
		}
		// get ruleset ids
		getCompartmentTemplatesFromSBML(name, rstmp);
		// add ruleset to compartment
		for(j = 0; j < ind; j++) {
			// find ruleset index
			for(k = 0; k < num_rulesets; k++)
				if(getSBMLIndex(ruleset[k]->getName()) == rstmp[j])
					break;
			if(k == num_rulesets) {
				cerr << "error: unknown template " << rstmp[j] << \
					" in compartment " << name << "\n";
				exit(EXIT_FAILURE);
			}
			// add ruleset
			addRuleSetToCompartment(k, i);
		}
		// free memory
		free(rstmp);
	}

	// set compartment information
	for(i = 0; i < num_compartments; i++) {
		// set growth and division functions if specified
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
			ind = compartment[i]->getRuleSetIndex(j);
			name1 = ruleset[ind]->getName();
			if(ruleset[ind]->isGrowthRuleSet()) {
				// set compartment volume if not set in compartment definition
				// and set in ruleset definition
				compartment[i]->setVolumeConstant(false);
				compartment[i]->setVolumeUnit(false);
				compartment[i]->setGrowthType(growth_type);
				if(ruleset[ind]->getVolume() != 1.0 && \
						compartment[i]->getInitialVolume() == 1.0) {
					compartment[i]->setVolume(ruleset[ind]->getVolume());
					compartment[i]->setInitialVolume(ruleset[ind]->getVolume());

				}
				if(growth_type == LINEAR_GROWTH || \
						growth_type == EXPONENTIAL_GROWTH) {
					// linear or exponential growth and division
					compartment[i]->setGenerationTime( \
							getGenerationTimeFromSBML(name1));
					compartment[i]->setDivisionVolume(2.0 * \
							compartment[i]->getInitialVolume());
					if(growth_type == LINEAR_GROWTH)
						compartment[i]->setVolumeGradient((double) \
								compartment[i]->getInitialVolume() / \
								compartment[i]->getGenerationTime());
					else
						compartment[i]->setVolumeGradient((double) \
								log(2.0) / compartment[i]->getGenerationTime());
				} else {
					// function based growth and division
					compartment[i]->initialiseGrowthDivision(name1, alphabet);
				}
			}
		}

		// set constant and distribution species
		for(j = 0; j < num_species; j++) {
			// set constant species
			if(alphabet->isConstant(j))
				compartment[i]->setLevelConstant(j);
			// treat distribution-based species as constant and nondividing
			if(alphabet->isDistribution(j))
				compartment[i]->setLevelConstant(j);
		}

		// create distribution-based species update list
		count = 0;
		for(ind = 0; ind < num_species; ind++) {
			if(!alphabet->isDistribution(ind))
				continue;
			for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
				for(rule_index = 0; rule_index < compartment[i]-> \
						getNumRules(j); rule_index++)
					for(k = 0; k < compartment[i]->getNumReactants(j, \
							rule_index); k++)
						if(compartment[i]->getReactantIndex(j, k, rule_index) \
								== ind) {
							count++;
							// break to top loop
							rule_index = compartment[i]->getNumRules(j);
							j = compartment[i]->getNumRuleSets();
							break;
						}
			}
		}
		if(!count)
			continue;
		compartment[i]->initialiseDistributionSpecies(count);
		count = 0;
		for(ind = 0; ind < num_species; ind++) {
			if(!alphabet->isDistribution(ind))
				continue;
			for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
				for(rule_index = 0; rule_index < compartment[i]-> \
						getNumRules(j); rule_index++)
					for(k = 0; k < compartment[i]->getNumReactants(j, \
							rule_index); k++)
						if(compartment[i]->getReactantIndex(j, k, rule_index) \
								== ind) {
							compartment[i]->setDistributionSpecies(count++, \
									ind);
							// break to top loop
							rule_index = compartment[i]->getNumRules(j);
							j = compartment[i]->getNumRuleSets();
							break;
						}
		}
	}

	// set targets for translocation rules
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
			ind = compartment[i]->getNumRules(j);
			ind3 = compartment[i]->getRuleSetIndex(j);
			for(k = 0; k < ind; k++) {
				rule = compartment[i]->getRule(j, k);
				if(rule->isChannel() && !rule->getTargetName().empty()) {
					rule_index = k;
					// rule specified by target compartment name
					if(rule->getTargetName() == "*") {
						if(compartment[i]->isEnclosedCompartment()) {
							// enclosed compartment
							ind2 = compartment[i]->getParentCompartmentIndex();
							setRuleTarget(i, ind3, rule_index, ind2);
						} else {
							// compartment on grid
							// valid for all neighbouring compartments
							// set target to left neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							if(periodic_x && (pos[0] == 0))
								pos[0] = xdim - 1;
							else
								pos[0]--;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								if(rule_index < ind) {
									setRuleTarget(i, ind3, rule_index, ind2);
									rule_index = ind;
								} else {
									rule_index = cloneRule(i, ind3, k);
									setRuleTarget(i, ind3, rule_index, ind2);
								}
							}
							// set target to right neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							pos[0]++;
							if(periodic_x && (pos[0] == xdim))
								pos[0] = 0;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								if(rule_index < ind) {
									setRuleTarget(i, ind3, rule_index, ind2);
									rule_index = ind;
								} else {
									rule_index = cloneRule(i, ind3, k);
									setRuleTarget(i, ind3, rule_index, ind2);
								}
							}
							// set target to top neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							if(periodic_y && (pos[1] == 0))
								pos[1] = ydim - 1;
							else
								pos[1]--;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								if(rule_index < ind) {
									setRuleTarget(i, ind3, rule_index, ind2);
									rule_index = ind;
								} else {
									rule_index = cloneRule(i, ind3, k);
									setRuleTarget(i, ind3, rule_index, ind2);
								}
							}
							// set target to bottom neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							pos[1]++;
							if(periodic_y && (pos[1] == ydim))
								pos[1] = 0;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								if(rule_index < ind) {
									setRuleTarget(i, ind3, rule_index, ind2);
									rule_index = ind;
								} else {
									rule_index = cloneRule(i, ind3, k);
									setRuleTarget(i, ind3, rule_index, ind2);
								}
							}
							if(neighbourhood == 8) {							
								// set target to top left neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								if(periodic_x && (pos[0] == 0))
									pos[0] = xdim - 1;
								else
									pos[0]--;
								if(periodic_y && (pos[1] == 0))
									pos[1] = ydim - 1;
								else
									pos[1]--;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to top right neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								pos[0]++;
								if(periodic_x && (pos[0] == xdim))
									pos[0] = 0;
								if(periodic_y && (pos[1] == 0))
									pos[1] = ydim - 1;
								else
									pos[1]--;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to bottom left neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								if(periodic_x && (pos[0] == 0))
									pos[0] = xdim - 1;
								else
									pos[0]--;
								pos[1]++;
								if(periodic_y && (pos[1] == ydim))
									pos[1] = 0;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to bottom right neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								pos[0]++;
								if(periodic_x && (pos[0] == xdim))
									pos[0] = 0;
								pos[1]++;
								if(periodic_y && (pos[1] == ydim))
									pos[1] = 0;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
							}
							// set target to enclosed compartments if exist
							for(ind2 = 0; ind2 < compartment[i]->\
										getNumChildCompartments(); ind2++) {
								ind4 = compartment[i]->\
										getChildCompartmentIndex(ind2);
								if(rule_index < ind) {
									setRuleTarget(i, ind3, rule_index, ind4);
									rule_index = ind;
								} else {
									rule_index = cloneRule(i, ind3, k);
									setRuleTarget(i, ind3, rule_index, ind4);
								}
							}
							// remove if no target exists
							if(rule_index < ind)
								compartment[i]->setReactionConstant(j, k, 0.0);
						}
					} else {
						// valid for specific target
						if(compartment[i]->isEnclosedCompartment()) {
							// enclosed compartment
							ind2 = compartment[i]->getParentCompartmentIndex();
							for(ind4 = 0; ind4 < compartment[ind2]-> \
										getNumRuleSets(); ind4++) {
								name = compartment[ind2]->getRuleSet(ind4)->\
										getName();
								name1 = name.substr(0, name.find(':'));
								if(name1 == rule->getTargetName())
									break;
							}
							if(ind4 == compartment[ind2]->getNumRuleSets())
								compartment[i]->setReactionConstant(j, k, 0.0);
							else
								setRuleTarget(i, ind3, rule_index, ind2);
						} else {
							// set target to left neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							if(periodic_x && (pos[0] == 0))
								pos[0] = xdim - 1;
							else
								pos[0]--;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								name = compartment[ind2]->getName();
								name1 = name.substr(0, name.find(':'));
								if(name1 == rule->getTargetName()) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
								}
							}
							// set target to right neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							pos[0]++;
							if(periodic_x && (pos[0] == xdim))
								pos[0] = 0;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								name = compartment[ind2]->getName();
								name1 = name.substr(0, name.find(':'));
								if(name1 == rule->getTargetName()) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
								}
							}
							// set target to top neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							if(periodic_y && (pos[1] == 0))
								pos[1] = ydim - 1;
							else
								pos[1]--;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								name = compartment[ind2]->getName();
								name1 = name.substr(0, name.find(':'));
								if(name1 == rule->getTargetName()) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
								}
							}
							// set target to bottom neighbour if exists
							compartment[i]->getPosition(pos);
							// adjust if boundary periodic
							pos[1]++;
							if(periodic_y && (pos[1] == ydim))
								pos[1] = 0;
							ind2 = getCompartmentIndex(pos);
							if(ind2 != num_compartments) {
								name = compartment[ind2]->getName();
								name1 = name.substr(0, name.find(':'));
								if(name1 == rule->getTargetName()) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
								}
							}
							if(neighbourhood == 8) {							
								// set target to top left neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								if(periodic_x && (pos[0] == 0))
									pos[0] = xdim - 1;
								else
									pos[0]--;
								if(periodic_y && (pos[1] == 0))
									pos[1] = ydim - 1;
								else
									pos[1]--;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to top right neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								pos[0]++;
								if(periodic_x && (pos[0] == xdim))
									pos[0] = 0;
								if(periodic_y && (pos[1] == 0))
									pos[1] = ydim - 1;
								else
									pos[1]--;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to bottom left neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								if(periodic_x && (pos[0] == 0))
									pos[0] = xdim - 1;
								else
									pos[0]--;
								pos[1]++;
								if(periodic_y && (pos[1] == ydim))
									pos[1] = 0;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
								// set target to bottom right neighbour if exists
								compartment[i]->getPosition(pos);
								// adjust if boundary periodic
								pos[0]++;
								if(periodic_x && (pos[0] == xdim))
									pos[0] = 0;
								pos[1]++;
								if(periodic_y && (pos[1] == ydim))
									pos[1] = 0;
								ind2 = getCompartmentIndex(pos);
								if(ind2 != num_compartments) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, ind2);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, ind2);
									}
									// scale reaction constant
									setReactionConstant(i, ind3, rule_index, \
											0.5 * getReactionConstant(i, ind3, \
											rule_index));
								}
							}							
							// set target to enclosed compartments if exist
							for(ind2 = 0; ind2 < compartment[i]->\
										getNumChildCompartments(); ind2++) {
								ind4 = compartment[i]->\
										getChildCompartmentIndex(ind2);
								name1 = compartment[ind4]->getName();
								if(name1 == rule->getTargetName()) {
									if(rule_index < ind) {
										setRuleTarget(i, ind3, rule_index, \
												ind4);
										rule_index = ind;
									} else {
										rule_index = cloneRule(i, ind3, k);
										setRuleTarget(i, ind3, rule_index, \
												ind4);
									}
								}
							}
							// remove if no target exists
							if(rule_index < ind)
								compartment[i]->setReactionConstant(j, k, 0.0);
						}
					}
				}
			}
		}

	// set initial amounts for species
	for(i = 0; i < sbml_model->getNumSpecies(); i++) {
		if(!isDegradation(sbml_model->getSpecies(i)->getName())) {
			name = sbml_model->getCompartment(sbml_model->getSpecies(i)-> \
					getCompartment())->getName();
			if(getCompartmentPositionFromSBML(name, pos)) {
				j = getCompartmentIndexFromId(sbml_model->getSpecies(i)-> \
						getCompartment());
				if(sbml_model->getSpecies(i)->isSetInitialAmount()) {
					compartment[j]->setLevel( \
							getSpeciesIndex(sbml_model->getSpecies(i)-> \
							getName()), (unsigned int) \
							sbml_model->getSpecies(i)->getInitialAmount());
				} else {
					if(sbml_model->getSpecies(i)->isSetInitialConcentration()) {
						compartment[j]->setLevel( \
								getSpeciesIndex(sbml_model->getSpecies(i)-> \
								getName()), (unsigned int) \
								sbml_model->getSpecies(i)->getInitialConcentration());
					} else {
						cerr << "error: initial amount or concentration not " \
								"set for species " << \
								sbml_model->getSpecies(i)->getName() << "\n";
						exit(EXIT_FAILURE);
					}
				}
				if(sbml_model->getSpecies(i)->getConstant())
					compartment[j]->setLevelConstant( \
							getSpeciesIndex(sbml_model->getSpecies(i)-> \
							getName()));
			}
		}
	}

	// if required, duplicate initial amounts of species
	if(duplicate_initial_amounts)
		for(i = 0; i < num_compartments; i++)
			for(k = 0; k < compartment[i]->getNumRuleSets(); k++) {
				// get index of compartment which defines template
				name1 = compartment[i]->getRuleSet(k)->getName();
				for(j = 0; j < sbml_model->getNumSpecies(); j++) {
					name2 = sbml_model->getCompartment(sbml_model-> \
							getSpecies(j)->getCompartment())->getName();
					if(name1 == name2 && !isDegradation( \
							sbml_model->getSpecies(j)->getName())) {
						if(sbml_model->getSpecies(j)->getInitialAmount() \
								!= 0.0) {
							compartment[i]->setLevel( \
								getSpeciesIndex(sbml_model->getSpecies(j)-> \
								getName()), (unsigned int) \
								sbml_model->getSpecies(j)->getInitialAmount());
							if(sbml_model->getSpecies(j)->getConstant())
								compartment[i]->setLevelConstant( \
									getSpeciesIndex(sbml_model->getSpecies(j) \
											->getName()));
						}
					}
				}
			}

	// adjust lattice size if specified
	if((parameter = sbml_model->getParameter("lx")))
		xdim = (unsigned long) parameter->getValue();
	if((parameter = sbml_model->getParameter("ly")))
		ydim = (unsigned long) parameter->getValue();
	if((parameter = sbml_model->getParameter("lz")))
		zdim = (unsigned long) parameter->getValue();
}

// create + populate cellular potts lattice
void Psystem::createLattice(SimulationParameters &parameters)
{
	unsigned long i, j, k, c, n, nn, ind, ind2, rule_index;
	LatticeSite *site;
	Simulation::Compartment *comp;
	Rule *rule;
	string name, name2;

	// load lattice from file
	lattice = new Lattice(parameters.lattice_file, \
			rng->randomNonnegativeInteger());

	// set dimensions in p system
	xdim = lattice->getXDim();
	ydim = lattice->getYDim();
	zdim = lattice->getZDim();

	// populate p system with compartments based on lattice
	for(c = 1; c <= lattice->getMaxSpin(); c++) {
		// create compartment
		name.empty();
		addCompartment(name);
		comp = compartment[c - 1];

		// setup list of neighbours
		std::list<unsigned long> neighs = std::list<unsigned long> ();

		// adjust compartment volumes
		comp->setVolume(0.0);
		bool first = true;
		for(i = 0; i < xdim; i++)
			for(j = 0; j < ydim; j++)
				for(k = 0; k < zdim; k++) {
					if(!(site = lattice->getSite(i, j, k)) || site->spin != c)
						continue;

					// add ruleset
					if(first) {
						string rsname = site->name;
						comp->setName(rsname);
						addRuleSetToCompartment(getRuleSetIndex(rsname), c - 1);
						first = false;
					}

					// adjust volume
					comp->increaseVolume(lattice->getSiteVolume());

					// add neighbours to neighbour list
					nn = lattice->calculateAlienNeighbours(site);
					if(nn)
						for(n = 0; n < nn; n++)
							if(lattice->getAlienNeighbourSpin(n))
								neighs.push_front(lattice->\
										getAlienNeighbourSpin(n) - 1);
				}

		// set list of neighbours
		neighs.sort();
		neighs.unique();
		comp->initialiseNeighbours(neighs.size());
		list<unsigned long>::iterator it = neighs.begin();
		for(i = 0; i < neighs.size(); i++)
			comp->setNeighbour(i, *it++);
	}

	// set targets for translocation rules
	for(i = 0; i < num_compartments; i++) {
		comp = compartment[i];
		for(j = 0; j < comp->getNumRuleSets(); j++) {
			ind = comp->getNumRules(j);
			ind2 = comp->getRuleSetIndex(j);
			for(k = 0; k < ind; k++) {
				rule = comp->getRule(j, k);
				if(rule->isChannel() && !rule->getTargetName().empty()) {
					rule_index = k;
					// rule specified by target compartment name
					if(rule->getTargetName() == "*") {
						// valid for all neighbouring compartments
						// set target to all neighbours
						nn = comp->getNumNeighbours();
						for(n = 0; n < nn; n++)
							if(rule_index < ind) {
								setRuleTarget(i, ind2, rule_index, \
										comp->getNeighbour(n));
								rule_index = ind;
							} else {
								rule_index = cloneRule(i, ind2, k);
								setRuleTarget(i, ind2, rule_index, \
										comp->getNeighbour(n));
							}
						// remove if no target exists
						if(rule_index < ind)
							comp->setReactionConstant(j, k, 0.0);
					} else {
						// valid for specific target
						// set target to specific neighbours
						nn = comp->getNumNeighbours();
						for(n = 0; n < nn; n++) {
							name = compartment[comp->getNeighbour(n)] \
									->getName();
							name2 = name.substr(0, name.find(':'));
							if(name2 != rule->getTargetName())
								continue;
							if(rule_index < ind) {
								setRuleTarget(i, ind2, rule_index, \
										comp->getNeighbour(n));
								rule_index = ind;
							} else {
								rule_index = cloneRule(i, ind2, k);
								setRuleTarget(i, ind2, rule_index, \
										comp->getNeighbour(n));
							}
						}
						// remove if no target exists
						if(rule_index < ind)
							comp->setReactionConstant(j, k, 0.0);
					}
				}
			}
		}
	}

	// set initial levels given in templates modules
	setInitialLevels();
}

// save list of initial compartments
void Psystem::saveInitialCompartments(void)
{
	register unsigned int i;

	// create initial compartment storage
	num_initial_compartments = num_compartments;
	if(!(initial_compartment = (Simulation::Compartment **) \
			calloc(num_compartments, sizeof(Simulation::Compartment *)))) {
		cerr << "error: couldn't allocate memory for initial compartments\n";
		exit(EXIT_FAILURE);
	}

	// copy list of compartments
	for(i = 0; i < num_compartments; i++)
		initial_compartment[i] = new Simulation::Compartment(*compartment[i]);
}

// reset compartments to initial state
void Psystem::resetCompartments(void)
{
	register unsigned int i;

	// remove previous list of compartments
	for(i = 0; i < num_compartments; i++)
		delete compartment[i];
	free(compartment);

	// create new compartment storage
	num_compartments = num_initial_compartments;
	simalg->setNumCompartments(num_compartments);
	if(!(compartment = (Simulation::Compartment **) \
			calloc(num_compartments, sizeof(Simulation::Compartment *)))) {
		cerr << "error: couldn't allocate memory for compartments\n";
		exit(EXIT_FAILURE);
	}

	// copy list of compartments
	for(i = 0; i < num_compartments; i++)
		compartment[i] = new Simulation::Compartment(*initial_compartment[i]);
}

// return index of compartment given id
unsigned long Psystem::getCompartmentIndexFromId(std::string s)
{
	register unsigned long i;

	for(i = 0; i < num_compartments; i++)
		if(compartment[i]->getId() == s)
			break;

	return i;
}

// return index of compartment given name
unsigned long Psystem::getCompartmentIndexFromName(std::string s)
{
	register unsigned long i;

	for(i = 0; i < num_compartments; i++)
		if(compartment[i]->getName() == s)
			break;

	return i;
}

// return index of compartment given position
unsigned long Psystem::getCompartmentIndex(unsigned long *pos)
{
	register unsigned long i, x, y, z;

	for(i = 0; i < num_compartments; i++) {
		if(compartment[i]->isEnclosedCompartment())
			continue;
		x = compartment[i]->getXPosition();
		y = compartment[i]->getYPosition();
		z = compartment[i]->getZPosition();
		if(pos[0] == x && pos[1] == y && pos[2] == z)
			break;
	}

	return i;
}

void Psystem::show(void)
{
	register unsigned int i;

	cout << "id=" << id << "\n";

	if(compartment) {
		cout << "\ncompartments=" << num_compartments << "\n\n";
		for(i = 0; i < num_compartments; i++) {
			cout << "compartment " << i << ": ";
			compartment[i]->show();
		}
		cout << "\n";
	}
	if(alphabet) {
		cout << "\nalphabet=" << getAlphabet()->size() << ":";
		for(i = 0; i < getAlphabet()->size(); i++)
			cout << " " << getAlphabet()->getName(i);
		cout << "\n";
	}
	if(ruleset) {
		cout << "\nrule templates=" << num_rulesets << "\n\n";
		for(i = 0; i < num_rulesets; i++) {
			cout << "ruleset " << i << ": ";
			ruleset[i]->show();
		}
		cout << "\n";
	}
	if(simalg) {
		cout << "\nsimulation algorithm: name=\"" << \
				getSimulationAlgorithm()->getName() << "\"\n";
	}
}

// calculate target compartment of rule
unsigned long Psystem::calculateRuleTarget(unsigned long rs, unsigned long r, \
		unsigned long c)
{
	unsigned long pos[3];
	long int off[3];

	// get compartment position
	pos[0] = compartment[c]->getXPosition();
	pos[1] = compartment[c]->getYPosition();
	pos[2] = compartment[c]->getZPosition();
	// get rule vector
	off[0] = compartment[c]->getRuleXVector(rs, r);
	off[1] = compartment[c]->getRuleYVector(rs, r);
	off[2] = compartment[c]->getRuleZVector(rs, r);
	// add vector to compartment position
	off[0] += pos[0];
	off[1] += pos[1];
	off[2] += pos[2];
	// allow for periodic boundary conditions
	if(periodic_x) {
		if(off[0] < 0)
			off[0] += xdim;
		else
			if(off[0] >= (long int) xdim)
				off[0] -= xdim;
	}
	if(periodic_y) {
		if(off[1] < 0)
			off[1] += ydim;
		else
			if(off[1] >= (long int) ydim)
				off[1] -= ydim;
	}
	if(periodic_z) {
		if(off[2] < 0)
			off[2] += zdim;
		else
			if(off[2] >= (long int) zdim)
				off[2] -= zdim;
	}
	pos[0] = (unsigned long) off[0];
	pos[1] = (unsigned long) off[1];
	pos[2] = (unsigned long) off[2];

	return getCompartmentIndex(pos);
}

// execute a specified rule
void Psystem::executeRule(unsigned long rs, unsigned long r, unsigned long c)
{
	register unsigned long i, j, reactind, prodind, num_species;
	register Simulation::Compartment *comp1, *comp2;
	register double time;

	i = c;
	comp1 = compartment[i];
	time = simalg->getTime();

	// increment number of reactions
	simalg->incrementNumReactions();

	// handle reactants
	num_species = comp1->getNumReactants(rs, r);
	for(j = 0; j < num_species; j++) {
		reactind = comp1->getReactantIndex(rs, j, r);
		if(!comp1->isLevelConstant(reactind))
			comp1->decrementLevel(reactind);
		if(log_reactions)
			simalg->writeReaction(time, reactind, comp1->getLevel(reactind), \
					i, comp1->getRuleSetIndex(rs), comp1->getRuleIndex(rs, r));
	}

	// handle products
	num_species = comp1->getNumProducts(rs, r);
	if(!num_species)
		return;  // degradation rule - finished
	if(comp1->isChannel(rs, r)) {
		i = comp1->getRuleTarget(rs, r);
		comp2 = compartment[i];
	} else
		comp2 = comp1;
	for(j = 0; j < num_species; j++) {
		prodind = comp1->getProductIndex(rs, j, r);
		if(!comp2->isLevelConstant(prodind))
			comp2->incrementLevel(prodind);
		if(log_reactions)
			simalg->writeReaction(time, prodind, comp2->getLevel(prodind), i, \
					comp1->getRuleSetIndex(rs), comp1->getRuleIndex(rs, r));
	}
}

// initialise p-system simulation
int Psystem::preprocess(void) { return simalg->preprocess(); }

// perform one iteration of p-system
int Psystem::execute(void) { return simalg->execute(); }

// perform one iteration of specified p-system compartment
void Psystem::execute(unsigned long ind)
{
	register unsigned long i, j;
	register Compartment *comp;

	comp = compartment[ind];

	// set level of distribution-based species
	for(i = 0; i < comp->getNumDistributionSpecies(); i++) {
		j = comp->getDistributionSpecies(i);
		// adjust in relation to compartment volume
/*		comp->setLevel(j, alphabet->getLevel(j) + \
				((fmod(simalg->getTime(), gt) / gt)) * \
				alphabet->getDistributionMean(j));
*/		comp->setLevel(j, alphabet->getLevel(j));
	}

	// iterate compartment
	simalg->execute(ind);
}

// iterate p-system for specified time and runs
void Psystem::execute(const double t) { simalg->execute(t); }

// save model as psys xml file
void Psystem::outputXML(char *outfile)
{
	FILE *fp;
	unsigned long i, j, k;

	if(!(fp = fopen(outfile, "w"))) {
		cerr << "error: couldn't open output file " << outfile << "\n";
		return;
	}

	// output header
	fprintf(fp, "<?xml version=\"1.0\"?>\n");
	fprintf(fp, "<PSystemModel name=\"%s\">\n\n", getId().c_str());

	// output alphabet of elementary object
	fprintf(fp, "<alphabetOfElementaryObjects>\n");
	for(i = 0; i < alphabet->size(); i++)
		fprintf(fp, "\t<elementaryObject name=\"%s\"/>\n", \
				alphabet->getName(i).c_str());
	fprintf(fp, "</alphabetOfElementaryObjects>\n\n");

	// output list of compartment labels
	fprintf(fp, "<listOfCompartmentLabels>\n");
	for(i = 0; i < num_rulesets; i++)
		fprintf(fp, "\t<compartmentLabel label=\"%s\"/>\n", \
				getCompartmentNameFromSBML(ruleset[i]->getName()).c_str());
	fprintf(fp, "</listOfCompartmentLabels>\n\n");

	// output list of compartments
	fprintf(fp, "<listOfCompartments>\n");
	for(i = 0; i < num_compartments; i++) {
		// output compartment information
		fprintf(fp, "\t<compartment name=\"%s\" label=\"%s\" " \
				"father=\"\" position=\"%ld,%ld,%ld\">\n", \
				compartment[i]->getId().c_str(), \
				getCompartmentNameFromSBML(compartment[i]->getName()).c_str(), \
				compartment[i]->getXPosition(), \
				compartment[i]->getYPosition(), \
				compartment[i]->getZPosition() \
				);

		// output compartment neighbour information
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++)
				if(compartment[i]->isChannel(j, k)) {
					fprintf(fp, "\t\t<listOfCompartmentNeighbours>\n");
					break;
				}
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++)
				if(compartment[i]->isChannel(j, k))
					fprintf(fp, "\t\t\t<compartmentNeighbour name=\"%s\"/>\n", \
							compartment[compartment[i]->getRuleTarget(j, k)]-> \
							getId().c_str());
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++)
				if(compartment[i]->isChannel(j, k)) {
					fprintf(fp, "\t\t</listOfCompartmentNeighbours>\n");
					break;
				}
		fprintf(fp, "\t</compartment>\n");
	}
	fprintf(fp, "</listOfCompartments>\n\n");

	// output list of initial multisets
	fprintf(fp, "<listOfInitialMultisets>\n");
	for(i = 0; i < num_rulesets; i++) {
		fprintf(fp, "\t<initialMultiset label=\"%s\">\n", \
				getCompartmentNameFromSBML( \
				ruleset[i]->getName()).c_str());
		for(j = 0; j < alphabet->size(); j++)
			fprintf(fp, "\t\t<object name=\"%s\"" \
					" multiplicity=\"%ld\"/>\n", \
					alphabet->getName(j).c_str(), \
					compartment[getCompartmentIndexFromId( \
					ruleset[i]->getId())]->getLevel(j));
		fprintf(fp, "\t</initialMultiset>\n");
	}
	fprintf(fp, "</listOfInitialMultisets>\n\n");

	// output list of rulesets
	fprintf(fp, "<listOfRuleSets>\n");
	for(i = 0; i < num_rulesets; i++) {
		fprintf(fp, "\t<ruleSet label=\"%s\">\n", \
				getCompartmentNameFromSBML(ruleset[i]->getName()).c_str());
		for(j = 0; j < ruleset[i]->getNumRules(); j++) {
			if(ruleset[i]->isChannel(j)) {
				fprintf(fp, "\t\t<rule name=\"%s\" type=\"translocation\" " \
						"constant=\"%f\" vector=\"%ld,%ld,%ld\">\n", \
						ruleset[i]->getRule(j)->getId().c_str(), \
						ruleset[i]->getRule(j)->getReactionConstant(), \
						ruleset[i]->getRuleXVector(j), \
						ruleset[i]->getRuleYVector(j), \
						ruleset[i]->getRuleZVector(j));
			} else
				fprintf(fp, "\t\t<rule name=\"%s\" type=\"boundary\" " \
						"constant=\"%f\">\n", \
						ruleset[i]->getRule(j)->getId().c_str(), \
						ruleset[i]->getRule(j)->getReactionConstant());
			fprintf(fp, "\t\t\t<lhs>\n");
			if(ruleset[i]->isChannel(j))
				fprintf(fp, "\t\t\t\t<listOfTranslocatedObjects " \
						"label=\"%s\">\n", getCompartmentNameFromSBML( \
						ruleset[i]->getName()).c_str());
			else
				fprintf(fp, "\t\t\t\t<listOfInsideObjects " \
						"label=\"%s\">\n", getCompartmentNameFromSBML( \
						ruleset[i]->getName()).c_str());
			for(k = 0; k < ruleset[i]->getNumReactants(j); k++) {
				fprintf(fp, "\t\t\t\t\t<object name=\"%s\"/>\n", \
						alphabet->getName(ruleset[i]-> \
						getReactantIndex(k, j)).c_str());
			}
			if(ruleset[i]->isChannel(j))
				fprintf(fp, "\t\t\t\t</listOfTranslocatedObjects>\n");
			else
				fprintf(fp, "\t\t\t\t</listOfInsideObjects>\n");
			fprintf(fp, "\t\t\t</lhs>\n");
			fprintf(fp, "\t\t\t<rhs>\n");
			if(ruleset[i]->isChannel(j))
				fprintf(fp, "\t\t\t\t<listOfTranslocatedObjects " \
						"label=\"%s\">\n", getCompartmentNameFromSBML( \
						ruleset[i]->getName()).c_str());
			else
				fprintf(fp, "\t\t\t\t<listOfInsideObjects " \
						"label=\"%s\">\n", getCompartmentNameFromSBML( \
						ruleset[i]->getName()).c_str());
			for(k = 0; k < ruleset[i]->getNumProducts(j); k++)
				fprintf(fp, "\t\t\t\t\t<object name=\"%s\"/>\n", \
						alphabet->getName(ruleset[i]-> \
						getProductIndex(k, j)).c_str());
			if(ruleset[i]->isChannel(j))
				fprintf(fp, "\t\t\t\t</listOfTranslocatedObjects>\n");
			else
				fprintf(fp, "\t\t\t\t</listOfInsideObjects>\n");
			fprintf(fp, "\t\t\t</rhs>\n");
			fprintf(fp, "\t\t</rule>\n");
		}
		fprintf(fp, "\t</ruleSet>\n");
	}
	fprintf(fp, "</listOfRuleSets>\n");

	// output trailer
	fprintf(fp, "\n</PSystemModel>\n");

	fclose(fp);
}

// get levels of species in compartment
unsigned long Psystem::getSpeciesLevel(unsigned long s, unsigned long c)
{
	if(c < 0 || c >= num_compartments || s < 0 || s >= num_species) {
		cerr << "warning: cannot get level of species " << s << \
				" in compartment " << c << "\n";
		return 0;
	}

	return compartment[c]->getLevel(s);
}

// calculate if a rule is valid for a compartment
bool Psystem::isValidRule(unsigned long rs, unsigned long r, unsigned long c)
{
	unsigned long pos[3];
	long int off[3];

 	if(!ruleset[rs]->isChannel(r))
		// same compartment if rule is a transformation rule
		return true;
	// get compartment position
	pos[0] = compartment[c]->getXPosition();
	pos[1] = compartment[c]->getYPosition();
	pos[2] = compartment[c]->getZPosition();
	// get rule vector
	off[0] = ruleset[rs]->getRuleXVector(r);
	off[1] = ruleset[rs]->getRuleYVector(r);
	off[2] = ruleset[rs]->getRuleZVector(r);
	// add vector to compartment position
	off[0] += pos[0];
	off[1] += pos[1];
	off[2] += pos[2];
	// allow for periodic boundary conditions
	if(periodic_x) {
		if(off[0] < 0)
			off[0] += xdim;
		else
			if(off[0] >= (long int) xdim)
				off[0] -= xdim;
	}
	if(periodic_y) {
		if(off[1] < 0)
			off[1] += ydim;
		else
			if(off[1] >= (long int) ydim)
				off[1] -= ydim;
	}
	if(periodic_z) {
		if(off[2] < 0)
			off[2] += zdim;
		else
			if(off[2] >= (long int) zdim)
				off[2] -= zdim;
	}
	pos[0] = (unsigned long) off[0];
	pos[1] = (unsigned long) off[1];
	pos[2] = (unsigned long) off[2];

	if(getCompartmentIndex(pos) == num_compartments)
		// no valid compartment found
		return false;
	// found a valid compartment
	return true;
}

// return true if compartment defines template
bool Psystem::isTemplateCompartment(string s)
{
	unsigned int count;
	string::size_type index;

	// find out if old or new representation
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}
	switch(count) {
		case 1:  // old representation
			if(s[0] == 'T')
				return true;
			break;

		case 2:  // openalea representation
			index = s.find(':');
			if(s[index + 1] != ':')
				return true;
			break;

		case 3:  // new representation
		case 4:
		case 5:
			index = s.find(':');
			if(s[index + 1] != ':')
				return true;
			break;

		default:
			break;
	}
	return false;
}

// test if compartment name valid
bool Psystem::validateCompartmentName(string s)
{
	unsigned long count;
	long val;
	string::size_type index;
	string sc, scc;
	char *endptr;
	double val2;
	bool template_compartment;

	// find out if old or new representation
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// check name attributes
	switch(count) {
		case 1:  // old representation
			// check compartment name
			if(!(s[0] == 'T' || s[0] == 't'))
				break;
			sc = s.substr(1, s.find(':') - 1);
			errno = 0;
			val = strtol(sc.c_str(), &endptr, 10);
			if((errno == ERANGE && val == LONG_MAX)
					|| (errno != 0 && val == 0) || (val < 0))
				break;
			if(endptr == sc.c_str() || *endptr != '\0')
				break;
			// check compartment position
			// x value
			if(s.find(',') == string::npos)
				break;
			sc = s.substr(s.find(':') + 1, s.find(',') - 1);
			errno = 0;
			val = strtol(sc.c_str(), &endptr, 10);
			if((errno == ERANGE && val == LONG_MAX)
					|| (errno != 0 && val == 0) || (val < 0))
				break;
			if(endptr == sc.c_str() || *endptr != ',')
				break;
			// y value
			sc = s.substr(s.find(',') + 1);
			errno = 0;
			val = strtol(sc.c_str(), &endptr, 10);
			if((errno == ERANGE && val == LONG_MAX)
					|| (errno != 0 && val == 0) || (val < 0))
				break;
			if(endptr == sc.c_str() || *endptr != '\0')
				break;
			// all ok
			return true;
			break;

		case 2: // openalea representation
			// check template definition
			sc = s.substr(s.find(':') + 1);
			if(sc[0] != ':') {
				scc = sc;
				do {
					errno = 0;
					val = strtol(scc.c_str(), &endptr, 10);
					if((errno == ERANGE && val == LONG_MAX)
							|| (errno != 0 && val == 0) || (val < 0))
						break;
					if(endptr == scc.c_str() || !(*endptr == ':' || *endptr == ','))
						break;
					if(*endptr == ':')
						break;
					sc = scc.substr(scc.find(',') + 1);
					scc = sc;
				} while(1);
				if(*endptr != ':')
					break;
			}
			// check template usage
			scc = sc.substr(sc.find(':') + 1);
			do {
				errno = 0;
				val = strtol(scc.c_str(), &endptr, 10);
				if((errno == ERANGE && val == LONG_MAX)
						|| (errno != 0 && val == 0) || (val < 0))
					break;
				if(endptr == scc.c_str() || !(*endptr == '\0' || *endptr == ','))
					break;
				if(*endptr == '\0')
					break;
				sc = scc.substr(scc.find(',') + 1);
				scc = sc;
			} while(1);
			if(*endptr != '\0')
				break;
			// all ok
			return true;
			break;

		case 3:  // new representation
		case 4:
		case 5:
			// check template definition
			sc = s.substr(s.find(':') + 1);
			if(sc[0] != ':') {
				errno = 0;
				val = strtol(sc.c_str(), &endptr, 10);
				if((errno == ERANGE && val == LONG_MAX)
						|| (errno != 0 && val == 0) || (val < 0))
					break;
				if(endptr == sc.c_str() || *endptr != ':')
					break;
				template_compartment = true;
			} else {
				template_compartment = false;
				// empty template definition and usage
//				if(sc[1] == ':')
//					break;
			}
			// check template usage
			scc = sc.substr(sc.find(':') + 1);
			do {
				errno = 0;
				val = strtol(scc.c_str(), &endptr, 10);
				if((errno == ERANGE && val == LONG_MAX)
						|| (errno != 0 && val == 0) || (val < 0))
					break;
				if(endptr == scc.c_str() || !(*endptr == ':' || *endptr == ','))
					break;
				if(*endptr == ':')
					break;
				sc = scc.substr(scc.find(',') + 1);
				scc = sc;
			} while(1);
			if(*endptr != ':')
				break;
			// check compartment position
			sc = scc.substr(scc.find(':') + 1);
			if(sc[0] == '\0')
				// empty position
				return true;
			// noempty position
			if(!template_compartment) {
				// x value
				if(sc.find(',') == string::npos)
					break;
				scc = sc.substr(0, sc.find(','));
				errno = 0;
				val = strtol(scc.c_str(), &endptr, 10);
				if((errno == ERANGE && val == LONG_MAX)
						|| (errno != 0 && val == 0) || (val < 0))
					break;
				if(endptr == scc.c_str() || *endptr != '\0')
					break;
				// y value
				scc = sc.substr(sc.find(',') + 1);
				errno = 0;
				val = strtol(scc.c_str(), &endptr, 10);
				if((errno == ERANGE && val == LONG_MAX)
						|| (errno != 0 && val == 0) || (val < 0))
					break;
				if(endptr == scc.c_str() || !(*endptr == '\0' || *endptr == ':'))
					break;
			}
			// check compartment growth rate
			if(sc.find(':') == string::npos)
				// no growth rate set
				return true;
			// check growth rate
			if(growth_type == LINEAR_GROWTH || \
					growth_type == EXPONENTIAL_GROWTH) {
				// linear or exponential growth
				scc = sc.substr(sc.find(':') + 1);
				errno = 0;
				val2 = strtod(scc.c_str(), &endptr);
				if(errno != 0 || endptr == scc.c_str() || *endptr != '\0' \
						|| val2 <= 0.0)
					break;
			} else {
				// function based growth and division
				scc = sc.substr(sc.find(':') + 1);
				if(scc.find(':') == string::npos)
					break;
			}
			// all ok
			return true;
			break;

		default:
			cerr << "error: invalid compartment name " << s << "\n";
			exit(EXIT_FAILURE);
			break;
	}
	cerr << "error: invalid compartment name " << s << "\n";
	exit(EXIT_FAILURE);
	return false;
}

// test if rule name is valid
bool Psystem::validateRuleName(string s)
{
	unsigned long count, val;
	string::size_type index;
	string sc;
	char *endptr;

	// check all attributes are in the name
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// check name attributes
	switch(count) {
		case 0:  // translation rule
			return true;
			break;

		case 1:
			// check rule position
			// x value
			return true;
			if(s.find(',') == string::npos)
				return true;
			sc = s.substr(s.find(':') + 1, s.find(',') - 1);
			errno = 0;
			val = strtoul(sc.c_str(), &endptr, 10);
			if((errno == ERANGE && val == ULONG_MAX)
					|| (errno != 0 && val == 0))
				break;
			if(endptr == sc.c_str() || *endptr != ',')
				break;
			// y value
			sc = s.substr(s.find(',') + 1);
			errno = 0;
			val = strtoul(sc.c_str(), &endptr, 10);
			if((errno == ERANGE && val == ULONG_MAX)
					|| (errno != 0 && val == 0))
				break;
			if(endptr == sc.c_str() || *endptr != '\0')
				break;
			// all ok
			return true;
			break;

		default:
			cerr << "error: invalid rule name " << s << "\n";
			exit(EXIT_FAILURE);
			break;
	}
	cerr << "error: invalid rule name " << s << "\n";
	exit(EXIT_FAILURE);
	return false;
}

// test if rule structure is valid
bool Psystem::validateRuleStructure(Reaction *r)
{
	unsigned int val;

	// only first and second order reactions
	val = r->getNumReactants();
	if(val == 1 || val == 2) {
		// with only zero, one, two or three products
		val = r->getNumProducts();
		if(val >= 0 && val <= 3)
			// with a kinetic law
			if(r->getKineticLaw())
				// that has one parameter specified
				if(r->getKineticLaw()->getNumParameters() == 1 || \
						r->getKineticLaw()->getNumParameters() == 3  || \
						r->getKineticLaw()->getNumParameters() == 4)
					return true;
	}
	cerr << "error: invalid rule specification for rule " << r->getName() << "\n";
	exit(EXIT_FAILURE);
	return false;
}

// validate distribution-based reaction constant
void Psystem::validateReactionDistribution(string s)
{
	unsigned long count;
	double val;
	string::size_type index;
	string sc, scc;
	char *endptr;

	// check all attributes are in the name
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}
	if(count != 2) {
		cerr << "error: invalid distribution parameter " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution
	sc = s.substr(0, s.find(':'));
	if(sc != "gaussian")  {
		cerr << "error: invalid parameter distribution " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution mean
	sc = s.substr(s.find(':') + 1);
	errno = 0;
	val = strtod(sc.c_str(), &endptr);
	if(errno != 0 || endptr == sc.c_str() || *endptr != ':' || val < 0.0) {
		cerr << "error: invalid distribution mean " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution standard deviation
	scc = sc.substr(sc.find(':') + 1);
	errno = 0;
	val = strtod(scc.c_str(), &endptr);
	if(errno != 0 || endptr == scc.c_str() || *endptr != '\0' || val < 0.0) {
		cerr << "error: invalid distribution standard deviation " << s << "\n";
		exit(EXIT_FAILURE);
	}
}

// validate distribution-based species
void Psystem::validateSpeciesName(string s)
{
	unsigned long count;
	double val;
	string::size_type index;
	string sc, scc;
	char *endptr;

	// check all attributes are in the name
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}
	if(!count)
		return;  // plain species name
	if(count != 3) {
		cerr << "error: invalid species name " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution
	sc = s.substr(s.find(':') + 1);
	scc = sc.substr(0, sc.find(':'));
	if(scc != "gaussian")  {
		cerr << "error: invalid species distribution " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution mean
	scc = sc.substr(sc.find(':') + 1);
	errno = 0;
	val = strtod(scc.c_str(), &endptr);
	if(errno != 0 || endptr == scc.c_str() || *endptr != ':' || val < 0.0) {
		cerr << "error: invalid species mean " << s << "\n";
		exit(EXIT_FAILURE);
	}

	// check distribution standard deviation
	sc = scc;
	scc = sc.substr(sc.find(':') + 1);
	errno = 0;
	val = strtod(scc.c_str(), &endptr);
	if(errno != 0 || endptr == scc.c_str() || *endptr != '\0' || val < 0.0) {
		cerr << "error: invalid species standard deviation " << s << "\n";
		exit(EXIT_FAILURE);
	}
}

// test if event name is valid
bool Psystem::validateEventName(string s)
{
	unsigned long count, val;
	string::size_type index;
	string sc;
	char *endptr;

	// check all attributes are in the name
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	if(count == 1) {
		// check event index
		sc = s.substr(s.find(':') + 1);
		errno = 0;
		val = strtoul(sc.c_str(), &endptr, 10);
		if(!((errno == ERANGE && val == ULONG_MAX) || (errno != 0 && val == 0) \
				|| (endptr == sc.c_str()) || (*endptr != '\0')))
			return true;  // all ok
	}

	cerr << "error: invalid event name " << s << "\n";
	exit(EXIT_FAILURE);
	return false;
}

// get event index from SBML
unsigned long Psystem::getEventIndexFromSBML(string s)
{
	return (unsigned long) atol(s.substr(s.find(':') + 1).c_str());
}

// get compartment position from SBML
unsigned long *Psystem::getCompartmentPositionFromSBML(string s, \
		unsigned long *pos)
{
	unsigned int count;
	string::size_type index;
	string sc;

	// find out if old or new representation
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// get position
	switch(count) {
		case 1:  // old representation
			sc = s.substr(s.find(':') + 1);
			if(sc[0] == '\0')
				return NULL;
			pos[0] = atol(sc.c_str());
			pos[1] = atol(sc.c_str() + sc.find(',') + 1);
			return pos;
			break;

		case 3:  // new representation
		case 4:
		case 5:
			// move to third colon
			count = 3;
			index = 0;
			while(count--)
				index = s.find(':', index) + 1;
			sc = s.substr(index);
			if(sc[0] == '\0' || sc[0] == ':')
				return NULL;
			if(sc.find(",") == string::npos)
				return NULL;
			pos[0] = atol(sc.c_str());
			pos[1] = atol(sc.c_str() + sc.find(',') + 1);
			return pos;
			break;

		default:
			break;
	}

	return NULL;
}

// get compartment template indices from SBML
void Psystem::getCompartmentTemplatesFromSBML(string s, unsigned long *temps)
{
	string::size_type index, count;
	string sc;

	// find out if old or new representation
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// get list of compartment templates
	switch(count) {
		case 1:  // old representation
			temps[0] = atoi(s.c_str() + 1);
			break;

		case 2:  // openalea representation
			index = s.find(':');
			index = s.find(':', index + 1);
			if(s[index + 1] == '\0')
				break;
			sc = s.substr(index + 1);
			temps[0] = atoi(sc.c_str());
			count = index = 0;
			while((index = sc.find(',', index)) != string::npos) {
				index++;
				temps[++count] = atoi(sc.c_str() + index);
			}
			break;

		case 3:  // new representation
		case 4:
		case 5:
			index = s.find(':');
			if(isTemplateCompartment(s))
				temps[0] = atoi(s.c_str() + index + 1);
			else {
				index = s.find(':', index + 1);
				sc = s.substr(index + 1);
				count = sc.find(':');
				sc = s.substr(index + 1, count);
				temps[0] = atoi(sc.c_str());
				count = index = 0;
				while((index = sc.find(',', index)) != string::npos) {
					index++;
					temps[++count] = atoi(sc.c_str() + index);
				}
			}
			break;

		default:
			break;
	}
}

// get number of compartment templates from SBML
unsigned int Psystem::getNumCompartmentTemplatesFromSBML(string s)
{
	string::size_type index, count;
	string sc;

	// find out if old or new representation
	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// get compartment templates
	switch(count) {
		case 1:  // event representation
			return 0;
			break;

		case 2:  // openalea representation
			index = s.find(':');
			index = s.find(':', index + 1);
			if(s[index + 1] == '\0')
				return 0;
			sc = s.substr(index + 1);
			count = index = 0;
			while((index = sc.find(',', index)) != string::npos) {
				index++;
				count++;
			}
			return(count + 1);
			break;

		case 3:  // new representation
		case 4:
		case 5:
			if(isTemplateCompartment(s))
				return 0;
			index = s.find(':');
			index = s.find(':', index + 1);
			if(s[index + 1] == ':')
				return 0;
			sc = s.substr(index + 1);
			count = sc.find(':');
			sc = s.substr(index + 1, count);
			count = index = 0;
			while((index = sc.find(',', index)) != string::npos) {
				index++;
				count++;
			}
			return(count + 1);
			break;

		default:
			break;
	}

	return 0;
}

// get name of compartment from SBML
string Psystem::getCompartmentNameFromSBML(string name)
{
	return name.substr(0, name.find(":"));
}

// get index of ruleset from name
unsigned long Psystem::getSBMLIndex(string name)
{
	string::size_type index, count;

	// find out if old or new representation
	count = index = 0;
	while((index = name.find(':', index)) != string::npos) {
		index++;
		count++;
	}

	// get position
	switch(count) {
		case 1:  // event representation
			index = name.find(':');
			return atoi(name.c_str() + index + 1);
			break;

		case 2:  // openalea representation
			index = name.find(':');
			return atoi(name.c_str() + index + 1);
			break;

		case 3:  // new representation
		case 4:
		case 5:
			index = name.find(':');
			return atoi(name.c_str() + index + 1);
			break;

		default:
			break;
	}
	cerr << "error: no index for ruleset with name " << name << "\n";
	exit(EXIT_FAILURE);
	return 0;
}

// get generation time from SBML
double Psystem::getGenerationTimeFromSBML(string s)
{
	unsigned int count;
	string::size_type index;
	char *endptr;

	count = index = 0;
	while((index = s.find(':', index)) != string::npos) {
		index++;
		count++;
	}
	if(count != 4) {
		cerr << "error: no generation time for compartment " << s << "\n";
		exit(EXIT_FAILURE);
	}

	return strtod(s.substr(s.rfind(':') + 1).c_str(), &endptr);
}

// save model as stochkit problem definition
// valid for following stochkit methods:
// tau-leaping with adaptive stepsize (poison, binomial)
// tau-leaping with fixed stepsize (implicit, explicit, trapezoidal)
// SSA direct method
void Psystem::outputStochKit(SimulationParameters &parameters, char *outfile)
{
	unsigned long i, j, k, l, i2, j2, k2, l2;
	unsigned long total_rules, species_index;
	unsigned long total_species, total_species2;
	unsigned long rule_target, rule_target2;
	unsigned long rule_index, rule_index2;
	int **sm, **dg, *sp;
	unsigned long *ia;
	double *src;
	Rule *rule, *rule2;
	FILE *fp;

	// open output file
	if(!(fp = fopen(outfile, "w"))) {
		cerr << "error: couldn't open output file " << outfile << "\n";
		exit(EXIT_FAILURE);
	}

	// create stoichiometric + dependency matrix, + reaction constant vector
	// find total number of rules
	total_rules = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			total_rules += compartment[i]->getNumRules(j);
	// initialise memory for stoichiometric matrix
	total_species = num_compartments * num_species;
	if(!(sm = (int **) calloc(total_rules, sizeof(int *)))) {
		cerr << "error: couldn't allocate memory for stoichiometric matrix\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < total_rules; i++)
		if(!(sm[i] = (int *) calloc(total_species, sizeof(int)))) {
			cerr << "error: couldn't allocate memory for stoichiometric " \
					"matrix\n";
			exit(EXIT_FAILURE);
		}
	// initialise memory for stochastic reaction constant vector
	if(!(src = (double *) calloc(total_rules, sizeof(double)))) {
		cerr << "error: couldn't allocate memory for reaction constants\n";
		exit(EXIT_FAILURE);
	}
	// initialise memory for dependency matrix
	if(!(dg = (int **) calloc(total_rules, sizeof(int *)))) {
		cerr << "error: couldn't allocate memory for dependency graph\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < total_rules; i++)
		if(!(dg[i] = (int *) calloc(total_rules, sizeof(int)))) {
			cerr << "error: couldn't allocate memory for dependency graph\n";
			exit(EXIT_FAILURE);
		}
	// initialise memory for actual species vector
	if(!(sp = (int *) calloc(total_species, sizeof(int)))) {
		cerr << "error: couldn't allocate memory for species vector\n";
		exit(EXIT_FAILURE);
	}
	// initialise memory for species initial amounts vector
	if(!(ia = (unsigned long *) calloc(total_species, \
			sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for initial amounts " \
				"vector\n";
		exit(EXIT_FAILURE);
	}

	// populate matrices + vectors
	rule_index = 0;
	for(i = 0; i < num_compartments; i++) {
		for(j = 0; j < num_species; j++)
			// initial amounts
			ia[i * num_species + j] = compartment[i]->getLevel(j);
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				// dependency graph
				// reactants
				rule_target = i;
				for(l = 0; l < rule->getNumReactants(); l++) {
					species_index = rule->getReactantIndex(l);
					rule_index2 = 0;
					for(i2 = 0; i2 < num_compartments; i2++)
						for(j2 = 0; j2 < compartment[i2]-> \
								getNumRuleSets(); j2++)
							for(k2 = 0; k2 < compartment[i2]-> \
									getNumRules(j2); k2++) {
								rule2 = compartment[i2]->getRule(j2, k2);
								// reactants
								rule_target2 = i2;
								for(l2 = 0; l2 < rule2->getNumReactants(); l2++)
									if(rule2->getReactantIndex(l2) == \
											species_index && rule_target2 \
											== rule_target)
										dg[rule_index][rule_index2] = 1;
								// products
								if(rule2->isChannel())
									rule_target2 = compartment[i2]-> \
											getRuleTarget(j2, k2);
								for(l2 = 0; l2 < rule2->getNumProducts(); l2++)
									if(rule2->getProductIndex(l2) == \
											species_index && rule_target2 \
											== rule_target)
										dg[rule_index][rule_index2] = 1;
								rule_index2++;
							}
				}
				// products
				if(rule->isChannel())
					rule_target = compartment[i]->getRuleTarget(j, k);
				for(l = 0; l < rule->getNumProducts(); l++) {
					species_index = rule->getProductIndex(l);
					rule_index2 = 0;
					for(i2 = 0; i2 < num_compartments; i2++)
						for(j2 = 0; j2 < compartment[i2]-> \
								getNumRuleSets(); j2++)
							for(k2 = 0; k2 < compartment[i2]-> \
									getNumRules(j2); k2++) {
								rule2 = compartment[i2]->getRule(j2, k2);
								// reactants
								rule_target2 = i2;
								for(l2 = 0; l2 < rule2->getNumReactants(); l2++)
									if(rule2->getReactantIndex(l2) == \
											species_index && rule_target2 \
											== rule_target)
										dg[rule_index][rule_index2] = 1;
								// products
								if(rule2->isChannel())
									rule_target2 = compartment[i2]-> \
											getRuleTarget(j2, k2);
								for(l2 = 0; l2 < rule2->getNumProducts(); l2++)
									if(rule2->getProductIndex(l2) == \
											species_index && rule_target2 \
											== rule_target)
										dg[rule_index][rule_index2] = 1;
								rule_index2++;
							}
				}

				// stoichiometric matrix + species vector
				// reactants
				rule_target = i;
				for(l = 0; l < rule->getNumReactants(); l++) {
					sm[rule_index][rule_target * num_species + \
							rule->getReactantIndex(l)]--;
					sp[rule_target * num_species + rule->getReactantIndex(l)] \
							= 1;
				}
				// products
				if(rule->isChannel())
					rule_target = compartment[i]->getRuleTarget(j, k);
				else
					rule_target = i;
				for(l = 0; l < rule->getNumProducts(); l++) {
					sm[rule_index][rule_target * num_species + \
							rule->getProductIndex(l)]++;
					sp[rule_target * num_species + rule->getProductIndex(l)] \
							= 1;
				}

				// stochastic reaction constant vector
				src[rule_index] = rule->getReactionConstant();

				rule_index++;
			}
	}

	// count actual number of species
	total_species2 = 0;
	for(i = 0; i < total_species; i++)
		total_species2 += sp[i];

	// set output datafile name
	string data_file(parameters.data_file);
	data_file = data_file.substr(0, data_file.find_last_of("."));
	data_file.append(".sk.data");

	// output stochkit c++ problem definition file

	// header
	fprintf(fp, "#include \"ProblemDefinition.h\"\n");

	// variables
	fprintf(fp, "\n// global variables\n");
	fprintf(fp, "double max_time = %f;\n", parameters.max_time);
	fprintf(fp, "unsigned int seed = %ld;\n", parameters.seed);
	fprintf(fp, "double log_interval = %f;\n", parameters.log_interval);
	fprintf(fp, "char outfile[] = \"%s\";\n", data_file.c_str());
	fprintf(fp, "unsigned long total_species = %ld;\n", total_species2);
	fprintf(fp, "unsigned long total_rules = %ld;\n", total_rules);

	// stochastic reaction constants
	fprintf(fp, "\n// stochastic reaction constants\n");
	fprintf(fp, "static double *c;\n");

	// Initialise function
	fprintf(fp, "\nVector Initialize()\n{\n");
	fprintf(fp, "\tVector x0(total_species, 0.0);\n\n");
	species_index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < num_species; j++)
			if(sp[i * num_species + j]) {
				fprintf(fp, "\tx0(%ld) = %ld;\t// %s", \
						species_index++, ia[i * num_species + j], \
						alphabet->getName(j).c_str());
				if(num_compartments > 1)
					fprintf(fp, ", compartment %ld", i);
				fprintf(fp, "\n");
			}
	fprintf(fp, "\n\tif(!(c = (double *) calloc(total_rules, " \
			"sizeof(double)))) {\n" \
			"\t\tfprintf(stderr, " \
			"\"couldn't allocate memory for reaction constants" \
			"\\n\");\n\t\texit(EXIT_FAILURE);\n\t}\n");
	for(i = 0; i < total_rules; i++)
		fprintf(fp, "\tc[%ld] = %f;\n", i, src[i]);
	fprintf(fp, "\n\treturn x0;\n}\n");

	// Stoichiometry function
	fprintf(fp, "\nMatrix Stoichiometry()\n{\n");
	fprintf(fp, "\tMatrix nu(total_species, total_rules, 0.0);\n\n");
	rule_index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				fprintf(fp, "\t// %s\n", rule->getId().c_str());
				species_index = 0;
				for(l = 0; l < total_species; l++) {
					if(sm[rule_index][l] != 0)
						fprintf(fp, "\tnu(%ld,%ld) = %d;\n", species_index, \
								rule_index, sm[rule_index][l]);
					species_index += sp[l];
				}
				fprintf(fp, "\n");
				rule_index++;
			}
	fprintf(fp, "\treturn nu;\n}\n");

	// DependentGrapth function
	fprintf(fp, "\nMatrix DependentGrapth()\n{\n");
	fprintf(fp, "\tMatrix DG(total_rules, total_rules, 0.0);\n\n");
	for(i = 0; i < total_rules; i++) {
		fprintf(fp, "\t");
		for(j = 0; j < total_rules; j++)
			fprintf(fp, "DG(%ld,%ld) = %d; ", i, j, dg[i][j]);
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n\treturn DG;\n}\n");

	// Propensity function
	fprintf(fp, "\nVector Propensity(const Vector &x)\n{\n");
	fprintf(fp, "\tVector a(total_rules);\n\n");
	rule_index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				fprintf(fp, "\ta(%ld) = c[%ld] * ", rule_index, rule_index);
				switch(rule->getNumReactants()) {
					case 1:
						i2 = i * num_species + rule->getReactantIndex(0);
						species_index = 0;
						while(i2--)
							if(sp[i2])
								species_index++;
							fprintf(fp, "x(%ld);\n", species_index);
						break;

					case 2:
						if(rule->getReactantIndex(0) == \
								rule->getReactantIndex(1)) {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "0.5 * x(%ld) * (x(%ld) - 1);\n", \
									species_index, species_index);
						} else {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "x(%ld) * ", species_index);
							i2 = i * num_species + rule->getReactantIndex(1);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "x(%ld);\n", species_index);
						}
						break;

					default:
						break;
				}
				rule_index++;
			}
	fprintf(fp, "\n\treturn a;\n}\n");

	// PartialProp function
	fprintf(fp, "\nVector PartialProp(int RIndex, const Vector &x, " \
			"const Matrix &dg, double &a0)\n{\n");
	fprintf(fp, "\tVector a(total_rules);\n\n");
	rule_index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				fprintf(fp, "\ta(%ld) = c[%ld] * ", rule_index, rule_index);
				switch(rule->getNumReactants()) {
					case 1:
						i2 = i * num_species + rule->getReactantIndex(0);
						species_index = 0;
						while(i2--)
							if(sp[i2])
								species_index++;
							fprintf(fp, "x(%ld);\n", species_index);
						break;

					case 2:
						if(rule->getReactantIndex(0) == \
								rule->getReactantIndex(1)) {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "0.5 * x(%ld) * (x(%ld) - 1);\n", \
									species_index, species_index);
						} else {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "x(%ld) * ", species_index);
							i2 = i * num_species + rule->getReactantIndex(1);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "x(%ld);\n", species_index);
						}
						break;

					default:
						break;
				}
				rule_index++;
			}
	fprintf(fp, "\n\treturn a;\n}\n");

	// PropensityJacobian function
	fprintf(fp, "\nMatrix PropensityJacobian(const Vector &x)\n{\n");
	fprintf(fp, "\tMatrix j(total_rules,total_species,0.0);\n\n");
	rule_index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				switch(rule->getNumReactants()) {
					case 1:
						i2 = i * num_species + rule->getReactantIndex(0);
						species_index = 0;
						while(i2--)
							if(sp[i2])
								species_index++;
							fprintf(fp, "\tj(%ld,%ld) = c[%ld];\n", \
									rule_index, species_index, rule_index);
						break;

					case 2:
						if(rule->getReactantIndex(0) == \
								rule->getReactantIndex(1)) {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "\tj(%ld,%ld) = c[%ld] * " \
									"(x(%ld) - 0.5);\n", rule_index, \
									species_index, rule_index, species_index);
						} else {
							i2 = i * num_species + rule->getReactantIndex(0);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							j2 = species_index;
							i2 = i * num_species + rule->getReactantIndex(1);
							species_index = 0;
							while(i2--)
								if(sp[i2])
									species_index++;
							fprintf(fp, "\tj(%ld,%ld) = c[%ld] * x(%ld);\n", \
									rule_index, j2, rule_index, species_index);
							fprintf(fp, "\tj(%ld,%ld) = c[%ld] * x(%ld);\n", \
									rule_index, species_index, rule_index, j2);
						}
						break;

					default:
						break;
				}
				rule_index++;
			}
	fprintf(fp, "\n\treturn j;\n}\n");

	// Equilibrium function
	fprintf(fp, "\nvoid Equilibrium(Vector &x, Vector &a, int rxn)\n{\n}\n");

	fclose(fp);

	// free memory
	for(i = 0; i < total_rules; i++) {
		free(sm[i]);
		free(dg[i]);
	}
	free(sm);
	free(src);
	free(dg);
	free(sp);
	free(ia);
}

// add a duplicate compartment
Simulation::Compartment *Psystem::addDuplicateCompartment(unsigned long c, \
		unsigned long *pos, char shift_direction)
{
	unsigned long i, cpos, ind, posmax;
	unsigned long pos2[3] = {pos[0], pos[1], pos[2]};

	Compartment *comp = compartment[c];

	// create duplicate compartment
	Compartment *comp2 = new Compartment(*comp);

	// shift compartments in shift direction
	switch(shift_direction) {
		case 'x':
			cpos = pos[0];
			i = 0;
			pos2[0] = xdim;
			// increase lattice dimensions if necessary
			if(pos[0] >= xdim)
				xdim = pos[0] + 1;
			break;

		case 'y':
			cpos = pos[1];
			i = 1;
			pos2[1] = ydim;
			// increase lattice dimensions if necessary
			if(pos[1] >= ydim)
				ydim = pos[1] + 1;
			break;

		case 'z':
			cpos = pos[2];
			i = 2;
			pos2[2] = zdim;
			// increase lattice dimensions if necessary
			if(pos[2] >= zdim)
				zdim = pos[2] + 1;
			break;

		default:
			break;
	}
	posmax = 0;
	while(--pos2[i] >= cpos) {
		if((ind = getCompartmentIndex(pos2)) == num_compartments)
			continue;  // no compartment at that position
		// move compartment up one position
		pos2[i]++;
		getCompartment(ind)->setPosition(pos2);
		// save position data if required
		if(log_volumes)
			simalg->writePosition(ind);
		if(pos2[i] > posmax)
			posmax = pos2[i];
		pos2[i]--;
	}

	// increase lattice dimensions if necessary
	switch(shift_direction) {
		case 'x':
			if(posmax >= xdim)
				xdim = posmax + 1;
			break;

		case 'y':
			if(posmax >= ydim)
				ydim = posmax + 1;
			break;

		case 'z':
			if(posmax >= zdim)
				zdim = posmax + 1;
			break;

		default:
			break;
	}

	// set new id of compartment
	comp2->setId(comp2->getId().append("c"));

	// set position of new compartment
	comp2->setPosition(pos);

	// set creation time
	comp2->setCreationTime(simalg->getTime());

	// add new compartment to compartment list
	// allocate memory
	if(!(compartment = (Simulation::Compartment **) realloc(compartment, \
			(num_compartments + 1) * sizeof(Simulation::Compartment *)))) {
		cerr << "error: couldn't allocate memory for compartments\n";
		exit(EXIT_FAILURE);
	}
	// add compartment
	compartment[num_compartments] = comp2;
	// save position data if required
	if(log_volumes)
		simalg->writePosition(num_compartments);
	// update simulation algorithm
	simalg->setNumCompartments(++num_compartments);

	// recalculate valid rules for both compartments
// 	comp2->freeValidRules();
// 	setValidRules(num_compartments - 1);
// 	comp->freeValidRules();
// 	setValidRules(c);

	// update transport rules for other compartments
	for(i = 0; i < num_compartments; i++) {
		compartment[i]->freeValidRules();
		setValidRules(i);
	}

	return comp2;
}

// set valid ruleset rules for compartment
void Psystem::setValidRules(unsigned long ind)
{
	RuleSet *rs;
	Compartment *comp;
	register unsigned long i, j, rsi, valid_rules, updates[1];

	comp = compartment[ind];
	for(i = 0; i < comp->getNumRuleSets(); i++) {
		rsi = comp->getRuleSetIndex(i);
		rs = ruleset[rsi];
		// calculate number of valid rules
		valid_rules = 0;
		for(j = 0; j < rs->getNumRules(); j++)
			if(isValidRule(rsi, j, ind))
				valid_rules++;
		// create valid rules storage
		comp->initialiseValidRules(i, valid_rules);
		// populate valid rules storage
		valid_rules = 0;
		for(j = 0; j < rs->getNumRules(); j++) {
			if(isValidRule(rsi, j, ind)) {
				comp->setValidRule(i, valid_rules, j);
				if(rs->isChannel(j)) {
					// set update compartments
					if((updates[0] = calculateRuleTarget(i, valid_rules, ind)) \
							!= num_compartments)
						comp->setUpdates(i, valid_rules, updates, 1);
				}
				valid_rules++;
			}
		}
	}
}

// check if string valid SMBML SId
bool Psystem::isValidSBMLSId(string s)
{
	int i;

	if(s.empty())
		return false;
	for(i = 0; i < s.size(); i++)
		if(!(isalnum(s[i]) || s[i] == '_'))
			return false;

	return true;
}

// write a fully complient sbml model
void Psystem::outputSBML(char *outfile)
{
	unsigned long i, j, k, l, m;
	char name[256];
	::Compartment *comp;
	SBMLWriter sbml_writer;
	Rule *rule;
	string sid;

	// create sbml document
	SBMLDocument *sbml_document = new SBMLDocument(2, 1);
	if(sbml_document->getNumErrors()) {
		cerr << "error creating sbml document\n";
		sbml_document->printErrors(cerr);
		exit(EXIT_FAILURE);
	}

	// create sbml model
	Model *sbml_model = sbml_document->createModel();
	sid = getId();
	if(!isValidSBMLSId(sid)) {
		cerr << "error: invalid SBML SId \"" << sid << "\"\n";
		exit(EXIT_FAILURE);
	}
	sbml_model->setId(getId());
	sbml_model->setName(getId());

	// create unit definitions
	UnitDefinition *unit_def;
	Unit *unit;
	unit_def = sbml_model->createUnitDefinition();
	unit_def->setId("per_second");
	unit = unit_def->createUnit();
	unit->setExponent(-1);
	unit->setKind(UNIT_KIND_SECOND);
	unit_def = sbml_model->createUnitDefinition();
	unit_def->setId("substance");
	unit = unit_def->createUnit();
	unit->setKind(UNIT_KIND_ITEM);

	// create compartments
	for(i = 0; i < num_compartments; i++) {
		comp = sbml_model->createCompartment();
		snprintf(name, 256, "c%ld", i);
		if(!isValidSBMLSId(name)) {
			cerr << "error: invalid SBML SId \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		comp->setId(name);
		snprintf(name, 256, "c%ld_%ld", \
				compartment[i]->getXPosition(), \
				compartment[i]->getYPosition());
		comp->setName(name);
		comp->setSize(compartment[i]->getVolume());
	}

	// create species
	Species *sp;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < num_species; j++) {
			for(k = 0; k < compartment[i]->getNumRuleSets(); k++)
				for(l = 0; l < compartment[i]->getNumRules(k); l++) {
					rule = compartment[i]->getRule(k, l);
					for(m = 0; m < rule->getNumReactants(); m++)
						if(rule->getReactantIndex(m) == j)
							goto found_species;
					for(m = 0; m < rule->getNumProducts(); m++)
						if(rule->getProductIndex(m) == j)
							goto found_species;
				}
			continue;
			found_species:
				sp = sbml_model->createSpecies();
				snprintf(name, 256, "s%ld_%ld", i, j);
				if(!isValidSBMLSId(name)) {
					cerr << "error: invalid SBML SId \"" << name << "\"\n";
					exit(EXIT_FAILURE);
				}
				sp->setId(name);
				sp->setName(alphabet->getName(j));
				snprintf(name, 256, "c%ld", i);
				sp->setCompartment(name);
				sp->setInitialAmount(compartment[i]->getLevel(j));
				sp->setHasOnlySubstanceUnits(true);
		}

	// create degradation species
	if(keep_degraded) {
		for(i = 0; i < num_compartments; i++)
			for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
				for(k = 0; k < compartment[i]->getNumRules(j); k++) {
					rule = compartment[i]->getRule(j, k);
					if(!rule->getNumProducts()) {
						sp = sbml_model->createSpecies();
						snprintf(name, 256, "s%ld_%ld_degraded", i, \
								rule->getReactantIndex(0));
						if(!isValidSBMLSId(name)) {
							cerr << "error: invalid SBML SId \"" << name << "\"\n";
							exit(EXIT_FAILURE);
						}
						sp->setId(name);
						snprintf(name, 256, "%s_degraded", \
								alphabet->getName(rule->getReactantIndex(0)).c_str());
						sp->setName(name);
						snprintf(name, 256, "c%ld", i);
						sp->setCompartment(name);
						sp->setInitialAmount(0);
						sp->setHasOnlySubstanceUnits(true);
					}
				}
	}

	// create reactions
	Reaction *re;
	SpeciesReference *spr;
	KineticLaw *kl;
	Parameter *pa;
	ASTNode *ast_math;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				rule = compartment[i]->getRule(j, k);
				// create reaction
				re = sbml_model->createReaction();
				snprintf(name, 256, "re%ld_%ld_%ld", i, \
						compartment[i]->getRuleSetIndex(j), \
						compartment[i]->getRuleIndex(j, k));
				if(!isValidSBMLSId(name)) {
					cerr << "error: invalid SBML SId \"" << name << "\"\n";
					exit(EXIT_FAILURE);
				}
				re->setId(name);
				re->setName(name);
				re->setReversible(false);
				// create kinetic law
				kl = re->createKineticLaw();
				string ast_str = \
						"<math xmlns=\"http://www.w3.org/1998/Math/MathML\">"
                        "<apply><times/>";
//				snprintf(name, 256, "<ci>c%ld</ci>", i);
//				ast_str += name;
				// create parameter
				pa = kl->createParameter();
				pa->setId("k1");
				pa->setName("k1");
				pa->setValue(compartment[i]->getReactionConstant(j, k));
				pa->setUnits("per_second");
				snprintf(name, 256, "<ci>k1</ci>");
				ast_str += name;
				// add reactants
				for(l = 0; l < rule->getNumReactants(); l++) {
					spr = re->createReactant();
					snprintf(name, 256, "s%ld_%ld", i, \
							rule->getReactantIndex(l));
					spr->setSpecies(name);
					if(l == 1) {
						if(rule->getReactantIndex(0) == \
								rule->getReactantIndex(1))
							snprintf(name, 256, "<apply><minus/>" \
									"<ci>s%ld_%ld</ci><cn type=\"integer\">" \
									"1</cn></apply><cn type=\"real\">0.5" \
									"</cn>", i, rule->getReactantIndex(l));
						else
							snprintf(name, 256, "<ci>s%ld_%ld</ci>", i, \
									rule->getReactantIndex(l));
					} else
						snprintf(name, 256, "<ci>s%ld_%ld</ci>", i, \
								rule->getReactantIndex(l));
					ast_str += name;
				}
				// add products
				if(!rule->getNumProducts() && keep_degraded) {
					spr = re->createProduct();
					snprintf(name, 256, "s%ld_%ld_degraded", i, \
							rule->getReactantIndex(0));
					spr->setSpecies(name);
				} else {
					for(l = 0; l < rule->getNumProducts(); l++) {
						spr = re->createProduct();
						if(rule->isChannel())
							snprintf(name, 256, "s%ld_%ld", \
									compartment[i]->getRuleTarget(j, k), \
									rule->getProductIndex(l));
						else
							snprintf(name, 256, "s%ld_%ld", i, \
									rule->getProductIndex(l));
						spr->setSpecies(name);
					}
				}
				// add kinetic law
				ast_str += "</apply></math>";
				ast_math = readMathMLFromString(ast_str.c_str());
				kl->setMath(ast_math);
				delete ast_math;
			}

	// validate model
	if(sbml_document->checkInternalConsistency() > 0) {
		cerr << "error validating sbml document\n";
		sbml_document->printErrors();
		exit(EXIT_FAILURE);
	}
// 	if(sbml_document->checkConsistency() > 0) {
// 		cerr << "error validating sbml document\n";
// 		sbml_document->printErrors();
// 		exit(EXIT_FAILURE);
// 	}

	// write sbml document
	sbml_writer.setProgramName("mcss");
	if(!sbml_writer.writeSBML(sbml_document, outfile)) {
		cerr << "error writing sbml document\n";
		exit(EXIT_FAILURE);
	}

	// tidy up
	delete sbml_document;
}

// dump model to binary format file
void Psystem::dump(char *dump_file)
{
	FILE *fp;
	unsigned long i, j, k;
	int val;
	unsigned int uval;
	float fval;
	Rule *rule;

	// open output file
	if(!(fp = fopen(dump_file, "w"))) {
		cerr << "error creating dump file\n";
		exit(EXIT_FAILURE);
	}

	// write model to binary file
	// write model parameters
	val = (int) num_compartments;
	if(fwrite(&val, sizeof(int), 1, fp) != 1) {
		cerr << "error writing dump file\n";
		exit(EXIT_FAILURE);
	}
	val = (int) num_species;
	if(fwrite(&val, sizeof(int), 1, fp) != 1) {
		cerr << "error writing dump file\n";
		exit(EXIT_FAILURE);
	}
	val = (int) num_rulesets;
	if(fwrite(&val, sizeof(int), 1, fp) != 1) {
		cerr << "error writing dump file\n";
		exit(EXIT_FAILURE);
	}

	// write species levels
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < num_species; j++) {
			uval = (unsigned int) compartment[i]->getLevel(j);
			if(fwrite(&uval, sizeof(unsigned int), 1, fp) != 1) {
				cerr << "error writing dump file\n";
				exit(EXIT_FAILURE);
			}
		}

	// write ruleset specifications
	for(i = 0; i < num_rulesets; i++) {
		val = (int) ruleset[i]->getNumRules();
		if(fwrite(&val, sizeof(int), 1, fp) != 1) {
			cerr << "error writing dump file\n";
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < ruleset[i]->getNumRules(); j++) {
			rule = ruleset[i]->getRule(j);
			fval = (float) rule->getReactionConstant();
			if(fwrite(&fval, sizeof(float), 1, fp) != 1) {
				cerr << "error writing dump file\n";
				exit(EXIT_FAILURE);
			}
			val = (int) rule->getNumReactants();
			switch(val) {
				case 1:
					val = (int) rule->getReactantIndex(0);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				case 2:
					val = (int) rule->getReactantIndex(0);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = (int) rule->getReactantIndex(1);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				default:
					cerr << "error writing dump file\n";
					exit(EXIT_FAILURE);
					break;
			}
			val = (int) rule->getNumProducts();
			switch(val) {
				case 0:
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				case 1:
					val = (int) rule->getProductIndex(0);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				case 2:
					val = (int) rule->getProductIndex(0);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = (int) rule->getProductIndex(1);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				case 3:
					val = (int) rule->getProductIndex(0);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = (int) rule->getProductIndex(1);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					val = (int) rule->getProductIndex(2);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
					break;

				default:
					cerr << "error writing dump file\n";
					exit(EXIT_FAILURE);
					break;
			}
		}
	}

	// write compartment rules
	for(i = 0; i < num_compartments; i++) {
		val = (int) compartment[i]->getNumRuleSets();
		if(fwrite(&val, sizeof(int), 1, fp) != 1) {
			cerr << "error writing dump file\n";
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
			val = (int) compartment[i]->getNumRules(j);
			if(fwrite(&val, sizeof(int), 1, fp) != 1) {
				cerr << "error writing dump file\n";
				exit(EXIT_FAILURE);
			}
		}
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				val = (int) compartment[i]->getRuleSetIndex(j);
				if(fwrite(&val, sizeof(int), 1, fp) != 1) {
					cerr << "error writing dump file\n";
					exit(EXIT_FAILURE);
				}
				val = (int) compartment[i]->getRuleIndex(j, k);
				if(fwrite(&val, sizeof(int), 1, fp) != 1) {
					cerr << "error writing dump file\n";
					exit(EXIT_FAILURE);
				}
				if(compartment[i]->isChannel(j, k)) {
					val = (int) compartment[i]->getRuleTarget(j, k);
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
				} else {
					val = -1;
					if(fwrite(&val, sizeof(int), 1, fp) != 1) {
						cerr << "error writing dump file\n";
						exit(EXIT_FAILURE);
					}
				}
			}
		}
	}

	// write grid parameters
	val = (int) xdim;
	if(fwrite(&val, sizeof(int), 1, fp) != 1) {
		cerr << "error writing dump file\n";
		exit(EXIT_FAILURE);
	}
	val = (int) ydim;
	if(fwrite(&val, sizeof(int), 1, fp) != 1) {
		cerr << "error writing dump file\n";
		exit(EXIT_FAILURE);
	}

	// write compartment positions
	for(i = 0; i < num_compartments; i++) {
		val = (int) compartment[i]->getXPosition();
		if(fwrite(&val, sizeof(int), 1, fp) != 1) {
			cerr << "error writing dump file\n";
			exit(EXIT_FAILURE);
		}
		val = (int) compartment[i]->getYPosition();
		if(fwrite(&val, sizeof(int), 1, fp) != 1) {
			cerr << "error writing dump file\n";
			exit(EXIT_FAILURE);
		}
	}

	// close output file
	fclose(fp);
}

// add a new compartment
unsigned long Psystem::addCompartment(string name, unsigned long *pos)
{
	Compartment *comp;
	ostringstream id;

	// close current logs if added during simulation
	if(simalg && simalg->isSimulationRunning()) {
		// close existing amounts dataset
		simalg->closeAmountsDataSet();

		// close existing volumes dataset
		simalg->closeVolumesDataSet();
	}

	// create compartment id
	id << "C" << num_compartments;

	// initialise compartment
	comp = new Simulation::Compartment(id.str());

	// set compartment name
	comp->setName(name);

	// set compartment position
	if(pos) {
		// check position not already occupied
		if(getCompartmentIndex(pos) != num_compartments) {
			cerr << "error: cannot add compartment " << name << \
					" at position (" << pos[0] << "," << pos[1] << "," << \
					pos[2] << ")\n";
			exit(EXIT_FAILURE);
		}
		// adjust lattice dimension if necessary
		if(pos[0] >= xdim)
			xdim = pos[0] + 1;
		if(pos[1] >= ydim)
			ydim = pos[1] + 1;
		if(pos[2] >= zdim)
			zdim = pos[2] + 1;
		// set position
		comp->setPosition(pos);
	}

	// initialise species storage
	comp->initialiseLevels(num_species);

	// set creation time if created during simulation_algorithm
	if(simalg)
		comp->setCreationTime(simalg->getTime());

	// add new compartment to compartment list
	// allocate memory
	if(!compartment) {
		// first compartment added to p system
		if(!(compartment = (Simulation::Compartment **) malloc( \
				sizeof(Simulation::Compartment *)))) {
			cerr << "error: couldn't allocate memory for compartments\n";
			exit(EXIT_FAILURE);
		}
	} else {
		if(!(compartment = (Simulation::Compartment **) realloc(compartment, \
				(num_compartments + 1) * sizeof(Simulation::Compartment *)))) {
			cerr << "error: couldn't allocate memory for compartments\n";
			exit(EXIT_FAILURE);
		}
	}

	// add compartment to compartment list
	compartment[num_compartments] = comp;

	// increase number of compartments
	num_compartments++;

	// recreate logs if added during simulation
	if(simalg) {
		// update simulation algorithm information
		simalg->setNumCompartments(num_compartments);

		if(simalg->isSimulationRunning()) {
			// save position data if required
			if(log_volumes && pos)
				simalg->writePosition(num_compartments - 1);

			// extend amounts dataset
			simalg->extendAmountsDataSet();

			// extend volumes dataset
			simalg->extendVolumesDataSet();
		}
	}

	return (num_compartments - 1);
}

// add a species
unsigned long Psystem::addSpecies(string name)
{
	unsigned long int i;

	// check if first species added
	if(!num_species) {
		// allocate memory for alphabet
		alphabet = new Simulation::Alphabet(1);
		alphabet->setName(0, name);
	} else {
		// append species to alphabet
		alphabet->addSpecies(name);
	}

	// increase number of species in p system
	num_species++;

	// increase compartment species storage
	for(i = 0; i < num_compartments; i++)
		compartment[i]->addSpecies();

	// update simulation algorithm
	if(simalg) {
		// check logging not turned on if adding to running simulation
		if(simalg->isLoggingOn()) {
			cerr << "error: cannot add species to simulation when logging" \
					" is turned on\n";
			exit(EXIT_FAILURE);
		}
		simalg->setNumSpecies(alphabet->size());
	}

	return (num_species - 1);
}

// add a ruleset
unsigned long Psystem::addRuleSet(string name)
{
	ostringstream id;

	// check logging not turned on if adding to running simulation
	if(simalg && simalg->isLoggingOn()) {
		cerr << "error: cannot add ruleset to simulation when logging" \
				" is turned on\n";
		exit(EXIT_FAILURE);
	}

	// check if first ruleset added
	if(!num_rulesets) {
		// allocate memory for ruleset
		if(!(ruleset = (RuleSet **) malloc(sizeof(RuleSet *)))) {
			cerr << "error: couldn't allocate memory for rulesets\n";
			exit(EXIT_FAILURE);
		}
	} else {
		// reallocate memory for ruleset
		if(!(ruleset = (RuleSet **) realloc(ruleset, \
					(num_rulesets + 1) * sizeof(RuleSet *)))) {
			cerr << "error: couldn't reallocate memory for rulesets\n";
			exit(EXIT_FAILURE);
		}
	}

	// initialise ruleset
	ruleset[num_rulesets] = new Simulation::RuleSet();
	ruleset[num_rulesets]->setName(name);
 	ruleset[num_rulesets]->setIndex(num_rulesets);
	id << "RS" << num_rulesets;
	ruleset[num_rulesets]->setId(id.str());

	// increase number of ruleset
	num_rulesets++;

	return (num_rulesets - 1);
}

// add a rule to a ruleset
unsigned long Psystem::addRule(unsigned long rsid, string name, \
		unsigned long num_reactants, unsigned long *reactants, \
		unsigned long num_products, unsigned long *products, \
		double reaction_constant)
{
	unsigned long i, rid;
	Rule *rule;
	ostringstream id;

	// check logging not turned on if adding to running simulation
	if(simalg && simalg->isLoggingOn()) {
		cerr << "error: cannot add rule to simulation when logging" \
				" is turned on\n";
		exit(EXIT_FAILURE);
	}

	// sanity check rule
	if(rsid >= num_rulesets) {
		cerr << "error: cannot add rule " << name << " to ruleset id=" << \
				rsid << " - ruleset not found\n";
		exit(EXIT_FAILURE);
	}
	if(num_reactants < 1 || num_reactants > 2) {
		cerr << "error: cannot add rule " << name << " to ruleset id=" << \
				rsid << " - invalid number of reactants\n";
		exit(EXIT_FAILURE);
	}
	if(num_products < 0 || num_products > 3) {
		cerr << "error: cannot add rule " << name << " to ruleset id=" << \
				rsid << " - invalid number of products\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_reactants; i++)
		if(reactants[i] < 0 || reactants[i] >= num_species) {
			cerr << "error: cannot add rule " << name << " to ruleset id=" \
					<< rsid << " - invalid product index " << reactants[i] \
					<< "\n";
			exit(EXIT_FAILURE);
		}
	for(i = 0; i < num_products; i++)
		if(products[i] < 0 || products[i] >= num_species) {
			cerr << "error: cannot add rule " << name << " to ruleset id=" \
					<< rsid << " - invalid product index " << products[i] \
					<< "\n";
			exit(EXIT_FAILURE);
		}
	if(reaction_constant < 0.0) {
		cerr << "error: cannot add rule " << name << " to ruleset id=" << \
				rsid << " - invalid reaction constant\n";
		exit(EXIT_FAILURE);
	}

	// create rule
	id << "R" << ruleset[rsid]->getNumRules();
	rule = new Rule(id.str());
	rule->setName(name);
	rule->setReactionConstant(reaction_constant);
	rule->setNumReactants(num_reactants);
	rule->setNumProducts(num_products);
	for(i = 0; i < num_reactants; i++)
		rule->setReactantIndex(i, reactants[i]);
	for(i = 0; i < num_products; i++)
		rule->setProductIndex(i, products[i]);

	// add rule to ruleset
	rid = ruleset[rsid]->addRule(rule);

	return rid;
}

// set rule to be a transportation rule
void Psystem::setChannel(unsigned long rsid, unsigned long rid)
{
	// sanity check
	if(rsid < 0 || rsid >= num_rulesets) {
		cerr << "error: cannot set rule id=" << rid << " in ruleset id=" \
				<< rsid << " as channel - invalid ruleset id\n";
		exit(EXIT_FAILURE);
	}
	if(rid < 0 || rid >= ruleset[rsid]->getNumRules()) {
		cerr << "error: cannot set rule id=" << rid << " in ruleset id=" \
				<< rsid << " as channel - invalid rule id\n";
		exit(EXIT_FAILURE);
	}

	// set as channel
	ruleset[rsid]->getRule(rid)->setChannel();
}

// add a ruleset to a compartment
void Psystem::addRuleSetToCompartment(unsigned long rsid, unsigned long cid)
{
	unsigned long i, nrs, ncid;

	// sanity check
	if(rsid < 0 || rsid >= num_rulesets) {
		cerr << "error: cannot add ruleset id=" << rsid << \
				" to compartment id=" << cid << " - invalid ruleset id\n";
		exit(EXIT_FAILURE);
	}
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot add ruleset id=" << rsid << \
				" to compartment id=" << cid << " - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}

	if(ruleset[rsid]->isEnclosedCompartment()) {
		ncid = addCompartment(ruleset[rsid]->getName().substr(0, \
				ruleset[rsid]->getName().find(":")));
		compartment[ncid]->setEnclosedCompartment(true);
		compartment[ncid]->setParentCompartmentIndex(cid);
		compartment[cid]->addChildCompartment(ncid);
		unsigned long pos[3] = {ULONG_MAX, ULONG_MAX, ULONG_MAX};
		compartment[ncid]->setPosition(pos);
	} else
		ncid = cid;

	if(ruleset[rsid]->isEventRuleSet()) {
		// add event to compartment
		addEventToCompartment(rsid, ncid);
	} else {
		// add ruleset to compartment
		compartment[ncid]->addRuleSet(rsid);

		// set pointer to ruleset in compartment
		nrs = compartment[ncid]->getNumRuleSets();
		compartment[ncid]->setRuleSet(nrs - 1, ruleset[rsid]);

		// initialise valid rules
		compartment[ncid]->initialiseValidRules(nrs - 1, \
				ruleset[rsid]->getNumRules());

		// set valid rules
		for(i = 0; i < ruleset[rsid]->getNumRules(); i++)
			compartment[ncid]->setValidRule(nrs - 1, i, i);

		// add child rulesets if any (recursive)
		if(ruleset[rsid]->getNumChildRuleSets())
			for(i = 0; i < ruleset[rsid]->getNumChildRuleSets(); i++)
				addRuleSetToCompartment(ruleset[rsid]->getChildRuleSet(i), ncid);
	}
}

// add an event to a compartment
void Psystem::addEventToCompartment(unsigned long rsid, unsigned long cid)
{
	compartment[cid]->addEvent(ruleset[rsid], alphabet);
}

// set target for translocation rule
void Psystem::setRuleTarget(unsigned long cid, unsigned long rsid, \
		unsigned long rid, unsigned long target_cid)
{
	unsigned long i, rsind, updates[1];

	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid source compartment\n";
		exit(EXIT_FAILURE);
	}
	if(target_cid < 0 || target_cid >= num_compartments) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid source compartment\n";
		exit(EXIT_FAILURE);
	}
	if(rsid < 0 || rsid >= num_rulesets) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid ruleset\n";
		exit(EXIT_FAILURE);
	}

	// find index of ruleset in compartment
	for(i = 0; i < compartment[cid]->getNumRuleSets(); i++)
		if(compartment[cid]->getRuleSetIndex(i) == rsid)
			break;
	if(i == compartment[cid]->getNumRuleSets()) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - ruleset not assigned to compartment\n";
		exit(EXIT_FAILURE);
	}

	// check rule id valid
	if(rid < 0 || rid >= compartment[cid]->getNumRules(i)) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid rule\n";
		exit(EXIT_FAILURE);
	}

	// set update for rule
	updates[0] = target_cid;
	compartment[cid]->setUpdates(i, rid, updates, 1);
}

// set simulation algorithm
void Psystem::setSimulationAlgorithm(string name, unsigned long seed)
{
	// set simulation algorithm
	if(name == "dm")
		setSimulationAlgorithm(new MulticompartmentGillespie(seed));
	if(name == "ldm")
		setSimulationAlgorithm(new LogarithmicDirectMethod(seed));
	if(name == "dmgd")
		setSimulationAlgorithm(new MulticompartmentGillespieGD(seed));
	if(name == "dmq2gd")
		setSimulationAlgorithm(new MulticompartmentGillespieQ2GD(seed));
	if(name == "dmq")
		setSimulationAlgorithm(new MulticompartmentGillespieQ(seed));
	if(name == "dmq2")
		setSimulationAlgorithm(new MulticompartmentGillespieQ2(seed));
	if(name == "dmqg")
		setSimulationAlgorithm(new MulticompartmentGillespieQG(seed));
	if(name == "dmq2g")
		setSimulationAlgorithm(new MulticompartmentGillespieQ2G(seed));
	if(name == "dmcp")
		setSimulationAlgorithm(new \
				MulticompartmentGillespieCellularPotts(seed));
	if(simalg == NULL) {
		cerr << "error: unknown simulation algorithm" << name << "\n";
		exit(EXIT_FAILURE);
	}

	// set simulation algorithm parameters
	simalg->setPsystem(this);
}

// set level of species in compartment
void Psystem::setSpeciesLevel(unsigned long cid, unsigned long sid, \
		unsigned long level)
{
	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot set level of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}
	if(sid < 0 || sid >= num_species) {
		cerr << "error: cannot set level of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid species id\n";
		exit(EXIT_FAILURE);
	}

	// set level of species
	compartment[cid]->setLevel(sid, level);

	// recalculate waiting times if simulation running
//	if(simalg && simalg->getTimeToNextReaction() != 0.0) {
//		if(simalg->getGlobalTimeToNextReaction() != 0.0)
//			simalg->setGlobalTimeToNextReaction(simalg->getTime());
//		simalg->preprocess();
//	}
}

// get level of species in compartment at timepoint
unsigned long Psystem::getSpeciesLevel(unsigned long cid, \
		unsigned long sid, unsigned int timepoint)
{
	// get concentration of species
	return simalg->getAmountFromBuffer(cid, sid, timepoint);
}

// set concentration of species in compartment
void Psystem::setSpeciesConcentration(unsigned long cid, unsigned long sid, \
		double concentration)
{
	unsigned long level;
	
	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot set concentration of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}
	if(sid < 0 || sid >= num_species) {
		cerr << "error: cannot set concentration of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid species id\n";
		exit(EXIT_FAILURE);
	}

	// set level of species
	level = (unsigned int) ceil((double) concentration * \
			getCompartmentVolume(cid) * GSL_CONST_NUM_AVOGADRO);
	compartment[cid]->setLevel(sid, level);

	// recalculate waiting times if simulation running
//	if(simalg && simalg->getTimeToNextReaction() != 0.0) {
//		if(simalg->getGlobalTimeToNextReaction() != 0.0)
//			simalg->setGlobalTimeToNextReaction(simalg->getTime());
//		simalg->preprocess();
//	}
}

// get concentration of species in compartment
double Psystem::getSpeciesConcentration(unsigned long cid, unsigned long sid)
{
	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot get concentration of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}
	if(sid < 0 || sid >= num_species) {
		cerr << "error: cannot get concentration of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid species id\n";
		exit(EXIT_FAILURE);
	}

	// get concentration of species
	return ((double) compartment[cid]->getLevel(sid) / \
			(getCompartmentVolume(cid) * GSL_CONST_NUM_AVOGADRO));
}

// get concentration of species in compartment at timepoint
double Psystem::getSpeciesConcentration(unsigned long cid, unsigned long sid, \
		unsigned int timepoint)
{
	// get concentration of species
	return ((double) simalg->getAmountFromBuffer(cid, sid, timepoint) / \
			(getCompartmentVolume(cid) * GSL_CONST_NUM_AVOGADRO));
}

// clone rule from ruleset in compartment
unsigned long Psystem::cloneRule(unsigned long cid, unsigned long rsid, \
		unsigned long rid)
{
	unsigned long i;

	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid source compartment\n";
		exit(EXIT_FAILURE);
	}
	if(rsid < 0 || rsid >= num_rulesets) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid ruleset\n";
		exit(EXIT_FAILURE);
	}
	if(rid < 0 || rid >= ruleset[rsid]->getNumRules()) {
		cerr << "error: cannot set rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid rule\n";
		exit(EXIT_FAILURE);
	}

	// find index of ruleset in compartment
	for(i = 0; i < compartment[cid]->getNumRuleSets(); i++)
		if(compartment[cid]->getRuleSetIndex(i) == rsid)
			break;
	if(i == compartment[cid]->getNumRuleSets()) {
		cerr << "error: cannot clone rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - ruleset not assigned to compartment\n";
		exit(EXIT_FAILURE);
	}

	// add rule to ruleset
	compartment[cid]->addValidRule(i, rid);

	return (compartment[cid]->getNumRules(i) - 1);
}

// set reaction constant of rule in compartment
void Psystem::setReactionConstant(unsigned long cid, unsigned long rsid, \
		unsigned long rid, double v)
{
	unsigned long i;

	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot set reaction constant rule id=" << rid << \
				" target in compartment id=" << cid << \
				" - invalid source compartment\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < compartment[cid]->getNumRuleSets(); i++)
		if(compartment[cid]->getRuleSetIndex(i) == rsid)
			break;
	if(i == compartment[cid]->getNumRuleSets()) {
		cerr << "error: cannot set reactions constant id=" << rid << \
				" target in compartment id=" << cid << \
				" - ruleset not assigned to compartment\n";
		exit(EXIT_FAILURE);
	}
	if(rid < 0 || rid >= compartment[cid]->getNumRules(i)) {
		cerr << "error: cannot set reaction constant id=" << rid << \
				" target in compartment id=" << cid << \
				" - rule not assigned to compartment\n";
		exit(EXIT_FAILURE);
	}

	compartment[cid]->setReactionConstant(i, rid, v);
}

// get ruleset index by name
unsigned long Psystem::getRuleSetIndex(string s)
{
	unsigned long i;

	for(i = 0; i < num_rulesets; i++)
		if(s == ruleset[i]->getName().substr(0, \
					ruleset[i]->getName().find(':')))
			break;

	return i;
}

// set initial levels given in templates modules
void Psystem::setInitialLevels(void)
{
	unsigned long i, j, k, ind, ind2, val;

	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++) {
			ind = compartment[i]->getRuleSetIndex(j);
			for(k = 0; k < ruleset[ind]->getNumRules(); k++) {
				for(ind2 = 0; ind2 < ruleset[ind]->getNumReactants(k); ind2++)
					if((val = ruleset[ind]->getRule(k)->getInitialReactant(ind2)))
						compartment[i]->setLevel(ruleset[ind]-> \
								getReactantIndex(ind2, k), val);
				for(ind2 = 0; ind2 < ruleset[ind]->getNumProducts(k); ind2++)
					if((val = ruleset[ind]->getRule(k)->getInitialProduct(ind2)))
						compartment[i]->setLevel(ruleset[ind]-> \
								getProductIndex(ind2, k), val);
			}
		}
}

// get rule vector from SBML
bool Psystem::getRuleVectorFromSBML(string n, long int *target)
{
	int ind;
	long int val;
	string s1, s2;
	char *endptr;

	// find second field in name
	if((ind = n.find(":")) == string::npos)
		return false;
	// find comma
	s1 = n.substr(ind + 1);
	if((ind = s1.find(",")) == string::npos)
		return false;
	// get x position
	s2 = s1.substr(0, ind);
	// check integer
	errno = 0;
	val = strtol(s2.c_str(), &endptr, 10);
	if((errno == ERANGE && val == LONG_MAX) || (errno != 0 && val == 0))
		return false;
	if(endptr == s2.c_str() || *endptr != '\0')
		return false;
	target[0] = val;
	// get y position
	s2 = s1.substr(ind + 1);
	// check integer
	errno = 0;
	val = strtol(s2.c_str(), &endptr, 10);
	if((errno == ERANGE && val == LONG_MAX) || (errno != 0 && val == 0))
		return false;
	if(endptr == s2.c_str() || *endptr != '\0')
		return false;
	target[1] = val;
	// all ok
	return true;
}

// get rule target from SBML
string Psystem::getRuleTargetFromSBML(string n)
{
	int ind;
	string s;

	s.clear();
	// find second field in name
	if((ind = n.find(":")) == string::npos)
		return s;
	// find comma
	s = n.substr(ind + 1);
	if((ind = s.find(",")) == string::npos)
		return s;
	// get target
	return s.substr(0, ind);
}

// get rule target type from SBML
string Psystem::getRuleTargetTypeFromSBML(string n)
{
	int ind;
	string s;

	s.clear();
	// find second field in name
	if((ind = n.find(":")) == string::npos)
		return s;
	// find comma
	s = n.substr(ind + 1);
	if((ind = s.find(",")) == string::npos) {
		s.clear();
		return s;
	}
	// get target
	return s.substr(ind + 1);
}

// get parent compartment name from SBML
string Psystem::getParentCompartmentFromSBML(string n)
{
	int ind;
	string s;

	s.clear();
	if((ind = n.find(':')) == string::npos)
		return s;
	if((ind = n.find(':', ind + 1)) == string::npos)
		return s;
	if((ind = n.find(':', ind + 1)) == string::npos)
		return s;
	if(n[ind + 1] == '\0')
		return s;
	return n.substr(ind + 1);
}

// create flattened representation of multicompartment model
void Psystem::flatten(void)
{
	unsigned long i, j, k, ind;
	
	// initialise data structures for flattened model
	// total number of species in flat model
	num_species_flat = num_species * num_compartments;

	// calculate total number of rules in flat model
	num_rules_flat = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			num_rules_flat += compartment[i]->getNumRules(j);
		
	// lookup table for species indices (flat to multicomp index)
	if(flat_species_species_lookup)
		free(flat_species_species_lookup);
	if(!(flat_species_species_lookup = (unsigned long *) \
				calloc(num_species_flat, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for flat species\n";
		exit(EXIT_FAILURE);
	}
	// populate lookup table
	ind = 0;
	for(i = 0; i < num_species_flat; i++) {
		if(ind == num_species)
			ind = 0;
		flat_species_species_lookup[i] = ind++;
	}
	
	// lookup table for compartment indices (flat to multicomp index)
	if(flat_species_compartment_lookup)
		free(flat_species_compartment_lookup);
	if(!(flat_species_compartment_lookup = (unsigned long *) \
				calloc(num_species_flat, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for flat compartments\n";
		exit(EXIT_FAILURE);
	}
	// populate lookup table
	ind = 0;
	for(i = 0; i < num_species_flat; i++) {
		if((i % num_species) == 0)
			ind++;
		flat_species_compartment_lookup[i] = ind - 1;
	}
	
	// lookup table for rule compartment indices (flat to multicomp index)
	if(flat_rule_compartment_lookup)
		free(flat_rule_compartment_lookup);
	if(!(flat_rule_compartment_lookup = (unsigned long *) \
				calloc(num_rules_flat, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for flat compartments\n";
		exit(EXIT_FAILURE);
	}
	// lookup table for rule ruleset indices (flat to multicomp index)
	if(flat_rule_ruleset_lookup)
		free(flat_rule_ruleset_lookup);
	if(!(flat_rule_ruleset_lookup = (unsigned long *) \
				calloc(num_rules_flat, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for flat rulesets\n";
		exit(EXIT_FAILURE);
	}
	// lookup table for rule rule indices (flat to multicomp index)
	if(flat_rule_rule_lookup)
		free(flat_rule_rule_lookup);
	if(!(flat_rule_rule_lookup = (unsigned long *) \
				calloc(num_rules_flat, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for flat rules\n";
		exit(EXIT_FAILURE);
	}
	// populate lookup tables
	ind = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < compartment[i]->getNumRuleSets(); j++)
			for(k = 0; k < compartment[i]->getNumRules(j); k++) {
				flat_rule_compartment_lookup[ind] = i;
				flat_rule_ruleset_lookup[ind] = j;
				flat_rule_rule_lookup[ind] = k;
				ind++;
			}

	// create vector of initial amounts
	flat_initial_amounts.clear();
	for(i = 0; i < num_species_flat; i++)
		flat_initial_amounts.push_back((int) getSpeciesLevel( \
				flat_species_species_lookup[i], \
				flat_species_compartment_lookup[i]));
}

// get number of species in flattened representation
unsigned long Psystem::getNumSpeciesFlat(void)
{
	return num_species_flat;
}

// get number of reactions in flattened representation
unsigned long Psystem::getNumReactionsFlat(void)
{
	return num_rules_flat;
}

// get initial state vector
std::vector<int>& Psystem::getInitialSpeciesFlat(void)
{
    return flat_initial_amounts;
}	

// get reaction constant in flattened representation
double Psystem::getReactionConstantFlat(unsigned long rid)
{
	return getReactionConstant(flat_rule_compartment_lookup[rid], \
			flat_rule_ruleset_lookup[rid], \
			flat_rule_rule_lookup[rid]);
}

// get reaction reactants in flattened representation
std::vector<int> Psystem::getReactantsFlat(unsigned long rid)
{
	std::vector<int> species;
	unsigned long i, c, r, rs, nr;
	
	c = flat_rule_compartment_lookup[rid];
	rs = compartment[c]->getRuleSetIndex(flat_rule_ruleset_lookup[rid]);
	r = compartment[c]->getRuleIndex(flat_rule_ruleset_lookup[rid], \
			flat_rule_rule_lookup[rid]);
	nr = ruleset[rs]->getNumReactants(r);
	for(i = 0; i < nr; i++) 
		species.push_back(c * num_species + ruleset[rs]->getReactantIndex(i, r));

	return species;
}

// get reaction products in flattened representation
std::vector<int> Psystem::getProductsFlat(unsigned long rid)
{
	std::vector<int> species;
	unsigned long i, c, r, rs, nr;
	
	c = flat_rule_compartment_lookup[rid];
	rs = compartment[c]->getRuleSetIndex(flat_rule_ruleset_lookup[rid]);
	r = compartment[c]->getRuleIndex(flat_rule_ruleset_lookup[rid], \
			flat_rule_rule_lookup[rid]);
	if(isChannel(c, rs, r))
		c = getTargetCompartment(c, rs, r);
	nr = ruleset[rs]->getNumProducts(r);
	for(i = 0; i < nr; i++) 
		species.push_back(c * num_species + ruleset[rs]->getProductIndex(i, r));

	return species;
}
