// ruleset class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "RuleSet.h"

#include <iostream>
#include <stdio.h>
#include <stdlib.h>

using namespace Simulation;

// constructor
RuleSet::RuleSet()
{
	// set initial values
	setInitialisationValues();
}

// constructor
RuleSet::RuleSet(const unsigned long nr)
{
	// set initial values
	setInitialisationValues();

	// initialise rule storage
	num_rules = nr;
	if(!(rule = (Rule **) calloc(num_rules, sizeof(Rule *)))) {
		std::cerr << "error: couldn't allocate memory for ruleset\n";
		exit(EXIT_FAILURE);
	}
	if(!(acum = (double *) calloc(num_rules, sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for ruleset\n";
		exit(EXIT_FAILURE);
	}
}

// set default values for initialised class
void RuleSet::setInitialisationValues(void)
{
	num_rules = 0;
	rule = NULL;
	acum = NULL;
	id = "";
	name = "";
	index = 0;
	density_ruleset = false;
	growth_ruleset = false;
	event_ruleset = false;
	volume = 1.0;
	num_child_rulesets = 0;
	child_ruleset = NULL;
	enclosed_compartment = false;
	parent_name = "";
	event_trigger = "";
	event_assignment_variable = "";
	event_assignment_expression = "";
}

// destructor
RuleSet::~RuleSet()
{
	unsigned int i;

	// free memory
	if(rule) {
		for(i = 0; i < num_rules; i++)
			if(rule[i])
				delete rule[i];
		free(rule);
	}
	if(acum)
		free(acum);
	if(child_ruleset)
		free(child_ruleset);
}

// general access methods
unsigned long RuleSet::getNumRules(void) { return num_rules; }
Simulation::Rule *RuleSet::getRule(const unsigned long i) { return rule[i]; }
void RuleSet::setRule(const unsigned long i, Rule *r) { rule[i] = r; }
std::string RuleSet::getId(void) { return id; }
void RuleSet::setId(std::string s) { id = s; }
std::string RuleSet::getName(void) { return name; }
void RuleSet::setName(std::string s) { name = s; }
void RuleSet::setIndex(const unsigned long i) { index = i; }
unsigned long RuleSet::getIndex(void) { return index; }
double *RuleSet::getCumulativePropensities(void) { return acum; }
unsigned long RuleSet::getReactantIndex(unsigned long ind, unsigned long i) { \
		return rule[i]->getReactantIndex(ind); }
unsigned long RuleSet::getProductIndex(unsigned long ind, unsigned long i) { \
		return rule[i]->getProductIndex(ind); }
int RuleSet::isChannel(unsigned long i) { \
		return rule[i]->isChannel(); }
const char *RuleSet::getRuleName(unsigned long i) { return rule[i]->getNameStr(); }
unsigned long RuleSet::getNumReactants(unsigned long i) \
		{ return rule[i]->getNumReactants(); }
unsigned long RuleSet::getNumProducts(unsigned long i) \
		{ return rule[i]->getNumProducts(); }
long int RuleSet::getRuleXVector(unsigned long i) \
		{ return rule[i]->getXVector(); }
long int RuleSet::getRuleYVector(unsigned long i) \
		{ return rule[i]->getYVector(); }
long int RuleSet::getRuleZVector(unsigned long i) \
		{ return rule[i]->getZVector(); }
double RuleSet::getPropensity(unsigned long i) \
		{ return rule[i]->getPropensity(); }
void RuleSet::setDensityRuleSet(bool b) { density_ruleset = b; }
bool RuleSet::isDensityRuleSet(void) { return density_ruleset; }
void RuleSet::setGrowthRuleSet(bool b) { growth_ruleset = b; }
bool RuleSet::isGrowthRuleSet(void) { return growth_ruleset; }
void RuleSet::setEventRuleSet(bool b) { event_ruleset = b; }
bool RuleSet::isEventRuleSet(void) { return event_ruleset; }
void RuleSet::setVolume(double v) { volume = v; }
double RuleSet::getVolume(void) { return volume; }
unsigned long RuleSet::getNumChildRuleSets(void) { return num_child_rulesets; }
unsigned long *RuleSet::getChildRuleSets(void) { return child_ruleset; }
unsigned long RuleSet::getChildRuleSet(unsigned long i) \
		{ return child_ruleset[i]; }
void RuleSet::setChildRuleSet(unsigned long i, unsigned long v) \
		{ child_ruleset[i] = v; }
bool RuleSet::isEnclosedCompartment(void) { return enclosed_compartment; }
void RuleSet::setEnclosedCompartment(bool b) { enclosed_compartment = b; }
std::string RuleSet::getParentCompartmentName(void) { return parent_name; }
void RuleSet::setParentCompartmentName(std::string s) { parent_name = s; }
std::string RuleSet::getEventTrigger(void) { return event_trigger; }
void RuleSet::setEventTrigger(std::string s) { event_trigger = s; }
std::string RuleSet::getEventAssignmentVariable(void) \
		{ return event_assignment_variable; }
void RuleSet::setEventAssignmentVariable(std::string s) \
		{ event_assignment_variable = s; }
std::string RuleSet::getEventAssignmentExpression(void) \
		{ return event_assignment_expression; }
void RuleSet::setEventAssignmentExpression(std::string s) \
		{ event_assignment_expression = s; }

void RuleSet::show(void)
{
	register unsigned long i;

	std::cout << "id=" << id << " index=" << index << " size=" << num_rules \
			<< "\n";
	for(i = 0; i < num_rules; i++) {
		std::cout << "\trule " << i << ": ";
		rule[i]->show();
	}
}

// append rule to a ruleset
unsigned long RuleSet::addRule(Rule *r)
{
	// check if first rule added
	num_rules++;
	if(!num_rules) {
		// allocate memory
		if(!(rule = (Rule **) calloc(num_rules, sizeof(Rule *)))) {
			std::cerr << "error: couldn't allocate memory for ruleset\n";
			exit(EXIT_FAILURE);
		}
		if(!(acum = (double *) calloc(num_rules, sizeof(double)))) {
			std::cerr << "error: couldn't allocate memory for ruleset\n";
			exit(EXIT_FAILURE);
		}
	} else {
		// reallocate memory
		if(!(rule = (Rule **) realloc(rule, num_rules * sizeof(Rule *)))) {
			std::cerr << "error: couldn't allocate memory for ruleset\n";
			exit(EXIT_FAILURE);
		}
		if(!(acum = (double *) realloc(acum, num_rules * sizeof(double)))) {
			std::cerr << "error: couldn't allocate memory for ruleset\n";
			exit(EXIT_FAILURE);
		}
	}

	// add rule
	rule[num_rules - 1] = r;

	return (num_rules - 1);
}

// set number of child rulesets
void RuleSet::setNumChildRuleSets(unsigned long nrs)
{
	num_child_rulesets = nrs;
	if(nrs) {
		if(child_ruleset) {
			if(!(child_ruleset = (unsigned long *) realloc(child_ruleset, nrs \
					* sizeof(unsigned long)))) {
				std::cerr << "error: couldn't reallocate memory for child " \
						"rulesets\n";
				exit(EXIT_FAILURE);
			}
		} else {
			if(!(child_ruleset = (unsigned long *) calloc(nrs, \
					sizeof(unsigned long)))) {
				std::cerr << "error: couldn't allocate memory for child " \
						"rulesets\n";
				exit(EXIT_FAILURE);
			}
		}
	}
}
