// simulation algorithm class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "SimulationAlgorithm.h"
#include "Psystem.h"

#include <iostream>
#include <stdio.h>
#include <gsl/gsl_sf_pow_int.h>
#include <gsl/gsl_sf_log.h>
#include <gsl/gsl_sf_gamma.h>
#include <math.h>
#include <time.h>
#include <ecsb/ecsb.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

using namespace Simulation;
using namespace ECSB;
using namespace SimulationDatasets;

// constructor
SimulationAlgorithm::SimulationAlgorithm(const char *s, \
		const unsigned long seed) {
	setName(s);  // set name of algorithm
	time = 0.0;  // set simulation time to 0
	tau0 = 0.0;
	tau0global = 0.0;
	num_reactions = 0;  // set reactions performed to 0
	log_type = LOG_TYPE_NONE;
	log_levels = true;
	log_propensities = false;
	log_volumes = false;
	log_memory = false;
	log_steady_state = true;
	log_lattice = false;
	num_timepoints = 0;
	show_progress = false;
	progress_interval = 0;
	last_progress_time = (unsigned int) std::time(NULL);
	num_runs = 1;
	num_rules = total_rules = 0;
	scale_translocation = false;
	first_time = true;
	volume_rate = 1;
	division_rate = 1;
	potts_rate = 1;
	static_lattice = false;
	events_defined = false;
	mpi_serial = false;
	num_event_compartments = 0;
	event_compartment = NULL;
	ode_algorithm = (gsl_odeiv_step_type *) gsl_odeiv_step_rk8pd;
	ode_eps_abs = 1.0E-6;
	ode_eps_rel = 0.0;
	ode_initial_step_size = 1.0E-6;
	ode_yscaling = 1.0;
	ode_dyscaling = 0.0;

	// initialise uniform random number generator
	rng = new RNG(seed);
	SimulationAlgorithm::seed = rng->getSeed();

	// initialise data buffers
	amount_chunk_size = 1000;
	amount_index = 0;
	amount_offset = 0;
	amount_buffer = NULL;
	volume_chunk_size = 1000;
	volume_index = 0;
	volume_offset = 0;
	volume_buffer = NULL;
	compartment_chunk_size = 1000;
	compartment_index = 0;
	reaction_chunk_size = 1000;
	reaction_index = 0;
	reaction_offset = 0;
	reaction_buffer = NULL;
	propensity_chunk_size = 1;
	propensity_index = 0;
	position_chunk_size = 1000;
	position_index = 0;
	lattice_spin_chunk_size = 1000;
	lattice_spin_index = 0;
	lattice_spin_offset = 0;
	lattice_spin_buffer = NULL;
	lattice_type_chunk_size = 1000;
	lattice_type_index = 0;
	lattice_type_offset = 0;
	lattice_type_buffer = NULL;
	#ifdef HAVE_MPI
	// set mpi thread id
	MPI_Comm_rank(MPI_COMM_WORLD, &tid);
	MPI_Comm_size(MPI_COMM_WORLD, &nthreads);
	#endif

	// hdf5 dataset compression
	compress = false;
	compression_level = 9;

	// setup signal handler
	pthread_mutex_init(&execute_running_mutex, NULL);
	execute_running = false;
	pthread_mutex_init(&hdf5_mutex, NULL);
}

// destructor
SimulationAlgorithm::~SimulationAlgorithm()
{
	// release random number generator
	delete rng;

	// free memory
	if(log_memory) {
		free3DArrayUInt(amount_buffer);
	}

	if(event_compartment)
		free(event_compartment);

	// signal handler
	pthread_mutex_destroy(&execute_running_mutex);
	pthread_mutex_destroy(&hdf5_mutex);
}

// general access methods
std::string SimulationAlgorithm::getName(void) { return name; }
void SimulationAlgorithm::setName(std::string s) { name = s; }
double SimulationAlgorithm::getTime(void) { return time; }
void SimulationAlgorithm::setTime(double t) { time = t; }
void SimulationAlgorithm::advanceTime(double t) { time += t; }
Psystem *SimulationAlgorithm::getPsystem(void) { return psystem; }
double SimulationAlgorithm::getTimeToNextReaction(void) { return tau0; }
double SimulationAlgorithm::getGlobalTimeToNextReaction(void) \
		{ return tau0global; }
void SimulationAlgorithm::setGlobalTimeToNextReaction(double t)
		{ tau0global = t; }
unsigned long SimulationAlgorithm::getNextCompartment(void) { return i0; }
unsigned long SimulationAlgorithm::getNextRuleset(void) { return rs0; }
unsigned long SimulationAlgorithm::getNumCompartments(void) { \
		return num_compartments; }
void SimulationAlgorithm::setNumCompartments(unsigned long i) { \
		num_compartments = i; }
void SimulationAlgorithm::setNumSpecies(unsigned long i) { \
		num_species = i; }
Simulation::Compartment *SimulationAlgorithm::getCompartment(unsigned long i) \
		{ return psystem->getCompartment(i); }
unsigned long SimulationAlgorithm::getLogType(void) { \
		return log_type; }
void SimulationAlgorithm::setLogType(unsigned long i) { \
		log_type = i; }
void SimulationAlgorithm::incrementNumReactions(void) { num_reactions++; }
hid_t SimulationAlgorithm::getDataFile(void) { return dfile; }
unsigned long SimulationAlgorithm::getNumTimepoints(void) { \
		return num_timepoints; }
void SimulationAlgorithm::setNumTimepoints(unsigned long i) { \
		num_timepoints = i; }
unsigned int ***SimulationAlgorithm::getAmountsDataset(void) \
		{ return amount_buffer; }
void SimulationAlgorithm::setShowProgress(bool b) \
		{ show_progress = b; }
void SimulationAlgorithm::setDivisionDirection(char c)
		{ division_direction = c; }
void SimulationAlgorithm::setKeepDivisions(bool b)
		{ keep_divisions = b; }
void SimulationAlgorithm::setCompressionLevel(unsigned int i)
		{ compression_level = i; }
void SimulationAlgorithm::setScaleTranslocation(bool b) \
		{ scale_translocation = b; }
double SimulationAlgorithm::getLogInterval(void) { return log_interval; }
void SimulationAlgorithm::setStaticLattice(bool b) { static_lattice = b; }
void SimulationAlgorithm::setEventsDefined(bool b) { events_defined = b; }
void SimulationAlgorithm::setMpiSerial(bool b) { mpi_serial = b; }
RNG *SimulationAlgorithm::getRNG(void) { return rng; }
#ifdef HAVE_MPI
int SimulationAlgorithm::getThreadID(void) { return tid; }
#endif
void SimulationAlgorithm::setOdeSolver(const gsl_odeiv_step_type *alg) \
		{ ode_algorithm = (gsl_odeiv_step_type *) alg; }
void SimulationAlgorithm::setAbsoluteError(double v) \
		{ ode_eps_abs = v; }
void SimulationAlgorithm::setRelativeError(double v) \
		{ ode_eps_rel = v; }
void SimulationAlgorithm::setInitialStepSize(double v) \
		{ ode_initial_step_size = v; }
void SimulationAlgorithm::setYScaling(double v) \
		{ ode_yscaling = v; }
void SimulationAlgorithm::setDYScaling(double v) \
		{ ode_dyscaling = v; }

// entry point for thread to execute simulation (no logging)
void *SimulationAlgorithm::execute_no_logging_thread(void *args)
{
	SimulationAlgorithm *sa = (SimulationAlgorithm *) args;

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to defer cancel
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

	// execute simulation
	sa->execute_no_logging();

	pthread_exit(NULL);
}

// execute simulation (no logging)
void SimulationAlgorithm::execute_no_logging(void)
{
	register int ret;
	register double next_time;

	#ifndef WIN32
	// set thread as running
	pthread_mutex_lock(&execute_running_mutex);
	execute_running = true;
	pthread_mutex_unlock(&execute_running_mutex);
	#endif

	// iterate simulation
	while((next_time = time + tau0) <= end_time && ret) {
		// execute next rule
		ret = execute();

		#ifndef WIN32
		// test if cancelled
		pthread_testcancel();
		#endif
	}

	#ifndef WIN32
	pthread_mutex_lock(&execute_running_mutex);

	simulation_status = SIMULATION_FINISHED;

	// set thread as running
	execute_running = false;

	// send USR2 signal to parent to indicate finished
	pthread_kill(ppid, SIGUSR2);

	pthread_mutex_unlock(&execute_running_mutex);
	#endif
}

// entry point for thread to execute simulation (logging)
void *SimulationAlgorithm::execute_logging_thread(void *args)
{
	SimulationAlgorithm *sa = (SimulationAlgorithm *) args;

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to defer cancel
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

	// execute simulation
	sa->execute_logging();

	pthread_exit(NULL);
}

// execute simulation (logging)
void SimulationAlgorithm::execute_logging(void)
{
	register int ret;
	register double current_interval, next_time;

	#ifndef WIN32
	// set thread as running
	pthread_mutex_lock(&execute_running_mutex);
	execute_running = true;
	pthread_mutex_unlock(&execute_running_mutex);
	#endif

	// iterate simulation
	current_interval = log_interval;  // interval for logging levels
	num_reactions = 0;  // reset number of reactions

	ret = 1;
	while((next_time = time + tau0) <= end_time && ret) {
		// write data to log for intermediate timesteps
		while(log_levels && next_time > current_interval) {
			writeAmount(current_interval);
			if(log_propensities)
				writePropensities(current_interval);
			if(log_volumes)
				writeVolume(current_interval);
			if(log_lattice) {
				writeLatticeSpin(current_interval);
				writeLatticeType(current_interval);
			}
			current_interval += log_interval;
		}

		// execute next rule
		ret = execute();

		#ifndef WIN32
		// test if cancelled
		pthread_testcancel();
		#endif
	}

	#ifndef WIN32
	pthread_mutex_lock(&execute_running_mutex);
	#endif

	// log up to max_time if specified
	if(log_steady_state) {
		while(log_levels && end_time >= current_interval) {
			writeAmount(current_interval);
			if(log_propensities)
				writePropensities(current_interval);
			if(log_volumes)
				writeVolume(current_interval);
			if(log_lattice) {
				writeLatticeSpin(current_interval);
				writeLatticeType(current_interval);
			}
			current_interval += log_interval;
		}
		time = end_time;
	} else {
		if(log_levels) {
			writeAmount(current_interval);
			if(log_propensities)
				writePropensities(current_interval);
			if(log_volumes)
				writeVolume(current_interval);
			if(log_lattice) {
				writeLatticeSpin(current_interval);
				writeLatticeType(current_interval);
			}
		}
		current_interval += log_interval;
	}

	// save statistics
	if(!log_memory) {
		writeAttribute("total_reactions_simulated", num_reactions, \
				dfile, run_group);
		writeAttribute("simulated_time", current_interval - \
				log_interval, dfile, run_group);
	}

	#ifndef WIN32
	simulation_status = SIMULATION_FINISHED;

	// set thread as running
	execute_running = false;

	// send USR2 signal to parent to indicate finished
	pthread_kill(ppid, SIGUSR2);

	pthread_mutex_unlock(&execute_running_mutex);
	#endif
}

// iterate p-system for specified time and runs
void SimulationAlgorithm::execute(double end_time)
{
	register int ret;
	register unsigned long from_run, to_run;
	char timestamp[256];
	pthread_t tid_execute;
	pthread_attr_t tattr;
	sigset_t sig_mask;  // current signal mask
	int sig;

	#ifndef WIN32
	// block all signals
	sigfillset(&sig_mask);
	pthread_sigmask(SIG_SETMASK, &sig_mask, &sig_mask_old);
	#endif

	SimulationAlgorithm::end_time = end_time;

	if(log_type == LOG_TYPE_NONE && num_runs == 1) {
		if(first_time) {
			// preprocess p system
			ret = psystem->preprocess();

			first_time = false;
		} else
			ret = 1;

		// finish if no reactions left to execute
		if(tau0 == 0.0)
			return;

		#ifdef WIN32
		execute_no_logging();
		#else
		// iterate simulation
		simulation_status = SIMULATION_CANCELLED;
		ppid = pthread_self();

		// run simulation in thread
		pthread_attr_init(&tattr);
		pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
		if(pthread_create(&tid_execute, &tattr, execute_no_logging_thread, \
				(void *) this)) {
			cerr << "error: unable to create execute thread\n";
			exit(EXIT_FAILURE);
		}
		// tidy up
		pthread_attr_destroy(&tattr);

		// wait for SIGUSR2 signal
		sigemptyset(&sig_mask);
		sigaddset(&sig_mask, SIGUSR2);
		sigwait(&sig_mask, &sig);

		pthread_mutex_lock(&execute_running_mutex);
		switch(simulation_status) {
			case SIMULATION_FINISHED:
				pthread_mutex_unlock(&execute_running_mutex);
				// wait for thread to finish
				pthread_join(tid_execute, NULL);
				break;

			case SIMULATION_CANCELLED:
				if(execute_running)
					pthread_cancel(tid_execute);
				pthread_mutex_unlock(&execute_running_mutex);
				// wait for thread to finish
				pthread_join(tid_execute, NULL);
				simulation_status = SIMULATION_CANCELLED;
				break;

			default:
				cerr << "error: unknown thread state\n";
				exit(EXIT_FAILURE);
				break;
		}
		#endif

		// all done
		return;
	}

	#ifdef HAVE_MPI
	if(mpi_serial) {
		from_run = 1;
		to_run = num_runs;
	} else {
		// calculate number of runs for each thread
		unsigned long total_runs, first_thread_runs;
		if(nthreads > num_runs) {
			total_runs = 1;
			first_thread_runs = 0;
			if(tid == 0) {
				from_run = 1;
				to_run = 0;
			} else {
				from_run = to_run = tid;
			}
		} else {
			if((num_runs % nthreads)) {
				#ifdef WIN32
				total_runs = (unsigned long) ceill((double) num_runs / \
						nthreads);
				#else
				total_runs = (unsigned long) ceill((long double) num_runs / \
						nthreads);
				#endif
				first_thread_runs = total_runs * nthreads - num_runs;
			} else {
				total_runs = num_runs / nthreads;
				first_thread_runs = total_runs;
			}
			if(tid == 0) {
				from_run = 1;
				to_run = from_run + first_thread_runs - 1;
			} else {
				from_run = first_thread_runs + (tid - 1) * total_runs + 1;
				to_run = from_run + total_runs - 1;
			}
		}
	}
	#else
	from_run = 1;
	to_run = num_runs;
	#endif
	if(log_type == LOG_TYPE_NONE) {
		for(run = from_run; run <= to_run; run++) {
			// reset simulation
			resetSimulation();
	
			// initialise simulation
			ret = psystem->preprocess();
			if(!ret)
				break;

			#ifdef WIN32
			execute_no_logging();
			#else
			simulation_status = SIMULATION_CANCELLED;
			ppid = pthread_self();

			// run simulation in thread
			pthread_attr_init(&tattr);
			pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
			if(pthread_create(&tid_execute, &tattr, execute_no_logging_thread, \
					(void *) this)) {
				cerr << "error: unable to create execute thread\n";
				exit(EXIT_FAILURE);
			}
			// tidy up
			pthread_attr_destroy(&tattr);

			// wait for SIGUSR2 signal
			sigemptyset(&sig_mask);
			sigaddset(&sig_mask, SIGUSR2);
			sigwait(&sig_mask, &sig);

			pthread_mutex_lock(&execute_running_mutex);
			switch(simulation_status) {
				case SIMULATION_FINISHED:
					pthread_mutex_unlock(&execute_running_mutex);
					// wait for thread to finish
					pthread_join(tid_execute, NULL);
					break;

				case SIMULATION_CANCELLED:
					if(execute_running)
						pthread_cancel(tid_execute);
					pthread_mutex_unlock(&execute_running_mutex);
					// wait for thread to finish
					pthread_join(tid_execute, NULL);

					pthread_mutex_lock(&execute_running_mutex);
					if(run == 1) {
						run++;
					} else {
						if(simulation_status != SIMULATION_CANCELLED)
							run++;
					}
					simulation_status = SIMULATION_CANCELLED;
					pthread_mutex_unlock(&execute_running_mutex);
					break;

				default:
					cerr << "error: unknown thread state\n";
					exit(EXIT_FAILURE);
					break;
			}

			if(simulation_status == SIMULATION_CANCELLED)
				break;
			#endif
		}
	} else {
		for(run = from_run; run <= to_run; run++) {
			// reset simulation
			if(!log_memory)
				resetSimulation();

			// create datasets for logging during run
			createLoggingDataset(end_time, run);

			if(!log_memory)
				writeAttribute("run_start_time", \
						generateTimestamp(timestamp), dfile, run_group);

			// initialise simulation
			if(!log_memory)
				writeAttribute("preprocess_start_time", \
						generateTimestamp(timestamp), dfile, run_group);
			ret = psystem->preprocess();
			if(!log_memory)
				writeAttribute("preprocess_end_time", \
						generateTimestamp(timestamp), dfile, run_group);

			// handle models with no executable reactions
			if(!ret) {
				if(!log_memory) {
					writeAttribute("main_loop_start_time", timestamp, \
							dfile, run_group);
					writeAttribute("run_end_time", timestamp, \
							dfile, run_group);
					writeAttribute("total_reactions_simulated", 0, \
									dfile, run_group);
					writeAttribute("simulated_time", 0.0, dfile, run_group);
				}
				closeLoggingDataSet();
				run++;
				break;
			}
						
			// save initial level data
			writeInitialDataSet();

			if(!log_memory)
				writeAttribute("main_loop_start_time", \
						generateTimestamp(timestamp), dfile, run_group);

			#ifdef WIN32
			execute_logging();
			simulation_status = SIMULATION_FINISHED;
			#else
			// iterate simulation
			simulation_status = SIMULATION_CANCELLED;
			ppid = pthread_self();

			// run simulation in thread
			pthread_attr_init(&tattr);
			pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
			if(pthread_create(&tid_execute, &tattr, execute_logging_thread, \
					(void *) this)) {
				cerr << "error: unable to create execute thread\n";
				exit(EXIT_FAILURE);
			}
			// tidy up
			pthread_attr_destroy(&tattr);

			// wait for SIGUSR2 signal
			sigemptyset(&sig_mask);
			sigaddset(&sig_mask, SIGUSR2);
			sigwait(&sig_mask, &sig);

			pthread_mutex_lock(&execute_running_mutex);
			#endif
			switch(simulation_status) {
				case SIMULATION_FINISHED:
					// write run information
					if(!log_memory) {
						writeAttribute("main_loop_end_time", \
								generateTimestamp(timestamp), dfile, run_group);
						writeAttribute("run_end_time", generateTimestamp(timestamp), \
								dfile, run_group);
					}
					// finish writing logs for run
					closeLoggingDataSet();
					#ifndef WIN32
					pthread_mutex_unlock(&execute_running_mutex);
					// wait for thread to finish
					pthread_join(tid_execute, NULL);
					#endif
					break;

				case SIMULATION_CANCELLED:
					#ifndef WIN32
					pthread_mutex_lock(&hdf5_mutex);
					if(execute_running)
						pthread_cancel(tid_execute);
					pthread_mutex_unlock(&hdf5_mutex);
					pthread_mutex_unlock(&execute_running_mutex);
					// wait for thread to finish
					pthread_join(tid_execute, NULL);

					pthread_mutex_lock(&execute_running_mutex);
					#endif
					if(run == 1) {
						if(simulation_status == SIMULATION_CANCELLED) {
							if(!log_memory) {
								writeAttribute("total_reactions_simulated", num_reactions, \
									dfile, run_group);
								writeAttribute("simulated_time", time, dfile, run_group);
							}
						}
						// keep completed part of first run
						// write run information
						if(!log_memory) {
							writeAttribute("main_loop_end_time", \
									generateTimestamp(timestamp), dfile, run_group);
							writeAttribute("run_end_time", generateTimestamp(timestamp), \
									dfile, run_group);
						}
						// finish writing logs for run
						closeLoggingDataSet();
						run++;
					} else {
						if(simulation_status == SIMULATION_CANCELLED) {
							// finish writing logs for run
							closeLoggingDataSet();
							// delete uncompleted run
							char name[64];
							snprintf(name, 64, "/run%lu", run);
							H5Ldelete(dfile, name, H5P_DEFAULT);
						} else {
							// finish writing logs for run
							closeLoggingDataSet();
							run++;
						}
					}
					simulation_status = SIMULATION_CANCELLED;
					#ifndef WIN32
					pthread_mutex_unlock(&execute_running_mutex);
					#endif
					break;

				default:
					cerr << "error: unknown thread state\n";
					exit(EXIT_FAILURE);
					break;
			}

			if(simulation_status == SIMULATION_CANCELLED)
				break;
		}
		run--;
		// finish writing dataset
		closeOutputDataSet();
	}

	// reset original signal mask
	#ifndef WIN32
	pthread_sigmask(SIG_SETMASK, &sig_mask_old, NULL);
	#endif
}

// create a daughter compartment from a parent compartment
void SimulationAlgorithm::divideCompartment(unsigned long c)
{
	register unsigned long i;
	register Compartment *compartment;

	compartment = getCompartment(c);

	// reset compartment volume
	compartment->resetVolume();

	// reset division flag
	compartment->setDividing(false);

	if(keep_divisions) {
		// add new compartment for daughter cell
		unsigned long pos[3];
		compartment->getPosition(pos);
		switch(division_direction) {
			case 'x':
				pos[0]++;
				break;

			case 'y':
				pos[1]++;
				break;

			case 'z':
				pos[2]++;
				break;

			default:
				break;
		}

		// close existing amounts dataset
		closeAmountsDataSet();

		// close existing volumes dataset
		closeVolumesDataSet();

		// add daughter compartment
		Compartment *compartment2 = psystem->addDuplicateCompartment( \
						c, pos, division_direction);

		// extend amounts dataset
		extendAmountsDataSet();

		// extend volumes dataset
		extendVolumesDataSet();

		// reduce levels of species in parent and daughter compartment
		for(i = 0; i < num_species; i++)
			if(!psystem->getAlphabet()->isNondividing(i)) {
				compartment2->setLevel(i, rng->randomHalf(compartment->getLevel(i)));
				compartment->setLevel(i, compartment->getLevel(i) - \
						compartment2->getLevel(i));
			}
	} else {
		// reduce levels of dividing species in parent compartment
		for(i = 0; i < num_species; i++)
			if(!psystem->getAlphabet()->isNondividing(i))
				compartment->setLevel(i, rng->randomHalf(compartment->getLevel(i)));
	}
}

// set p-system for simulation algorithm
void SimulationAlgorithm::setPsystem(Psystem *ps)
{
	unsigned int i, j;

	// set pointer to p-system
	psystem = ps;

	// set some variables for quick access during simulation
	num_compartments = psystem->getNumCompartments();
	if(psystem->getAlphabet())
		num_species = psystem->getNumSpecies();
	else
		num_species = 0;

	num_rules = 0;
	for(i = 0; i < psystem->getNumRuleSets(); i++)
		num_rules += psystem->getRuleSet(i)->getNumRules();
	total_rules = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < psystem->getCompartment(i)->getNumRuleSets(); j++)
			total_rules += psystem->getCompartment(i)->getNumRules(j);
}

// reset simulation
void SimulationAlgorithm::resetSimulation(void)
{
	time = tau0 = tau0global = 0.0;  // set simulation time to 0
	num_reactions = 0;  // set reactions performed to 0

	// reset data buffers
	amount_index = 0;
	amount_offset = 0;
	volume_index = 0;
	volume_offset = 0;
	compartment_index = 0;
	reaction_index = 0;
	reaction_offset = 0;
	propensity_index = 0;
	position_index = 0;
	lattice_spin_index = 0;
	lattice_spin_offset = 0;
	lattice_type_index = 0;
	lattice_type_offset = 0;
	first_time = true;

	// reset compartments and volume changes list
	if(run != 1 || log_memory)
		psystem->resetCompartments();
}

// create output dataset for saving results
void SimulationAlgorithm::createOutputDataset(SimulationParameters &parameters)
{
	unsigned int i, j;
	unsigned long index, entries;
	char timestamp[256];

	// setup data logging
	if(strcmp(parameters.log_type, "reactions") == 0) {
		log_type = LOG_TYPE_REACTIONS;
		log_levels = false;
		psystem->logReactions(true);
	} else {
		if(strcmp(parameters.log_type, "levels") == 0) {
			log_type = LOG_TYPE_LEVELS;
			psystem->logReactions(false);
			log_levels = true;
			log_interval = parameters.log_interval;
			psystem->logReactions(false);
		} else {
			if(strcmp(parameters.log_type, "none") == 0) {
				log_type = LOG_TYPE_NONE;
				psystem->logReactions(false);
				log_interval = parameters.log_interval;
			}
		}
	}
	log_propensities = parameters.log_propensities;
	log_volumes = parameters.log_volumes;
	log_memory = parameters.log_memory;
	log_steady_state = parameters.log_steady_state;
	log_lattice = parameters.log_lattice;
	show_progress = parameters.show_progress;
	progress_interval = (unsigned int) parameters.progress_interval;
	if(log_memory && log_type != LOG_TYPE_LEVELS && !log_propensities) {
		cerr << "error: can only log to memory if logging levels\n";
		exit(EXIT_FAILURE);
	}

	// if logging to memory or file all done
	if(parameters.log_memory || log_type == LOG_TYPE_NONE)
		return;

	// log to file
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace;  // hdf5 dataspace
	hid_t dtype;  // hdf5 datatype

	// create datafile
	#ifdef HAVE_MPI
	if(mpi_serial)
		dfile = createDatafile(parameters.data_file);
	else {
		// create output file
		if(tid == 0) {
			if((dfile = H5Fcreate(parameters.data_file, H5F_ACC_TRUNC, \
					H5P_DEFAULT, H5P_DEFAULT)) < 0) {
				cerr << "error: couldn't create file " << \
						parameters.data_file << "\n";
				exit(EXIT_FAILURE);
			}
		} else {
			char filename[1024];
			snprintf(filename, 1024, "%s.%d", parameters.data_file, tid);
			if((dfile = H5Fcreate(filename, H5F_ACC_TRUNC, \
					H5P_DEFAULT, H5P_DEFAULT)) < 0) {
				cerr << "error: couldn't create file " << filename << "\n";
				exit(EXIT_FAILURE);
			}
		}
		// return if not controller thread
		if(tid != 0)
			return;
	}
	#else
	dfile = createDatafile(parameters.data_file);
	#endif

	// write general attributes
	writeAttribute("mcss_version", VERSION, dfile);
	writeAttribute("simulation_start_time", generateTimestamp(timestamp), dfile);
	writeAttribute("simulation_algorithm", parameters.simulation_algorithm, \
			dfile);
	writeAttribute("simulation_algorithm_name", name.c_str(), dfile);
	writeAttribute("model_input_file", parameters.model_file, dfile);
	writeAttribute("model_format", parameters.model_format, dfile);
	writeAttribute("data_file", parameters.data_file, dfile);
	writeAttribute("log_type", parameters.log_type, dfile);
	writeAttribute("log_interval", parameters.log_interval, dfile);
	writeAttribute("number_of_species", num_species, dfile);
	writeAttribute("lattice_x_dimension", psystem->getLatticeX(), dfile);
	writeAttribute("lattice_y_dimension", psystem->getLatticeY(), dfile);
	writeAttribute("lattice_z_dimension", psystem->getLatticeZ(), dfile);
	writeAttribute("max_time", parameters.max_time, dfile);
	writeAttribute("seed", seed, dfile);
	writeAttribute("duplicate_initial_amounts", \
			parameters.duplicate_initial_amounts, dfile);
	writeAttribute("log_propensities", parameters.log_propensities, dfile);
	writeAttribute("log_volumes", parameters.log_volumes, dfile);
	writeAttribute("log_degraded", parameters.log_degraded, dfile);
	writeAttribute("log_lattice", parameters.log_lattice, dfile);
	writeAttribute("periodic_x", parameters.periodic_x, dfile);
	writeAttribute("periodic_y", parameters.periodic_y, dfile);
	writeAttribute("periodic_z", parameters.periodic_z, dfile);
	writeAttribute("growth_type", parameters.growth_type, dfile);
	writeAttribute("keep_divisions", parameters.keep_divisions, dfile);
	writeAttribute("division_direction", parameters.division_direction, dfile);
	writeAttribute("number_of_rule_templates", \
			psystem->getNumRuleSets(), dfile);
	writeAttribute("number_of_rules_in_templates", num_rules, dfile);
	writeAttribute("total_number_of_rules", total_rules, dfile);

	// save species information
	// create datatype
	dtype = createSpeciesDatatype();

	// create dataspace
	entries = num_species;
	dspace = createDataspace(entries);

	// create dataset
	dset = createDataset("species_information", dfile, dtype, dspace);

	// create data buffer
	SpeciesEntry *species_buffer = (SpeciesEntry *) \
			createDataBuffer(entries, dtype);

	// create data entry
	SpeciesEntry *species_entry = (SpeciesEntry *) \
			createDataEntry(dtype);

	// write data to buffer
	index = 0;
	for(i = 0; i < entries; i++) {
		// assign data to entry
		species_entry->si = i;
		strncpy(species_entry->sn, psystem->getAlphabet()-> \
				getName(i).c_str(), 255);

		// copy data to buffer
		writeDataset(species_entry, species_buffer, dset, dtype, \
				dspace, index, entries);
	}

	// close dataset
	closeDataset(dfile, dset, dtype, species_buffer, index, entries);

	// free data entry memory
	freeDataEntry(species_entry);

	// free data buffer memory
	freeDataBuffer(species_buffer);

	// save ruleset information
	// create datatype
	dtype = createRulesetDatatype();

	// create dataspace
	entries = psystem->getNumRuleSets();
	dspace = createDataspace(entries);

	// create dataset
	dset = createDataset("ruleset_information", dfile, dtype, dspace);

	// create data buffer
	RulesetEntry *ruleset_buffer = (RulesetEntry *) \
			createDataBuffer(entries, dtype);

	// create data entry
	RulesetEntry *ruleset_entry = (RulesetEntry *) \
			createDataEntry(dtype);

	// write data to buffer
	index = 0;
	for(i = 0; i < entries; i++) {
		// assign data to entry
		ruleset_entry->rsi = i;
		strncpy(ruleset_entry->rsn, psystem->getRuleSet(i)-> \
				getName().c_str(), 255);
		strncpy(ruleset_entry->rsc, psystem->getRuleSet(i)-> \
				getId().c_str(), 255);
		ruleset_entry->nr = psystem->getRuleSet(i)->getNumRules();

		// copy data to buffer
		writeDataset(ruleset_entry, ruleset_buffer, dset, dtype, \
				dspace, index, entries);
	}

	// close dataset
	closeDataset(dfile, dset, dtype, ruleset_buffer, index, entries);

	// free data entry memory
	freeDataEntry(ruleset_entry);

	// free data buffer memory
	freeDataBuffer(ruleset_buffer);

	// create datatype
	dtype = createRuleDatatype();

	// create dataspace
	entries = num_rules;
	dspace = createDataspace(entries);

	// create dataset
	dset = createDataset("rule_information", dfile, dtype, dspace);

	// create data buffer
	RuleEntry *rule_buffer = (RuleEntry *) createDataBuffer(entries, dtype);

	// create data entry
	RuleEntry *rule_entry = (RuleEntry *) createDataEntry(dtype);

	// write data to buffer
	index = 0;
	for(i = 0; i < psystem->getNumRuleSets(); i++)
		for(j = 0; j < psystem->getRuleSet(i)->getNumRules(); j++) {
			// assign data to entry
			rule_entry->rsi = i;
			rule_entry->ri = j;
			strncpy(rule_entry->rid, psystem->getRuleSet(i)-> \
					getRule(j)->getId().c_str(), 255);
			strncpy(rule_entry->rn, psystem->getRuleSet(i)-> \
					getRule(j)->getName().c_str(), 255);
			rule_entry->rx = (long) psystem->getRuleSet(i)-> \
					getRule(j)->getXVector();
			rule_entry->ry = (long) psystem->getRuleSet(i)-> \
					getRule(j)->getYVector();
			strncpy(rule_entry->rstr, psystem->getRuleSet(i)-> \
					getRule(j)->getRuleAsString( \
							psystem->getAlphabet()).c_str(), 255);

			// copy data to buffer
			writeDataset(rule_entry, rule_buffer, dset, dtype, dspace, \
					index, num_rules);
		}

	// close dataset
	closeDataset(dfile, dset, dtype, rule_buffer, index, num_rules);

	// free data entry memory
	freeDataEntry(rule_entry);

	// free data buffer memory
	freeDataBuffer(rule_buffer);
}

// create dataset for logging during simulation run
void SimulationAlgorithm::createLoggingDataset(const double end_time, \
		const unsigned long run)
{
	unsigned int i, j;
	unsigned long num_rules, index, entries;
	char dset_name[64];
	hid_t cparams;

	// check if logging to memory or file
	if(log_memory) {
		if(log_type == LOG_TYPE_LEVELS) {
			entries = (unsigned long) ceil((double) \
					end_time / log_interval) + 1;
			// create buffer if doesn't exist or size changed
			if(!amount_buffer || entries != amount_chunk_size) {
				// close existing dataset
				closeAmountsDataSet();
				// create storage
				amount_chunk_size = (unsigned long) ceil((double) \
						end_time / log_interval) + 1;
				hsize_t dims[3] = {num_species, num_compartments, \
						amount_chunk_size};
				amount_buffer = create3DArrayUInt(dims);
			}
		}
		if(log_volumes && !volume_buffer) {
			// close existing dataset
			closeVolumesDataSet();
			// create storage
			volume_chunk_size = (unsigned long) ceil((double) \
					end_time / log_interval) + 1;
			hsize_t dims[2] = {num_compartments, amount_chunk_size};
			volume_buffer = create2DArrayDouble(&dims[1]);
		}
		return;  // all done
	}

	// create group for run
	snprintf(run_group, 64, "/run%u", (unsigned int) run);
	dgroup = createGroup(run_group, dfile);

	// create compartment dataset
	// create datatype
	compartment_dtype = createCompartmentDatatype();

	// create dataspace
	entries = compartment_chunk_size;
	compartment_dspace = createDataspaceScaled(&entries);
	compartment_chunk_size = entries;

	// create dataset
	snprintf(dset_name, 64, "/run%u/compartment_information", \
		(unsigned int) run);
	compartment_dset = createDataset(dset_name, dfile, \
			compartment_dtype, compartment_dspace);

	// create data buffer
	compartment_buffer = (CompartmentEntry *) createDataBuffer(entries, \
			compartment_dtype);

	// create data entry
	compartment_entry = (CompartmentEntry *) createDataEntry(compartment_dtype);

	// save reaction data
	if(log_type == LOG_TYPE_REACTIONS) {
		// create datatype
		reaction_dtype = createReactionDatatype();

		// fit dataset into available memory
		hsize_t dims[1] = {reaction_chunk_size};
		hsize_t maxdims[1] = {H5S_UNLIMITED};
		do {
			// create property list
			dims[0] = reaction_chunk_size;
			cparams = createChunkedPropertyList(1, dims, reaction_dtype);
			reaction_chunk_size = dims[0];

			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}

			// create dataspace
			reaction_dspace = createDataspaceScaled(1, dims, maxdims);

			// test if dataspace same size as chunk size
			if(dims[0] != reaction_chunk_size) {
				H5Pclose(cparams);
				closeDataspace(reaction_dspace);
			}
		} while(dims[0] != reaction_chunk_size);

		// create dataset
		snprintf(dset_name, 64, "/run%u/reactions", (unsigned int) run);
		if((reaction_dset = H5Dcreate2(dfile, dset_name, reaction_dtype, \
				reaction_dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create reactions dataset\n";
			exit(EXIT_FAILURE);
		}

		// create data buffer
		reaction_buffer = (ReactionEntry *) createDataBuffer(entries, \
				reaction_dtype);

		// free resources
		H5Pclose(cparams);
	}

	// save amounts data
	if(log_type == LOG_TYPE_LEVELS) {
		// create datatype
		amount_dtype = createAmountDatatype();

		// fit dataset into available memory
		hsize_t dims[3] = {num_species, num_compartments, amount_chunk_size};
		if(!dims[0])  // handle models with no species
			dims[0] = 1;
		hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
		do {
			// create property list
			dims[2] = amount_chunk_size;
			cparams = createChunkedPropertyList(3, dims, amount_dtype);
			amount_chunk_size = dims[2];

			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}

			// create dataspace
			amount_dspace = createDataspaceScaled(3, dims, maxdims);

			// test if dataspace same size as chunk size
			if(dims[2] != amount_chunk_size) {
				H5Pclose(cparams);
				closeDataspace(amount_dspace);
			}
		} while(dims[2] != amount_chunk_size);

		// create dataset
		snprintf(dset_name, 64, "/run%u/amounts", (unsigned int) run);
		if((amount_dset = H5Dcreate2(dfile, dset_name, amount_dtype, \
				amount_dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create amounts dataset\n";
			exit(EXIT_FAILURE);
		}

		// create data buffer
		amount_buffer = create3DArrayUInt(dims);

		// free resources
		H5Pclose(cparams);
	}

	// save reaction propensity data
	if(log_propensities) {
		// create datatype
		propensity_dtype = createPropensityDatatype(total_rules);

		// create dataspace
		entries = propensity_chunk_size;
		propensity_dspace = createDataspaceScaled(&entries);
		propensity_chunk_size = entries;

		// create dataset
		snprintf(dset_name, 64, "/run%u/propensities", (unsigned int) run);
		propensity_dset = createDataset(dset_name, dfile, propensity_dtype, \
				propensity_dspace);

		// create data buffer
		propensity_buffer = createDataBuffer(entries, propensity_dtype);

		// create data entry
		propensity_entry = createDataEntry(propensity_dtype);
	}

	// save compartment volumes data
	if(log_volumes) {
		// create volumes dataset
		// create datatype
		volume_dtype = createVolumeDatatype();

		// fit dataset into available memory
		hsize_t dims[2] = {num_compartments, volume_chunk_size};
		hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
		do {
			// create property list
			dims[1] = volume_chunk_size;
			cparams = createChunkedPropertyList(2, dims, volume_dtype);
			volume_chunk_size = dims[1];

			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}

			// create dataspace
			volume_dspace = createDataspaceScaled(2, dims, maxdims);

			// test if dataspace same size as chunk size
			if(dims[1] != volume_chunk_size) {
				H5Pclose(cparams);
				closeDataspace(volume_dspace);
			}
		} while(dims[1] != volume_chunk_size);

		// create dataset
		snprintf(dset_name, 64, "/run%u/volumes", (unsigned int) run);
		if((volume_dset = H5Dcreate2(dfile, dset_name, volume_dtype, \
				volume_dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create volumes dataset\n";
			exit(EXIT_FAILURE);
		}

		// create data buffer
		volume_buffer = create2DArrayDouble(dims);

		// create positions dataset
		// create datatype
		position_dtype = createPositionDatatype();

		// create dataspace
		entries = position_chunk_size;
		position_dspace = createDataspaceScaled(&entries);
		position_chunk_size = entries;

		// create dataset
		snprintf(dset_name, 64, "/run%u/positions", (unsigned int) run);
		position_dset = createDataset(dset_name, dfile, \
				position_dtype, position_dspace);

		// create data buffer
		position_buffer = (PositionEntry *) createDataBuffer(entries, \
				position_dtype);

		// create data entry
		position_entry = (PositionEntry *) createDataEntry(position_dtype);

		// free resources
		H5Pclose(cparams);
	}

	// save amounts data
	if(log_lattice) {
		// create datatype
		lattice_spin_dtype = createLatticeSpinDatatype();

		// fit dataset into available memory
		hsize_t dims[3] = {psystem->getLatticeX(), psystem->getLatticeY(), \
				lattice_spin_chunk_size};
		hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
		do {
			// create property list
			dims[2] = lattice_spin_chunk_size;
			cparams = createChunkedPropertyList(3, dims, lattice_spin_dtype);
			lattice_spin_chunk_size = dims[2];

			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}

			// create dataspace
			lattice_spin_dspace = createDataspaceScaled(3, dims, maxdims);

			// test if dataspace same size as chunk size
			if(dims[2] != lattice_spin_chunk_size) {
				H5Pclose(cparams);
				closeDataspace(lattice_spin_dspace);
			}
		} while(dims[2] != lattice_spin_chunk_size);

		// create dataset
		snprintf(dset_name, 64, "/run%u/lattice_spin", (unsigned int) run);
		if((lattice_spin_dset = H5Dcreate2(dfile, dset_name, \
				lattice_spin_dtype, lattice_spin_dspace, H5P_DEFAULT, cparams, \
				H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create lattice spin dataset\n";
			exit(EXIT_FAILURE);
		}

		// create data buffer
		lattice_spin_buffer = create3DArrayUInt(dims);

		// free resources
		H5Pclose(cparams);

		// create datatype
		lattice_type_dtype = createLatticeTypeDatatype();

		// fit dataset into available memory
		dims[2] = lattice_type_chunk_size;
		do {
			// create property list
			dims[2] = lattice_type_chunk_size;
			cparams = createChunkedPropertyList(3, dims, lattice_type_dtype);
			lattice_type_chunk_size = dims[2];

			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}

			// create dataspace
			lattice_type_dspace = createDataspaceScaled(3, dims, maxdims);

			// test if dataspace same size as chunk size
			if(dims[2] != lattice_type_chunk_size) {
				H5Pclose(cparams);
				closeDataspace(lattice_type_dspace);
			}
		} while(dims[2] != lattice_type_chunk_size);

		// create dataset
		snprintf(dset_name, 64, "/run%u/lattice_type", (unsigned int) run);
		if((lattice_type_dset = H5Dcreate2(dfile, dset_name, \
				lattice_type_dtype, lattice_type_dspace, H5P_DEFAULT, cparams, \
				H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create lattice type dataset\n";
			exit(EXIT_FAILURE);
		}

		// create data buffer
		lattice_type_buffer = create3DArrayUInt(dims);

		// free resources
		H5Pclose(cparams);
	}
}

// extend amounts dataset
void SimulationAlgorithm::extendAmountsDataSet(void)
{
	char dset_name[64];

	// handle logging to memory
	if(log_memory) {
		// free old buffer if still exists
		if(amount_buffer)
			free3DArrayUInt(amount_buffer);
		// create data buffer if first run
		hsize_t dims[3] = {num_species, num_compartments, amount_chunk_size};
		amount_buffer = create3DArrayUInt(dims);
		// reset index
		amount_index = 0;
		return;  // all done
	}

	pthread_mutex_lock(&hdf5_mutex);
	// reopen dataset
	snprintf(dset_name, 64, "/run%u/amounts", (unsigned int) run);
	if((amount_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	amount_offset += amount_index;
	hsize_t dims[3] = {num_species, num_compartments, amount_offset};
	if(H5Dset_extent(amount_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[2] = amount_chunk_size;
	amount_buffer = create3DArrayUInt(dims);

	// reassign dataspace
	hsize_t maxdims[3] = {num_species, H5S_UNLIMITED, H5S_UNLIMITED};
	if((amount_dspace = H5Screate_simple(3, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create amounts dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	amount_index = 0;
	pthread_mutex_unlock(&hdf5_mutex);
}

// extend volumes dataset
void SimulationAlgorithm::extendVolumesDataSet(void)
{
	char dset_name[64];

	// handle logging to memory
	if(log_memory) {
		// free old buffer if still exists
		if(volume_buffer)
			free2DArrayDouble(volume_buffer);
		// create data buffer if first run
		hsize_t dims[2] = {num_compartments, volume_chunk_size};
		volume_buffer = create2DArrayDouble(dims);
		// reset index
		volume_index = 0;
		return;  // all done
	}

	pthread_mutex_lock(&hdf5_mutex);
	// reopen dataset
	snprintf(dset_name, 64, "/run%u/volumes", (unsigned int) run);
	if((volume_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	volume_offset += volume_index;
	hsize_t dims[2] = {num_compartments, volume_offset};
	if(H5Dset_extent(volume_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[1] = volume_chunk_size;
	volume_buffer = create2DArrayDouble(dims);

	// reassign dataspace
	hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
	if((volume_dspace = H5Screate_simple(2, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create volume dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	volume_index = 0;
	pthread_mutex_unlock(&hdf5_mutex);
}

// extend lattice spin dataset
void SimulationAlgorithm::extendLatticeSpinDataSet(void)
{
	char dset_name[64];

	// handle logging to memory
	if(log_memory) {
		// free old buffer if still exists
		if(lattice_spin_buffer)
			free3DArrayUInt(lattice_spin_buffer);
		// create data buffer if first run
		hsize_t dims[3] = {psystem->getLatticeX(), psystem->getLatticeY(), \
				lattice_spin_chunk_size};
		lattice_spin_buffer = create3DArrayUInt(dims);
		// reset index
		lattice_spin_index = 0;
		return;  // all done
	}

	pthread_mutex_lock(&hdf5_mutex);
	// reopen dataset
	snprintf(dset_name, 64, "/run%u/lattice_spin", (unsigned int) run);
	if((lattice_spin_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	lattice_spin_offset += lattice_spin_index;
	hsize_t dims[3] = {psystem->getLatticeX(), psystem->getLatticeY(), \
			lattice_spin_offset};
	if(H5Dset_extent(lattice_spin_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[2] = lattice_spin_chunk_size;
	lattice_spin_buffer = create3DArrayUInt(dims);

	// reassign dataspace
	hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
	if((lattice_spin_dspace = H5Screate_simple(3, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	lattice_spin_index = 0;
	pthread_mutex_unlock(&hdf5_mutex);
}

// extend lattice type dataset
void SimulationAlgorithm::extendLatticeTypeDataSet(void)
{
	char dset_name[64];

	// handle logging to memory
	if(log_memory) {
		// free old buffer if still exists
		if(lattice_type_buffer)
			free3DArrayUInt(lattice_type_buffer);
		// create data buffer if first run
		hsize_t dims[3] = {psystem->getLatticeX(), psystem->getLatticeY(), \
				lattice_type_chunk_size};
		lattice_type_buffer = create3DArrayUInt(dims);
		// reset index
		lattice_type_index = 0;
		return;  // all done
	}

	pthread_mutex_lock(&hdf5_mutex);
	// reopen dataset
	snprintf(dset_name, 64, "/run%u/lattice_type", (unsigned int) run);
	if((lattice_type_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	lattice_type_offset += lattice_type_index;
	hsize_t dims[3] = {psystem->getLatticeX(), psystem->getLatticeY(), \
			lattice_type_offset};
	if(H5Dset_extent(lattice_type_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[2] = lattice_type_chunk_size;
	lattice_type_buffer = create3DArrayUInt(dims);

	// reassign dataspace
	hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
	if((lattice_type_dspace = H5Screate_simple(3, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	lattice_type_index = 0;
	pthread_mutex_unlock(&hdf5_mutex);
}

// write reaction data
void SimulationAlgorithm::writeReaction(double time, \
		unsigned long species, unsigned long level, unsigned long comp, \
		unsigned long ruleset, unsigned long reaction)
{
	register unsigned long i, j;
	register unsigned int progress_time;

	// output progress
	if(show_progress) {
		if(progress_interval) {
			progress_time = (unsigned int) std::time(NULL);
			if((progress_time - last_progress_time) > progress_interval) {
				std::cout << time << " " << run << "\n";
				last_progress_time = progress_time;
			}
		}
		else {
			// output every time data logged
			std::cout << time << " " << run << "\n";
		}
	}

	// write data to buffer
	pthread_mutex_lock(&hdf5_mutex);
	reaction_buffer[reaction_index].t = time;
	reaction_buffer[reaction_index].si = species;
	reaction_buffer[reaction_index].sl = level;
	reaction_buffer[reaction_index].ci = comp;
	reaction_buffer[reaction_index].rsi = ruleset;
	reaction_buffer[reaction_index].ri = reaction;
	reaction_index++;

	if(!log_memory && reaction_index == reaction_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[1] = {reaction_offset};
		hsize_t dims1[1] = {reaction_offset + reaction_chunk_size};
		hsize_t dims2[1] = {reaction_chunk_size};
		if(H5Dset_extent(reaction_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend reactions dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(reaction_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write reactions dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(reaction_dset, reaction_dtype, reaction_dspace, fspace, \
				H5P_DEFAULT, &reaction_buffer[0]) < 0) {
			std::cerr << "error: couldn't write reactions dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		reaction_offset += reaction_chunk_size;
		reaction_index = 0;
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write level data
void SimulationAlgorithm::writeAmount(double time)
{
	register unsigned long i, j;
	register unsigned int progress_time;

	// output progress
	if(show_progress) {
		if(progress_interval) {
			progress_time = (unsigned int) std::time(NULL);
			if((progress_time - last_progress_time) > progress_interval) {
				std::cout << time << " " << run << "\n";
				last_progress_time = progress_time;
			}
		}
		else {
			// output every time data logged
			std::cout << time << " " << run << "\n";
		}
	}

	// write data to buffer
	pthread_mutex_lock(&hdf5_mutex);
	for(i = 0; i < num_species; i++)
		for(j = 0; j < num_compartments; j++)
			amount_buffer[i][j][amount_index] = (unsigned int) \
					psystem->getSpeciesLevel(i, j);
	amount_index++;

	if(!log_memory && amount_index == amount_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[3] = {0, 0, amount_offset};
		hsize_t dims1[3] = {num_species, num_compartments, amount_offset + \
				amount_chunk_size};
		hsize_t dims2[3] = {num_species, num_compartments, amount_chunk_size};
		if(H5Dset_extent(amount_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend amounts dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(amount_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write amounts dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(amount_dset, amount_dtype, amount_dspace, fspace, \
				H5P_DEFAULT, &amount_buffer[0][0][0]) < 0) {
			std::cerr << "error: couldn't write amounts dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		amount_offset += amount_chunk_size;
		amount_index = 0;
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write volume data
void SimulationAlgorithm::writeVolume(double t)
{
	register unsigned long i, j;

	pthread_mutex_lock(&hdf5_mutex);
	// write data to buffer
	for(i = 0; i < num_compartments; i++)
		volume_buffer[i][volume_index] = \
				psystem->getCompartmentVolume(i);
	volume_index++;

	if(!log_memory && volume_index == volume_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[2] = {0, volume_offset};
		hsize_t dims1[2] = {num_compartments, volume_offset + \
				volume_chunk_size};
		hsize_t dims2[2] = {num_compartments, volume_chunk_size};
		if(H5Dset_extent(volume_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend volumes dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(volume_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write volumes dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(volume_dset, volume_dtype, volume_dspace, fspace, \
				H5P_DEFAULT, &volume_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't write volume dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		volume_offset += volume_chunk_size;
		volume_index = 0;
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write lattice spin data
void SimulationAlgorithm::writeLatticeSpin(double time)
{
	register unsigned long i, j, xdim, ydim;
	Lattice *lat;

	pthread_mutex_lock(&hdf5_mutex);
	xdim = psystem->getLatticeX();
	ydim = psystem->getLatticeY();
	lat = psystem->getLattice();
	// write data to buffer
	for(i = 0; i < xdim; i++)
		for(j = 0; j < ydim; j++)
			lattice_spin_buffer[i][j][lattice_spin_index] = \
					(unsigned int) lat->getSpin(i, j);
	lattice_spin_index++;

	if(!log_memory && lattice_spin_index == lattice_spin_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[3] = {0, 0, lattice_spin_offset};
		hsize_t dims1[3] = {xdim, ydim, lattice_spin_offset + \
				lattice_spin_chunk_size};
		hsize_t dims2[3] = {xdim, ydim, lattice_spin_chunk_size};
		if(H5Dset_extent(lattice_spin_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(lattice_spin_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(lattice_spin_dset, lattice_spin_dtype, \
				lattice_spin_dspace, fspace, H5P_DEFAULT, \
				&lattice_spin_buffer[0][0][0]) < 0) {
			std::cerr << "error: couldn't write lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		lattice_spin_offset += lattice_spin_chunk_size;
		lattice_spin_index = 0;
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write lattice type data
void SimulationAlgorithm::writeLatticeType(double time)
{
	register unsigned long i, j, xdim, ydim;
	Lattice *lat;

	pthread_mutex_lock(&hdf5_mutex);
	xdim = psystem->getLatticeX();
	ydim = psystem->getLatticeY();
	lat = psystem->getLattice();
	// write data to buffer
	for(i = 0; i < xdim; i++)
		for(j = 0; j < ydim; j++)
			lattice_type_buffer[i][j][lattice_type_index] = \
					(unsigned int) lat->getType(i, j);
	lattice_type_index++;

	if(!log_memory && lattice_type_index == lattice_type_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[3] = {0, 0, lattice_type_offset};
		hsize_t dims1[3] = {xdim, ydim, lattice_type_offset + \
				lattice_type_chunk_size};
		hsize_t dims2[3] = {xdim, ydim, lattice_type_chunk_size};
		if(H5Dset_extent(lattice_type_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(lattice_type_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(lattice_type_dset, lattice_type_dtype, \
				lattice_type_dspace, fspace, H5P_DEFAULT, \
				&lattice_type_buffer[0][0][0]) < 0) {
			std::cerr << "error: couldn't write lattice dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		lattice_type_offset += lattice_type_chunk_size;
		lattice_type_index = 0;
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write position data
void SimulationAlgorithm::writePosition(unsigned long ind)
{
	pthread_mutex_lock(&hdf5_mutex);
	// assign data
	position_entry->ci = ind;
	position_entry->cx = getCompartment(ind)->getXPosition();
	position_entry->cy = getCompartment(ind)->getYPosition();
	position_entry->ct = time;

	// copy data to buffer
	if(!log_memory) {
		writeDataset(position_entry, position_buffer, position_dset, \
				position_dtype, position_dspace, position_index, \
				position_chunk_size);
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// write propensities data
void SimulationAlgorithm::writePropensities(double time)
{
	register unsigned long i, j, k, index;
	PropensityEntry pe;

	pthread_mutex_lock(&hdf5_mutex);
	// assign data to entry
	convertPropensityEntry(propensity_entry, propensity_dtype, pe);
	// write time
	*pe.t = time;
	// write compartment indices
	index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < psystem->getCompartment(i)->getNumRuleSets(); j++)
			for(k = 0; k < psystem->getCompartment(i)->getNumRules(j); k++)
				pe.ci[index++] = i;
	// write ruleset indices
	index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < psystem->getCompartment(i)->getNumRuleSets(); j++)
			for(k = 0; k < psystem->getCompartment(i)->getNumRules(j); k++)
				pe.rsi[index++] = psystem->getCompartment(i)->\
						getRuleSetIndex(j);
	// write rule indices
	index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < psystem->getCompartment(i)->getNumRuleSets(); j++)
			for(k = 0; k < psystem->getCompartment(i)->getNumRules(j); k++)
				pe.ri[index++] = psystem->getCompartment(i)->getRuleIndex(j, k);
	// write propensities
	index = 0;
	for(i = 0; i < num_compartments; i++)
		for(j = 0; j < psystem->getCompartment(i)->getNumRuleSets(); j++)
			for(k = 0; k < psystem->getCompartment(i)->getNumRules(j); k++)
				pe.rp[index++] = calculatePropensity(j, k, i);

	// copy data to buffer
	if(!log_memory) {
		writeDataset(propensity_entry, propensity_buffer, propensity_dset, \
				propensity_dtype, propensity_dspace, propensity_index, \
				propensity_chunk_size);
	}
	pthread_mutex_unlock(&hdf5_mutex);
}

// save initial levels
void SimulationAlgorithm::writeInitialDataSet(void)
{
	register unsigned int i, j, con;

	if(log_type == LOG_TYPE_REACTIONS) {
		for(i = 0; i < num_compartments; i++)
			for(j = 0; j < psystem->getAlphabet()->size(); j++)
				if((con = getCompartment(i)->getLevel(j)))
					writeReaction(time, j, con, i, 0, 0);
	} else
		if(log_type == LOG_TYPE_LEVELS) {
			writeAmount(time);
			if(log_propensities)
				writePropensities(time);
		}
	if(log_volumes) {
		// write intial volumes
		writeVolume(time);

		// write initial compartment positions
		for(i = 0; i < num_compartments; i++)
			writePosition(i);
	}
	if(log_lattice) {
		// write intial lattice
		writeLatticeSpin(time);
		writeLatticeType(time);
	}
}

// finish writing simulation data
void SimulationAlgorithm::closeOutputDataSet(void)
{
	char timestamp[256];

	if(!log_memory) {
		#ifndef HAVE_MPI
		// write remaining attributes
		writeAttribute("number_of_runs", run, dfile);
		writeAttribute("simulation_end_time", generateTimestamp(timestamp), \
				dfile);
		// close datafile
		closeDatafile(dfile);
		#else
		if(mpi_serial) {
			// write remaining attributes
			writeAttribute("number_of_runs", run, dfile);
			writeAttribute("simulation_end_time", \
					generateTimestamp(timestamp), dfile);
			// close datafile
			closeDatafile(dfile);
		} else {
			// collate data for parallel runs
			if(tid == 0) {
				int count = nthreads - 1;
				int thread;
				MPI_Status status;
				hid_t dfile2, dgroup;
				H5G_info_t ginfo;
				char filename[1024], tfilename[1024], gname[1024];
				hsize_t i;
				// get base filename of datafile
				if(H5Fget_name(dfile, filename, 1024) < 0) {
					cerr << "error: couldn't collate data\n";
					exit(EXIT_FAILURE);
				}
				while(count) {
					// wait for other threads to finish
					MPI_Recv(&thread, 1, MPI_INT, MPI_ANY_SOURCE, 0, \
							MPI_COMM_WORLD, &status);
					// ignore unused threads
					if(thread >= nthreads)
						continue;
					// once each thread has finished collate data from output file
					snprintf(tfilename, 1024, "%s.%d", filename, thread);
					// open thread datafile
					if((dfile2 = openDatafile(tfilename)) < 0) {
						cerr << "error: couldn't collate data\n";
						exit(EXIT_FAILURE);
					}
					// open root group
					if((dgroup = openGroup("/", dfile2)) < 0) {
						cerr << "error: couldn't collate data\n";
						exit(EXIT_FAILURE);
					}
					// get list of groups in thread datafile
					if(H5Gget_info(dgroup, &ginfo) < 0) {
						cerr << "error: couldn't collate data\n";
						exit(EXIT_FAILURE);
					}
					// copy each group from thread datafile
					for(i = 0; i < ginfo.nlinks; i++) {
						// get group name
						if(H5Lget_name_by_idx(dgroup, ".", H5_INDEX_NAME, \
								H5_ITER_INC, i, gname, 1024, H5P_DEFAULT) < 0) {
							cerr << "error: couldn't collate data\n";
							exit(EXIT_FAILURE);
						}
						// copy group
						if(H5Ocopy(dfile2, gname, dfile, gname, H5P_DEFAULT, \
								H5P_DEFAULT) < 0) {
							cerr << "error: couldn't collate data\n";
							exit(EXIT_FAILURE);
						}
					}
					// close thread root group
					closeGroup(dgroup);
					// close thread datafile
					closeDatafile(dfile2);
					// delete thread datafile
					if(unlink(tfilename) < 0) {
						cerr << "error: couldn't collate data\n";
						exit(EXIT_FAILURE);
					}
					count--;
				}
				// write remaining attributes
				writeAttribute("number_of_runs", num_runs, dfile);
				writeAttribute("simulation_end_time", \
						generateTimestamp(timestamp), dfile);
				// close master datafile
				closeDatafile(dfile);
			} else {
				// close datafile
				closeDatafile(dfile);
				// send message to indicate finished
				MPI_Send(&tid, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
			}
		}
		#endif
	}
}

// close reactions dataset
void SimulationAlgorithm::closeReactionsDataSet(void)
{
	unsigned long i, j, k;

	if(log_memory) {
		// free data buffer memory
		if(reaction_buffer)
			freeDataBuffer(reaction_buffer);
		reaction_index = reaction_offset = 0;
		return;
	}

	if(reaction_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[1] = {reaction_index};
		ReactionEntry *reaction_buffer_tmp = (ReactionEntry *) \
				createDataBuffer(dims2[0], reaction_dtype);

		// copy over remaining data
		for(i = 0; i < reaction_index; i++) {
			reaction_buffer_tmp[i].t = reaction_buffer[i].t;
			reaction_buffer_tmp[i].si = reaction_buffer[i].si;
			reaction_buffer_tmp[i].sl = reaction_buffer[i].sl;
			reaction_buffer_tmp[i].ci = reaction_buffer[i].ci;
			reaction_buffer_tmp[i].rsi = reaction_buffer[i].rsi;
			reaction_buffer_tmp[i].ri = reaction_buffer[i].ri;
		}

		// adjust dataset size
		hsize_t offset[1] = {reaction_offset};
		hsize_t dims1[1] = {reaction_offset + reaction_index};
		hsize_t maxdims[1] = {H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(1, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[0] > 0 && dims1[0] < reaction_chunk_size && !reaction_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(1, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(reaction_dset, name, 512);
			H5Dclose(reaction_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 1, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((reaction_dset = H5Dcreate2(dfile, name, reaction_dtype, \
					dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(reaction_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(reaction_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(reaction_dset, reaction_dtype, dspace, fspace, H5P_DEFAULT, \
				&reaction_buffer_tmp[0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(reaction_dspace);

		// free memory
		freeDataBuffer(reaction_buffer_tmp);
	}

	// close dataset
	H5Dclose(reaction_dset);

	// free data buffer memory
	freeDataBuffer(reaction_buffer);
}

// close amounts dataset
void SimulationAlgorithm::closeAmountsDataSet(void)
{
	unsigned long i, j, k;

	if(log_memory) {
		// free data buffer memory
//		if(amount_buffer)
//			free3DArrayUInt(amount_buffer);
//		amount_index = amount_offset = 0;
		return;
	}

	if(amount_index && num_species) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[3] = {num_species, num_compartments, amount_index};
		unsigned int ***amount_buffer_tmp = create3DArrayUInt(dims2);

		// copy over remaining data
		for(i = 0; i < num_species; i++)
			for(j = 0; j < num_compartments; j++)
				for(k = 0; k < amount_index; k++)
					amount_buffer_tmp[i][j][k] = amount_buffer[i][j][k];

		// adjust dataset size
		hsize_t offset[3] = {0, 0, amount_offset};
		hsize_t dims1[3] = {num_species, num_compartments, amount_offset + \
				amount_index};
		hsize_t maxdims[3] = {num_species, H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(3, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[2] > 0 && dims1[2] < amount_chunk_size && !amount_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(3, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(amount_dset, name, 512);
			H5Dclose(amount_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 3, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((amount_dset = H5Dcreate2(dfile, name, amount_dtype, \
					dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(amount_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(amount_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(amount_dset, amount_dtype, dspace, fspace, H5P_DEFAULT, \
				&amount_buffer_tmp[0][0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(amount_dspace);

		// free memory
		free3DArrayUInt(amount_buffer_tmp);
	}

	// close dataset
	H5Dclose(amount_dset);

	// free data buffer memory
	free3DArrayUInt(amount_buffer);
}

// close volumes dataset
void SimulationAlgorithm::closeVolumesDataSet(void)
{
	unsigned long i, j;

	if(log_memory) {
		// free data buffer memory
		if(volume_buffer)
			free2DArrayDouble(volume_buffer);
		volume_index = volume_offset = 0;
		return;
	}

	if(volume_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[2] = {num_compartments, volume_index};
		double **volume_buffer_tmp = create2DArrayDouble(dims2);

		// copy over remaining data
		for(i = 0; i < num_compartments; i++)
			for(j = 0; j < volume_index; j++)
				volume_buffer_tmp[i][j] = volume_buffer[i][j];

		// adjust dataset size
		hsize_t offset[2] = {0, volume_offset};
		hsize_t dims1[2] = {num_compartments, volume_offset + \
				volume_index};
		hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(2, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[1] > 0 && dims1[1] < volume_chunk_size && !volume_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(2, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(volume_dset, name, 512);
			H5Dclose(volume_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 2, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((volume_dset = H5Dcreate2(dfile, name, volume_dtype, \
					dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(volume_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(volume_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(volume_dset, volume_dtype, dspace, fspace, H5P_DEFAULT, \
				&volume_buffer_tmp[0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(volume_dspace);

		// free memory
		free2DArrayDouble(volume_buffer_tmp);
	}

	// close dataset
	H5Dclose(volume_dset);

	// free data buffer memory
	free2DArrayDouble(volume_buffer);
}

// close lattice spin dataset
void SimulationAlgorithm::closeLatticeSpinDataSet(void)
{
	unsigned long i, j, k, xdim, ydim;

	if(log_memory) {
		// free data buffer memory
		if(lattice_spin_buffer)
			free3DArrayUInt(lattice_spin_buffer);
		lattice_spin_index = lattice_spin_offset = 0;
		return;
	}

	xdim = psystem->getLatticeX();
	ydim = psystem->getLatticeY();
	if(lattice_spin_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[3] = {xdim, ydim, lattice_spin_index};
		unsigned int ***lattice_spin_buffer_tmp = create3DArrayUInt(dims2);

		// copy over remaining data
		for(i = 0; i < xdim; i++)
			for(j = 0; j < ydim; j++)
				for(k = 0; k < lattice_spin_index; k++)
					lattice_spin_buffer_tmp[i][j][k] = \
							lattice_spin_buffer[i][j][k];

		// adjust dataset size
		hsize_t offset[3] = {0, 0, lattice_spin_offset};
		hsize_t dims1[3] = {xdim, ydim, lattice_spin_offset + \
				lattice_spin_index};
		hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(3, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[2] > 0 && dims1[2] < lattice_spin_chunk_size && \
					!lattice_spin_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(3, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(lattice_spin_dset, name, 512);
			H5Dclose(lattice_spin_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 3, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((lattice_spin_dset = H5Dcreate2(dfile, name, \
					lattice_spin_dtype, dspace, H5P_DEFAULT, cparams, \
					H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(lattice_spin_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(lattice_spin_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(lattice_spin_dset, lattice_spin_dtype, dspace, fspace, \
				H5P_DEFAULT, &lattice_spin_buffer_tmp[0][0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(lattice_spin_dspace);

		// free memory
		free3DArrayUInt(lattice_spin_buffer_tmp);
	}

	// close dataset
	H5Dclose(lattice_spin_dset);

	// free data buffer memory
	free3DArrayUInt(lattice_spin_buffer);
}

// close lattice type dataset
void SimulationAlgorithm::closeLatticeTypeDataSet(void)
{
	unsigned long i, j, k, xdim, ydim;

	if(log_memory) {
		// free data buffer memory
		if(lattice_type_buffer)
			free3DArrayUInt(lattice_type_buffer);
		lattice_type_index = lattice_type_offset = 0;
		return;
	}

	xdim = psystem->getLatticeX();
	ydim = psystem->getLatticeY();
	if(lattice_type_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[3] = {xdim, ydim, lattice_type_index};
		unsigned int ***lattice_type_buffer_tmp = create3DArrayUInt(dims2);

		// copy over remaining data
		for(i = 0; i < xdim; i++)
			for(j = 0; j < ydim; j++)
				for(k = 0; k < lattice_type_index; k++)
					lattice_type_buffer_tmp[i][j][k] = \
							lattice_type_buffer[i][j][k];

		// adjust dataset size
		hsize_t offset[3] = {0, 0, lattice_type_offset};
		hsize_t dims1[3] = {xdim, ydim, lattice_type_offset + \
				lattice_type_index};
		hsize_t maxdims[3] = {H5S_UNLIMITED, H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(3, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[2] > 0 && dims1[2] < lattice_type_chunk_size && \
					!lattice_type_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(3, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(lattice_type_dset, name, 512);
			H5Dclose(lattice_type_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 3, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(compress)
				if(H5Pset_deflate(cparams, compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((lattice_type_dset = H5Dcreate2(dfile, name, \
					lattice_type_dtype, dspace, H5P_DEFAULT, cparams, \
					H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(lattice_type_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(lattice_type_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(lattice_type_dset, lattice_type_dtype, dspace, fspace, \
				H5P_DEFAULT, &lattice_type_buffer_tmp[0][0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(lattice_type_dspace);

		// free memory
		free3DArrayUInt(lattice_type_buffer_tmp);
	}

	// close dataset
	H5Dclose(lattice_type_dset);

	// free data buffer memory
	free3DArrayUInt(lattice_type_buffer);
}

// finish writing logs for run
void SimulationAlgorithm::closeLoggingDataSet(void)
{
	unsigned long i;

	// no files to close if logging to memory
	if(log_memory) {
		if(log_type == LOG_TYPE_LEVELS) {
			closeAmountsDataSet();
			// save number of timepoints
			num_timepoints = amount_index;
		}
		// close volumes dataset
		if(log_volumes)
			closeVolumesDataSet();
		return;
	}

	if(log_type == LOG_TYPE_REACTIONS) {
		// close dataset
		closeReactionsDataSet();

		// save number of timepoints
		writeAttribute("number_of_timepoints", num_timepoints = reaction_offset \
				+ reaction_index, dfile, run_group);
	}

	if(log_type == LOG_TYPE_LEVELS) {
		// close amounts dataset
		closeAmountsDataSet();

		// save number of timepoints
		writeAttribute("number_of_timepoints", num_timepoints = amount_offset \
				+ amount_index, dfile, run_group);
	}

	if(log_volumes) {
		// close volumes dataset
		closeVolumesDataSet();

		// close positions dataset
		closeDataset(dfile, position_dset, position_dtype, \
				position_buffer, position_index, position_chunk_size);

		// free data entry memory
		freeDataEntry(position_entry);

		// free data buffer memory
		freeDataBuffer(position_buffer);
	}

	if(log_lattice) {
		// close lattice spin dataset
		closeLatticeSpinDataSet();

		// close lattice type dataset
		closeLatticeTypeDataSet();
	}

	if(log_propensities) {
		// close dataset
		closeDataset(dfile, propensity_dset, propensity_dtype, \
				propensity_buffer, propensity_index, propensity_chunk_size);

		// free data entry memory
		freeDataEntry(propensity_entry);

		// free data buffer memory
		freeDataBuffer(propensity_buffer);
	}

	// write data for compartments which still exist
	writeAttribute("number_of_compartments", num_compartments, dfile, \
			run_group);
	for(i = 0; i < num_compartments; i++) {
		// assign data to entry
		compartment_entry->ci = i;
		strncpy(compartment_entry->cid, psystem->getCompartment(i)-> \
				getId().c_str(), 255);
		strncpy(compartment_entry->cn, psystem->getCompartment(i)-> \
				getName().c_str(), 255);
		compartment_entry->cx = psystem->getCompartment(i)-> \
				getXPosition();
		compartment_entry->cy = psystem->getCompartment(i)-> \
				getYPosition();
		if(psystem->getCompartment(i)->getNumRuleSets())
			compartment_entry->ct = psystem->getCompartment(i)-> \
					getRuleSetIndex(0);
		else
			compartment_entry->ct = 0;
		compartment_entry->cst = psystem->getCompartment(i)-> \
				getCreationTime();
		compartment_entry->cet = psystem->getCompartment(i)-> \
				getDestructionTime();
		if(compartment_entry->cet < compartment_entry->cst)
			compartment_entry->cet = time;

		// copy data to buffer
		writeDataset(compartment_entry, compartment_buffer, compartment_dset, \
				compartment_dtype, compartment_dspace, compartment_index, \
				compartment_chunk_size);
	}

	// close dataset
	closeDataset(dfile, compartment_dset, compartment_dtype, \
			compartment_buffer, compartment_index, compartment_chunk_size);

	// free data entry memory
	freeDataEntry(compartment_entry);

	// free data buffer memory
	freeDataBuffer(compartment_buffer);

	// close group
	closeGroup(dgroup);
}

// calculate propensity for given reaction in given compartment
double SimulationAlgorithm::calculatePropensity(unsigned long rs, \
		unsigned long r, unsigned long c)
{
	register double propensity;
	register Simulation::Rule *rule;
	register Simulation::Compartment *comp;
	register unsigned long ri1, ri2;
	register unsigned long level1, level2;
#ifdef VARIABLE_RATE_CONSTANTS
	register double dt, brdt;
	register unsigned int N;
#endif

	comp = psystem->getCompartment(c);
	rule = comp->getRule(rs, r);
	switch(rule->getNumReactants()) {
		case 1:
			ri1 = rule->getReactantIndex(0);
			level1 = comp->getLevel(ri1);
#ifdef VARIABLE_RATE_CONSTANTS
			if(rule->isDiffusion()) {  // time dependent diffusion reaction
				// calculate time dependent reaction constant
				dt = time;
				if(dt == 0.0)
					dt = rule->getDiffusionStart();
				brdt = rule->getDiffusionBaseRate() * dt;
				N = rule->getDiffusionNumber();
				propensity = 1.0;
				while(--N)
					propensity += (gsl_sf_pow_int(brdt, N) / \
							gsl_sf_fact(N));
				propensity = (brdt - gsl_sf_log(propensity)) / dt;
				// and now calculate propensity
				propensity *= (double) level1;
			} else  // non-diffusion reaction
#endif
			propensity = comp->getReactionConstant(rs, r) * level1;
			// scale propensity based on distance if translocation rule
			if(rule->isChannel() && scale_translocation)
				propensity /= (0.5 * (comp->getVolume() + \
						getCompartment(comp->getRuleTarget(rs, r))-> \
						getVolume()));
			comp->setPropensity(rs, r, propensity);
			return propensity;
			break;

		case 2:
			ri1 = rule->getReactantIndex(0);
			ri2 = rule->getReactantIndex(1);
			level1 = comp->getLevel(ri1);
			if(ri1 == ri2)
				propensity = 0.5 * comp->getReactionConstant(rs, r) * \
						level1 * (level1 - 1.0);
			else {
				level2 = comp->getLevel(ri2);
				propensity = comp->getReactionConstant(rs, r) * level1 * level2;
			}
//			if(!comp->isVolumeUnit())
//				propensity /= comp->getVolume();
			comp->setPropensity(rs, r, propensity);
			return propensity;
			break;

		default:
			break;
	}
	return 0.0;
}

// output information about state of simulation
void SimulationAlgorithm::sigusr1Handler(void)
{
	// output time
	std::cout << time << " " << run << "\n";
	fflush(stdout);
}

// gracefully exit upon sigusr2
void SimulationAlgorithm::sigusr2Handler(void)
{
	char timestamp[256];

	// ignore signal handler while writing hdf5
//	SignalHandler::ignore_handler(SIGINT);

	// fix variables
	run++;

	// save statistics
	if(log_type != LOG_TYPE_NONE) {
		if(!log_memory) {
			writeAttribute("main_loop_end_time", generateTimestamp(timestamp), \
					dfile, run_group);
			writeAttribute("total_reactions_simulated", num_reactions, \
					dfile, run_group);
			writeAttribute("simulated_time", time, dfile, run_group);
			writeAttribute("run_end_time", generateTimestamp(timestamp), \
					dfile, run_group);
		}

		// finish writing log for run
		closeLoggingDataSet();

		// finish writing simulation dataset
		closeOutputDataSet();
	}

	exit(EXIT_SUCCESS);
}

// set hdf5 dataset compression
void SimulationAlgorithm::setCompress(bool b)
{
	unsigned int filter_info;

	compress = false;
	if(b) {
		// check compression available
		if(!H5Zfilter_avail(H5Z_FILTER_DEFLATE)) {
			cerr << "warning: dataset compression unavailable\n";
			return;
		}
		H5Zget_filter_info(H5Z_FILTER_DEFLATE, &filter_info);
		if(!(filter_info & H5Z_FILTER_CONFIG_ENCODE_ENABLED) || \
				!(filter_info & H5Z_FILTER_CONFIG_DECODE_ENABLED)) {
			cerr << "warning: dataset compression unavailable\n";
			return;
		}
		// turn compression on
		compress = true;
	}
}

// check if simulation running
bool SimulationAlgorithm::isSimulationRunning(void)
{
	return !(time == 0.0 && tau0 == 0.0);
}

// returns true if logging switched on
bool SimulationAlgorithm::isLoggingOn(void)
{
	return !(log_type == LOG_TYPE_NONE);
}

// set rate of volume update
void SimulationAlgorithm::setVolumeRate(unsigned long r)
{
	volume_rate = r;
}

// set rate of division update
void SimulationAlgorithm::setDivisionRate(unsigned long r)
{
	division_rate = r;
}

// set rate of cellular potts update
void SimulationAlgorithm::setPottsRate(unsigned long r)
{
	potts_rate = r;
}

// initialise event handling
void SimulationAlgorithm::initialiseEvents(void)
{
	unsigned long i;
	Compartment *compartment;

	// set compartments with events
	num_event_compartments = 0;
	for(i = 0; i < num_compartments; i++)
		if(getCompartment(i)->getNumEvents())
			num_event_compartments++;

	if(num_event_compartments) {
		// allocate memory for list of compartments containing events
		if(!(event_compartment = (unsigned long *) calloc( \
				num_event_compartments, sizeof(unsigned long)))) {
			cerr << "error: unable to allocate memory for event " \
					"compartment\n";
			exit(EXIT_FAILURE);
		}

		// set list of compartments containing events
		num_event_compartments = 0;
		for(i = 0; i < num_compartments; i++)
			if(getCompartment(i)->getNumEvents())
				event_compartment[num_event_compartments++] = i;

		// set events defined flag
		events_defined = true;
	}
}

void SimulationAlgorithm::setNumRuns(unsigned long i)
{
	num_runs = i;
	#ifdef HAVE_MPI
	if(nthreads > num_runs && !mpi_serial)
		nthreads = num_runs + 1;
	#endif
}

// get amount from amount buffer
unsigned int SimulationAlgorithm::getAmountFromBuffer(unsigned long cid, \
		unsigned long sid, unsigned long timepoint)
{
	// sanity check
	if(cid < 0 || cid >= num_compartments) {
		cerr << "error: cannot get level of species id=" << sid << \
				" in compartment id=" << cid << \
				" at timepoint " << timepoint << " - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}
	if(sid < 0 || sid >= num_species) {
		cerr << "error: cannot get level of species id=" << sid << \
				" in compartment id=" << cid << \
				" at timepoint " << timepoint << " - invalid species id\n";
		exit(EXIT_FAILURE);
	}
	if(timepoint < 0 || timepoint >= amount_index) {
		cerr << "error: cannot get level of species id=" << sid << \
				" in compartment id=" << cid << \
				" at timepoint " << timepoint << " - invalid timepoint\n";
		exit(EXIT_FAILURE);
	}
	
	return amount_buffer[sid][cid][timepoint];
}
