// simulation algorithm class definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#ifndef SIMULATIONALGORITHM_H
#define SIMULATIONALGORITHM_H

#include <string>
#include <hdf5.h>
#include <float.h>
#include <limits.h>
#include <ecsb/ecsb.h>
#ifdef HAVE_MPI
	#include <mpi.h>
#endif
#include <gsl/gsl_odeiv.h>

#include "SimulationParameters.h"
#include "SimulationDatasets.h"
#include "Compartment.h"

using namespace ECSB;
using namespace SimulationDatasets;

// identifiers for different log types
#define LOG_TYPE_REACTIONS	0
#define LOG_TYPE_LEVELS		1
#define LOG_TYPE_NONE		2

#define TAUMAX		DBL_MAX
#define RULESETMAX	ULONG_MAX
#define RULEMAX		ULONG_MAX

#define SIMULATION_FINISHED		0
#define SIMULATION_CANCELLED	1

namespace Simulation {

// simulation algorithms abstract class
class SimulationAlgorithm {
	protected:
		std::string name;  // name of algorithm
		double time;  // current simulation time
		double tau0;  // time to next reaction
		double tau0global;  // time of next reaction
		double end_time;  // time to stop simulation
		unsigned long i0;  // compartment containing next reaction
		unsigned long mu0;  // index of next reaction
		unsigned long rs0;  // index of ruleset of next reaction

		unsigned long num_compartments;  // number of compartments simulated
		unsigned long num_species;  // number of species simulated
		unsigned long num_rules;  // number of rules in templates
		unsigned long total_rules;  // total number of rules in initial model
		unsigned long num_timepoints; // number of timepoints recorded
		unsigned long num_runs;  // number of runs to perform
		unsigned long long num_reactions;  // total reactions performed
		unsigned long seed;  // seed used to initialise random number generator
		unsigned long run;  // current run
		char run_group[64];  // name of group containing current run data

		class Psystem *psystem;  // simulated p-system
		bool show_progress;  // output current log interval to screen
		unsigned int progress_interval;  // time interval to output progress
		unsigned int last_progress_time;  // last time progress output
		void sigusr1Handler(void);  // handle USR1 signal
		void sigusr2Handler(void);  // handle USR2 signal
		RNG *rng;  // random number generator

		bool keep_divisions;  // keep dividing cells
		char division_direction;  // direction of cell division
		bool scale_translocation;  // scale translocation rules with volume
		bool first_time;  // flag indicating if first time called
		unsigned long volume_rate;  // rate of volume update
		unsigned long division_rate;  // rate of division update
		unsigned long potts_rate;  // rate of cellular potts update
		bool static_lattice;  // use a static lattice
		bool events_defined;  // events defined
		bool mpi_serial;  // don't write parallel mpi files

		unsigned long num_event_compartments;  // compartments containing events
		unsigned long *event_compartment;

		// signal handling
		sigset_t sig_mask_old;  // original signal mask
		int simulation_status;  // status of simulation
		pthread_t ppid;  // parent process id
		pthread_mutex_t execute_running_mutex;
		bool execute_running;
		pthread_mutex_t hdf5_mutex;

		// data logging
		bool compress;  // compress datasets
		unsigned int compression_level;  // compression level (0-9, 9 best)
		hid_t dfile;  // output data file
		hid_t dgroup;  // output data group
		unsigned long log_type;  // type of log to save
		double log_interval;  // time interval (seconds) to log species levels
		bool log_levels;  // flag indicating if species levels are logged
		bool log_propensities;  // flag indicating if propensities are logged
		bool log_volumes;  // flag indicating if volumes are logged
		bool log_memory;  // don't log output to file but store in memory
		bool log_steady_state;  // log up to max_time if steady state reached
		bool log_lattice;  // log lattice
		// data buffers
		// reactions dataset
		ReactionEntry *reaction_buffer;  // buffer to store data
		hid_t reaction_dtype;  // hdf5 datatype
		hid_t reaction_dset;  // hdf5 dataset
		hid_t reaction_dspace;  // hdf5 dataspace
		unsigned long reaction_chunk_size;  // chunk size
		unsigned long reaction_index;
		unsigned long reaction_offset;
		// amounts dataset
		unsigned int ***amount_buffer;  // buffer to store data
		hid_t amount_dtype;  // hdf5 datatype
		hid_t amount_dset;  // hdf5 dataset
		hid_t amount_dspace;  // hdf5 dataspace
		unsigned long amount_chunk_size;
		unsigned long amount_index;
		unsigned long amount_offset;
		// volumes dataset
		double **volume_buffer;  // buffer to store data
		hid_t volume_dtype;  // hdf5 datatype
		hid_t volume_dset;  // hdf5 dataset
		hid_t volume_dspace;  // hdf5 dataspace
		unsigned long volume_chunk_size;
		unsigned long volume_index;
		unsigned long volume_offset;
		// propensities dataset
		void *propensity_buffer;  // buffer to store data
		void *propensity_entry;  // single entry in buffer
		hid_t propensity_dtype;  // hdf5 datatype
		hid_t propensity_dset;  // hdf5 dataset
		hid_t propensity_dspace;  // hdf5 dataspace
		unsigned long propensity_chunk_size;
		unsigned long propensity_index;
		// compartment dataset
		CompartmentEntry *compartment_buffer;  // buffer to store data
		CompartmentEntry *compartment_entry;  // single entry in buffer
		hid_t compartment_dtype;  // hdf5 datatype
		hid_t compartment_dset;  // hdf5 dataset
		hid_t compartment_dspace;  // hdf5 dataspace
		unsigned long compartment_chunk_size;
		unsigned long compartment_index;
		// positions dataset
		PositionEntry *position_buffer;  // buffer to store data
		PositionEntry *position_entry;  // single entry in buffer
		hid_t position_dtype;  // hdf5 datatype
		hid_t position_dset;  // hdf5 dataset
		hid_t position_dspace;  // hdf5 dataspace
		unsigned long position_chunk_size;
		unsigned long position_index;
		// lattice spin dataset
		unsigned int ***lattice_spin_buffer;  // buffer to store data
		hid_t lattice_spin_dtype;  // hdf5 datatype
		hid_t lattice_spin_dset;  // hdf5 dataset
		hid_t lattice_spin_dspace;  // hdf5 dataspace
		unsigned long lattice_spin_chunk_size;
		unsigned long lattice_spin_index;
		unsigned long lattice_spin_offset;
		// lattice type dataset
		unsigned int ***lattice_type_buffer;  // buffer to store data
		hid_t lattice_type_dtype;  // hdf5 datatype
		hid_t lattice_type_dset;  // hdf5 dataset
		hid_t lattice_type_dspace;  // hdf5 dataspace
		unsigned long lattice_type_chunk_size;
		unsigned long lattice_type_index;
		unsigned long lattice_type_offset;

		// deterministic ode solver parameters
		// ode solver parameters
		gsl_odeiv_step_type *ode_algorithm;  // ode solver algorithm
		double ode_eps_abs;  // absolute error
		double ode_eps_rel;  // relative error
		double ode_initial_step_size;  // initial step size
		double ode_yscaling;  // y scaling factor
		double ode_dyscaling;  // y scaling factor

		#ifdef HAVE_MPI
		int tid;  // thread id for mpi process
		int nthreads;  // number of threads
		#endif

	public:
		// constructor
		SimulationAlgorithm(const char *, const unsigned long seed = 0);

		// destructor
		virtual ~SimulationAlgorithm();

		// general access methods
		std::string getName(void);
		void setName(std::string);
		double getTime(void);
		void setTime(double);
		void advanceTime(double);
		Psystem *getPsystem(void);
		void setPsystem(Psystem *);
		double getTimeToNextReaction(void);
		double getGlobalTimeToNextReaction(void);
		void setGlobalTimeToNextReaction(double);
		unsigned long getNextCompartment(void);
		unsigned long getNextRuleset(void);
		unsigned long getNumCompartments(void);
		void setNumCompartments(unsigned long);
		void setNumSpecies(unsigned long);
		void setNumRuns(unsigned long);
		Simulation::Compartment *getCompartment(unsigned long);
		void incrementNumReactions(void);
		unsigned long getLogType(void);
		void setLogType(unsigned long);
		hid_t getDataFile(void);
		unsigned long getNumTimepoints(void);
		void setNumTimepoints(unsigned long);
		unsigned int ***getAmountsDataset(void);
		void setShowProgress(bool);
		void setDivisionDirection(char);
		void setKeepDivisions(bool);
		void setCompress(bool);
		void setCompressionLevel(unsigned int);
		void setScaleTranslocation(bool);
		double getLogInterval(void);
		void setStaticLattice(bool);
		void setEventsDefined(bool);
		void setMpiSerial(bool);
		RNG *getRNG(void);
		#ifdef HAVE_MPI
		int getThreadID(void);
		#endif
		void setOdeSolver(const gsl_odeiv_step_type *alg);
		void setAbsoluteError(double v);
		void setRelativeError(double v);
		void setInitialStepSize(double v);
		void setYScaling(double v);
		void setDYScaling(double v);

		// initialise algorithm
		virtual int preprocess(void) = 0;

		// perform one iteration of algorithm for all compartments
		virtual int execute(void) = 0;

		// perform one iteration of algorithm for given compartment
		virtual void execute(unsigned long) = 0;

		// iterate p-system for specified time and runs
		void execute(double end_time);

		// calculate propensity for given reaction in given compartment
		double calculatePropensity(unsigned long, unsigned long, unsigned long);

		// create output dataset for saving results
		void createOutputDataset(SimulationParameters &);

		// create dataset for logging during simulation run
		void createLoggingDataset(const double end_time, \
				const unsigned long run);

		// extend amounts dataset
		void extendAmountsDataSet(void);

		// extend volumes dataset
		void extendVolumesDataSet(void);

		// extend lattice spin dataset
		void extendLatticeSpinDataSet(void);

		// extend lattice type dataset
		void extendLatticeTypeDataSet(void);

		// save initial simulation data
		void writeInitialDataSet(void);

		// finish writing simulation data
		void closeOutputDataSet(void);

		// close reactions dataset
		void closeReactionsDataSet(void);

		// close amounts dataset
		void closeAmountsDataSet(void);

		// close volumes dataset
		void closeVolumesDataSet(void);

		// close lattice spin dataset
		void closeLatticeSpinDataSet(void);

		// close lattice type dataset
		void closeLatticeTypeDataSet(void);

		// finish writing compartment dataset
		void closeLoggingDataSet(void);

		// write reaction data to dataset
		void writeReaction(double, unsigned long, unsigned long, \
				unsigned long, unsigned long, unsigned long);

		// write level data to dataset
		void writeAmount(double);

		// write volume data to dataset
		void writeVolume(double);

		// write lattice spin data
		void writeLatticeSpin(double);

		// write lattice type data
		void writeLatticeType(double);

		// write position data
		void writePosition(unsigned long);

		// write propensities data to dataset
		void writePropensities(double);

		// reset simulation
		void resetSimulation(void);

		// create a daughter compartment from a parent compartment
		void divideCompartment(unsigned long);

		// check if simulation running
		bool isSimulationRunning(void);

		// returns true if logging switched on
		bool isLoggingOn(void);

		// set rate of volume updates
		void setVolumeRate(unsigned long);

		// set rate of division updates
		void setDivisionRate(unsigned long);

		// set rate of cellular potts updates
		void setPottsRate(unsigned long);

		// initialise event handling
		void initialiseEvents(void);

		// entry point for thread to execute simulation (no logging)
		static void *execute_no_logging_thread(void *args);

		// execute simulation (no logging)
		void execute_no_logging(void);

		// entry point for thread to execute simulation (logging)
		static void *execute_logging_thread(void *args);

		// execute simulation (logging)
		void execute_logging(void);

		// get amount from amount buffer
		unsigned int getAmountFromBuffer(unsigned long cid, \
			unsigned long sid, unsigned long timepoint);
};

}

#endif
