// simulation query definitions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "SimulationQuery.h"
#include "SimulationDatasets.h"

using namespace Simulation;
using namespace SimulationDatasets;
using namespace ECSB;

#include <stdio.h>
#include <unistd.h>
#include <gsl/gsl_statistics_double.h>

// set sbml paramter
void Simulation::sbml_set_parameter(const char *model, const char *parameter, \
		double value)
{
	SBMLDocument *sbml_document;
	SBMLWriter sbml_writer;
	Model *sbml_model;
	ListOfParameters *sbml_parameters;
	Parameter *sbml_parameter;
	ListOfReactions *sbml_reactions;
	Reaction *sbml_reaction;
	bool flag;
	unsigned int i, j;

	// read sbml document
	sbml_document = readSBML(model);
	if(sbml_document->getNumErrors()) {
		cerr << "error opening sbml document " << model << "\n";
		sbml_document->printErrors(cerr);
		exit(EXIT_FAILURE);
	}

	// get model from document
	if(!(sbml_model = sbml_document->getModel())) {
		cerr << "error fetching sbml model in document " << model << "\n";
		exit(EXIT_FAILURE);
	}

	flag = false;  // set to true if parameter found

	// search global parameters and set to specified value if found
	sbml_parameters = sbml_model->getListOfParameters();
	for(i = 0; i < sbml_parameters->size(); i++) {
		sbml_parameter = sbml_parameters->get(i);
		if(sbml_parameter->isSetId())
			if(strcmp(sbml_parameter->getId().c_str(), parameter) == 0) {
				// parameter found
				sbml_parameter->setValue(value);
				flag = true;
			}
	}

	// search parameters in reactions and set to specified value if found
	sbml_reactions = sbml_model->getListOfReactions();
	for(i = 0; i < sbml_reactions->size(); i++) {
		sbml_reaction = sbml_reactions->get(i);
		if(sbml_reaction->isSetKineticLaw()) {
			sbml_parameters = sbml_reaction->getKineticLaw()-> \
					getListOfParameters();
			for(j = 0; j < sbml_parameters->size(); j++) {
				sbml_parameter = sbml_parameters->get(j);
				if(sbml_parameter->isSetId())
					if(strcmp(sbml_parameter->getId().c_str(), parameter) == 0) {
						// parameter found
						sbml_parameter->setValue(value);
						flag = true;
					}
			}
		}
	}

	// warn if parameter not found
	if(!flag) {
		cerr << "warning: parameter " << parameter << \
				" not found in document " << model << "\n";
		delete sbml_document;
		return;
	}

	// write sbml document
	ofstream outfile(model, ios::out);
	if(!outfile.is_open()) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	if(!sbml_writer.writeSBML(sbml_document, outfile)) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	outfile.flush();
	outfile.close();

	// tidy up
	delete sbml_document;
}

// set sbml initial amount
void Simulation::sbml_set_initial_amount(const char *model, const char *species, \
		const char *compartment, unsigned int amount)
{
	SBMLDocument *sbml_document;
	SBMLWriter sbml_writer;
	Model *sbml_model;
	ListOfCompartments *sbml_compartments;
	::Compartment *sbml_compartment;
	ListOfSpecies *sbml_species;
	Species *sbml_specie;
	bool flag;
	unsigned int i;

	// read sbml document
	sbml_document = readSBML(model);
	if(sbml_document->getNumErrors()) {
		cerr << "error opening sbml document " << model << "\n";
		sbml_document->printErrors(cerr);
		exit(EXIT_FAILURE);
	}

	// get model from document
	if(!(sbml_model = sbml_document->getModel())) {
		cerr << "error fetching sbml model in document " << model << "\n";
		exit(EXIT_FAILURE);
	}

	// search compartments and find compartment id
	flag = false;  // set to true if compartment found
	sbml_compartments = sbml_model->getListOfCompartments();
	for(i = 0; i < sbml_compartments->size(); i++) {
		sbml_compartment = sbml_compartments->get(i);
		if(sbml_compartment->isSetId() && sbml_compartment->isSetName())
			if(strcmp(sbml_compartment->getName().c_str(), compartment) == 0) {
				// compartment found
				flag = true;
				break;
			}
	}

	// warn if compartment not found
	if(!flag) {
		cerr << "warning: compartment " << compartment << \
				" not found in document " << model << "\n";
		delete sbml_document;
		return;
	}

	// search compartment for species and set to specified value if found
	flag = false;  // set to true if species found
	sbml_species = sbml_model->getListOfSpecies();
	for(i = 0; i < sbml_species->size(); i++) {
		sbml_specie = sbml_species->get(i);
		if(sbml_specie->isSetId() && sbml_specie->isSetName() && \
					sbml_specie->isSetCompartment())
			if(strcmp(sbml_specie->getCompartment().c_str(), \
					sbml_compartment->getId().c_str()) == 0 && \
					strcmp(sbml_specie->getName().c_str(), species) == 0) {
				if(sbml_specie->isSetInitialConcentration())
					sbml_specie->setInitialConcentration((double) amount);
				else
					sbml_specie->setInitialAmount((double) amount);
				flag = true;
			}
	}

	// warn if species not found
	if(!flag) {
		cerr << "warning: species " << species << " in compartment " \
				<< compartment << " not found in document " << model << "\n";
		delete sbml_document;
		return;
	}

	// write sbml document
	ofstream outfile(model, ios::out);
	if(!outfile.is_open()) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	if(!sbml_writer.writeSBML(sbml_document, outfile)) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	outfile.flush();
	outfile.close();

	// tidy up
	delete sbml_document;
}

// set assignment in sbml event
void Simulation::sbml_set_event_assignment(const char *model, \
		const char *variable, const char *event, long double value)
{
	SBMLDocument *sbml_document;
	SBMLWriter sbml_writer;
	Model *sbml_model;
	ListOfEvents *sbml_events;
	::Event *sbml_event;
	ListOfEventAssignments *sbml_event_assignments;
	EventAssignment *sbml_event_assignment;
	ASTNode *astnode;
	bool flag;
	unsigned int i, j;

	// read sbml document
	sbml_document = readSBML(model);
	if(sbml_document->getNumErrors()) {
		cerr << "error opening sbml document " << model << "\n";
		sbml_document->printErrors(cerr);
		exit(EXIT_FAILURE);
	}

	// get model from document
	if(!(sbml_model = sbml_document->getModel())) {
		cerr << "error fetching sbml model in document " << model << "\n";
		exit(EXIT_FAILURE);
	}

	// search events and find variable assignment
	flag = false;  // set to true if event found
	sbml_events = sbml_model->getListOfEvents();
	for(i = 0; i < sbml_events->size(); i++) {
		sbml_event = sbml_events->get(i);
		if(sbml_event->isSetName())
			if(strcmp(sbml_event->getName().c_str(), event) == 0) {
				sbml_event_assignments = sbml_event->getListOfEventAssignments();
				for(j = 0; j < sbml_event_assignments->size(); j++) {
					sbml_event_assignment = sbml_event_assignments->get(j);
					if(sbml_event_assignment->isSetVariable())
						if(strcmp(sbml_event_assignment->getVariable().c_str() \
									, variable) == 0 && \
									sbml_event_assignment->isSetMath()) {
							// variable found
							astnode = (ASTNode *) sbml_event_assignment->getMath();
							if(astnode->isInteger())
								astnode->setValue((long int) value);
							if(astnode->isReal())
								astnode->setValue((double) value);
							flag = true;
						}
				}
			}
	}

	// warn if variable not found
	if(!flag) {
		cerr << "warning: variable " << variable << " in event " \
				<< event << " not found in document " << model << "\n";
		delete sbml_document;
		return;
	}

	// write sbml document
	ofstream outfile(model, ios::out);
	if(!outfile.is_open()) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	if(!sbml_writer.writeSBML(sbml_document, outfile)) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	outfile.flush();
	outfile.close();

	// tidy up
	delete sbml_document;
}

// set sbml compartment_volume
void Simulation::sbml_set_compartment_volume(const char *model, \
		const char *compartment, double value)
{
	SBMLDocument *sbml_document;
	SBMLWriter sbml_writer;
	Model *sbml_model;
	ListOfCompartments *sbml_compartments;
	::Compartment *sbml_compartment;
	bool flag;
	unsigned int i;

	// read sbml document
	sbml_document = readSBML(model);
	if(sbml_document->getNumErrors()) {
		cerr << "error opening sbml document " << model << "\n";
		sbml_document->printErrors(cerr);
		exit(EXIT_FAILURE);
	}

	// get model from document
	if(!(sbml_model = sbml_document->getModel())) {
		cerr << "error fetching sbml model in document " << model << "\n";
		exit(EXIT_FAILURE);
	}

	// search compartments and find compartment id
	flag = false;  // set to true if compartment found
	sbml_compartments = sbml_model->getListOfCompartments();
	for(i = 0; i < sbml_compartments->size(); i++) {
		sbml_compartment = sbml_compartments->get(i);
		if(sbml_compartment->isSetId() && sbml_compartment->isSetName())
			if(strcmp(sbml_compartment->getName().c_str(), compartment) == 0) {
				// compartment found
				flag = true;
				break;
			}
	}

	// warn if compartment not found
	if(!flag) {
		cerr << "warning: compartment " << compartment << \
				" not found in document " << model << "\n";
		delete sbml_document;
		return;
	}

	// set compartment volume
	sbml_compartment->setVolume(value);

	// write sbml document
	ofstream outfile(model, ios::out);
	if(!outfile.is_open()) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	if(!sbml_writer.writeSBML(sbml_document, outfile)) {
		cerr << "error writing sbml document %s " << model << "\n";
		exit(EXIT_FAILURE);
	}
	outfile.flush();
	outfile.close();

	// tidy up
	delete sbml_document;
}

// get index of species
int Simulation::mcss_species_index(hid_t dfile, const char *species)
{
	hid_t dset, dtype;
	SpeciesEntry *species_buffer;
	int index, buffer_size;

 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer( \
			(unsigned long) buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// find species
	for(index = 0; index < buffer_size; index++)
		if(strcmp(species, species_buffer[index].sn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(species_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(index < buffer_size)
		return(index);
	else
		return(-1);
}

// get index of compartment
int Simulation::mcss_compartment_index(hid_t dfile, const char *compartment, \
		unsigned int run)
{
	hid_t dset, dtype;
	CompartmentEntry *compartment_buffer;
	int index, buffer_size;
	char dset_name[1024];

 	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 1024, "/run%u/compartment_information", run);
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) createDataBuffer((unsigned long) \
			buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// find compartment
	for(index = 0; index < buffer_size; index++)
		if(strcmp(compartment, compartment_buffer[index].cn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(compartment_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(index < buffer_size)
		return(index);
	else
		return(-1);
}

// get average level of species
void Simulation::mcss_average_level(const char *data_file, const char *species, \
		const char *compartment, double time, double *mean, double *sd)
{
	hsize_t dims[1], coord[1][3];
	hid_t dfile, dset, dattr, dspace, dgroup, mspace, dtype;
	char dset_name[128], attribute[128], group[128];
	unsigned int value, run;
	unsigned long nr;
	double *values, log_interval;
	int species_index, compartment_index, buffer_size;
	H5E_auto2_t old_func;
	void *old_client_data;
	SpeciesEntry *species_buffer;
	CompartmentEntry *compartment_buffer;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	snprintf(attribute, 128, "number_of_runs");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if(H5Aread(dattr, H5T_NATIVE_ULONG, &nr) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	snprintf(attribute, 128, "log_interval");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &log_interval) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	// get species index
 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer( \
			(unsigned long) buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// find species
	for(species_index = 0; species_index < buffer_size; species_index++)
		if(strcmp(species, species_buffer[species_index].sn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(species_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(species_index == buffer_size) {
		H5Fclose(dfile);
		fprintf(stderr, "error: couldn't find species %s in %s\n", \
				species, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// get compartment index
 	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 128, "/run%u/compartment_information", run);
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) createDataBuffer((unsigned long) \
			buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// find compartment
	for(compartment_index = 0; compartment_index < buffer_size; \
				compartment_index++)
		if(strcmp(compartment, compartment_buffer[compartment_index].cn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(compartment_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(compartment_index == buffer_size) {
		H5Fclose(dfile);
		fprintf(stderr, "error: couldn't find compartment %s in %s\n", \
				compartment, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// allocate memory for values
	if(!(values = (double *) calloc(nr, sizeof(double)))) {
		fprintf(stderr, "error: couldn't allocate memory for %s\n", data_file);
		exit(EXIT_FAILURE);
	}

	dims[0] = 1;
	if((mspace = H5Screate_simple(1, dims, NULL)) < 0) {
		fprintf(stderr, "error: unable to create mspace for %s\n", \
				data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	coord[0][0] = (hsize_t) species_index;  // species
	coord[0][1] = (hsize_t) compartment_index;  // compartment
	coord[0][2] = (hsize_t) (time / log_interval);  // timepoint
	for(run = 1; run <= nr; run++) {
		snprintf(dset_name, 128, "/run%u/amounts", run);
		if((dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
			fprintf(stderr, "error: unable to open dataset %s in %s\n", \
					dset_name, data_file);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}
		dspace = H5Dget_space(dset);
		dtype = H5Dget_type(dset);

		if(H5Sselect_elements(dspace, H5S_SELECT_SET, 1, \
					(const hsize_t *) coord) < 0) {
			fprintf(stderr, "error: unable to select data at timepoint %f " \
					"from %s\n", time, data_file);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}

		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, &value) < 0) {
			fprintf(stderr, "error: unable to read data at timepoint %f " \
					"from %s\n", time, data_file);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}
		values[run - 1] = (double) value;

		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
	}

	H5Sclose(mspace);
	H5Fclose(dfile);

	if(nr == 1) {
		*mean = values[0];
		*sd = 0.0;
	} else {
		*mean = gsl_stats_mean(values, 1, nr);
		*sd = gsl_stats_sd(values, 1, nr);
	}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
	free(values);
}

// get average level of species for multiple timepoints
void Simulation::mcss_average_level(const char *data_file, \
		const char *species, const char *compartment, double *time, \
		unsigned int num_timepoints, double *means, double *sds)
{
 	hsize_t dims[2], *coord;
	hid_t dfile, dset, dattr, dspace, dgroup, mspace, dtype;
	char dset_name[128], attribute[128], group[128];
	unsigned int **values, i, run, ind;
	unsigned long nr;
	double log_interval, mean, sd;
	int species_index, compartment_index, buffer_size;
	H5E_auto2_t old_func;
	void *old_client_data;
	SpeciesEntry *species_buffer;
	CompartmentEntry *compartment_buffer;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	snprintf(attribute, 128, "number_of_runs");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if(H5Aread(dattr, H5T_NATIVE_ULONG, &nr) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	snprintf(attribute, 128, "log_interval");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &log_interval) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	// get species index
 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer( \
			(unsigned long) buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// find species
	for(species_index = 0; species_index < buffer_size; species_index++)
		if(strcmp(species, species_buffer[species_index].sn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(species_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(species_index == buffer_size) {
		fprintf(stderr, "error: couldn't find species %s in %s\n", \
				species, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// get compartment index
 	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 128, "/run%u/compartment_information", 1);
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) createDataBuffer((unsigned long) \
			buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// find compartment
	for(compartment_index = 0; compartment_index < buffer_size; \
				compartment_index++)
		if(strcmp(compartment, compartment_buffer[compartment_index].cn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(compartment_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(compartment_index == buffer_size) {
		fprintf(stderr, "error: couldn't find compartment %s in %s\n", \
				compartment, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// allocate memory for values and timepoints
	dims[0] = nr;
	dims[1] = num_timepoints;
	values = create2DArrayUInt(dims);
	if(!(coord = (hsize_t *) malloc(num_timepoints * 3 * sizeof(hsize_t)))) {
		fprintf(stderr, "error: unable to allocate memory for coord\n");
		exit(EXIT_FAILURE);
	}
	ind = 0;
	for(i = 0; i < num_timepoints; i++) {
		// set timepoints
		coord[ind++] = (hsize_t) species_index;  // species
		coord[ind++] = (hsize_t) compartment_index;  // compartment
		coord[ind++] = (hsize_t) (time[i] / log_interval);  // timepoint
	}

	dims[0] = num_timepoints;
	if((mspace = H5Screate_simple(1, dims, NULL)) < 0) {
		fprintf(stderr, "error: unable to create mspace for %s\n", \
				data_file);
		H5Fclose(dfile);
		free(coord);
		free2DArrayUInt(values);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	for(run = 1; run <= nr; run++) {
		snprintf(dset_name, 128, "/run%u/amounts", run);
		if((dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
			fprintf(stderr, "error: unable to open dataset %s in %s\n", \
					dset_name, data_file);
			H5Sclose(mspace);
			H5Fclose(dfile);
			free(coord);
			free2DArrayUInt(values);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}
		dspace = H5Dget_space(dset);
		dtype = H5Dget_type(dset);

		if(H5Sselect_elements(dspace, H5S_SELECT_SET, num_timepoints, \
					(const hsize_t *) coord) < 0) {
			fprintf(stderr, "error: unable to select data at timepoints " \
					"from %s\n", data_file);
			H5Tclose(dtype);
			H5Sclose(dspace);
			H5Dclose(dset);
			H5Sclose(mspace);
			H5Fclose(dfile);
			free(coord);
			free2DArrayUInt(values);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}

		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, values[run - 1]) < 0) {
			fprintf(stderr, "error: unable to read data at timepoints " \
					"from %s\n", data_file);
			H5Tclose(dtype);
			H5Sclose(dspace);
			H5Dclose(dset);
			H5Sclose(mspace);
			H5Fclose(dfile);
			free(coord);
			free2DArrayUInt(values);
			// turn back on printing of hdf5 errors
			H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
			return;
		}

		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
	}

	H5Sclose(mspace);
	H5Fclose(dfile);
	for(i = 0; i < num_timepoints; i++)
		if(nr == 1) {
			means[i] = (double) values[i][0];
			if(sds)
				sds[i] = 0.0;
		} else {
			mean = 0.0;
			for(run = 0; run < nr; run++)
				mean += (double) values[run][i];
			mean /= (double) nr;
			means[i] = mean;
			if(sds) {
				sd = 0.0;
				for(run = 0; run < nr; run++)
					sd += (((double) values[run][i] - mean) * \
							((double) values[run][i] - mean));
				sd = sqrt(sd) / ((double) nr - 1.0);
				sds[i] = sd;
			}
		}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
	free(coord);
	free2DArrayUInt(values);
}

// get level of species for multiple timepoints
void Simulation::mcss_level(const char *data_file, const unsigned int run, \
		const char *species, const char *compartment, double *time, \
		unsigned int num_timepoints, double *levels)
{
 	hsize_t dims[2], *coord;
	hid_t dfile, dset, dattr, dspace, dgroup, mspace, dtype;
	char dset_name[128], attribute[128], group[128];
	unsigned int *values, i, ind;
	double log_interval;
	int species_index, compartment_index, buffer_size;
	H5E_auto2_t old_func;
	void *old_client_data;
	SpeciesEntry *species_buffer;
	CompartmentEntry *compartment_buffer;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	snprintf(attribute, 128, "log_interval");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &log_interval) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	// get species index
 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer( \
			(unsigned long) buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// find species
	for(species_index = 0; species_index < buffer_size; species_index++)
		if(strcmp(species, species_buffer[species_index].sn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(species_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(species_index == buffer_size) {
		fprintf(stderr, "error: couldn't find species %s in %s\n", \
				species, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// get compartment index
 	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 128, "/run%u/compartment_information", 1);
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) createDataBuffer((unsigned long) \
			buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// find compartment
	for(compartment_index = 0; compartment_index < buffer_size; \
				compartment_index++)
		if(strcmp(compartment, compartment_buffer[compartment_index].cn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(compartment_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(compartment_index == buffer_size) {
		fprintf(stderr, "error: couldn't find compartment %s in %s\n", \
				compartment, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	// allocate memory for values and timepoints
	if(!(values = (unsigned int *) malloc(num_timepoints * \
				sizeof(unsigned int)))) {
		fprintf(stderr, "error: unable to allocate memory for values\n");
		exit(EXIT_FAILURE);
	}
	if(!(coord = (hsize_t *) malloc(num_timepoints * 3 * sizeof(hsize_t)))) {
		fprintf(stderr, "error: unable to allocate memory for coord\n");
		exit(EXIT_FAILURE);
	}
	ind = 0;
	for(i = 0; i < num_timepoints; i++) {
		// set timepoints
		coord[ind++] = (hsize_t) species_index;  // species
		coord[ind++] = (hsize_t) compartment_index;  // compartment
		coord[ind++] = (hsize_t) (time[i] / log_interval);  // timepoint
	}

	dims[0] = num_timepoints;
	if((mspace = H5Screate_simple(1, dims, NULL)) < 0) {
		fprintf(stderr, "error: unable to create mspace for %s\n", \
				data_file);
		H5Fclose(dfile);
		free(coord);
		free(values);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	snprintf(dset_name, 128, "/run%u/amounts", run);
	if((dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open dataset %s in %s\n", \
				dset_name, data_file);
		H5Sclose(mspace);
		H5Fclose(dfile);
		free(coord);
		free(values);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}
	dspace = H5Dget_space(dset);
	dtype = H5Dget_type(dset);

	if(H5Sselect_elements(dspace, H5S_SELECT_SET, num_timepoints, \
				(const hsize_t *) coord) < 0) {
		fprintf(stderr, "error: unable to select data at timepoints " \
				"from %s\n", data_file);
		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
		H5Sclose(mspace);
		H5Fclose(dfile);
		free(coord);
		free(values);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, values) < 0) {
		fprintf(stderr, "error: unable to read data at timepoints " \
				"from %s\n", data_file);
		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
		H5Sclose(mspace);
		H5Fclose(dfile);
		free(coord);
		free(values);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return;
	}

	H5Tclose(dtype);
	H5Sclose(dspace);
	H5Dclose(dset);

	H5Sclose(mspace);
	H5Fclose(dfile);

	for(i = 0; i < num_timepoints; i++)
		levels[i] = (double) values[i];

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
	free(coord);
	free(values);
}

// get level of species for single timepoint
unsigned int Simulation::mcss_level(const char *data_file, \
		const unsigned int run, const char *species, const char *compartment, \
		double time)
{
 	hsize_t dims[2], coord[3];
	hid_t dfile, dset, dattr, dspace, dgroup, mspace, dtype;
	char dset_name[128], attribute[128], group[128];
	unsigned int value, i, ind;
	double log_interval;
	int species_index, compartment_index, buffer_size;
	H5E_auto2_t old_func;
	void *old_client_data;
	SpeciesEntry *species_buffer;
	CompartmentEntry *compartment_buffer;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}

	snprintf(attribute, 128, "log_interval");
	snprintf(group, 128, "/");
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &log_interval) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		H5Fclose(dfile);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	// get species index
 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer( \
			(unsigned long) buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// find species
	for(species_index = 0; species_index < buffer_size; species_index++)
		if(strcmp(species, species_buffer[species_index].sn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(species_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(species_index == buffer_size) {
		fprintf(stderr, "error: couldn't find species %s in %s\n", \
				species, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}

	// get compartment index
 	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 128, "/run%u/compartment_information", 1);
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = (int) getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) createDataBuffer((unsigned long) \
			buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// find compartment
	for(compartment_index = 0; compartment_index < buffer_size; \
				compartment_index++)
		if(strcmp(compartment, compartment_buffer[compartment_index].cn) == 0)
			break;

	// free memory for data buffer
	freeDataBuffer(compartment_buffer);

	// close hdf objects
	H5Tclose(dtype);
	closeDataset(dset);

	if(compartment_index == buffer_size) {
		fprintf(stderr, "error: couldn't find compartment %s in %s\n", \
				compartment, data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}

	// set timepoints
	coord[0] = (hsize_t) species_index;  // species
	coord[1] = (hsize_t) compartment_index;  // compartment
	coord[2] = (hsize_t) (time / log_interval);  // timepoint

	dims[0] = 1;
	if((mspace = H5Screate_simple(1, dims, NULL)) < 0) {
		fprintf(stderr, "error: unable to create mspace for %s\n", \
				data_file);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}
	snprintf(dset_name, 128, "/run%u/amounts", run);
	if((dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open dataset %s in %s\n", \
				dset_name, data_file);
		H5Sclose(mspace);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}
	dspace = H5Dget_space(dset);
	dtype = H5Dget_type(dset);

	if(H5Sselect_elements(dspace, H5S_SELECT_SET, 1, \
				(const hsize_t *) coord) < 0) {
		fprintf(stderr, "error: unable to select data at timepoints " \
				"from %s\n", data_file);
		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
		H5Sclose(mspace);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}

	if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, &value) < 0) {
		fprintf(stderr, "error: unable to read data at timepoints " \
				"from %s\n", data_file);
		H5Tclose(dtype);
		H5Sclose(dspace);
		H5Dclose(dset);
		H5Sclose(mspace);
		H5Fclose(dfile);
		// turn back on printing of hdf5 errors
		H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
		return 0;
	}

	H5Tclose(dtype);
	H5Sclose(dspace);
	H5Dclose(dset);

	H5Sclose(mspace);
	H5Fclose(dfile);

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
	return value;
}


// get number of reactions simulated
unsigned int Simulation::mcss_reactions_simulated(const char *data_file)
{
	hid_t dfile, dgroup, dattr;
	unsigned long value;
	char group[128] = "/";
	char attribute[] = "total_reactions_simulated";
	unsigned int nr, i, reactions;

	nr = mcss_num_runs(data_file);
	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		return 0;
	}
	reactions = 0;
	for(i = 1; i <= nr; i++) {
		snprintf(group, 128, "/run%u", i);
		if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
			fprintf(stderr, "error: unable to open group %s in %s\n", \
					group, data_file);
			return 0;
		}
		if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
			H5Gclose(dgroup);
			fprintf(stderr, "error: unable to open attribute %s in %s\n", \
					attribute, data_file);
			return 0;
		}
		if(H5Aread(dattr, H5T_NATIVE_ULONG, &value) < 0) {
			H5Aclose(dattr);
			H5Gclose(dgroup);
			fprintf(stderr, "error: unable to read attribute %s in %s\n", \
					attribute, data_file);
			return 0;
		}
		H5Aclose(dattr);
		H5Gclose(dgroup);
		reactions += (unsigned int) value;
	}
	H5Fclose(dfile);

	return(reactions);
}

// get simulated time for run
double Simulation::mcss_simulated_time(const char *data_file, unsigned int run)
{
	hid_t dfile, dgroup, dattr;
	double value;
	char group[64];
	char attribute[] = "simulated_time";

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		return 0;
	}
	snprintf(group, 64, "/run%u", run);
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		return 0;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		return 0;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &value) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		return 0;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);
	H5Fclose(dfile);

	return(value);
}

// get number of runs simulated
unsigned int Simulation::mcss_num_runs(const char *data_file)
{
	hid_t dfile, dgroup, dattr;
	unsigned long value;
	char group[] = "/";
	char attribute[] = "number_of_runs";

	if((dfile = H5Fopen(data_file, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open %s\n", data_file);
		return 0;
	}
	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		fprintf(stderr, "error: unable to open group %s in %s\n", \
				group, data_file);
		return 0;
	}
	if((dattr = H5Aopen(dgroup, attribute, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		fprintf(stderr, "error: unable to open attribute %s in %s\n", \
				attribute, data_file);
		return 0;
	}
	if(H5Aread(dattr, H5T_NATIVE_ULONG, &value) < 0) {
		H5Aclose(dattr);
		H5Gclose(dgroup);
		fprintf(stderr, "error: unable to read attribute %s in %s\n", \
				attribute, data_file);
		return 0;
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);
	H5Fclose(dfile);

	return((unsigned int) value);
}
