// boost Psystem class wrapper
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#include "mcss/mcss.h"

using namespace Simulation;

#include <boost/python.hpp>
#include <boost/python/tuple.hpp>

using namespace boost::python;

#include <string.h>

// boost declarations

unsigned long wrap_add_rule(Psystem &ps, unsigned long rsid, string name, \
		const tuple &reactants, const tuple &products, double rc)
{
	unsigned long i, *re, *pr, rid;

	// convert reactants
	if(!(re = (unsigned long *) calloc(len(reactants), sizeof(unsigned long)))) {
		cerr << "error converting tuple\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < (unsigned long) len(reactants); i++)
		re[i] = ps.getSpeciesIndex(extract<string>(reactants[i]));

	// convert products
	if(!(pr = (unsigned long *) calloc(len(products), sizeof(unsigned long)))) {
		cerr << "error converting tuple\n";
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < (unsigned long) len(products); i++)
		pr[i] = ps.getSpeciesIndex(extract<string>(products[i]));

	// call method
	rid = ps.addRule(rsid, name, (unsigned long) len(reactants), re, \
			(unsigned long) len(products), pr, rc);

	// tidy up
	free(pr);
	free(re);

	return rid;
}

void wrap_set_simulation_algorithm(Psystem &ps, string name)
{
	ps.setSimulationAlgorithm(name);
}

unsigned long wrap_add_compartment(Psystem &ps, string name)
{
	return ps.addCompartment(name);
}

void wrap_execute(Psystem &ps, double end_time)
{
	ps.execute(end_time);
}

unsigned long wrap_species_level(Psystem &ps, unsigned long cid, unsigned long sid)
{
	// sanity check
	if(cid < 0 || cid >= ps.getNumCompartments()) {
		cerr << "error: cannot get level of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid compartment id\n";
		exit(EXIT_FAILURE);
	}
	if(sid < 0 || sid >= ps.getNumSpecies()) {
		cerr << "error: cannot get level of species id=" << sid << \
				" in compartment id=" << cid << \
				" - invalid species id\n";
		exit(EXIT_FAILURE);
	}

	return ps.getSpeciesLevel(sid, cid);
}

BOOST_PYTHON_MODULE(Psystem)
{
	// member function pointers for overloaded methods
	unsigned long (Psystem::*num_rulesets1)(void) = &Psystem::getNumRuleSets;
	unsigned long (Psystem::*num_rulesets2)(unsigned long) = \
			&Psystem::getNumRuleSets;
	unsigned long (Psystem::*ruleset_index1)(string) = \
			&Psystem::getRuleSetIndex;
	unsigned long (Psystem::*ruleset_index2)(unsigned long, unsigned long) = \
			&Psystem::getRuleSetIndex;

	// class wrappers
	class_<Psystem>("Psystem", "A P system")
		.def(init<SimulationParameters &>())
		.def(init<char *>())
		.def("species_level", &wrap_species_level, \
				"Get the level of a species in a given compartment")
		.def("set_id", &Psystem::setId, "Set ID of P system")
		.def("id", &Psystem::getId, "Get ID of P system")
		.def("num_compartments", &Psystem::getNumCompartments, \
				"Get number of compartments in P system")
		.def("set_log_reactions", &Psystem::logReactions, "\
				Set reaction logging")
		.def("show", &Psystem::show, "Print P System")
		.def("lattice_x", &Psystem::getLatticeX, \
				"Get x dimension of lattice")
		.def("lattice_y", &Psystem::getLatticeY, \
				"Get y dimension of lattice")
		.def("lattice_z", &Psystem::getLatticeZ, \
				"Get z dimension of lattice")
		.def("num_species", &Psystem::getNumSpecies, \
				"Get number of species in P system")
		.def("species_index", &Psystem::getSpeciesIndex, \
				"Get index of species with given name")
		.def("is_degradation", &Psystem::isDegradation, \
				"Test if species is a degradation species")
		.def("num_timepoints", &Psystem::getNumTimepoints, \
				"Get number of timepoints logged")
		.def("compartment_volume", &Psystem::getCompartmentVolume, \
				"Get volume of a compartment with given index")
		.def("set_compartment_volume", &Psystem::setCompartmentVolume, \
				"Set volume of a compartment with given index")
		.def("calculate_rule_target", &Psystem::calculateRuleTarget, \
				"Calculate rule target")
		.def("execute_rule", &Psystem::executeRule, "Execute a specific rule")
		.def("preprocess", &Psystem::preprocess, "Initialise simulation")
		.def("execute", &wrap_execute, \
				"Execute P system up to a given time")
		.def("output_xml", &Psystem::outputXML, "Save model as a psxml file")
		.def("is_valid_rule", &Psystem::isValidRule, \
				"Calculate if rule is valid for a given compartment")
		.def("save_initial_compartments", &Psystem::saveInitialCompartments, \
				"Save initial state of compartments")
		.def("reset_compartments", &Psystem::resetCompartments, \
				"Reset compartments to initial state")
		.def("set_valid_rules", &Psystem::setValidRules, \
				"Set valid ruleset rules for compartment")
		.def("output_sbml", &Psystem::outputSBML, \
				"Write a fully compliant sbml model")
		.def("dump", &Psystem::dump, "Dump model to binary format file")
		.def("add_species", &Psystem::addSpecies, "Add species to P system")
		.def("add_ruleset", &Psystem::addRuleSet, "Add ruleset to P system")
		.def("add_rule", &wrap_add_rule, "Add rule to P system ruleset")
		.def("set_channel", &Psystem::setChannel, \
				"Set rule as translocation rule")
		.def("add_compartment", &wrap_add_compartment, \
				"Add compartment to P system")
		.def("add_ruleset_to_compartment", \
				&Psystem::addRuleSetToCompartment, "Add ruleset to compartment")
		.def("set_rule_target", &Psystem::setRuleTarget, \
				"Set target of translocation rule")
		.def("clone_rule", &Psystem::cloneRule, "Clone a rule in a ruleset")
		.def("set_species_level", &Psystem::setSpeciesLevel, \
				"Set level of species in a compartment")
		.def("set_simulation_algorithm", &wrap_set_simulation_algorithm, \
				"Set simulation algorithm")
		.def("set_reaction_constant", &Psystem::setReactionConstant, \
				"Set reaction constant of rule in a ruleset")
		.def("reaction_constant", &Psystem::getReactionConstant, \
				"Get reaction constant of rule in a ruleset")
		.def("ruleset_index", ruleset_index1, \
				"Get ruleset index from ruleset name")
		.def("ruleset_index", ruleset_index2, \
				"Get ruleset index for ruleset in given compartment")
		.def("num_rulesets", num_rulesets1, \
				"Get number of rulesets in P system")
		.def("num_rulesets", num_rulesets2, \
				"Get number of rulesets in compartment")
		.def("num_rules", &Psystem::getNumRules, \
				"Get number of rules in rulesets in compartment")
		.def("is_channel", &Psystem::isChannel, \
				"Return true if rule is channel")
		.def("target_name", &Psystem::getTargetName, \
				"Get target name of rule")
		.def("target_type", &Psystem::getTargetType, \
				"Get target type of rule")
		.def("target_compartment", &Psystem::getTargetCompartment, \
				"Get target compartment of rule")
		.def("set_initial_levels", &Psystem::setInitialLevels, \
				"Set initial levels of species in compartment from templates")
		.def("log_interval", &Psystem::getLogInterval, \
				"Get log interval")
		.def("compartment_name", &Psystem::getCompartmentName, \
				"Get name of compartment")
		.def("simulation_time", &Psystem::getSimulationTime, \
				"Get simulation time")
		;
}
