// postprocess output of mcss-cmaes simulation
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include <ecsb/ecsb.h>
#include <mcss/mcss.h>
#include <mcss/McssCmaes.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <iostream>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_cdf.h>

using namespace std;
using namespace ECSB;

// operation modes
#define MODE_FITNESS			1
#define MODE_INFO				2
#define MODE_PARAMETERS			3
#define MODE_ATTRIBUTE_INFO		4
#define MODE_BEST_INDIVIDUAL	5
#define MODE_MEAN_FITNESS		6

// width of model id field
#define ID_WIDTH			7

// hdf5 file
hid_t dfile;

// data buffers
static double **fitness_buffer;
static double **parameters_buffer;

static unsigned long population_size, num_parameters, num_iterations, runs;
static unsigned long num_optimisation_runs;
static unsigned long chunk_size = 1000;
static double best_fitness;
unsigned long best_individual, best_iteration;
static double *best_parameters;
static bool legacy = false;
static bool average_runs = true;
static unsigned long output_run = 0;
static char run_group[64];

// command line parameters
static char infile[1024];
static int mode = MODE_INFO;
static unsigned long *output_individuals;
static unsigned long *output_iterations;
static unsigned long num_output_individuals = 0;
static unsigned long num_output_iterations = 0;
static bool output_header = true;
static FILE *ofp = stdout;
static char *input_individuals = NULL;
static char *input_iterations = NULL;
static char separator = ' ';
static char attribute[1024];
static bool best_flag = false;
static bool best_overall_flag = false;
static bool parallel_runs;

static void parse_command_line(int argc, char **argv)
{
	int c;
	char *ptr, *tptr, *bptr;
	unsigned long i, ind;

	// parse that command line...
	while((c = getopt(argc, argv, \
			"A:i:e:o:s:r:t:bBfFgpqh?")) != -1)
		switch(c) {
			case 'A':  // get attribute value
				mode = MODE_ATTRIBUTE_INFO;
				strncpy(attribute, optarg, 1024);
				break;

			case 's':  // output column separator
				separator = optarg[0];
				break;

			case 'i':  // individuals to output
				if(!(input_individuals = (char *) \
						calloc(strlen(optarg) + 1, sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"individuals\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_individuals, optarg);
				break;

			case 'b':  // output best individual
				best_flag = true;
				break;

			case 'B':  // output best individual overall
				best_overall_flag = true;
				mode = MODE_BEST_INDIVIDUAL;
				break;

			case 'e':  // read levels in one chunk
				chunk_size = atol(optarg);
				break;

			case 'f':  // output fitness data
				mode = MODE_FITNESS;
				break;

			case 'F':  // output mean fitness data
				mode = MODE_MEAN_FITNESS;
				break;

			case 'g':  // output parameter data
				mode = MODE_PARAMETERS;
				break;

			case 'o':  // output to file
				if(!(ofp = fopen(optarg, "w"))) {
					fprintf(stderr, "error: couldn't open output file %s\n", \
							optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case 'p':  // don't output headers with data
				output_header = false;
				break;

			case 'r':  // iterations to output
				if(!(input_iterations = (char *) \
						calloc(strlen(optarg) + 1, sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"iterations\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_iterations, optarg);
				break;

			case 't':  // run to output data for
				output_run = atol(optarg);
				break;
				
			case 'h':
			case '?':
			usage:
				#ifdef HAVE_MPI
				fprintf(stderr, "mcss-cmaes-postprocess version %s (MPI)\n", \
						VERSION);
				#else
				fprintf(stderr, "mcss-cmaes-postprocess version %s\n", VERSION);
				#endif
				fprintf(stderr,
						"usage: mcss-cmaes-postprocess [OPTIONS] INFILE\n" \
						"\t-A ATTRIBUTE - output attribute value\n"
						"\t-b - output best individual\n" \
						"\t-B - output best individual overall\n" \
						"\t-e CHUNK_SIZE - set chunk size\n"
						"\t-f - output fitness data\n" \
						"\t-g - output parameters data\n"
						"\t-h - display this usage message\n" \
						"\t-i i1,i2,... - individuals to output\n" \
						"\t-o FILENAME - save output to file\n"
						"\t-p - don't output header with data\n"
						"\t-r it1,it2,... - iterations to output\n" \
						"\t-s CHAR - output column separator\n"
						"\t-t RUN - optimisation run to output\n"
						);
				exit(EXIT_FAILURE);
				break;

			default:
				goto usage;
				break;
		}
	if(optind != (argc - 1))
		goto usage;
	strncpy(infile, argv[optind], 1024);
}

static void initialise_input(void)
{
	unsigned long i, j, buffer_size, ind;
	char dset_name[1024], *str;
	FILE *fp;
	bool flag;

	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dtype;  // hdf5 datatype

	// open input file
	dfile = openDatafile(infile);

	// setup output group
	if(attributeExists("num_optimisation_runs", dfile)) {
		readAttribute("num_optimisation_runs", num_optimisation_runs, dfile);
		if(output_run > num_optimisation_runs) {
			cerr << "error: run " << output_run << " not found\n";
			exit(EXIT_FAILURE);
		}
		average_runs = false;
		if(num_optimisation_runs == 1)
			output_run = 1;
	} else {
		legacy = true;
		num_optimisation_runs = 1;
		if(output_run > num_optimisation_runs) {
			cerr << "error: run " << output_run << " not found\n";
			exit(EXIT_FAILURE);
		}
		average_runs = false;
	}
	if(output_run)
		snprintf(run_group, 64, "/run%lu", output_run);
	
	// read general attributes
	readAttribute("population_size", population_size, dfile);
	readAttribute("num_parameters", num_parameters, dfile);
	readAttribute("runs", runs, dfile);
	if(legacy)
		readAttribute("num_iterations", num_iterations, dfile);
	readAttribute("parallel_runs", parallel_runs, dfile);

	// set no chunking if specified
//	if(!chunk_size)
//		chunk_size = num_iterations;

	// set output individuals
	if(input_individuals) {
		char *ptr = input_individuals;
		num_output_individuals = 1;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_output_individuals++;
		}
		if(!(output_individuals = (unsigned long *) \
				calloc(num_output_individuals, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output " \
					"individuals\n";
			exit(EXIT_FAILURE);
		}
		ptr = input_individuals;
		for(i = 0; i < num_output_individuals; i++) {
			output_individuals[i] = atol(ptr);
			if(output_individuals[i] < 1 || \
					output_individuals[i] > population_size) {
					cerr << "error: invalid individuals index " << \
							output_individuals[i] << "\n";
					exit(EXIT_FAILURE);
				}
			ptr = strchr(ptr, ',') + 1;
			output_individuals[i]--;
		}
	} else {
		// not set on command line so output all individuals
		num_output_individuals = population_size;
		if(!(output_individuals = (unsigned long *) \
				calloc(num_output_individuals, sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output individuals\n";
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < num_output_individuals; i++)
			output_individuals[i] = i;
	}

	// set output iterations
	if(input_iterations) {
		char *ptr = input_iterations;
		num_output_iterations = 1;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_output_iterations++;
		}
		if(!(output_iterations = (unsigned long *) \
				calloc(num_output_iterations, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output " \
					"iterations\n";
			exit(EXIT_FAILURE);
		}
		ptr = input_iterations;
		for(i = 0; i < num_output_iterations; i++) {
			output_iterations[i] = atol(ptr);
			if(output_iterations[i] < 1 || \
					output_iterations[i] > num_iterations) {
					cerr << "error: invalid iterations index " << \
							output_iterations[i] << "\n";
					exit(EXIT_FAILURE);
				}
			ptr = strchr(ptr, ',') + 1;
			output_iterations[i]--;
		}
	} else {
		// not set on command line so output all iterations
		if(legacy)
			num_output_iterations = num_iterations;
		else
			readAttribute("num_iterations", num_output_iterations, dfile, \
					"/run1");
		if(!(output_iterations = (unsigned long *) \
				calloc(num_output_iterations, sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output iterations\n";
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < num_output_iterations; i++)
			output_iterations[i] = i;
	}

	// sanity check command line options
	switch(mode) {
		default:
			break;
	}
}

// output fitness data
void output_fitness(void)
{
	register unsigned long i, j, ind, iteration, entries, num_iterations;
	char dset_name[1024];

	// save data header
	if(output_header && !best_overall_flag) {
		fprintf(ofp, "iteration");
		if(best_flag)
			printf("%cfitness%cindividual\n", separator, separator);
		else {
			for(i = 0; i < num_output_individuals; i++)
				fprintf(ofp, "%cindividual_%ld", separator, \
						output_individuals[i] + 1);
			fprintf(ofp, "\n");
		}
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[2];

	// open dataset
	if(legacy)
		snprintf(dset_name, 1024, "/%s", "fitness");
	else {
		snprintf(dset_name, 1024, "/run%lu/%s", output_run, "fitness");
		readAttribute("num_iterations", num_iterations, dfile, \
				run_group);
	}
	dset = openDataset(dset_name, dfile);

	// create datatype
	if((dtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) {
		fprintf(stderr, "error: couldn't create datatype\n");
		exit(EXIT_FAILURE);
	}

	// reduce chunk size if greater than total number of iterations
	if(num_iterations < chunk_size)
		chunk_size = num_iterations;

	// create data buffer
	dims[0] = population_size;
	dims[1] = chunk_size;
	fitness_buffer = create2DArrayDouble(dims);

	// create memspace
	mspace = createDataspace(2, dims);

	// create dataspace
	dims[1] = num_iterations;
	dspace = createDataspace(2, dims);

	// output iterations
	iteration = 0;
	entries = chunk_size;
	hsize_t count[2] = {population_size, chunk_size};
	hsize_t offset[2] = {0, 0};
	if(best_overall_flag) {
		best_fitness = DBL_MAX;
		best_individual = 0;
	}
	best_iteration = 0;
	while(iteration < num_iterations) {
		// adjust memspace size if remaining iterations are less than chunk size
		if((num_iterations - iteration) < chunk_size) {
			entries = num_iterations - iteration;
			free2DArrayDouble(fitness_buffer);
			dims[1] = entries;
			fitness_buffer = create2DArrayDouble(dims);
			closeDataspace(mspace);
			mspace = createDataspace(2, dims);
			count[1] = entries;
		}

		// read in data
		offset[1] = iteration;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
				count, NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&fitness_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		// output data
		ind = 0;
		while(ind < entries) {
			if(best_flag) {
				best_fitness = fitness_buffer[output_individuals[0]][ind];
				best_individual = output_individuals[0];
			}
			for(i = 0; i < num_output_iterations; i++)
				if(iteration == output_iterations[i]) {
					if(best_flag || best_overall_flag) {
						for(j = 0; j < num_output_individuals; j++)
							if(fitness_buffer[output_individuals[j]][ind] \
										< best_fitness) {
								best_fitness = fitness_buffer[ \
										output_individuals[j]][ind];
								best_individual = output_individuals[j];
								best_iteration = iteration;
							}
					} else {
						fprintf(ofp, "%ld", iteration + 1);
						for(j = 0; j < num_output_individuals; j++)
							fprintf(ofp, "%c%E", separator, \
									fitness_buffer[output_individuals[j]][ind]);
						fprintf(ofp, "\n");
					}
					break;
				}
			if(best_flag && i != num_output_iterations)
					fprintf(ofp, "%ld%c%E%c%ld\n", iteration + 1, separator, \
							best_fitness, separator, best_individual + 1);
			ind++;
			iteration++;
		}
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free2DArrayDouble(fitness_buffer);
}

// output mean fitness data
void output_mean_fitness(void)
{
	register unsigned long i, j, ind, iteration, entries;
	char dset_name[1024];
	double **fitness, mean, sd;

	// allocate memory for computing statistics
	if(!(fitness = (double **) calloc(num_output_iterations, \
			sizeof(double *)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_output_iterations; i++)
		if(!(fitness[i] = (double *) calloc(num_output_individuals, \
				sizeof(double)))) {
			fprintf(stderr, "error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}

	// save data header
	if(output_header)
		printf("iteration%cfitness_mean%cfitness_sd\n", separator, \
				separator, separator);

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[2];

	// open dataset
	snprintf(dset_name, 1024, "/%s", "fitness");
	dset = openDataset(dset_name, dfile);

	// create datatype
	if((dtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) {
		fprintf(stderr, "error: couldn't create datatype\n");
		exit(EXIT_FAILURE);
	}

	// reduce chunk size if greater than total number of iterations
	if(num_iterations < chunk_size)
		chunk_size = num_iterations;

	// create data buffer
	dims[0] = population_size;
	dims[1] = chunk_size;
	fitness_buffer = create2DArrayDouble(dims);

	// create memspace
	mspace = createDataspace(2, dims);

	// create dataspace
	dims[1] = num_iterations;
	dspace = createDataspace(2, dims);

	// output iterations
	iteration = 0;
	entries = chunk_size;
	hsize_t count[2] = {population_size, chunk_size};
	hsize_t offset[2] = {0, 0};
	while(iteration < num_iterations) {
		// adjust memspace size if remaining iterations are less than chunk size
		if((num_iterations - iteration) < chunk_size) {
			entries = num_iterations - iteration;
			free2DArrayDouble(fitness_buffer);
			dims[1] = entries;
			fitness_buffer = create2DArrayDouble(dims);
			closeDataspace(mspace);
			mspace = createDataspace(2, dims);
			count[1] = entries;
		}

		// read in data
		offset[1] = iteration;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
				count, NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&fitness_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		//  data
		ind = 0;
		while(ind < entries) {
			for(i = 0; i < num_output_iterations; i++)
				if(iteration == output_iterations[i]) {
					for(j = 0; j < num_output_individuals; j++)
						fitness[i][j] = \
								fitness_buffer[output_individuals[j]][ind];
					break;
				}
			ind++;
			iteration++;
		}
	}

	// calculate statistics
	for(i = 0; i < num_output_iterations; i++) {
		mean = 0.0;
		for(j = 0; j < num_output_individuals; j++)
			mean += fitness[i][j];
		mean /= num_output_individuals;
		sd = 0.0;
		for(j = 0; j < num_output_individuals; j++)
			sd += ((fitness[i][j] - mean) * (fitness[i][j] - mean));
		sd = sqrt(sd / (num_output_individuals - 1));
		fprintf(ofp, "%ld%c%E%c%E\n", output_iterations[i] + 1, separator, \
				mean, separator, sd);
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free(fitness);
	free2DArrayDouble(fitness_buffer);
}

// output parameter data
void output_parameters(void)
{
	register unsigned long i, j, k, l, ind, iteration, entries;
	char dset_name[1024];
	double **parameters;

	// save data header
	if(output_header && !best_overall_flag) {
		fprintf(ofp, "iteration");
		for(i = 0; i < num_output_individuals; i++)
			for(j = 0; j < num_parameters; j++)
				fprintf(ofp, "%cindividual_%ld_%ld", separator, \
						output_individuals[i] + 1, j + 1);
		fprintf(ofp, "\n");
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[2];

	// open dataset
	if(legacy)
		snprintf(dset_name, 1024, "/%s", "parameters");
	else {
		snprintf(dset_name, 1024, "/run%lu/%s", output_run, "parameters");
		readAttribute("num_iterations", num_iterations, dfile, \
				run_group);
	}
	dset = openDataset(dset_name, dfile);

	// create datatype
	if((dtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) {
		fprintf(stderr, "error: couldn't create datatype\n");
		exit(EXIT_FAILURE);
	}

	// reduce chunk size if greater than total number of iterations
	if(num_iterations < chunk_size)
		chunk_size = num_iterations;

	// create data buffer
	dims[0] = population_size * num_parameters;
	dims[1] = chunk_size;
	parameters_buffer = create2DArrayDouble(dims);

	// create memspace
	mspace = createDataspace(2, dims);

	// create dataspace
	dims[1] = num_iterations;
	dspace = createDataspace(2, dims);

	// output iterations
	iteration = 0;
	entries = chunk_size;
	hsize_t count[2] = {population_size * num_parameters, chunk_size};
	hsize_t offset[2] = {0, 0};
	while(iteration < num_iterations) {
		// adjust memspace size if remaining iterations are less than chunk size
		if((num_iterations - iteration) < chunk_size) {
			entries = num_iterations - iteration;
			free2DArrayDouble(parameters_buffer);
			dims[1] = entries;
			parameters_buffer = create2DArrayDouble(dims);
			closeDataspace(mspace);
			mspace = createDataspace(2, dims);
			count[1] = entries;
		}

		// read in data
		offset[1] = iteration;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
				count, NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&parameters_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		// output data
		ind = 0;
		while(ind < entries) {
			for(i = 0; i < num_output_iterations; i++)
				if(iteration == output_iterations[i]) {
					if(!best_overall_flag)
						fprintf(ofp, "%ld", iteration + 1);
					for(j = 0; j < num_output_individuals; j++)
						for(k = 0; k < num_parameters; k++)
							if(best_overall_flag) {
								if(iteration == best_iteration && \
										output_individuals[j] == \
										best_individual)
									best_parameters[k] = parameters_buffer \
											[output_individuals[j] \
											* num_parameters + k][ind];
							}
							else
								fprintf(ofp, "%c%E", separator, \
										parameters_buffer[output_individuals[j] \
										* num_parameters + k][ind]);
					if(!best_overall_flag)
						fprintf(ofp, "\n");
				}
			ind++;
			iteration++;
		}
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free2DArrayDouble(parameters_buffer);
}

void output_best_individual(void)
{
	unsigned long i, start_run, end_run;

	// allocate memory to store best parameters
	if(!(best_parameters = (double *) calloc(num_parameters, sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for best parameters\n";
		exit(EXIT_FAILURE);
	}

	// find fitness of best individual
	output_fitness();

	// find parameters of best individual
	output_parameters();

	if(parallel_runs) {
		start_run = best_iteration * population_size * runs + \
				best_individual * runs + 1;
		end_run = start_run + runs - 1;
	} else {
		start_run = end_run = best_iteration * population_size + \
				best_individual + 1;
	}
	if(output_header) {
		fprintf(ofp, "individual%citeration%cfitness%crun_index", separator, \
				separator, separator, separator);
		for(i = 0; i < num_parameters; i++)
			fprintf(ofp, "%cparameter_%ld", separator, i + 1);
		fprintf(ofp, "\n");
	}
	fprintf(ofp, "%ld%c%ld%c%E%c%.*ld", best_individual + 1, separator, \
				best_iteration + 1, separator, best_fitness, separator, \
				ID_WIDTH, start_run);
	if(start_run != end_run)
		fprintf(ofp, "-%.*ld", ID_WIDTH, end_run);
	for(i = 0; i < num_parameters; i++)
		fprintf(ofp, "%c%E", separator, best_parameters[i]);
	fprintf(ofp, "\n");

	// tidy up
	free(best_parameters);
}

// output information on dataset
void output_info(void)
{
	char *str, *str1, *str2, timestamp[256];
	unsigned long ulong;
	double dble;
	bool b;
	int i;
	unsigned int usint;

 	readAttribute("mcss_cmaes_version", str, dfile);
 	printf("%s: %s\n", "mcss_cmaes_version", str);
 	free(str);
 	readAttribute("function_file", str, dfile);
 	printf("%s: %s\n", "function_file", str);
 	free(str);
 	readAttribute("initial_x", str, dfile);
 	printf("%s: %s\n", "initial_x", str);
 	free(str);
 	readAttribute("mcss_parameter_file", str, dfile);
 	printf("%s: %s\n", "mcss_parameter_file", str);
 	free(str);
		readAttribute("runs", ulong, dfile);
		printf("%s: %lu\n", "runs", ulong);
	if(!legacy) {
		readAttribute("optimisation_runs", i, dfile);
		printf("%s: %d\n", "optimisation_runs", i);
	}
	if(legacy) {
		readAttribute("num_iterations", ulong, dfile);
		printf("%s: %lu\n", "num_iterations", ulong);
	}
 	readAttribute("parallel_runs", b, dfile);
 	printf("%s: %d\n", "parallel_runs", b);
 	readAttribute("lower_bound", dble, dfile);
 	printf("%s: %f\n", "lower_bound", dble);
 	readAttribute("upper_bound", dble, dfile);
 	printf("%s: %f\n", "upper_bound", dble);
 	readAttribute("target_data_file", str, dfile);
 	printf("%s: %s\n", "target_data_file", str);
 	free(str);
 	readAttribute("data_file", str, dfile);
 	printf("%s: %s\n", "data_file", str);
 	free(str);
 	readAttribute("population_size", i, dfile);
 	printf("%s: %d\n", "population_size", i);
 	readAttribute("num_parameters", i, dfile);
 	printf("%s: %d\n", "num_parameters", i);
 	readAttribute("max_iterations", dble, dfile);
 	printf("%s: %d\n", "max_iterations", (int) dble);
 	readAttribute("max_function_evals", dble, dfile);
 	printf("%s: %f\n", "max_function_evals", dble);
 	readAttribute("stop_fitness", dble, dfile);
 	printf("%s: %f\n", "stop_fitness", dble);
 	readAttribute("stop_fitness_diff", dble, dfile);
 	printf("%s: %f\n", "stop_fitness_diff", dble);
 	readAttribute("stop_fitness_diff_best", dble, dfile);
 	printf("%s: %f\n", "stop_fitness_diff_best", dble);
 	readAttribute("stop_step_size", dble, dfile);
 	printf("%s: %f\n", "stop_step_size", dble);
 	readAttribute("stop_sd", dble, dfile);
 	printf("%s: %f\n", "stop_sd", dble);
 	readAttribute("seed", usint, dfile);
 	printf("%s: %u\n", "seed", usint);
 	readAttribute("start_time", str1, dfile);
 	printf("%s: %s\n", "start_time", str1);
 	readAttribute("end_time", str2, dfile);
 	printf("%s: %s\n", "end_time", str2);
 	printf("total_time: %s\n", convert_time(timestamp_diff(str1, str2), \
			timestamp));
	free(str1);
	free(str2);
}

// output attribute value
void output_attribute_value(void)
{
	char *str;
	H5E_auto2_t old_func;
	void *old_client_data;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if(attributeAsString(attribute, str, dfile)) {
		printf("%s\n", str);
		free(str);
	} else {
		if(attributeAsString(attribute, str, dfile)) {
			printf("%s\n", str);
			free(str);
		} else {
			cerr << "error: unable to find attribute \"" << attribute << "\"\n";
			exit(EXIT_FAILURE);
		}
	}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
}

int main(int argc, char **argv)
{
	// parse command line
	parse_command_line(argc, argv);

	// read initial data
	initialise_input();

	// perform specified operation on dataset
	switch(mode) {
		case MODE_INFO:
			// output information on optimisation
			output_info();
			break;

		case MODE_FITNESS:
			// output fitness data
			output_fitness();
			break;

		case MODE_MEAN_FITNESS:
			// output mean fitness data
			output_mean_fitness();
			break;

		case MODE_PARAMETERS:
			// output parameters data
			output_parameters();
			break;

		case MODE_BEST_INDIVIDUAL:
			// output data on best individual
			output_best_individual();
			break;

		case MODE_ATTRIBUTE_INFO:
			// output attribute value
			output_attribute_value();
			break;

		default:
			break;
	}

	fclose(ofp);

	return EXIT_SUCCESS;
}
