// mcss and cma-es parameter optimisation
// copyright 2008, 2009, 2010 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

// show debugging output
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <ctime>
#include <mcss/mcss.h>
#include <unistd.h>
//#include <regex.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <hdf5.h>
#include <queue>
#include <dlfcn.h>
#include "mcss/McssCmaes.h"
#include "cmaes.h"
#include <ecsb/ecsb.h>

using namespace Simulation;
using namespace ECSB;

McssCmaesParameters *cmaesparameters;
SimulationParameters *mcssparameters;

// user defined functions
void *plugin_handle;
typedef void (*initialise_t)(void);
typedef void (*terminate_t)(void);
typedef void (*set_initial_x_t)(double *, double *);
typedef void (*set_model_parameters_t)(char *, double *, int);
typedef double (*calculate_run_fitness_t)(char *);
typedef double (*calculate_total_fitness_t)(double *);
typedef double (*calculate_fitness_t)(char **);
typedef int (*is_feasible_t)(double *);
typedef int (*perform_run_t)(char *, char *, int);
initialise_t initialise;
terminate_t terminate;
set_initial_x_t set_initial_x;
set_model_parameters_t set_model_parameters;
calculate_run_fitness_t calculate_run_fitness;
calculate_total_fitness_t calculate_total_fitness;
calculate_fitness_t calculate_fitness;
is_feasible_t is_feasible;
perform_run_t perform_run;

static bool resample = false;
static char *model_base;
static int iteration;

static gsl_rng *rng;  // random number generator

static hid_t dfile;  // results data file
// fitness dataset
static double **fitness_buffer;  // buffer to store data
static hid_t fitness_dtype;
static hid_t fitness_dset;  // hdf5 dataset
static hid_t fitness_dspace;  // hdf5 dataspace
static unsigned long fitness_chunk_size = 1000;
static unsigned long fitness_index;
static unsigned long fitness_offset;
// parameters dataset
static double **parameters_buffer;  // buffer to store data
static hid_t parameters_dtype;
static hid_t parameters_dset;  // hdf5 dataset
static hid_t parameters_dspace;  // hdf5 dataspace
static unsigned long parameters_chunk_size = 1000;
static unsigned long parameters_index;
static unsigned long parameters_offset;

// message ids
#define MESSAGE_SHUTDOWN	0
#define MESSAGE_SIMULATE	1
#define MESSAGE_SIMULATED	2

#define SIMULATION_FINISHED		0
#define SIMULATION_CANCELLED	1

// width of model id field
#define ID_WIDTH			7

// openmpi variables
#ifdef HAVE_MPI
static int tid;  // thread id
static int mpinthreads;  // thread id
//static sigaction old_sigintHandler;  // old signal handler
#endif
int nthreads;  // number of threads

// timer thread data
struct TimerData {
	unsigned int max_time;
	int tnum;
	pthread_t ppid;
	pthread_t tid;
};

// mcss thread data
struct McssData {
	Psystem *psystem;
	double max_time;
	int simulation_status;
	int tnum;
	pthread_t ppid;
};

// pthread variables
#ifndef HAVE_MPI
static int *ptnum;
static pthread_t **ptid;
static pthread_mutex_t **message_status_mutex;
static pthread_cond_t **message_status;
static queue<int> **message;
#endif
static pthread_mutex_t **mcss_status_mutex;
static pthread_mutex_t **mcss_execute_running_mutex;
static bool *mcss_execute_running;
static pthread_mutex_t **timer_running_mutex;
static bool *timer_running;
static pthread_mutex_t signal_handler_running_mutex = PTHREAD_MUTEX_INITIALIZER;
static bool cancel = false;

// function prototypes
unsigned long initialise_rng(unsigned long seed);
int mcss(const char *param_file, const char *params);
void pthread_recv(void *buf, int count, int tid);
void pthread_send(void *buf, int count, int tid);
void close_results_file(void);
void shutdown_threads(void);

// create dataset to store fitness values
void create_fitness_dataset(unsigned int optimisation_run)
{
	hsize_t dims[2], maxdims[2];
	char dset_name[64];
	hid_t cparams;

	// create datatype
	if((fitness_dtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) {
		fprintf(stderr, "error: couldn't create datatype\n");
		exit(EXIT_FAILURE);
	}
	// fit dataset into available memory
	dims[0] = cmaesparameters->population_size;
	dims[1] = fitness_chunk_size;
	maxdims[0] = H5S_UNLIMITED;
	maxdims[1] = H5S_UNLIMITED;
	do {
		// create property list
		dims[1] = fitness_chunk_size;
		cparams = createChunkedPropertyList(2, dims, fitness_dtype);
		fitness_chunk_size = dims[1];

		// set dataset compression if specified
		if(cmaesparameters->compress)
			if(H5Pset_deflate(cparams, cmaesparameters->compression_level) < 0) {
				std::cerr << "error: unable to set dataset compression\n";
				exit(EXIT_FAILURE);
			}

		// create dataspace
		fitness_dspace = createDataspaceScaled(2, dims, maxdims);

		// test if dataspace same size as chunk size
		if(dims[1] != fitness_chunk_size) {
			H5Pclose(cparams);
			closeDataspace(fitness_dspace);
		}
	} while(dims[1] != fitness_chunk_size);

	// create dataset
	snprintf(dset_name, 64, "/run%u/fitness", optimisation_run);
	if((fitness_dset = H5Dcreate2(dfile, dset_name, fitness_dtype, \
			fitness_dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't create fitness dataset\n";
		exit(EXIT_FAILURE);
	}

	// create data buffer
	fitness_buffer = create2DArrayDouble(dims);
	fitness_index = fitness_offset = 0;

	// free resources
	H5Pclose(cparams);
}

// extend fitness dataset
void extend_fitness_dataset(unsigned int optimisation_run)
{
	char dset_name[64];

	// reopen dataset
	snprintf(dset_name, 64, "/run%u/fitness", optimisation_run);
	if((fitness_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	fitness_offset += fitness_index;
	hsize_t dims[2] = {cmaesparameters->population_size, fitness_offset};
	if(H5Dset_extent(fitness_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[1] = fitness_chunk_size;
	fitness_buffer = create2DArrayDouble(dims);

	// reassign dataspace
	hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
	if((fitness_dspace = H5Screate_simple(2, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create fitness dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	fitness_index = 0;
}

// write fitness data
void write_fitness(double *fitness)
{
	register unsigned long i, j;

	// write data to buffer
	for(i = 0; i < cmaesparameters->population_size; i++)
		fitness_buffer[i][fitness_index] = fitness[i];
	fitness_index++;

	if(fitness_index == fitness_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[2] = {0, fitness_offset};
		hsize_t dims1[2] = {cmaesparameters->population_size, fitness_offset + \
				fitness_chunk_size};
		hsize_t dims2[2] = {cmaesparameters->population_size, \
				fitness_chunk_size};
		if(H5Dset_extent(fitness_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend fitness dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(fitness_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write fitness dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(fitness_dset, fitness_dtype, fitness_dspace, fspace, \
				H5P_DEFAULT, &fitness_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't write fitness dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		fitness_offset += fitness_chunk_size;
		fitness_index = 0;
	}
}

// close fitness dataset
void close_fitness_dataset(void)
{
	unsigned long i, j;

	if(fitness_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[2] = {cmaesparameters->population_size, fitness_index};
		double **fitness_buffer_tmp = create2DArrayDouble(dims2);

		// copy over remaining data
		for(i = 0; i < cmaesparameters->population_size; i++)
			for(j = 0; j < fitness_index; j++)
				fitness_buffer_tmp[i][j] = fitness_buffer[i][j];

		// adjust dataset size
		hsize_t offset[2] = {0, fitness_offset};
		hsize_t dims1[2] = {cmaesparameters->population_size, fitness_offset + \
				fitness_index};
		hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(2, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[1] > 0 && dims1[1] < fitness_chunk_size && !fitness_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(2, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(fitness_dset, name, 512);
			H5Dclose(fitness_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 2, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(cmaesparameters->compress)
				if(H5Pset_deflate(cparams, cmaesparameters->compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((fitness_dset = H5Dcreate2(dfile, name, fitness_dtype, \
					dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(fitness_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(fitness_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(fitness_dset, fitness_dtype, dspace, fspace, H5P_DEFAULT, \
				&fitness_buffer_tmp[0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(fitness_dspace);

		// free memory
		free2DArrayDouble(fitness_buffer_tmp);
	}

	// close dataset
	H5Dclose(fitness_dset);

	// free data buffer memory
	free2DArrayDouble(fitness_buffer);
}

// create dataset to store parameters values
void create_parameters_dataset(unsigned int optimisation_run)
{
	hsize_t dims[2], maxdims[2];
	char dset_name[64];
	hid_t cparams;

	// create datatype
	if((parameters_dtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) {
		fprintf(stderr, "error: couldn't create datatype\n");
		exit(EXIT_FAILURE);
	}
	// fit dataset into available memory
	dims[0] = cmaesparameters->population_size * cmaesparameters->num_parameters;
	dims[1] = parameters_chunk_size;
	maxdims[0] = H5S_UNLIMITED;
	maxdims[1] = H5S_UNLIMITED;
	do {
		// create property list
		dims[1] = parameters_chunk_size;
		cparams = createChunkedPropertyList(2, dims, parameters_dtype);
		parameters_chunk_size = dims[1];

		// set dataset compression if specified
		if(cmaesparameters->compress)
			if(H5Pset_deflate(cparams, cmaesparameters->compression_level) < 0) {
				std::cerr << "error: unable to set dataset compression\n";
				exit(EXIT_FAILURE);
			}

		// create dataspace
		parameters_dspace = createDataspaceScaled(2, dims, maxdims);

		// test if dataspace same size as chunk size
		if(dims[1] != parameters_chunk_size) {
			H5Pclose(cparams);
			closeDataspace(parameters_dspace);
		}
	} while(dims[1] != parameters_chunk_size);

	// create dataset
	snprintf(dset_name, 64, "/run%u/parameters", optimisation_run);
	if((parameters_dset = H5Dcreate2(dfile, dset_name, parameters_dtype, \
			parameters_dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't create parameters dataset\n";
		exit(EXIT_FAILURE);
	}

	// create data buffer
	parameters_buffer = create2DArrayDouble(dims);
	parameters_index = parameters_offset = 0;

	// free resources
	H5Pclose(cparams);
}

// extend parameters dataset
void extend_parameters_dataset(unsigned int optimisation_run)
{
	char dset_name[64];

	// reopen dataset
	snprintf(dset_name, 64, "/run%u/parameters", optimisation_run);
	if((parameters_dset = H5Dopen2(dfile, dset_name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// extend dataset
	parameters_offset += parameters_index;
	hsize_t dims[2] = {cmaesparameters->population_size * \
			cmaesparameters->num_parameters, parameters_offset};
	if(H5Dset_extent(parameters_dset, dims) < 0) {
		std::cerr << "error: couldn't extend dataset \"" << dset_name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// create new data buffer
	dims[1] = parameters_chunk_size;
	parameters_buffer = create2DArrayDouble(dims);

	// reassign dataspace
	hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
	if((parameters_dspace = H5Screate_simple(2, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create parameters dataspace\n";
		exit(EXIT_FAILURE);
	}

	// reset index
	parameters_index = 0;
}

// write parameters data
void write_parameters(double **parameters)
{
	register unsigned long i, j;

	// write data to buffer
	for(i = 0; i < cmaesparameters->population_size; i++)
		for(j = 0; j < cmaesparameters->num_parameters; j++)
			parameters_buffer[i * cmaesparameters->num_parameters + j] \
					[parameters_index] = pow(10.0, parameters[i][j]);
	parameters_index++;

	if(parameters_index == parameters_chunk_size) {
		// write data in buffer to disk
		hsize_t offset[2] = {0, parameters_offset};
		hsize_t dims1[2] = {cmaesparameters->population_size * \
				cmaesparameters->num_parameters, parameters_offset + \
				parameters_chunk_size};
		hsize_t dims2[2] = {cmaesparameters->population_size * \
				cmaesparameters->num_parameters, \
				parameters_chunk_size};
		if(H5Dset_extent(parameters_dset, dims1) < 0) {
			std::cerr << "error: couldn't extend parameters dataset\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(parameters_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			std::cerr << "error: couldn't write parameters dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(parameters_dset, parameters_dtype, parameters_dspace, fspace, \
				H5P_DEFAULT, &parameters_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't write parameters dataset\n";
			exit(EXIT_FAILURE);
		}
		closeDataspace(fspace);
		parameters_offset += parameters_chunk_size;
		parameters_index = 0;
	}
}

// close parameters dataset
void close_parameters_dataset(void)
{
	unsigned long i, j, k;

	if(parameters_index) {
		// allocate memory for buffer to store remaining data
		hsize_t dims2[2] = {cmaesparameters->population_size * \
				cmaesparameters->num_parameters, parameters_index};
		double **parameters_buffer_tmp = create2DArrayDouble(dims2);

		// copy over remaining data
		for(i = 0; i < cmaesparameters->population_size; i++)
			for(j = 0; j < cmaesparameters->num_parameters; j++)
				for(k = 0; k < parameters_index; k++)
					parameters_buffer_tmp[i * cmaesparameters->num_parameters \
							+ j][k] = parameters_buffer[i * \
							cmaesparameters->num_parameters + j][k];
		for(i = 0; i < cmaesparameters->num_parameters; i++)
			for(j = 0; j < parameters_index; j++)
				parameters_buffer_tmp[i][j] = parameters_buffer[i][j];

		// adjust dataset size
		hsize_t offset[2] = {0, parameters_offset};
		hsize_t dims1[2] = {cmaesparameters->population_size * \
				cmaesparameters->num_parameters, parameters_offset + \
				parameters_index};
		hsize_t maxdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
		hid_t dspace;
		if((dspace = H5Screate_simple(2, dims2, maxdims)) < 0) {
			std::cerr << "error: couldn't create dataspace\n";
			exit(EXIT_FAILURE);
		}
		if(dims1[1] > 0 && dims1[1] < parameters_chunk_size && !parameters_offset) {
			// resize dataset if smaller than chunk size
			closeDataspace(dspace);
			if((dspace = H5Screate_simple(2, dims1, maxdims)) < 0) {
				std::cerr << "error: couldn't create dataspace\n";
				exit(EXIT_FAILURE);
			}
			char name[512];
			H5Iget_name(parameters_dset, name, 512);
			H5Dclose(parameters_dset);
			H5Ldelete(dfile, name, H5P_DEFAULT);
			hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
			if(H5Pset_chunk(cparams, 2, dims2) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			// set dataset compression if specified
			if(cmaesparameters->compress)
				if(H5Pset_deflate(cparams, cmaesparameters->compression_level) < 0) {
					std::cerr << "error: unable to set dataset compression\n";
					exit(EXIT_FAILURE);
				}
			if((parameters_dset = H5Dcreate2(dfile, name, parameters_dtype, \
					dspace, H5P_DEFAULT, cparams, H5P_DEFAULT)) < 0) {
				std::cerr << "error: couldn't create dataset\n";
				exit(EXIT_FAILURE);
			}
			H5Pclose(cparams);
		} else
			// otherwise extend the dataset
			if(H5Dset_extent(parameters_dset, dims1) < 0) {
				std::cerr << "error: couldn't write dataset\n";
				exit(EXIT_FAILURE);
			}

		// write remaining data
		hid_t fspace = H5Dget_space(parameters_dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
				NULL) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(parameters_dset, parameters_dtype, dspace, fspace, H5P_DEFAULT, \
				&parameters_buffer_tmp[0][0]) < 0) {
			std::cerr << "error: couldn't write dataset\n";
			exit(EXIT_FAILURE);
		}

		// close dataspaces
		closeDataspace(fspace);
		closeDataspace(dspace);
		closeDataspace(parameters_dspace);

		// free memory
		free2DArrayDouble(parameters_buffer_tmp);
	}

	// close dataset
	H5Dclose(parameters_dset);

	// free data buffer memory
	free2DArrayDouble(parameters_buffer);
}

// create file to store optimisation results
void create_results_file(void)
{
	dfile = createDatafile(cmaesparameters->data_file);
}

void close_results_file(void)
{
	closeDatafile(dfile);
}

void write_cmaes_parameters(void)
{
	writeAttribute("mcss_cmaes_version", VERSION, dfile);
	writeAttribute("optimisation_runs", cmaesparameters->optimisation_runs, dfile);
	writeAttribute("function_file", cmaesparameters->function_file, dfile);
	writeAttribute("initial_x", cmaesparameters->initial_x, dfile);
	writeAttribute("mcss_parameter_file", cmaesparameters->mcss_parameter_file, dfile);
	writeAttribute("parallel_runs", cmaesparameters->parallel_runs, dfile);
	writeAttribute("runs", cmaesparameters->runs, dfile);
	writeAttribute("lower_bound", cmaesparameters->lower_bound, dfile);
	writeAttribute("upper_bound", cmaesparameters->upper_bound, dfile);
	writeAttribute("column_separator", cmaesparameters->column_separator, dfile);
	writeAttribute("target_data_file", cmaesparameters->target_data_file, dfile);
	writeAttribute("data_file", cmaesparameters->data_file, dfile);
	writeAttribute("population_size", cmaesparameters->population_size, dfile);
	writeAttribute("num_parameters", cmaesparameters->num_parameters, dfile);
	writeAttribute("max_iterations", cmaesparameters->max_iterations, dfile);
	writeAttribute("max_function_evals", cmaesparameters->max_function_evals, dfile);
	writeAttribute("stop_fitness", cmaesparameters->stop_fitness, dfile);
	writeAttribute("stop_fitness_diff", cmaesparameters->stop_fitness_diff, dfile);
	writeAttribute("stop_fitness_diff_best", cmaesparameters->stop_fitness_diff_best, dfile);
	writeAttribute("stop_step_size", cmaesparameters->stop_step_size, dfile);
	writeAttribute("stop_sd", cmaesparameters->stop_sd, dfile);
	writeAttribute("seed", cmaesparameters->seed, dfile);
	writeAttribute("max_eigen_decomposition_time", cmaesparameters->max_eigen_decomposition_time, dfile);
	writeAttribute("updatecov", cmaesparameters->updatecov, dfile);
	writeAttribute("updatecov_multiplier", cmaesparameters->updatecov_multiplier, dfile);
	writeAttribute("resume_distribution", cmaesparameters->resume_distribution, dfile);
	writeAttribute("used_parameter_file", cmaesparameters->used_parameter_file, dfile);
	writeAttribute("weights", cmaesparameters->weights, dfile);
	writeAttribute("mu", cmaesparameters->mu, dfile);
	writeAttribute("fac_cs", cmaesparameters->fac_cs, dfile);
	writeAttribute("fac_damps", cmaesparameters->fac_damps, dfile);
	writeAttribute("ccumcov", cmaesparameters->ccumcov, dfile);
	writeAttribute("mucov", cmaesparameters->mucov, dfile);
	writeAttribute("fac_ccov", cmaesparameters->fac_ccov, dfile);
	writeAttribute("diagonal_covariance_matrix", cmaesparameters->diagonal_covariance_matrix, dfile);
	writeAttribute("fac_max_func_evals", cmaesparameters->fac_max_func_evals, dfile);
}

void create_logging_dataset(unsigned int optimisation_run)
{
	char run_group[64];
	hid_t dgroup;
	
	// create group for run
	snprintf(run_group, 64, "/run%u", (unsigned int) optimisation_run);
	dgroup = createGroup(run_group, dfile);
	closeGroup(dgroup);

	create_fitness_dataset(optimisation_run);
	create_parameters_dataset(optimisation_run);
}

void close_logging_dataset(void)
{
	close_parameters_dataset();
	close_fitness_dataset();
}

unsigned long initialise_rng(unsigned long seed)
{
	FILE *fp;
	unsigned long newseed;

	// initialise uniform random number generator
	rng = gsl_rng_alloc(gsl_rng_mt19937);

	// seed generator randomly if seed set to zero
	newseed = seed;
	if(!newseed) {
		if((fp = fopen("/dev/urandom", "r"))) {
			// set seed from /dev/urandom
			if(fread(&newseed, sizeof(newseed), 1, fp) != 1)
				// fall back on setting by time
				gsl_rng_set(rng, newseed);
			fclose(fp);
		} else
			// set seed from time
			newseed = (unsigned long) std::time(0);
	}
	gsl_rng_set(rng, newseed);

	return newseed;
}

// release memory for random number generator
void free_rng(void)
{
	gsl_rng_free(rng);
}

#ifndef WIN32
// signal handler thread
void *signal_handler(void *args)
{
	int sig;

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// wait for signal
	sigset_t sig_mask;
	sigemptyset(&sig_mask);
	sigaddset(&sig_mask, SIGINT);
	sigwait(&sig_mask, &sig);

	pthread_mutex_lock(&signal_handler_running_mutex);
	printf("shutting down gracefully, please wait...\n");
	cancel = true;
	#ifndef HAVE_MPI
	unsigned int i;
	// stop simulations in slave threads running
	for(i = 1; i < nthreads; i++) {
		pthread_mutex_lock(mcss_status_mutex[i]);
		// send USR2 signal to slave to indicate finished
		pthread_kill(*ptid[i], SIGUSR2);
		pthread_mutex_unlock(mcss_status_mutex[i]);
	}
	#endif

	// set thread as finished
	pthread_mutex_unlock(&signal_handler_running_mutex);

	pthread_exit(NULL);
}

// timer thread handler
void *timer(void *args)
{
	// get arguments
	struct TimerData *timer_data = (struct TimerData *) args;

	// set thread as running
	pthread_mutex_lock(timer_running_mutex[timer_data->tnum]);
	timer_running[timer_data->tnum] = true;
	pthread_mutex_unlock(timer_running_mutex[timer_data->tnum]);

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// sleep for specified runtime
	sleep_ptime(timer_data->max_time, timer_data->tid);

	pthread_mutex_lock(mcss_status_mutex[timer_data->tnum]);

	#if DEBUG > 1
	printf("slave %d timer: timer fired\n", timer_data->tnum);
	#endif

	// set thread as finished
	pthread_mutex_lock(timer_running_mutex[timer_data->tnum]);
	timer_running[timer_data->tnum] = false;
	pthread_mutex_unlock(timer_running_mutex[timer_data->tnum]);

	// send USR2 signal to parent to indicate finished
	pthread_mutex_lock(mcss_execute_running_mutex[timer_data->tnum]);
	if(mcss_execute_running[timer_data->tnum])
		pthread_kill(timer_data->ppid, SIGUSR2);
	pthread_mutex_unlock(mcss_execute_running_mutex[timer_data->tnum]);

	pthread_mutex_unlock(mcss_status_mutex[timer_data->tnum]);

	pthread_exit(NULL);
}
#endif

void *mcss_execute(void *args)
{
	// get arguments
	struct McssData *mcss_data = (struct McssData *) args;

	// set thread as running
	pthread_mutex_lock(mcss_execute_running_mutex[mcss_data->tnum]);
	mcss_execute_running[mcss_data->tnum] = true;
	pthread_mutex_unlock(mcss_execute_running_mutex[mcss_data->tnum]);

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// execute simulation
	mcss_data->psystem->execute(mcss_data->max_time);

	pthread_mutex_lock(mcss_status_mutex[mcss_data->tnum]);

	// finished ok
	mcss_data->simulation_status = SIMULATION_FINISHED;

	#if DEBUG > 1
	printf("slave %d mcss_execute: finished with status %d\n", \
			mcss_data->tnum, mcss_data->simulation_status);
	#endif

	// set thread as finished
	pthread_mutex_lock(mcss_execute_running_mutex[mcss_data->tnum]);
	mcss_execute_running[mcss_data->tnum] = false;
	pthread_mutex_unlock(mcss_execute_running_mutex[mcss_data->tnum]);

	// send USR2 signal to parent to indicate finished
	#ifndef WIN32
	if(cmaesparameters->max_runtime != 0.0) {
		pthread_mutex_lock(timer_running_mutex[mcss_data->tnum]);
		if(timer_running[mcss_data->tnum])
			pthread_kill(mcss_data->ppid, SIGUSR2);
		pthread_mutex_unlock(timer_running_mutex[mcss_data->tnum]);
	} else
		pthread_kill(mcss_data->ppid, SIGUSR2);
	#endif

	pthread_mutex_unlock(mcss_status_mutex[mcss_data->tnum]);

	pthread_exit(NULL);
}

int mcss(const char *param_file, const char *params, int tnum)
{
	int argc, i, sig, status;
	char **argv, *params2, *ptr;
	pthread_t tid_mcss_execute, tid_timer;
	pthread_attr_t tattr;
	struct McssData mcss_data;
	struct TimerData timer_data;
	sigset_t sig_mask;

	#if DEBUG > 1
	printf("slave %d: starting mcss\n", tnum);
	#endif

	// create parameters
	if(!(params2 = (char *) calloc(strlen(params) + 1, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameters\n");
		exit(EXIT_FAILURE);
	}
	strcpy(params2, params);
	argc = 0;
	if(!(argv = (char **) calloc(2, sizeof(char *)))) {
		fprintf(stderr, "error: unable to allocate memory for parameters\n");
		exit(EXIT_FAILURE);
	}
	if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameter\n");
		exit(EXIT_FAILURE);
	}
	strncpy(argv[argc++], "mcss", 128);
	if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
		fprintf(stderr, "error: unable to allocate memory for parameter\n");
		exit(EXIT_FAILURE);
	}
	strncpy(argv[argc++], param_file, 128);
	while((ptr = strsep(&params2, " "))) {
		if(strcmp(ptr, "") == 0)
			continue;
		if(!(argv = (char **) realloc(argv, sizeof(char *) * (argc + 1)))) {
			fprintf(stderr, \
					"error: unable to reallocate memory for parameters\n");
			exit(EXIT_FAILURE);
		}
		if(!(argv[argc] = (char *) calloc(128, sizeof(char)))) {
			fprintf(stderr, "error: unable to allocate memory for parameter\n");
			exit(EXIT_FAILURE);
		}
		strncpy(argv[argc++], ptr, 128);
	}

	// read in parameters
	SimulationParameters *parameters = new SimulationParameters();
	parameters->read(argc, argv);

	// free memory
	for(i = 0; i < argc; i++)
		free(argv[i]);
	free(argv);
	free(params2);

	#ifdef HAVE_MPI
	// set mpi to operate in serial mode
	parameters->mpi_serial = true;
	#endif

	// create p-system from parameters specified in parameter file
	Psystem *psystem = new Psystem(*parameters);

	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
	// run simulation in thread
	mcss_data.psystem = psystem;
	mcss_data.simulation_status = SIMULATION_CANCELLED;
	mcss_data.max_time = parameters->max_time;
	mcss_data.ppid = pthread_self();
	mcss_data.tnum = tnum;
	if(pthread_create(&tid_mcss_execute, &tattr, mcss_execute, \
				(void *) &mcss_data)) {
		fprintf(stderr, "error: unable to create mcss_execute thread\n");
		exit(EXIT_FAILURE);
	}
	#ifdef WIN32
	// run timer if specified
	if(cmaesparameters->max_runtime != 0.0)
		fprintf(stderr, "warning: max_runtime not available under windows\n");
	#else
	// run timer if specified
	if(cmaesparameters->max_runtime != 0.0) {
		// start timer thread
		timer_data.max_time = (unsigned int) cmaesparameters->max_runtime;
		timer_data.ppid = pthread_self();
		timer_data.tnum = tnum;
		timer_data.tid = tid_mcss_execute;
		if(pthread_create(&tid_timer, &tattr, timer, (void *) &timer_data)) {
			fprintf(stderr, "error: unable to create timer thread\n");
			exit(EXIT_FAILURE);
		}
	}
	#endif

	// tidy up
	pthread_attr_destroy(&tattr);

	#if DEBUG > 1
	printf("slave %d: waiting for mcss\n", tnum);
	#endif

	#ifdef WIN32
	pthread_join(tid_mcss_execute, NULL);
	#else
	// wait for SIGUSR2 signal
	sigemptyset(&sig_mask);
	sigaddset(&sig_mask, SIGUSR2);
	sigwait(&sig_mask, &sig);
	#endif

	pthread_mutex_lock(mcss_status_mutex[tnum]);
	status = mcss_data.simulation_status;
	#if DEBUG > 1
	printf("slave %d: mcss finished with status %d\n", tnum, status);
	#endif
	switch(status) {
		case SIMULATION_FINISHED:
			if(cmaesparameters->max_runtime != 0.0) {
				pthread_mutex_lock(timer_running_mutex[tnum]);
				if(timer_running[tnum])
					pthread_cancel(tid_timer);  // kill thread
				pthread_mutex_unlock(timer_running_mutex[tnum]);
			}
			break;

		case SIMULATION_CANCELLED:
			// cancel mcss execute thread
			pthread_mutex_lock(mcss_execute_running_mutex[tnum]);
			if(mcss_execute_running[tnum])
				pthread_kill(tid_mcss_execute, SIGUSR2);  // cancel simulation
			pthread_mutex_unlock(mcss_execute_running_mutex[tnum]);

			// cancel timer
			if(cmaesparameters->max_runtime != 0.0) {
				pthread_mutex_lock(timer_running_mutex[tnum]);
				if(timer_running[tnum])
					pthread_cancel(tid_timer);  // cancel thread
				pthread_mutex_unlock(timer_running_mutex[tnum]);
			}
			break;

		default:
			fprintf(stderr, "error: unknown thread state\n");
			exit(EXIT_FAILURE);
			break;
	}
	pthread_mutex_unlock(mcss_status_mutex[tnum]);

	// wait for threads to finish
	pthread_join(tid_mcss_execute, NULL);
	if(cmaesparameters->max_runtime != 0.0)
		pthread_join(tid_timer, NULL);

	// tidy up
	delete parameters;
	delete psystem;

	#if DEBUG > 1
	printf("slave %d: finished running mcss status=%d\n", tnum, status);
	#endif

	return status;
}

// shutdown slave threads
void shutdown_threads(void)
{
	int i, mid[4];

	// send message to slave thread to shutdown
	for(i = 1; i < nthreads; i++) {
		mid[0] = MESSAGE_SHUTDOWN;
		#ifdef HAVE_MPI
		MPI_Send(mid, 4, MPI_INT, i, 0, MPI_COMM_WORLD);
		#else
		pthread_send(mid, 4, i);
		#endif
	}
	#ifndef HAVE_MPI
	// wait for threads to shutdown
	for(i = 1; i < nthreads; i++)
		pthread_join(*ptid[i], NULL);
	#endif
}

#ifndef HAVE_MPI
// wait to receive message from thread
void pthread_recv(void *buf, int count, int tid)
{
	int i, *ptr;

	ptr = (int *) buf;
	pthread_mutex_lock(message_status_mutex[tid]);
	while(message[tid]->size() < count)
		pthread_cond_wait(message_status[tid], message_status_mutex[tid]);
	for(i = 0; i < count; i++) {
		ptr[i] = message[tid]->front();
		message[tid]->pop();
	}
	pthread_mutex_unlock(message_status_mutex[tid]);
}
#endif

#ifndef HAVE_MPI
// send message to thread
void pthread_send(void *buf, int count, int tid)
{
	int i, *ptr;

	ptr = (int *) buf;
	pthread_mutex_lock(message_status_mutex[tid]);
	for(i = 0; i < count; i++)
		message[tid]->push(ptr[i]);
	pthread_cond_signal(message_status[tid]);
	pthread_mutex_unlock(message_status_mutex[tid]);
}
#endif

// master thread (tid = 0)
// this thread runs the cma-es algorithm
// it instructs slave threads to simulate and calculate individual fitness
int execute_master(void)
{
	int mid[4];  // message id
	int i, j, individual, run, runid, *runs_completed, ret;
	double **parameters, **run_fitness;
	char run_model_file[1024], run_data_file[1024], ***run_data_files;
	char timestamp[256], run_group[64];
	#ifdef HAVE_MPI
	MPI_Status status;
	#endif
	int num_parameters, pop_size, num_runs, best_individual;
	cmaes_t evo;
	double *fitness, *initial_x, *initial_stddev, best_fitness;
	pthread_t tid_signal;
	pthread_attr_t tattr;
	unsigned int optimisation_run, optimisation_runs;

	// initialise rng
	cmaesparameters->seed = initialise_rng(cmaesparameters->seed);

	// initialise cmaes
	num_parameters = cmaesparameters->num_parameters;
	pop_size = cmaesparameters->population_size;
	optimisation_runs = cmaesparameters->optimisation_runs;
	if(cmaesparameters->parallel_runs || cmaesparameters->subruns)
		num_runs = (int) cmaesparameters->runs;
	else
		num_runs = 1;
	if(!(initial_x = (double *) calloc(num_parameters, \
				sizeof(double)))) {
		fprintf(stderr, "error: unable to initialise memory for initial x\n");
		shutdown_threads();
		return EXIT_FAILURE;
	}
	if(!(initial_stddev = (double *) calloc(num_parameters, \
				sizeof(double)))) {
		fprintf(stderr, "error: unable to initialise memory for initial stddev\n");
		shutdown_threads();
		return EXIT_FAILURE;
	}
	if(strcmp(cmaesparameters->initial_x, "user") == 0)
		set_initial_x(initial_x, initial_stddev);
	else {
		for(i = 0; i < num_parameters; i++) {
			if(strcmp(cmaesparameters->initial_x, "random") == 0) {
				initial_x[i] = gsl_rng_uniform_pos(rng) * \
						(cmaesparameters->upper_bound - \
						cmaesparameters->lower_bound) + \
						cmaesparameters->lower_bound;
				initial_stddev[i] = (cmaesparameters->upper_bound - \
						cmaesparameters->lower_bound) / 3.0;
			}
			if(strcmp(cmaesparameters->initial_x, "fixed") == 0) {
				initial_x[i] = 0.5;
				initial_stddev[i] = 0.3;
			}
		}
	}

	// create results file
	create_results_file();
	// output header and initial values
	write_cmaes_parameters();
	writeAttribute("start_time", generateTimestamp(timestamp), dfile);

	// initialise memory for run fitnesses
	if(calculate_run_fitness) {
		if(!(run_fitness = (double **) calloc(pop_size, sizeof(double *)))) {
			fprintf(stderr, "error: unable to initialise memory for run"
					" fitness\n");
			shutdown_threads();
			return EXIT_FAILURE;
		}
		for(i = 0; i < pop_size; i++)
			if(!(run_fitness[i] = (double *) calloc(num_runs, \
						sizeof(double)))) {
				fprintf(stderr, "error: unable to initialise memory for run" \
						" fitness\n");
				shutdown_threads();
				return EXIT_FAILURE;
			}
	} else {
		if(!(run_data_files = (char ***) calloc(pop_size, sizeof(char **)))) {
			fprintf(stderr, "error: unable to initialise memory for run data" \
					" files\n");
			shutdown_threads();
			return EXIT_FAILURE;
		}
		for(i = 0; i < pop_size; i++) {
			if(!(run_data_files[i] = (char **) calloc(num_runs, \
						sizeof(char *)))) {
				fprintf(stderr, "error: unable to initialise memory for run" \
						" data files\n");
				shutdown_threads();
				return EXIT_FAILURE;
			}
			for(j = 0; j < num_runs; j++)
				if(!(run_data_files[i][j] = (char *) calloc(128, \
							sizeof(char)))) {
					fprintf(stderr, "error: unable to initialise memory for "
							" run data files\n");
					shutdown_threads();
					return EXIT_FAILURE;
				}
		}
	}
	// initialise memory for runs completed
	if(!(runs_completed = (int *) calloc(pop_size, sizeof(int)))) {
		fprintf(stderr, "error: unable to initialise memory for runs\n");
		shutdown_threads();
		return EXIT_FAILURE;
	}

	#ifndef WIN32
	// start signal handler thread
 	pthread_attr_init(&tattr);
 	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
 	if(pthread_create(&tid_signal, &tattr, signal_handler, NULL)) {
 		fprintf(stderr, "error: unable to create signal handler thread\n");
 		exit(EXIT_FAILURE);
 	}
 	// tidy up
 	pthread_attr_destroy(&tattr);
	#endif
		
	for(optimisation_run = 1; optimisation_run <= optimisation_runs; \
			optimisation_run++) {
		snprintf(run_group, 64, "/run%u", optimisation_run);
		
		// generate random initial starting point
		if(strcmp(cmaesparameters->initial_x, "random") == 0)
			for(i = 0; i < num_parameters; i++) {
				initial_x[i] = gsl_rng_uniform_pos(rng) * \
						(cmaesparameters->upper_bound - \
						cmaesparameters->lower_bound) + \
						cmaesparameters->lower_bound;
				initial_stddev[i] = (cmaesparameters->upper_bound - \
						cmaesparameters->lower_bound) / 3.0;
			}
		
		// initialise cmaes
		fitness = cmaes_init(&evo, cmaesparameters->num_parameters, initial_x, \
				initial_stddev, cmaesparameters->seed, \
				cmaesparameters->population_size, NULL);

		#if DEBUG != 0
		printf("master: %s\n", cmaes_SayHello(&evo));
		#endif

		// output header and initial values
		// create logging datasets
		create_logging_dataset(optimisation_run);
		// write parameters
		writeAttribute("start_time", generateTimestamp(timestamp), dfile, \
				run_group);

		// initialise user supplied routines if necessary
		if(initialise)
			initialise();

		// run parameter optimisation
		iteration = 1;
		while(!cmaes_TestForTermination(&evo) && !cancel) {
			#if DEBUG != 0
			printf("master: iteration %d...\n", iteration);
			#endif

			// get parameter set
			parameters = cmaes_SamplePopulation(&evo);

			pthread_mutex_lock(&signal_handler_running_mutex);
			for(individual = 0; individual < pop_size; individual++) {
				// resample until solution feasible
				if(resample)
					while(!is_feasible(parameters[individual]))
						cmaes_ReSampleSingle(&evo, individual);

				// perform simulations
				for(run = 0; run < num_runs; run++) {
					// create model file
					runid = (iteration - 1) * num_runs * pop_size + \
							individual * num_runs + run + 1;
					snprintf(run_model_file, 1024, "%s.%.*d.%.*d.sbml", \
							model_base, ID_WIDTH, optimisation_run, ID_WIDTH, \
							runid);
					copy_file(mcssparameters->model_file, run_model_file);

					// set parameters in model file
					set_model_parameters(run_model_file, \
							parameters[individual], run);

					// flush file buffers
	//				sync();

					// ask thread to simulate model
					#if DEBUG > 1
					printf("master: simulating individual=%d run=%d id=%.*d " \
							"thread=%d iteration=%d optimisation_run=%d\n", \
							individual + 1, run + 1,  ID_WIDTH, runid, \
							individual * num_runs + run + 1, iteration, \
							optimisation_run);
					#endif
					mid[0] = MESSAGE_SIMULATE;
					mid[1] = runid;
					mid[2] = (int) gsl_rng_uniform_int(rng, INT_MAX);
					mid[3] = optimisation_run;
					#ifdef HAVE_MPI
					MPI_Send(mid, 4, MPI_INT, individual * num_runs + run + 1, \
							0, MPI_COMM_WORLD);
					#else
					pthread_send(mid, 4, individual * num_runs + run + 1);
					#endif
				}
			}
			pthread_mutex_unlock(&signal_handler_running_mutex);

			// collect results
			i = nthreads - 1;
			while(i) {
				#if DEBUG != 0
				printf("master: waiting for %d/%d simulations to finish " \
						"iteration=%d optimisation_run=%d...\n", i, \
						nthreads - 1, iteration, optimisation_run);
				#endif
				// wait for message from thread
				#ifdef HAVE_MPI
				MPI_Recv(mid, 4, MPI_INT, MPI_ANY_SOURCE, 0, \
						MPI_COMM_WORLD, &status);
				#else
				pthread_recv(mid, 4, 0);
				#endif

				// calculate fitness of individual
				runid = mid[1];
				ret = mid[2];
				run = (runid - 1 - (iteration - 1) * num_runs * pop_size) % \
						num_runs;
				individual = ((runid - 1 - (iteration - 1) * num_runs * \
						pop_size) - run) / num_runs;
				#if DEBUG != 0
				printf("master: received result from thread=%d iteration=%d\n", \
						individual * num_runs + run + 1, iteration);
				#endif
				if(calculate_total_fitness)
					snprintf(run_data_file, 1024, "%s.%.*d.%.*d.h5", \
							model_base, ID_WIDTH, optimisation_run, \
							ID_WIDTH, runid);
				else
					snprintf(run_data_files[individual][run], 1024, \
							"%s.%.*d.%.*d.h5", model_base, ID_WIDTH, \
							optimisation_run, ID_WIDTH, runid);
				// calculate fitness for individual run
				if(calculate_run_fitness) {
					run_fitness[individual][runs_completed[individual]] = \
							calculate_run_fitness(run_data_file);
					#if DEBUG > 1
					printf("master: calculating fitness for individual=%d " \
							"run=%d fitness=%f id=%.*d thread=%d iteration=%d" \
							" optimisation_run=%d\n", \
							individual + 1, run + 1, \
							run_fitness[individual][runs_completed[individual]], \
							ID_WIDTH, runid, individual * num_runs + run + 1, \
							iteration, optimisation_run);
					#endif
					if(cmaesparameters->show_progress)
						printf("individual %d: run=%d fitness=%E\n", \
								individual + 1, run + 1, \
								run_fitness[individual][runs_completed \
								[individual]]);
				}
				if(++runs_completed[individual] == num_runs) {
					if(calculate_total_fitness)
						// all runs completed, so calculate overall fitness
						fitness[individual] = calculate_total_fitness( \
								run_fitness[individual]);
					else
						fitness[individual] = calculate_fitness( \
								run_data_files[individual]);
					#if DEBUG > 1
					printf("master: calculating overall fitness for " \
							"individual=%d fitness=%f thread=%d " \
							"iteration=%d optimisation_run=%d\n", \
							individual + 1, fitness[individual], \
							individual * num_runs + run + 1, iteration, \
							optimisation_run);
					#endif
					runs_completed[individual] = 0;
					if(cmaesparameters->show_progress)
						printf("individual %d: overall_fitness=%E\n", \
								individual + 1, fitness[individual]);
				}
				i--;
			}

			// save fitness and parameters data
			#if DEBUG > 1
			printf("master: writing data iteration=%d\n", iteration);
			#endif
			write_fitness(fitness);
			write_parameters(parameters);

			if(cmaesparameters->show_progress) {
				best_fitness = fitness[0];
				best_individual = 0;
				for(i = 1; i < cmaesparameters->population_size; i++)
					if(fitness[i] < best_fitness) {
						best_fitness = fitness[i];
						best_individual = i;
					}
				printf("iteration %d optimisation_run=%d: best_individual=%d" \
						" fitness=%E parameters=[", iteration, \
						optimisation_run, best_individual + 1, best_fitness);
				for(i = 0; i < num_parameters; i++)
					printf(" %E", pow(10.0, parameters[best_individual][i]));
				printf(" ]\n");
			}

			// update the search distribution
			cmaes_UpdateDistribution(&evo, fitness);

			iteration++;
		}

		// terminate user supplied routines if necessary
		if(::terminate)
			::terminate();

		// write attributes
		writeAttribute("num_iterations", iteration - 1, dfile, run_group);
		writeAttribute("end_time", generateTimestamp(timestamp), dfile, run_group);
		close_logging_dataset();

		// release memory
		cmaes_exit(&evo);
	}
	
	// close results file
	writeAttribute("num_optimisation_runs", optimisation_run - 1, dfile);
	writeAttribute("end_time", generateTimestamp(timestamp), dfile);
	close_results_file();
	
	// shutdown slave threads
	shutdown_threads();

	// tidy up
	free_rng();
	free(initial_stddev);
	free(initial_x);
	free(runs_completed);
	if(calculate_run_fitness) {
		for(i = 0; i < pop_size; i++)
			free(run_fitness[i]);
		free(run_fitness);
	}

	return EXIT_SUCCESS;
}

// slave thread
// each slave thread simulates a specified model
#ifdef HAVE_MPI
void execute_slave(void)
#else
void *execute_slave(void *ptr)
#endif
{
	int mid[4], ret;  // message id
	int flag = 1;  // set to false if thread to exit
	char parameters[1024];
	unsigned long seed;
	#ifdef HAVE_MPI
	MPI_Status status;
	#else
	int tid = *((int *) ptr);
	#endif

	while(flag) {
		// wait for message from master
		#ifdef HAVE_MPI
		MPI_Recv(mid, 4, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
		#else
		pthread_recv(mid, 4, tid);
		#endif
		// parse message
		switch(mid[0]) {
			case MESSAGE_SHUTDOWN:
				// master finished, so stop slave
				flag = 0;
				break;

			case MESSAGE_SIMULATE:
				// simulate a model
				// construct command to execute
				if(cmaesparameters->subruns)
					snprintf(parameters, 1024, "model_file=%s.%.*d.%.*d.sbml" \
							" data_file=%s.%.*d.%.*d.h5 seed=%d runs=%ld" \
							" max_runtime=0", \
							model_base, ID_WIDTH, mid[3], ID_WIDTH, mid[1], \
							model_base, ID_WIDTH, mid[3], ID_WIDTH, mid[1], \
							mid[2], cmaesparameters->subruns);
				else {
					if(cmaesparameters->parallel_runs)
						snprintf(parameters, 1024, "model_file=%s.%.*d.%.*d" \
								".sbml data_file=%s.%.*d.%.*d.h5 seed=%d " \
								"runs=%d max_runtime=0", \
								model_base, ID_WIDTH, mid[3], ID_WIDTH, \
								mid[1], model_base, ID_WIDTH, mid[3], \
								ID_WIDTH, mid[1], mid[2], 1);
					else
						snprintf(parameters, 1024, "model_file=%s.%.*d.%.*d" \
								".sbml data_file=%s.%.*d.%.*d.h5 seed=%d " \
								"runs=%ld max_runtime=0", \
								model_base, ID_WIDTH, mid[3], ID_WIDTH, \
								mid[1], model_base, ID_WIDTH, mid[3], \
								ID_WIDTH, mid[1], mid[2], \
								cmaesparameters->runs);
				}

				// run simulation
				#if DEBUG > 1
				printf("slave %d: simulating individual id %.*d " \
						"optimisation_run=%d\n", tid, ID_WIDTH, \
						mid[1], mid[3]);
				#endif
				if(perform_run)
					ret = perform_run(cmaesparameters->mcss_parameter_file, \
							parameters, mid[1]);
				else
					ret = mcss(cmaesparameters->mcss_parameter_file, \
							parameters, tid);
				#if DEBUG > 1
				printf("slave %d: finished simulating individual id %.*d " \
						"optimisation_run=%d\n", tid, ID_WIDTH, mid[1], mid[3]);
				#endif
				// flush file buffers
//				sync();

				// send back exit status of simulation
				mid[0] = MESSAGE_SIMULATED;
				mid[2] = ret;
				#ifdef HAVE_MPI
				MPI_Send(mid, 4, MPI_INT, 0, 0, MPI_COMM_WORLD);
				#else
				pthread_send(mid, 4, 0);
				#endif
				break;

			default:
				fprintf(stderr, "error: unknown message id %d received by" \
						" thread %d\n", mid[0], tid);
				break;
		}
	}
	#ifndef HAVE_MPI
	pthread_exit(NULL);
	#endif
}

// load user specified functions
void load_user_functions(void)
{
	plugin_handle = dlopen(cmaesparameters->function_file, RTLD_LAZY);
	if(!plugin_handle) {
		fprintf(stderr, "error: cannot open function file %s: %s\n", \
				cmaesparameters->function_file, dlerror());
		exit(EXIT_FAILURE);
	}
	initialise = (initialise_t) dlsym(plugin_handle, "initialise");
	::terminate = (terminate_t) dlsym(plugin_handle, "terminate");
	set_initial_x = (set_initial_x_t) dlsym(plugin_handle, "set_initial_x");
	if(!set_initial_x && strcmp(cmaesparameters->initial_x, "user") == 0) {
		fprintf(stderr, "error: initial_x set to \"user\" but" \
				"set_initial_x function not defined\n", \
				cmaesparameters->function_file, dlerror());
		exit(EXIT_FAILURE);
	}
	set_model_parameters = (set_model_parameters_t) dlsym(plugin_handle, \
			"set_model_parameters");
	if(!set_model_parameters) {
		fprintf(stderr, "error: set_model_parameters function not defined\n");
		exit(EXIT_FAILURE);
	}
	calculate_run_fitness = (calculate_run_fitness_t) dlsym(plugin_handle, \
			"calculate_run_fitness");
	calculate_total_fitness = (calculate_total_fitness_t) dlsym(plugin_handle, \
			"calculate_total_fitness");
	calculate_fitness = (calculate_fitness_t) dlsym(plugin_handle, \
			"calculate_fitness");
	if(!(calculate_fitness || (calculate_run_fitness && \
			calculate_total_fitness))) {
		fprintf(stderr, "error: calculate_fitness function or both "
				"calculate_run_fitness and calculate_total_fitness "
				"not defined\n");
		exit(EXIT_FAILURE);
	}
	is_feasible = (is_feasible_t) dlsym(plugin_handle, "is_feasible");
	if(is_feasible)
		resample = true;
	perform_run = (perform_run_t) dlsym(plugin_handle, "perform_run");
}

int main(int argc, char **argv)
{
	int i, ret;
	char *ptr;
	sigset_t sig_mask;

	#ifdef HAVE_MPI
	// initialise mpi
	MPI_Init(&argc, &argv);
	atexit((void (*)()) MPI_Finalize);
	#endif

	// read in parameters
	// parse command line and read in parameters
	if(argc == 1)
		#ifdef HAVE_MPI
		fprintf(stderr, "mcss-cmaes version %s (MPI)\n", VERSION);
		#else
		fprintf(stderr, "mcss-cmaes version %s\n", VERSION);
		#endif
	cmaesparameters = cmaesParameters();
	mcssparameters = mcssParameters();
	cmaesparameters->read(argc, argv);
	mcssparameters->read(cmaesparameters->mcss_parameter_file);
	if(!cmaesparameters->runs)
		cmaesparameters->runs = mcssparameters->runs;
	else
		mcssparameters->runs = cmaesparameters->runs;
	if(!cmaesparameters->max_runtime)
		cmaesparameters->max_runtime = mcssparameters->max_runtime;
	mcssparameters->max_runtime = 0.0;

	// set global variables
	if(cmaesparameters->parallel_runs || cmaesparameters->subruns)
		nthreads = cmaesparameters->population_size * cmaesparameters->runs + 1;
	else
		nthreads = cmaesparameters->population_size + 1;
	if(cmaesparameters->subruns && cmaesparameters->parallel_runs) {
		#ifdef HAVE_MPI
		if(tid == 0)
		#endif
		fprintf(stderr, "error: cannont perform subruns with parallel " \
				"runs enabled\n");
		exit(EXIT_FAILURE);
	}
	if(strlen(mcssparameters->model_file) == 0) {
		#ifdef HAVE_MPI
		if(tid == 0)
		#endif
		fprintf(stderr, "error: model_file not found\n");
		exit(EXIT_FAILURE);
	}
	if(!(model_base = (char *) calloc(strlen(mcssparameters->model_file) + 1, \
				sizeof(char)))) {
		#ifdef HAVE_MPI
		if(tid == 0)
		#endif
		fprintf(stderr, "error: unable to allocate memory for model_base\n");
		exit(EXIT_FAILURE);
	}
	strncpy(model_base, mcssparameters->model_file, strlen(mcssparameters->model_file) + 1);
	if((ptr = strrchr(model_base, '.')))
		*ptr = '\0';

	#ifndef WIN32
	// block all signals
	sigfillset(&sig_mask);
	pthread_sigmask(SIG_SETMASK, &sig_mask, NULL);
	#endif

	// load target data if given
	if(strcmp(cmaesparameters->target_data_file, "") != 0)
		readTargetData();

	// load user specified functions
	load_user_functions();

	// initialise mcss threads
	if(!(mcss_status_mutex = (pthread_mutex_t **) \
				calloc(nthreads, sizeof(pthread_mutex_t *)))) {
		fprintf(stderr, "error: unable to create mcss_status_mutex\n");
		exit(EXIT_FAILURE);
	}
	if(!(mcss_execute_running_mutex = (pthread_mutex_t **) \
				calloc(nthreads, sizeof(pthread_mutex_t *)))) {
		fprintf(stderr, "error: unable to create mcss_execute_mutex\n");
		exit(EXIT_FAILURE);
	}
	if(!(timer_running_mutex = (pthread_mutex_t **) \
				calloc(nthreads, sizeof(pthread_mutex_t *)))) {
		fprintf(stderr, "error: unable to create timer_running_mutex\n");
		exit(EXIT_FAILURE);
	}
	if(!(timer_running = (bool *) calloc(nthreads, sizeof(bool)))) {
		fprintf(stderr, "error: unable to create timer_running\n");
		exit(EXIT_FAILURE);
	}
	if(!(mcss_execute_running = (bool *) calloc(nthreads, sizeof(bool)))) {
		fprintf(stderr, "error: unable to create mcss_execute_running\n");
		exit(EXIT_FAILURE);
	}
	for(i = 1; i < nthreads; i++) {
		if(!(mcss_status_mutex[i] = (pthread_mutex_t *) \
					malloc(sizeof(pthread_mutex_t)))) {
			fprintf(stderr, "error: unable to create mcss_status_mutex\n");
			exit(EXIT_FAILURE);
		}
		if(!(mcss_execute_running_mutex[i] = (pthread_mutex_t *) \
					malloc(sizeof(pthread_mutex_t)))) {
			fprintf(stderr, "error: unable to create mcss_execute_mutex\n");
			exit(EXIT_FAILURE);
		}
		if(!(timer_running_mutex[i] = (pthread_mutex_t *) \
				malloc(sizeof(pthread_mutex_t)))) {
			fprintf(stderr, "error: unable to create timer_running_mutex\n");
			exit(EXIT_FAILURE);
		}
		timer_running[i] = mcss_execute_running[i] = false;
		pthread_mutex_init(mcss_status_mutex[i], NULL);
		pthread_mutex_init(mcss_execute_running_mutex[i], NULL);
		pthread_mutex_init(timer_running_mutex[i], NULL);
	}

	#ifdef HAVE_MPI
	// get thread id
	MPI_Comm_rank(MPI_COMM_WORLD, &tid);

	// get number of threads
	MPI_Comm_size(MPI_COMM_WORLD, &mpinthreads);

	// must run with sufficient number of threads
	if(nthreads > mpinthreads) {
		if(tid == 0)
			fprintf(stderr, "error: must run with at least %d threads\n", nthreads);
		exit(EXIT_FAILURE);
	}

	// more threads than needed, so terminate unneeded ones
	if(tid >= nthreads) {
		exit(EXIT_SUCCESS);
	}

	if(tid == 0)
		ret = execute_master();  // master thread
	else
		execute_slave();  // slave thread
	#else
	// allocate storage for threads
	if(!(ptid = (pthread_t **) calloc(nthreads, sizeof(pthread_t *)))) {
		fprintf(stderr, "error: unable to create ptid\n");
		exit(EXIT_FAILURE);
	}
	if(!(ptnum = (int *) calloc(nthreads, sizeof(int)))) {
		fprintf(stderr, "error: unable to create ptnum\n");
		exit(EXIT_FAILURE);
	}
	if(!(message_status_mutex = (pthread_mutex_t **) calloc(nthreads, \
				sizeof(pthread_mutex_t *)))) {
		fprintf(stderr, "error: unable to create message_status_mutex\n");
		exit(EXIT_FAILURE);
	}
	if(!(message_status = (pthread_cond_t **) calloc(nthreads, \
				sizeof(pthread_cond_t *)))) {
		fprintf(stderr, "error: unable to create message_status\n");
		exit(EXIT_FAILURE);
	}
	if(!(message = (queue<int> **) calloc(nthreads, sizeof(queue<int> *)))) {
		fprintf(stderr, "error: unable to create message buffer\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < nthreads; i++) {
		if(!(ptid[i] = (pthread_t *) malloc(sizeof(pthread_t)))) {
			fprintf(stderr, "error: unable to create ptid\n");
			exit(EXIT_FAILURE);
		}
		if(!(message_status_mutex[i] = (pthread_mutex_t *) malloc( \
					sizeof(pthread_mutex_t)))) {
			fprintf(stderr, "error: unable to create message_status_mutex\n");
			exit(EXIT_FAILURE);
		}
		if(!(message_status[i] = (pthread_cond_t *) malloc( \
					sizeof(pthread_cond_t)))) {
			fprintf(stderr, "error: unable to create message_status\n");
			exit(EXIT_FAILURE);
		}
		if(!(message[i] = new queue<int>())) {
			fprintf(stderr, "error: unable to create message queue\n");
			exit(EXIT_FAILURE);
		}
		pthread_mutex_init(message_status_mutex[i], NULL);
		pthread_cond_init(message_status[i], NULL);
	}

	// start slave threads
	ptnum[0] = 0;
	for(i = 1; i < nthreads; i++) {
		ptnum[i] = i;
		ret = pthread_create(ptid[i], NULL, execute_slave, (void *) &ptnum[i]);
		if(ret != 0) {
			fprintf(stderr, "error: unable to create thread %d\n", i);
			exit(EXIT_FAILURE);
		}
	}

	// start master
	ret = execute_master();

	// tidy up
	for(i = 0; i < nthreads; i++)
		delete message[i];
	free(message);
	free(message_status);
	free(message_status_mutex);
	free(ptnum);
	free(ptid);
	#endif

	// tidy up
//	pthread_mutex_destroy(&mcss_status_mutex);
//	pthread_cond_destroy(&mcss_status);

	dlclose(plugin_handle);

	#ifdef HAVE_MPI
	if(tid != 0)
		return 0;
	#endif
	return ret;
}
