// postprocess output of mcss simulation
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include <ecsb/ecsb.h>
#include <mcss/mcss.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <iostream>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_cdf.h>

using namespace std;
using namespace SimulationDatasets;
using namespace ECSB;

// operation modes
#define MODE_LEVELS				1
#define MODE_INFO				2
#define MODE_COMPARTMENT_INFO	3
#define MODE_SPECIES_INFO		4
#define MODE_LEVELS_RATIO		5
#define MODE_PROPENSITIES		6
#define MODE_RULES_INFO			7
#define MODE_LEVEL_TIME			8
#define MODE_VOLUMES			9
#define MODE_ATTRIBUTE_INFO		10
#define MODE_LEVEL_TIME_MEAN	11
#define MODE_REACTION_FREQUENCY	12

// hdf5 file
hid_t dfile;

// data buffers
static SpeciesEntry *species_buffer;
static CompartmentEntry *compartment_buffer;
static RuleEntry *rule_buffer;
static unsigned int ***amount_buffer;
static double **volume_buffer;
static void *propensity_buffer;

static unsigned long num_species, num_compartments, total_rules, num_rules;
static unsigned long num_timepoints, num_runs;
static double log_interval;
#ifdef WIN32
static double cifactor = 0.0;
#else
static long double cifactor = 0.0;
#endif
static unsigned long chunk_size = 1000;

// command line parameters
static char infile[1024];
static bool average_runs = true;
static int mode = MODE_INFO;
static unsigned long *output_species;
static unsigned long *output_compartments;
static unsigned long num_output_species = 0;
static unsigned long num_output_compartments = 0;
static unsigned long *top_ratio_compartments;
static unsigned long *bottom_ratio_compartments;
static unsigned long num_top_ratio_compartments = 0;
static unsigned long num_bottom_ratio_compartments = 0;
static bool sum_species = false;
static bool sum_compartments = false;
static bool sum_nothing = true;
static bool output_header = true;
static FILE *ofp = stdout;
static double cidegree = 0.95;
static long target_level = 0;
static unsigned long output_run = 0;
static char *input_species = NULL;
static char *input_compartments = NULL;
static char *input_ratios = NULL;
static bool use_species_indices = false;
static bool use_compartment_indices = false;
static bool use_ratio_indices = false;
static double start_time = -1.0;
static double end_time = -1.0;
static unsigned long start_timepoint = 0;
static unsigned long end_timepoint = 0;
static unsigned long num_timepoints_selected = 0;
static char separator = ' ';
static bool upper_lower_bounds = false;
static bool output_concentrations = false;
static char attribute[1024];
static double output_interval = -1.0;
static unsigned long xdim, ydim;
static int time_precision;
static bool output_sds = true;

// calculate precision of floating point number
#ifdef WIN32
static int get_precision(double val)
#else
static int get_precision(long double val)
#endif
{
	#ifdef WIN32
	double frac, iptr, val2;
	#else
	long double frac, iptr, val2;
	#endif
	int precision;

	#ifdef WIN32
	frac = modf(val, &iptr);
	#else
	frac = modfl(val, &iptr);
	#endif
	if(frac == 0.0)
		return 0;
	val2 = 1.0;
	precision = 0;
	while((frac < val2) != 0.0) {
		val2 = val2 / 10.0;
		precision++;
	}
	return precision;
}

static void parse_command_line(int argc, char **argv)
{
	int c;
	char *ptr, *tptr, *bptr;
	unsigned long i, ind;

	// parse that command line...
	while((c = getopt(argc, argv, \
			"A:b:c:C:s:S:o:r:R:f:g:G:t:e:u:v:z:FwxydmnlijkpqQah?")) != -1)
		switch(c) {
			case 'A':  // get attribute value
				mode = MODE_ATTRIBUTE_INFO;
				strncpy(attribute, optarg, 1024);
				break;

			case 'b':  // output column separator
				separator = optarg[0];
				break;

			case 'c':  // compartments to calculate levels for
				if(!(input_compartments = (char *) \
						calloc(strlen(optarg) + 1, sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"compartments\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_compartments, optarg);
				break;

			case 'C':  // compartments to calculate levels for
				if(!(input_compartments = (char *) \
						calloc(strlen(optarg) + 1, sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"compartments\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_compartments, optarg);
				use_compartment_indices = true;
				break;

			case 'r':  // compartments to calculate level ratios for
				mode = MODE_LEVELS_RATIO;
				if(!(input_ratios = (char *) calloc(strlen(optarg) + 1, \
						sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"ratios\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_ratios, optarg);
				break;

			case 'R':  // compartments to calculate level ratios for
				mode = MODE_LEVELS_RATIO;
				if(!(input_ratios = (char *) calloc(strlen(optarg) + 1, \
						sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"ratios\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_ratios, optarg);
				use_ratio_indices = true;
				break;

			case 's':  // species to calculate levels for
				if(!(input_species = (char *) calloc(strlen(optarg) + 1, \
						sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"species\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_species, optarg);
				break;

			case 'S':  // species to calculate levels for
				if(!(input_species = (char *) calloc(strlen(optarg) + 1, \
						sizeof(char)))) {
					cerr << "error: couldn't allocate memory for output " \
							"species\n";
					exit(EXIT_FAILURE);
				}
				strcpy(input_species, optarg);
				use_species_indices = true;
				break;

			case 'g':  // calculate time taken to reach level
				mode = MODE_LEVEL_TIME;
				target_level = atol(optarg);
				break;

			case 'F':  // calculate reaction frequencies
				mode = MODE_REACTION_FREQUENCY;
				break;

			case 'G':  // calculate time taken to reach level
				mode = MODE_LEVEL_TIME_MEAN;
				target_level = atol(optarg);
				break;

			case 't':  // run to output data for
				output_run = atol(optarg);
				break;

			case 'a':  // sum species in a compartment
				sum_species = true;
				sum_nothing = false;
				break;

			case 'e':  // read levels in one chunk
				chunk_size = atol(optarg);
				break;

			case 'p':  // don't output headers with data
				output_header = false;
				break;

			case 'l':  // output species levels
				mode = MODE_LEVELS;
				break;

			case 'x':  // output concentrations
				output_concentrations = true;
				break;

			case 'i':  // output dataset information
				mode = MODE_INFO;
				break;

			case 'j':  // output dataset information
				mode = MODE_PROPENSITIES;
				break;

			case 'k':  // output rule information
				mode = MODE_RULES_INFO;
				break;

			case 'w':  // output compartment volume
				mode = MODE_VOLUMES;
				break;

			case 'm':  // output compartment information
				mode = MODE_COMPARTMENT_INFO;
				break;

			case 'n':  // output species information
				mode = MODE_SPECIES_INFO;
				break;

			case 'f':  // degree of confidence for confidence intervals
				cidegree = atof(optarg);
				break;

			case 'd':  // sum species over compartments
				sum_compartments = true;
				sum_nothing = false;
				break;

			case 'u':  // set start time
				start_time = atof(optarg);
				break;

			case 'v':  // set end time
				end_time = atof(optarg);
				break;

			case 'q':  // output upper and lower bounds for means
				upper_lower_bounds = true;
				break;

			case 'Q':  // don't output standard deviations
				output_sds = false;
				break;

			case 'y':  // use compartment indices in output
				use_compartment_indices = true;
				break;

			case 'z':  // set output time interval
				output_interval = atof(optarg);
				break;

			case 'o':  // output to file
				if(!(ofp = fopen(optarg, "w"))) {
					fprintf(stderr, "error: couldn't open output file %s\n", \
							optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case 'h':
			case '?':
			usage:
				#ifdef HAVE_MPI
				fprintf(stderr, "mcss-postprocess version %s (MPI)\n", VERSION);
				#else
				fprintf(stderr, "mcss-postprocess version %s\n", VERSION);
				#endif
				fprintf(stderr,
						"usage: mcss-postprocess [OPTIONS] INFILE\n" \
						"\t-a - sum levels for species\n"
						"\t-A ATTRIBUTE - output attribute value\n"
						"\t-b CHAR - output column separator\n"
						"\t-c [x1,y1],[x2,y2];... - compartments to output\n" \
						"\t-C c1,c2;... - compartments to output " \
								"(use compartment indices)\n" \
						"\t-d - sum species over compartments\n"
						"\t-e CHUNK_SIZE - set chunk size\n"
						"\t-f DEGREE - confidence interval degree\n"
						"\t-F - output reaction frequencies\n"
						"\t-g LEVEL - output time taken to reach level\n" \
						"\t-G LEVEL - output time taken to reach mean level\n" \
						"\t-h - display this usage message\n" \
						"\t-i - output information on dataset\n"
						"\t-j - output propensities data\n"
						"\t-k - output information on rules\n"
						"\t-l - output levels data\n" \
						"\t-m - output information on compartments\n"
						"\t-n - output information on species\n"
						"\t-o FILENAME - save output to file\n"
						"\t-p - don't output header with data\n"
						"\t-q - output upper and lower bounds for means\n"
						"\t-Q - don't output standard deviations\n"
						"\t-r [x1,y1],...:[x2,y2],... - output ratio of " \
								"levels data\n" \
						"\t-R c1,...:c2,... - output ratio of levels data " \
								"(use compartment indices)\n" \
						"\t-s s1,s2,... - species to output\n" \
						"\t-S s1,s2,... - species to output " \
								"(use species indices)\n" \
						"\t-t RUN - run to output\n" \
						"\t-u TIME - time at which to start output\n" \
						"\t-v TIME - time at which to end output\n" \
						"\t-w - output compartment volumes\n" \
						"\t-x - output concentrations\n" \
						"\t-y - use compartment indices in output\n" \
						"\t-z TIME - output time step\n" \
						);
				exit(EXIT_FAILURE);
				break;

			default:
				goto usage;
				break;
		}
	if(optind != (argc - 1))
		goto usage;
	strncpy(infile, argv[optind], 1024);
}

static void initialise_input(void)
{
	unsigned long i, j, buffer_size, ind;
	char run_group[64], dset_name[1024], *str;
	FILE *fp;
	bool flag;

	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dtype;  // hdf5 datatype

	// open input file
	dfile = openDatafile(infile);

	// setup output group
	readAttribute("number_of_runs", num_runs, dfile);
	if(output_run) {
		if(output_run > num_runs) {
			cerr << "error: run " << output_run << " not found\n";
			exit(EXIT_FAILURE);
		}
		average_runs = false;
	}
	if(num_runs == 1) {
		average_runs = false;
		output_run = 1;
	}
	snprintf(run_group, 64, "/run%u", (unsigned int) (output_run ? output_run : 1));

	// get cumulative distribution function value for given degrees of freedom
	if(average_runs)
		cifactor = gsl_cdf_tdist_Pinv(1.0 - (1.0 - cidegree) / 2.0, \
				num_runs - 1) / sqrt((double) num_runs);

	// read general attributes
	readAttribute("number_of_species", num_species, dfile);
	readAttribute("number_of_compartments", num_compartments, dfile, run_group);
	readAttribute("number_of_timepoints", num_timepoints, dfile, run_group);
	readAttribute("log_interval", log_interval, dfile);
	readAttribute("total_number_of_rules", total_rules, dfile);
	readAttribute("number_of_rules_in_templates", num_rules, dfile);
 	readAttribute("lattice_x_dimension", xdim, dfile);
 	readAttribute("lattice_y_dimension", ydim, dfile);

	// set time precision
	time_precision = get_precision(log_interval);

	// don't output positions if using cellular potts algorithm
 	readAttribute("simulation_algorithm", str, dfile);
	if(strcmp(str, "dmcp") == 0)
		use_compartment_indices = true;
 	free(str);

	// set no chunking if specified
	if(!chunk_size)
		chunk_size = num_timepoints;

 	// open and read species dataset
	// open dataset
	dset = openDataset("species_information", dfile);

	// create datatype
	dtype = createSpeciesDatatype();

	// create data buffer
	buffer_size = getDatasetSize(dset);
	species_buffer = (SpeciesEntry *) createDataBuffer(buffer_size, dtype);

	// read data to buffer
	readDataset(species_buffer, dset);

	// close dataset
	closeDataset(dset);

	// open and read compartment dataset
	// open dataset
	snprintf(dset_name, 1024, "%s/%s", run_group, "compartment_information");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createCompartmentDatatype();

	// create data buffer
	buffer_size = getDatasetSize(dset);
	compartment_buffer = (CompartmentEntry *) \
			createDataBuffer(buffer_size, dtype);

	// read data to buffer
	readDataset(compartment_buffer, dset);

	// close dataset
	closeDataset(dset);

 	// open and read rule dataset
	// open dataset
	dset = openDataset("rule_information", dfile);

	// create datatype
	dtype = createRuleDatatype();

	// create data buffer
	buffer_size = getDatasetSize(dset);
	rule_buffer = (RuleEntry *) createDataBuffer(buffer_size, dtype);

	// read data to buffer
	readDataset(rule_buffer, dset);

	// close dataset
	closeDataset(dset);

	// remove distribution information from species names
	for(i = 0; i < num_species; i++)
		if(strchr(species_buffer[i].sn, ':'))
			*strchr(species_buffer[i].sn, ':') = '\0';

	// set start and end output timepoints
	if(start_time < 0.0)
		start_timepoint = 0;
	else {
		start_timepoint = (unsigned long) ((double) start_time / log_interval);
		if(start_timepoint < 0 || start_timepoint > num_timepoints) {
			cerr << "error: start time " << start_time << " out-of-range\n";
			exit(EXIT_FAILURE);
		}
	}
	if(end_time < 0.0)
		end_timepoint = num_timepoints;
	else {
		end_timepoint = (unsigned long) ((double) (end_time / log_interval)) + 1;
		if(end_timepoint < 0 || end_timepoint > num_timepoints || \
				end_timepoint < start_timepoint) {
			cerr << "error: end time " << end_time << " out-of-range\n";
			exit(EXIT_FAILURE);
		}
	}
	num_timepoints_selected = end_timepoint - start_timepoint;

	// set output species
	if(input_species) {
		char *ptr = input_species;
		num_output_species = 1;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_output_species++;
		}
		if(!(output_species = (unsigned long *) \
				calloc(num_output_species, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output " \
					"species\n";
			exit(EXIT_FAILURE);
		}
		if(use_species_indices) {
			ptr = input_species;
			for(i = 0; i < num_output_species; i++) {
				output_species[i] = atoi(ptr);
				if(output_species[i] < 0 || \
						output_species[i] >= num_species) {
					cerr << "error: invalid species index " << \
							output_species[i] << "\n";
					exit(EXIT_FAILURE);
				}
				ptr = strchr(ptr, ',') + 1;
			}
		} else {
			ptr = input_species;
			char sn[128], *ptr2;
			for(i = 0; i < num_output_species; i++) {
				ptr2 = strchr(ptr, ',');
				if(!ptr2)
					ptr2 = ptr + strlen(ptr);
				strncpy(sn, ptr, ptr2 - ptr);
				sn[ptr2 - ptr] = '\0';
				for(j = 0; j < num_species; j++)
					if(strcmp(sn, species_buffer[j].sn) == 0)
						break;
				if(j == num_species) {
					cerr << "error: invalid species name " << sn << "\n";
					exit(EXIT_FAILURE);
				}
				output_species[i] = j;
				ptr = ptr2 + 1;
			}
		}
	} else {
		// not set on command line so output all species
		num_output_species = num_species;
		if(!(output_species = (unsigned long *) \
				calloc(num_output_species, sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output species\n";
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < num_output_species; i++)
			output_species[i] = i;
	}

	// set output compartments
	if(input_compartments) {
		char *ptr = input_compartments;
		num_output_compartments = 1;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_output_compartments++;
		}
		if(!use_compartment_indices) {
			if((num_output_compartments % 2) != 0) {
				cerr << "error: invalid compartment specification " << \
						input_compartments << "\n";
				exit(EXIT_FAILURE);
			}
			num_output_compartments = num_output_compartments / 2;
		}
		if(!(output_compartments = (unsigned long *) \
				calloc(num_output_compartments, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output " \
					"compartments\n";
			exit(EXIT_FAILURE);
		}
		if(use_compartment_indices) {
			ptr = input_compartments;
			for(i = 0; i < num_output_compartments; i++) {
				output_compartments[i] = atol(ptr);
				if(output_compartments[i] < 0 || \
						output_compartments[i] >= num_compartments) {
					cerr << "error: invalid compartment index " << \
							output_compartments[i] << "\n";
					exit(EXIT_FAILURE);
				}
				ptr = strchr(ptr, ',') + 1;
			}
		} else {
			ptr = input_compartments;
			unsigned long cx, cy;
			for(i = 0; i < num_output_compartments; i++) {
				cx = atol(ptr + 1);
				ptr = strchr(ptr, ',') + 1;
				cy = atoi(ptr);
				for(j = 0; j < num_compartments; j++)
					if(compartment_buffer[j].cx == cx && \
							compartment_buffer[j].cy == cy)
						break;
				if(j == num_compartments) {
					cerr << "error: invalid compartment position [" << \
						cx << "," << cy << "]\n";
					exit(EXIT_FAILURE);
				}
				output_compartments[i] = j;
				ptr = strchr(ptr, ',') + 1;
			}
		}
	} else {
		// not set on command line so output all compartments
		num_output_compartments = num_compartments;
		if(!(output_compartments = (unsigned long *) \
				calloc(num_output_compartments, sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for output compartments\n";
			exit(EXIT_FAILURE);
		}
		// try to order compartments in position order if 1D model
		flag = true;
		if(xdim == 1) {
			i = 0;
			while(i < num_compartments) {
				for(j = 0; j < num_compartments; j++) {
					if(compartment_buffer[j].cy == i)
						break;
				}
				if(j == num_compartments)
					break;
				output_compartments[i++] = j;
			}
			if(i == num_compartments)
				flag = false;
		} else {
			if(ydim == 1) {
				i = 0;
				while(i < num_compartments) {
					for(j = 0; j < num_compartments; j++) {
						if(compartment_buffer[j].cx == i)
							break;
					}
					if(j == num_compartments)
						break;
					output_compartments[i++] = j;
				}
				if(i == num_compartments)
					flag = false;
			}
		}
		// otherwise output in any order
		if(flag)
			for(i = 0; i < num_output_compartments; i++)
				output_compartments[i] = i;
	}

	// set compartments if calculating ratio of levels
	if(mode == MODE_LEVELS_RATIO) {
		char *ptr;
		char *tptr = input_ratios;
		char *bptr = strchr(input_ratios, ':');
		if(!bptr) {
			cerr << "error: invalid ratio specification " << input_ratios \
					<< "\n";
			exit(EXIT_FAILURE);
		}
		*bptr++ = '\0';
		num_top_ratio_compartments = 1;
		ptr = tptr;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_top_ratio_compartments++;
		}
		if(!use_compartment_indices) {
			if((num_top_ratio_compartments % 2) != 0) {
				cerr << "error: invalid compartment specification " << \
						input_ratios << "\n";
				exit(EXIT_FAILURE);
			}
			num_top_ratio_compartments = num_top_ratio_compartments / 2;
		}
		if(!(top_ratio_compartments = (unsigned long *) \
				calloc(num_top_ratio_compartments, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for top " \
					"ratio compartments\n";
			exit(EXIT_FAILURE);
		}
		num_bottom_ratio_compartments = 1;
		ptr = bptr;
		while((ptr = strchr(ptr, ','))) {
			ptr++;
			num_bottom_ratio_compartments++;
		}
		if(!use_compartment_indices) {
			if((num_bottom_ratio_compartments % 2) != 0) {
				cerr << "error: invalid compartment specification " << \
						input_ratios << "\n";
				exit(EXIT_FAILURE);
			}
			num_bottom_ratio_compartments = num_bottom_ratio_compartments / 2;
		}
		if(!(bottom_ratio_compartments = (unsigned long *) \
				calloc(num_bottom_ratio_compartments, \
						sizeof(unsigned long)))) {
			cerr << "error: couldn't allocate memory for bottom " \
					"ratio compartments\n";
			exit(EXIT_FAILURE);
		}
		if(use_ratio_indices) {
			ptr = tptr;
			for(i = 0; i < num_top_ratio_compartments; i++) {
				top_ratio_compartments[i] = atol(ptr);
				if(top_ratio_compartments[i] < 0 || \
						top_ratio_compartments[i] >= num_compartments) {
					cerr << "error: invalid compartment index " << \
							top_ratio_compartments[i] << "\n";
					exit(EXIT_FAILURE);
				}
				ptr = strchr(ptr, ',') + 1;
			}
			ptr = bptr;
			for(i = 0; i < num_bottom_ratio_compartments; i++) {
				bottom_ratio_compartments[i] = atol(ptr);
				if(bottom_ratio_compartments[i] < 0 || \
						bottom_ratio_compartments[i] >= num_compartments) {
					cerr << "error: invalid compartment index " << \
							bottom_ratio_compartments[i] << "\n";
					exit(EXIT_FAILURE);
				}
				ptr = strchr(ptr, ',') + 1;
			}
		} else {
			unsigned long cx, cy;
			ptr = tptr;
			for(i = 0; i < num_top_ratio_compartments; i++) {
				cx = atol(ptr + 1);
				ptr = strchr(ptr, ',') + 1;
				cy = atoi(ptr);
				for(j = 0; j < num_compartments; j++)
					if(compartment_buffer[j].cx == cx && \
							compartment_buffer[j].cy == cy)
						break;
				if(j == num_compartments) {
					cerr << "error: invalid compartment position (" << \
						cx << "," << cy << ")\n";
					exit(EXIT_FAILURE);
				}
				top_ratio_compartments[i] = j;
				ptr = strchr(ptr, ',') + 1;
			}
			ptr = bptr;
			for(i = 0; i < num_bottom_ratio_compartments; i++) {
				cx = atol(ptr + 1);
				ptr = strchr(ptr, ',') + 1;
				cy = atoi(ptr);
				for(j = 0; j < num_compartments; j++)
					if(compartment_buffer[j].cx == cx && \
							compartment_buffer[j].cy == cy)
						break;
				if(j == num_compartments) {
					cerr << "error: invalid compartment position (" << \
						cx << "," << cy << ")\n";
					exit(EXIT_FAILURE);
				}
				bottom_ratio_compartments[i] = j;
				ptr = strchr(ptr, ',') + 1;
			}
		}
	}

	// sanity check command line options
// 	switch(mode) {
// 		case MODE_LEVEL_TIME:
// 		case MODE_LEVEL_TIME_MEAN:
// 			if(!(num_output_species == 1 && num_output_compartments == 1)) {
// 				cerr << "error: can only calculate time taken to reach " \
// 						"level for one species in one compartment\n";
// 				exit(EXIT_FAILURE);
// 			}
// 			break;
//
// 		default:
// 			break;
// 	}

	// set output interval
	if(output_interval == -1.0)
		output_interval = log_interval;
}

// calculate ratio of species levels for single run
void output_levels_ratio(void)
{
	register unsigned long i, j, k, ind, timepoint;
	unsigned long entries, os, oc, top_species_total, bottom_species_total;
	double ratio;
	char dset_name[1024];

	// save data header
	if(output_header) {
		string header_name = "";
		for(i = 0; i < num_output_species; i++) {
			header_name.append(species_buffer[output_species[i]].sn);
			header_name.append("_");
		}
		char header_index[64];
		for(i = 0; i < num_top_ratio_compartments; i++) {
			snprintf(header_index, 64, "%ld", top_ratio_compartments[i]);
			header_name.append(header_index);
			header_name.append("_");
		}
		for(i = 0; i < num_bottom_ratio_compartments; i++) {
			snprintf(header_index, 64, "%ld", bottom_ratio_compartments[i]);
			header_name.append(header_index);
			if(i != num_bottom_ratio_compartments - 1)
				header_name.append("_");
		}
		fprintf(ofp, "time%c%s_ratio\n", separator, header_name.c_str());
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3];

	// open dataset
	snprintf(dset_name, 1024, "/run%u/%s", (unsigned int) output_run, "amounts");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createAmountDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = chunk_size;
	amount_buffer = create3DArrayUInt(dims);

	// create memspace
	mspace = createDataspace(3, dims);

	// create dataspace
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// output levels
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {num_species, num_compartments, chunk_size};
	hsize_t offset[3] = {0, 0, start_timepoint};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			free3DArrayUInt(amount_buffer);
			dims[2] = entries;
			amount_buffer = create3DArrayUInt(dims);
			closeDataspace(mspace);
			mspace = createDataspace(3, dims);
		}

		// read in data
		offset[2] = timepoint;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, count, \
				NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		// read chunk from dataset
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&amount_buffer[0][0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		ind = 0;
		while(ind < entries) {
			if(sum_nothing) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(j = 0; j < num_output_species; j++) {
						os = output_species[j];
						top_species_total = 0;
						for(i = 0; i < num_top_ratio_compartments; i++)
							top_species_total += amount_buffer[os] \
									[top_ratio_compartments[i]][ind];
						bottom_species_total = 0;
						for(i = 0; i < num_bottom_ratio_compartments; i++)
							bottom_species_total += amount_buffer[os] \
									[bottom_ratio_compartments[i]][ind];
						if(bottom_species_total)
							ratio = (double) top_species_total / \
									bottom_species_total;
						else
							ratio = 0.0;
						fprintf(ofp, "%c%f", separator, ratio);
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
				continue;
			}

			if(sum_species) {
				top_species_total = 0;
				for(i = 0; i < num_top_ratio_compartments; i++) {
					oc = top_ratio_compartments[i];
					for(j = 0; j < num_output_species; j++)
						top_species_total += \
								amount_buffer[output_species[j]][oc][ind];
				}
				bottom_species_total = 0;
				for(i = 0; i < num_bottom_ratio_compartments; i++) {
					oc = bottom_ratio_compartments[i];
					for(j = 0; j < num_output_species; j++)
						bottom_species_total += \
								amount_buffer[output_species[j]][oc][ind];
				}
				if(bottom_species_total)
					ratio = (double) top_species_total / bottom_species_total;
				else
					ratio = 0.0;
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval)
					#ifdef WIN32
					fprintf(ofp, "%f%c%f\n", (double) timepoint * \
							log_interval, separator, ratio);
					#else
					fprintf(ofp, "%Lf%c%f\n", (long double) timepoint * \
							log_interval, separator, ratio);
					#endif
				ind++;
				timepoint++;
				continue;
			}
		}
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free3DArrayUInt(amount_buffer);
}

// calculate average ratio of species levels
void output_average_levels_ratio(void)
{
	register unsigned long i, j, k, entries, timepoint, ind, run;
	unsigned long os, oc, top_species_total, bottom_species_total;
	#ifdef WIN32
	register double mean, sd, total;
	#else
	register long double mean, sd, total;
	#endif
	double ***values;
	char dset_name[1024];

	// allocate memory for computing ratios
	if(!(values = (double ***) calloc(num_output_species, \
			sizeof(double **)))) {
		fprintf(stderr, "error: couldn't allocate memory for ratios\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_output_species; i++) {
		if(!(values[i] = (double **) calloc(num_runs, sizeof(double *)))) {
			fprintf(stderr, "error: couldn't allocate memory for ratios\n");
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < num_runs; j++)
			if(!(values[i][j] = (double *) \
					calloc(chunk_size, sizeof(double)))) {
				fprintf(stderr, "error: couldn't allocate memory for ratios\n");
				exit(EXIT_FAILURE);
			}
	}

	// save data header
	if(output_header) {
		string header_name = "";
		if(output_sds) {
			for(i = 0; i < num_output_species; i++) {
				header_name.append(species_buffer[output_species[i]].sn);
				header_name.append("_");
			}
			char header_index[64];
			for(i = 0; i < num_top_ratio_compartments; i++) {
				snprintf(header_index, 64, "%ld", top_ratio_compartments[i]);
				header_name.append(header_index);
				header_name.append("_");
			}
			for(i = 0; i < num_bottom_ratio_compartments; i++) {
				snprintf(header_index, 64, "%ld", bottom_ratio_compartments[i]);
				header_name.append(header_index);
				if(i != num_bottom_ratio_compartments - 1)
					header_name.append("_");
			}
			fprintf(ofp, "time%c%s_ratio_mean%c%s_ratio_sd%c%s_ratio_ci\n", \
					separator, header_name.c_str(), separator, \
					header_name.c_str(), separator, header_name.c_str());
		} else {
			fprintf(ofp, "time%c%s_ratio_mean\n", \
					separator, header_name.c_str());
		}
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3];

	// create datatype
	dtype = createAmountDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = chunk_size;
	amount_buffer = create3DArrayUInt(dims);

	// create memspace
	mspace = createDataspace(3, dims);

	// create dataspace
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// calculate ratios
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {num_species, num_compartments, chunk_size};
	hsize_t offset[3] = {0, 0, start_timepoint};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			free3DArrayUInt(amount_buffer);
			dims[2] = entries;
			amount_buffer = create3DArrayUInt(dims);
			closeDataspace(mspace);
			mspace = createDataspace(3, dims);
			count[2] = entries;
		}

		// calculate ratios
		for(run = 0; run < num_runs; run++) {
			// read in data
			snprintf(dset_name, 1024, "run%ld/%s", run + 1, "amounts");
			dset = openDataset(dset_name, dfile);
			offset[2] = timepoint;
			if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
					count, NULL) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			// read chunk from dataset
			if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
					&amount_buffer[0][0][0]) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			closeDataset(dset);

			// calculate ratios
			ind = 0;
			while(ind < entries) {
				for(j = 0; j < num_output_species; j++) {
					os = output_species[j];
					top_species_total = 0;
					for(i = 0; i < num_top_ratio_compartments; i++)
						top_species_total += amount_buffer[os] \
								[top_ratio_compartments[i]][ind];
					bottom_species_total = 0;
					for(i = 0; i < num_bottom_ratio_compartments; i++)
						bottom_species_total += amount_buffer[os] \
								[bottom_ratio_compartments[i]][ind];
					if(bottom_species_total)
						values[j][run][ind] = (double) top_species_total \
								/ bottom_species_total;
					else
						values[j][run][ind] = 0.0;
				}
				ind++;
			}
		}

		// calculate statistics
		ind = 0;
		while(ind < entries) {
			if(sum_nothing) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(j = 0; j < num_output_species; j++) {
						mean = 0.0;
						for(i = 0; i < num_runs; i++)
							mean += values[j][i][ind];
						mean /= num_runs;
						sd = 0.0;
						for(i = 0; i < num_runs; i++)
							sd += ((values[j][i][ind] - mean) * \
									(values[j][i][ind] - mean));
						sd = sqrt(sd / (num_runs - 1));
						if(output_sds) {
							#ifdef WIN32
							fprintf(ofp, "%c%f%c%f%c%f", separator, mean, \
									separator, sd, separator, cifactor * sd);
							#else
							fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, mean, \
									separator, sd, separator, cifactor * sd);
							#endif
						} else {
							#ifdef WIN32
							fprintf(ofp, "%c%f", separator, mean);
							#else
							fprintf(ofp, "%c%Lf", separator, mean);
							#endif
						}
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
				continue;
			}

			if(sum_species) {
				mean = 0.0;
				for(i = 0; i < num_runs; i++)
					for(j = 0; j < num_output_species; j++)
						mean += values[j][i][ind];
				mean /= num_runs;
				sd = 0.0;
				for(i = 0; i < num_runs; i++) {
					total = 0.0;
					for(j = 0; j < num_output_species; j++)
						total += values[j][i][ind];
					sd += ((total - mean) * (total - mean));
				}
				sd = sqrt(sd / (num_runs - 1));
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval)
					if(output_sds) {
						#ifdef WIN32
						fprintf(ofp, "%f%c%f%c%f%c%f\n", \
								(double) timepoint * log_interval, \
								separator, mean, separator, sd, separator, \
								cifactor * sd);
						#else
						fprintf(ofp, "%Lf%c%Lf%c%Lf%c%Lf\n", \
								(long double) timepoint * log_interval, \
								separator, mean, separator, sd, separator, \
								cifactor * sd);
						#endif
					} else {
						#ifdef WIN32
						fprintf(ofp, "%f%c%f\n", \
								(double) timepoint * log_interval, \
								separator, mean);
						#else
						fprintf(ofp, "%Lf%c%Lf\n", \
								(long double) timepoint * log_interval, \
								separator, mean);
						#endif
					}
				ind++;
				timepoint++;
				continue;
			}
		}
	}

	// tidy up
	free3DArrayUInt(amount_buffer);

	for(i = 0; i < num_output_species; i++) {
		for(j = 0; j < num_runs; j++)
			free(values[i][j]);
		free(values[i]);
	}
	free(values);
}

// output species levels for single run
void output_levels(void)
{
	register unsigned long i, j, k, ind, timepoint, entries, os, oc, total;
	register double total_volume;
	char dset_name[1024];
	hid_t vdset;  // hdf5 dataset for volumes
	hid_t vdspace, vmspace;  // hdf5 dataspaces for volumes
	hid_t vdtype;  // hdf5 datatype for volumes
	hsize_t vdims[2], vcount[2], voffset[2];

	// save data header
	if(output_header) {
		fprintf(ofp, "time");

		if(sum_nothing) {
			for(i = 0; i < num_output_compartments; i++)
				for(j = 0; j < num_output_species; j++) {
					if(use_species_indices)
						fprintf(ofp, "%c%ld_", separator, output_species[j]);
					else
						fprintf(ofp, "%c%s_", separator, \
								species_buffer[output_species[j]].sn);
					if(use_compartment_indices)
						fprintf(ofp, "%ld", output_compartments[i]);
					else
						fprintf(ofp, "[%ld,%ld]", \
								compartment_buffer[output_compartments[i]].cx, \
								compartment_buffer[output_compartments[i]].cy);
				}
			fprintf(ofp, "\n");
		}

		if(sum_species) {
			string header_name = "";
			char header_index[64];
			for(i = 0; i < num_output_species; i++) {
				if(use_species_indices) {
					snprintf(header_index, 64, "%ld", output_species[i]);
					header_name.append(header_index);
				}
				else
					header_name.append(species_buffer[output_species[i]].sn);
				if(i != num_output_species - 1)
					header_name.append("_");
			}
			for(i = 0; i < num_output_compartments; i++)
				if(use_compartment_indices)
					fprintf(ofp, "%c%s_%ld", separator, \
							header_name.c_str(), \
							output_compartments[i]);
				else
					fprintf(ofp, "%c%s_[%ld,%ld]", separator, \
							header_name.c_str(), \
							compartment_buffer[output_compartments[i]].cx, \
							compartment_buffer[output_compartments[i]].cy);
			fprintf(ofp, "\n");
		}

		if(sum_compartments) {
			string header_name = "";
			char header_index[64];
			for(i = 0; i < num_output_compartments; i++) {
				if(use_compartment_indices)
					snprintf(header_index, 64, "%ld", output_compartments[i]);
				else
					snprintf(header_index, 64, "[%ld,%ld]", \
							compartment_buffer[output_compartments[i]].cx, \
							compartment_buffer[output_compartments[i]].cy);
				header_name.append(header_index);
				if(i != num_output_compartments - 1)
					header_name.append("_");
			}
			for(i = 0; i < num_output_species; i++)
				if(use_species_indices)
					fprintf(ofp, "%c%ld_%s", separator, output_species[i], \
							header_name.c_str());
				else
					fprintf(ofp, "%c%s_%s", separator, \
							species_buffer[output_species[i]].sn, \
							header_name.c_str());
			fprintf(ofp, "\n");
		}
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3];

	// open dataset
	snprintf(dset_name, 1024, "/run%u/%s", (unsigned int) output_run, "amounts");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createAmountDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = chunk_size;
	amount_buffer = create3DArrayUInt(dims);

	// create memspace
	mspace = createDataspace(3, dims);

	// create dataspace
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	if(output_concentrations) {
		// open dataset
		snprintf(dset_name, 1024, "/run%u/%s", (unsigned int) output_run, \
				"volumes");
		vdset = openDataset(dset_name, dfile);

		// create datatype
		vdtype = createVolumeDatatype();

		// create data buffer
		vdims[0] = num_compartments;
		vdims[1] = chunk_size;
		volume_buffer = create2DArrayDouble(vdims);

		// create memspace
		vmspace = createDataspace(2, vdims);

		// create dataspace
		vdims[1] = num_timepoints;
		vdspace = createDataspace(2, vdims);

		vcount[0] = num_compartments;
		vcount[1] = chunk_size;
		voffset[0] = 0;
		voffset[1] = start_timepoint;
	}

	// output levels
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {num_species, num_compartments, chunk_size};
	hsize_t offset[3] = {0, 0, start_timepoint};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			free3DArrayUInt(amount_buffer);
			dims[2] = entries;
			amount_buffer = create3DArrayUInt(dims);
			closeDataspace(mspace);
			mspace = createDataspace(3, dims);
			count[2] = entries;
			if(output_concentrations) {
				free2DArrayDouble(volume_buffer);
				vdims[1] = entries;
				volume_buffer = create2DArrayDouble(vdims);
				closeDataspace(vmspace);
				vmspace = createDataspace(2, vdims);
				vcount[1] = entries;
			}
		}

		// read in data
		offset[2] = timepoint;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
				count, NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&amount_buffer[0][0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		if(output_concentrations) {
			// read in data
			voffset[1] = timepoint;
			if(H5Sselect_hyperslab(vdspace, H5S_SELECT_SET, voffset, NULL, \
					vcount, NULL) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			// read chunk from dataset
			if(H5Dread(vdset, vdtype, vmspace, vdspace, H5P_DEFAULT, \
					&volume_buffer[0][0]) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
		}

		// output data
		if(sum_nothing) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%.*Lf", time_precision, \
							(long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_compartments; i++) {
						oc = output_compartments[i];
						for(j = 0; j < num_output_species; j++)
							if(output_concentrations)
								fprintf(ofp, "%c%f", separator, \
										(volume_buffer[oc][ind] == 0.0) ? 0.0 :
										((double) \
										amount_buffer[output_species[j]] \
										[oc][ind] / volume_buffer[oc][ind]));
							else
								fprintf(ofp, "%c%u", separator, \
										amount_buffer[output_species[j]] \
										[oc][ind]);
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}

		if(sum_species) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_compartments; i++) {
						oc = output_compartments[i];
						total = 0;
						for(j = 0; j < num_output_species; j++)
							total += amount_buffer[output_species[j]][oc][ind];
						if(output_concentrations)
							fprintf(ofp, "%c%f", separator, \
									(volume_buffer[oc][ind] == 0.0) ? 0.0 : \
									((double) total / volume_buffer[oc][ind]));
						else
							fprintf(ofp, "%c%ld", separator, total);
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}

		if(sum_compartments) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_species; i++) {
						os = output_species[i];
						total = 0;
						total_volume = 0.0;
						for(j = 0; j < num_output_compartments; j++) {
							total += amount_buffer[os][ \
									output_compartments[j]][ind];
							if(output_concentrations)
								total_volume += \
										volume_buffer[output_compartments[j]] \
												[ind];
						}
						if(output_concentrations)
							fprintf(ofp, "%c%f", separator, \
									(total_volume == 0.0) ? 0.0 : \
									((double) total / total_volume));
						else
							fprintf(ofp, "%c%ld", separator, total);
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free3DArrayUInt(amount_buffer);

	if(output_concentrations) {
		// close dataset
		closeDataset(vdset);

		// free data buffer memory
		free2DArrayDouble(volume_buffer);
	}
}

// output average species levels over multiple runs
void output_average_levels(void)
{
	register unsigned long i, j, k, entries, timepoint, ind;
	#ifdef WIN32
	register double mean, sd, total;
	#else
	register long double mean, sd, total;
	#endif
	double ***vvalues;
	unsigned int ****values;
	char dset_name[1024];
	hid_t vdset;  // hdf5 dataset for volumes
	hid_t vdspace, vmspace;  // hdf5 dataspaces for volumes
	hid_t vdtype;  // hdf5 datatype for volumes
	hsize_t vdims[2], vcount[2], voffset[2];

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// allocate memory for computing statistics
	if(!(values = (unsigned int ****) calloc(num_output_species, \
			sizeof(unsigned int ***)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_output_species; i++) {
		if(!(values[i] = (unsigned int ***) calloc(num_output_compartments, \
				sizeof(unsigned int **)))) {
			fprintf(stderr, "error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < num_output_compartments; j++) {
			if(!(values[i][j] = (unsigned int **) calloc(num_runs, \
					sizeof(unsigned int *)))) {
				fprintf(stderr, \
						"error: couldn't allocate memory for statistics\n");
				exit(EXIT_FAILURE);
			}
			for(k = 0; k < num_runs; k++)
				if(!(values[i][j][k] = (unsigned int *) \
						calloc(chunk_size, sizeof(unsigned int)))) {
					fprintf(stderr, \
							"error: couldn't allocate memory for statistics\n");
					exit(EXIT_FAILURE);
				}
		}
	}

	if(output_concentrations) {
		if(!(vvalues = (double ***) calloc(num_output_compartments, \
				sizeof(double **)))) {
			fprintf(stderr, "error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < num_output_compartments; i++) {
			if(!(vvalues[i] = (double **) calloc(num_runs, \
					sizeof(double *)))) {
				fprintf(stderr, \
						"error: couldn't allocate memory for statistics\n");
				exit(EXIT_FAILURE);
			}
			for(j = 0; j < num_runs; j++) {
				if(!(vvalues[i][j] = (double *) calloc(chunk_size, \
						sizeof(double)))) {
					fprintf(stderr, \
							"error: couldn't allocate memory for statistics\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	}

	// save data header
	if(output_header) {
		fprintf(ofp, "time");

		if(sum_nothing) {
			for(i = 0; i < num_output_compartments; i++)
				for(j = 0; j < num_output_species; j++) {
					string header_name = "";
					char header_index[64];
					if(use_species_indices) {
						snprintf(header_index, 64, "%ld", output_species[j]);
						header_name.append(header_index);
					}
					else
						header_name.append( \
								species_buffer[output_species[j]].sn);
					header_name.append("_");
					if(use_compartment_indices)
						snprintf(header_index, 64, "%ld", \
								output_compartments[i]);
					else
						snprintf(header_index, 64, "[%ld,%ld]", \
								compartment_buffer[output_compartments[i]].cx, \
								compartment_buffer[output_compartments[i]].cy);
					header_name.append(header_index);
					if(output_sds) {
						if(upper_lower_bounds)
							fprintf(ofp, "%c%s_mean%c%s_sdl%c%s_sdu", \
									separator, header_name.c_str(), separator, \
									header_name.c_str(), separator, \
									header_name.c_str());
						else
							fprintf(ofp, "%c%s_mean%c%s_sd%c%s_ci", \
									separator, header_name.c_str(), separator, \
									header_name.c_str(), separator, \
									header_name.c_str());
					} else {
						fprintf(ofp, "%c%s_mean", separator, \
								header_name.c_str());
					}
				}
			fprintf(ofp, "\n");
		}

		if(sum_species) {
			string header_name = "";
			char header_index[64];
			for(i = 0; i < num_output_species; i++) {
				if(use_species_indices) {
					snprintf(header_index, 64, "%ld", output_species[i]);
					header_name.append(header_index);
				}
				else
					header_name.append(species_buffer[output_species[i]].sn);
				if(i != num_output_species - 1)
					header_name.append("_");
			}
			for(i = 0; i < num_output_compartments; i++)
				if(use_compartment_indices) {
					if(output_sds) {
						if(upper_lower_bounds)
							fprintf(ofp, "%c%s_%ld_mean%c%s_%ld_sdl%c%s_%ld" \
									"_sdu", separator, header_name.c_str(), \
									output_compartments[i], separator, \
									header_name.c_str(), \
									output_compartments[i], separator, \
									header_name.c_str(), \
									output_compartments[i]);
						else
							fprintf(ofp, "%c%s_%ld_mean%c%s_%ld_sd%c%s_%ld" \
									"_ci", separator, header_name.c_str(), \
									output_compartments[i], separator, \
									header_name.c_str(), \
									output_compartments[i], separator, \
									header_name.c_str(), \
									output_compartments[i]);
					} else {
						fprintf(ofp, "%c%s_%ld_mean_sdu", separator, \
								header_name.c_str(), output_compartments[i]);
					}
				} else {
					if(output_sds) {
						if(upper_lower_bounds)
							fprintf(ofp, "%c%s_[%ld,%ld]_mean%c%s_[%ld,%ld]" \
									"_sdl%c" "%s_[%ld,%ld]_sdu", \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy, \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy, \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy);
						else
							fprintf(ofp, "%c%s_[%ld,%ld]_mean%c%s_[%ld,%ld]_sd%c" \
									"%s_[%ld,%ld]_ci", \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy, \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy, \
									separator, header_name.c_str(), \
									compartment_buffer[ \
									output_compartments[i]].cx, \
									compartment_buffer[ \
									output_compartments[i]].cy);
					} else {
						fprintf(ofp, "%c%s_[%ld,%ld]_mean", \
								separator, header_name.c_str(), \
								compartment_buffer[ \
								output_compartments[i]].cx, \
								compartment_buffer[ \
								output_compartments[i]].cy);
					}
			}
			fprintf(ofp, "\n");
		}

		if(sum_compartments) {
			string header_name = "";
			char header_index[64];
			for(i = 0; i < num_output_compartments; i++) {
				if(use_compartment_indices)
					snprintf(header_index, 64, "%ld", output_compartments[i]);
				else
					snprintf(header_index, 64, "[%ld,%ld]", \
							compartment_buffer[output_compartments[i]].cx, \
							compartment_buffer[output_compartments[i]].cy);
				header_name.append(header_index);
				if(i != num_output_compartments - 1)
					header_name.append("_");
			}
			for(i = 0; i < num_output_species; i++)
				if(use_species_indices) {
					if(output_sds) {
						if(upper_lower_bounds)
							fprintf(ofp, "%c%ld_%s_mean%c%ld_%s_sdl%c%ld_%s" \
									"_sdu", separator, output_species[i], \
									header_name.c_str(), separator, \
									output_species[i], header_name.c_str(), \
									separator, output_species[i], \
									header_name.c_str());
						else
							fprintf(ofp, "%c%ld_%s_mean%c%ld_%s_sd%c%ld_%s" \
									"_ci", separator, output_species[i], \
									header_name.c_str(), separator, \
									output_species[i], header_name.c_str(), \
									separator, output_species[i], \
									header_name.c_str());
					} else {
						fprintf(ofp, "%c%ld_%s_mean" "_sdu", separator, \
								output_species[i], header_name.c_str());
					}
				} else {
					if(output_sds) {
						if(upper_lower_bounds)
							fprintf(ofp, "%c%s_%s_mean%c%s_%s_sdl%c%s_%s_sdu", \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str(), \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str(), \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str());
						else
							fprintf(ofp, "%c%s_%s_mean%c%s_%s_sd%c%s_%s_ci", \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str(), \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str(), \
									separator, \
									species_buffer[output_species[i]].sn, \
									header_name.c_str());
					} else {
						fprintf(ofp, "%c%s_%s_mean", separator, \
								species_buffer[output_species[i]].sn, \
								header_name.c_str());
					}
				}
			fprintf(ofp, "\n");
		}
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dtype;  // hdf5 datatype
	hid_t dspace, mspace; // hdf5 dataspaces
	hsize_t dims[3], offset[3];

	// create datatype
	dtype = createAmountDatatype();

	// create memspace
	dims[0] = chunk_size;
	mspace = createDataspace(1, dims);

	// create dataspace
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// setup concentrations output if required
	if(output_concentrations) {
		// create datatype
		vdtype = createVolumeDatatype();

		// create memspace
		vdims[0] = chunk_size;
		vmspace = createDataspace(1, vdims);

		// create dataspace
		vdims[0] = num_compartments;
		vdims[1] = num_timepoints;
		vdspace = createDataspace(2, vdims);

		vcount[0] = 1;
		vcount[1] = chunk_size;
	}

	// calculate average levels
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {1, 1, entries};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			count[2] = entries;
			dims[0] = entries;
			closeDataspace(mspace);
			mspace = createDataspace(1, dims);
			if(output_concentrations) {
				vcount[1] = entries;
				vdims[0] = entries;
				closeDataspace(vmspace);
				vmspace = createDataspace(1, vdims);
			}
		}

		// read in levels data
		offset[2] = timepoint;
		for(i = 0; i < num_output_compartments; i++) {
			offset[1] = output_compartments[i];
			for(j = 0; j < num_output_species; j++) {
				offset[0] = output_species[j];
				for(k = 0; k < num_runs; k++) {
					snprintf(dset_name, 1024, "/run%u/amounts", \
							(unsigned int) k + 1);
					dset = openDataset(dset_name, dfile);
					if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, \
							NULL, count, NULL) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
							values[j][i][k]) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					closeDataset(dset);
				}
			}
		}

		// read in volumes data
		if(output_concentrations) {
			voffset[1] = timepoint;
			for(i = 0; i < num_output_compartments; i++) {
				voffset[0] = output_compartments[i];
				for(k = 0; k < num_runs; k++) {
					snprintf(dset_name, 1024, "/run%u/volumes", \
							(unsigned int) k + 1);
					vdset = openDataset(dset_name, dfile);
					if(H5Sselect_hyperslab(vdspace, H5S_SELECT_SET, voffset, \
							NULL, vcount, NULL) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					if(H5Dread(vdset, vdtype, vmspace, vdspace, H5P_DEFAULT, \
							vvalues[i][k]) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					closeDataset(vdset);
				}
			}
		}

		// calculate statistics
		if(sum_nothing) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_compartments; i++) {
						for(j = 0; j < num_output_species; j++) {
							mean = 0.0;
							for(k = 0; k < num_runs; k++)
								if(output_concentrations) {
									if(vvalues[i][k][ind] \
											!= 0.0)
										mean += (values[j][i][k][ind] / \
												vvalues[i][k][ind]);
								} else
									mean += values[j][i][k][ind];
							mean /= num_runs;
							sd = 0.0;
							for(k = 0; k < num_runs; k++)
								if(output_concentrations) {
									if(vvalues[i][k][ind] != 0.0)
										sd += (((values[j][i][k][ind] / \
												vvalues[i][k][ind]) - mean) * \
												((values[j][i][k][ind] /
												vvalues[i][k][ind]) - mean));
								} else
									sd += ((values[j][i][k][ind] - mean) * \
											(values[j][i][k][ind] - mean));
							sd = sqrt(sd / (num_runs - 1));
							if(output_sds) {
								if(upper_lower_bounds)
									#ifdef WIN32
									fprintf(ofp, "%c%f%c%f%c%f", separator, \
											mean, separator, mean - sd, \
											separator, mean + sd);
									#else
									fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
											mean, separator, mean - sd, \
											separator, mean + sd);
									#endif
								else
									#ifdef WIN32
									fprintf(ofp, "%c%f%c%f%c%f", separator, \
											(double) mean, separator, \
											(double) sd, separator, \
											(double) cifactor * sd);
									#else
									fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
											mean, separator, sd, separator, \
											cifactor * sd);
									#endif
							} else {
								#ifdef WIN32
								fprintf(ofp, "%c%f", separator, mean);
								#else
								fprintf(ofp, "%c%Lf", separator, mean);
								#endif
							}
						}
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}

		if(sum_species) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_compartments; i++) {
						mean = 0.0;
						for(k = 0; k < num_runs; k++)
							for(j = 0; j < num_output_species; j++)
								mean += values[j][i][k][ind];
						mean /= num_runs;
						sd = 0.0;
						for(k = 0; k < num_runs; k++) {
							total = 0.0;
							for(j = 0; j < num_output_species; j++)
								total += values[j][i][k][ind];
							sd += ((total - mean) * (total - mean));
						}
						sd = sqrt(sd / (num_runs - 1));
						if(output_sds) {
							if(upper_lower_bounds)
								#ifdef WIN32
								fprintf(ofp, "%c%f%c%f%c%f", separator, mean, \
										separator, mean - sd, separator, \
										mean + sd);
								#else
								fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
										mean, separator, mean - sd, \
										separator, mean + sd);
								#endif
							else
								#ifdef WIN32
								fprintf(ofp, "%c%f%c%f%c%f", separator, mean, \
										separator, sd, separator, \
										cifactor * sd);
								#else
								fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
										mean, separator, sd, separator, \
										cifactor * sd);
								#endif
						} else {
							#ifdef WIN32
							fprintf(ofp, "%c%f", separator, mean);
							#else
							fprintf(ofp, "%c%Lf", separator, mean);
							#endif
						}
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}

		if(sum_compartments) {
			ind = 0;
			while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
					#ifdef WIN32
					fprintf(ofp, "%f", (double) timepoint * log_interval);
					#else
					fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
					#endif
					for(i = 0; i < num_output_species; i++) {
						mean = 0.0;
						for(k = 0; k < num_runs; k++)
							for(j = 0; j < num_output_compartments; j++)
								mean += values[i][j][k][ind];
						mean /= num_runs;
						sd = 0.0;
						for(k = 0; k < num_runs; k++) {
							total = 0.0;
							for(j = 0; j < num_output_compartments; j++)
								total += values[i][j][k][ind];
							sd += ((total - mean) * (total - mean));
						}
						sd = sqrt(sd / (num_runs - 1));
						if(output_sds) {
							if(upper_lower_bounds)
								#ifdef WIN32
								fprintf(ofp, "%c%f%c%f%c%f", separator, mean, \
										separator, mean - sd, separator, \
										mean + sd);
								#else
								fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
										mean, separator, mean - sd, \
										separator, mean + sd);
								#endif
							else
								#ifdef WIN32
								fprintf(ofp, "%c%f%c%f%c%f", separator, mean, \
										separator, sd, separator, \
										cifactor * sd);
								#else
								fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
										mean, separator, sd, separator, \
										cifactor * sd);
								#endif
						} else {
							#ifdef WIN32
							fprintf(ofp, "%c%f", separator, mean);
							#else
							fprintf(ofp, "%c%Lf", separator, mean);
							#endif
						}
					}
					fprintf(ofp, "\n");
				}
				ind++;
				timepoint++;
			}
			continue;
		}
	}

	// free memory
	for(i = 0; i < num_output_species; i++) {
		for(j = 0; j < num_output_compartments; j++) {
			for(k = 0; k < num_runs; k++)
				free(values[i][j][k]);
			free(values[i][j]);
		}
		free(values[i]);
	}
	free(values);

	if(output_concentrations) {
		for(i = 0; i < num_output_compartments; i++) {
			for(j = 0; j < num_runs; j++)
				free(vvalues[i][j]);
			free(vvalues[i]);
		}
		free(vvalues);
	}
}

// output average rule frequencies over multiple runs
void output_average_rule_frequencies(void)
{
	register unsigned long i, j, k, l, entries, timepoint;
	#ifdef WIN32
	register double mean, sd, total, last_time;
	#else
	register long double mean, sd, total, last_time;
	#endif
	unsigned int **values;
	char dset_name[1024];

	// allocate memory for computing statistics
	if(!(values = (unsigned int **) calloc(num_rules, \
			sizeof(unsigned int *)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_rules; i++)
		if(!(values[i] = (unsigned int *) calloc(num_runs, \
				sizeof(unsigned int)))) {
			fprintf(stderr, "error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}

	// save data header
	if(output_header) {
		for(i = 0; i < num_rules; i++)
			if(i == 0)
				fprintf(ofp, "%s", rule_buffer[i].rstr);
			else
				fprintf(ofp, "%c%s", separator, rule_buffer[i].rstr);
		fprintf(ofp, "\n");
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dtype;  // hdf5 datatype
	hid_t dspace, mspace; // hdf5 dataspaces
	hsize_t dims[1], offset[1];

	// create datatype
	dtype = createReactionDatatype();

	// create memspace
	dims[0] = chunk_size;
	mspace = createDataspace(1, dims);

	dims[0] = chunk_size;
	ReactionEntry *reaction_buffer = (ReactionEntry *) \
			createDataBuffer(dims[0], dtype);

	hsize_t count[1];
	// calculate average frequencies
	for(i = 0; i < num_runs; i++) {
		// create dataspace
		snprintf(dset_name, 1024, "/run%u", (unsigned int) i + 1);
		readAttribute("number_of_timepoints", num_timepoints, dfile, dset_name);
		dims[0] = num_timepoints;
		dspace = createDataspace(1, dims);

		timepoint = 0;
		end_timepoint = num_timepoints;
		entries = chunk_size;
		count[0] = entries;
		last_time = 0.0;
		while(timepoint < end_timepoint) {
			// adjust memspace size if remaining timepoints are less than chunk size
			if((end_timepoint - timepoint) < chunk_size) {
				entries = end_timepoint - timepoint;
				count[0] = entries;
				dims[0] = entries;
				closeDataspace(mspace);
				mspace = createDataspace(1, dims);
			}

			// read in levels data
			offset[0] = timepoint;

			// read data
			snprintf(dset_name, 1024, "/run%u/reactions", (unsigned int) i + 1);
			dset = openDataset(dset_name, dfile);
			if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, \
					NULL, count, NULL) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
					&reaction_buffer[0]) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			closeDataset(dset);
			for(j = 0; j < entries; j++) {
				if(reaction_buffer[j].t == 0.0)
					continue;
				for(k = 0; k < num_output_compartments; k++)
					if(reaction_buffer[j].ci == output_compartments[k]) {
						for(l = 0; l < num_rules; l++)
							if(rule_buffer[l].rsi == reaction_buffer[j].rsi && \
										rule_buffer[l].ri == reaction_buffer[j].ri) {
								if(last_time != reaction_buffer[j].t) {
									values[l][i]++;
									last_time = reaction_buffer[j].t;
									break;
								}
							}
					}
			}
			timepoint += entries;
		}
		closeDataspace(dspace);
	}

	// calculate statistics
	for(i = 0; i < num_rules; i++) {
		mean = 0.0;
		for(j = 0; j < num_runs; j++)
			mean += values[i][j];
		mean /= num_runs;
		sd = 0.0;
		if(num_runs > 1) {
			for(j = 0; j < num_runs; j++)
				sd += ((values[i][j] - mean) * (values[i][j] - mean));
			sd = sqrt(sd / (num_runs - 1));
		}
		if(upper_lower_bounds) {
			#ifdef WIN32
			if(i == 0)
				fprintf(ofp, "%f%c%f%c%f", mean, separator, mean - sd, \
						separator, mean + sd);
			else
				fprintf(ofp, "%c%f%c%f%c%f", separator, mean, separator, \
						mean - sd, separator, mean + sd);
			#else
			if(i == 0)
				fprintf(ofp, "%Lf%c%Lf%c%Lf", mean, separator, mean - sd, \
						separator, mean + sd);
			else
				fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, mean, separator, \
						mean - sd, separator, mean + sd);
			#endif
		} else {
			#ifdef WIN32
			if(i == 0)
				fprintf(ofp, "%f%c%f%c%f", (double) mean, separator, \
					(double) sd, separator, (double) cifactor * sd);
			else
				fprintf(ofp, "%c%f%c%f%c%f", separator, (double) mean, \
						separator, (double) sd, separator, \
						(double) cifactor * sd);
			#else
			if(i == 0)
				fprintf(ofp, "%Lf", mean);
//				fprintf(ofp, "%Lf%c%Lf%c%Lf", mean, separator, sd, separator,
//					cifactor * sd);
			else
				fprintf(ofp, "%c%Lf", separator, mean);
//				fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, mean, separator,
//						sd, separator, cifactor * sd);
			#endif
		}
	}
	fprintf(ofp, "\n");

	// free memory
	for(i = 0; i < num_rules; i++)
		free(values[i]);
	free(values);
	freeDataBuffer(reaction_buffer);
}

// find time taken to reach level
void output_level_time(void)
{
	register unsigned long i, timepoint, entries;
	unsigned int *levels;
	bool fall = false;
	char dset_name[1024];

	// if target_level negative find time taken to fall (as opposed to rise)
	// to level
	if(target_level <= 0) {
		fall = true;
		target_level *= -1;
	}

	// prepare to read hdf5 dataset
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3];

	// open dataset
	snprintf(dset_name, 1024, "run%ld/%s", output_run, "amounts");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createAmountDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	// allocate memory for storing levels
	if(!(levels = (unsigned int *) calloc(chunk_size, \
			sizeof(unsigned int)))) {
		fprintf(stderr, "error: couldn't allocate memory for levels\n");
		exit(EXIT_FAILURE);
	}

	// create memspace
	dims[0] = chunk_size;
	mspace = createDataspace(1, dims);

	// create dataspace
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// find time taken to reach level
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {1, 1, chunk_size};
	hsize_t offset[3] = {output_species[0], output_compartments[0], \
			start_timepoint};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			dims[0] = entries;
			count[2] = entries;
			closeDataspace(mspace);
			mspace = createDataspace(1, dims);
		}

		// read in data
		offset[2] = timepoint;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, count, \
				NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, levels) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		// find time at which level exceeded
		for(i = 0; i < entries; i++)
			if(fall) {
				if(levels[i] <= (unsigned int) target_level)
					break;
			} else {
				if(levels[i] >= (unsigned int) target_level)
					break;
			}
		if(i != entries)
			break;  // found time

		timepoint += entries;
	}

	if(output_header)
		fprintf(ofp, "species%ccompartment%clevel%ctime\n", \
				separator, separator, separator);
	if(use_species_indices)
		fprintf(ofp, "%ld", output_species[0]);
	else
		fprintf(ofp, "%s", species_buffer[output_species[0]].sn);
	if(use_compartment_indices)
		fprintf(ofp, "%c%ld", separator, output_compartments[0]);
	else
		fprintf(ofp, "%c[%ld,%ld]", separator, \
				compartment_buffer[output_compartments[0]].cx,
				compartment_buffer[output_compartments[0]].cy);
	if(timepoint < end_timepoint)
		#ifdef WIN32
		fprintf(ofp, "%c%ld%c%0.8f\n", separator, target_level, separator, \
				(double) (timepoint + i) * log_interval);
		#else
		fprintf(ofp, "%c%ld%c%0.8Lf\n", separator, target_level, separator, \
				(long double) (timepoint + i) * log_interval);
		#endif
	else
		fprintf(ofp, "%c%ld%c-1\n", separator, target_level, separator);

	// tidy up
	closeDataset(dset);
	free(levels);
}

// find time taken to reach level
void output_average_level_time(void)
{
	register unsigned long i, timepoint, entries, total, run;
	unsigned int *levels;
	double *values;
	#ifdef WIN32
	double mean, sd;
	#else
	long double mean, sd;
	#endif
	bool fall = false;
	char dset_name[1024];

	// if target_level negative find time taken to fall (as opposed to rise)
	// to level
	if(target_level <= 0) {
		fall = true;
		target_level *= -1;
	}

	// prepare to read hdf5 dataset
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3];

	// create datatype
	dtype = createAmountDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	// allocate memory for storing levels
	if(!(levels = (unsigned int *) calloc(chunk_size, \
			sizeof(unsigned int)))) {
		fprintf(stderr, "error: couldn't allocate memory for levels\n");
		exit(EXIT_FAILURE);
	}
	if(!(values = (double *) calloc(num_runs, \
			sizeof(double)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}

	// create dataspace
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// read in level times
	total = 0;
	for(run = 0; run < num_runs; run++) {
		// open dataset
		snprintf(dset_name, 1024, "run%ld/%s", run + 1, "amounts");
		dset = openDataset(dset_name, dfile);

		// create memspace
		dims[0] = chunk_size;
		mspace = createDataspace(1, dims);

		// find time taken to reach level
		timepoint = start_timepoint;
		entries = chunk_size;
		hsize_t count[3] = {1, 1, chunk_size};
		hsize_t offset[3] = {output_species[0], output_compartments[0], \
				start_timepoint};
		while(timepoint < end_timepoint) {
			// adjust memspace size
			if((end_timepoint - timepoint) < chunk_size) {
				entries = end_timepoint - timepoint;
				dims[0] = entries;
				closeDataspace(mspace);
				mspace = createDataspace(1, dims);
				count[2] = entries;
			}

			// read in data
			offset[2] = timepoint;
			if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
					count, NULL) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}
			if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, levels) < 0) {
				std::cerr << "error: couldn't read dataset\n";
				exit(EXIT_FAILURE);
			}

			// find time at which level exceeded
			for(i = 0; i < entries; i++)
				if(fall) {
					if(levels[i] <= (unsigned int) target_level)
						break;
				} else {
					if(levels[i] >= (unsigned int) target_level)
						break;
				}
			if(i != entries)
				break;  // found time

			timepoint += entries;
		}

		if(timepoint < end_timepoint) {
			values[run] = (double) (timepoint + i) * log_interval;
			total++;
		} else
			values[run] = -1.0;

		// tidy up
		closeDataspace(mspace);
		closeDataset(dset);
	}

	// calculate statistics
	if(output_header)
		fprintf(ofp, "species%ccompartment%clevel%ctime_mean%ctime_sd%c" \
				"time_ci%cruns\n", separator, separator, separator, separator, \
				separator, separator);
	if(use_species_indices)
		fprintf(ofp, "%ld", output_species[0]);
	else
		fprintf(ofp, "%s", species_buffer[output_species[0]].sn);
	if(use_compartment_indices)
		fprintf(ofp, "%c%ld", separator, output_compartments[0]);
	else
		fprintf(ofp, "%c[%ld,%ld]", separator, \
				compartment_buffer[output_compartments[0]].cx,
				compartment_buffer[output_compartments[0]].cy);
	if(total) {
		mean = 0.0;
		for(i = 0; i < num_runs; i++)
			if(values[i] >= 0.0)
				mean += values[i];
		mean /= total;
		if(total == 1)
			sd = 0.0;
		else {
			sd = 0.0;
			for(i = 0; i < num_runs; i++)
				if(values[i] >= 0.0)
					sd += ((values[i] - mean) * (values[i] - mean));
			sd = sqrt(sd / (total - 1));
		}
		#ifdef WIN32
		fprintf(ofp, "%c%ld%c%f%c%f%c%f%c%ld\n", separator, target_level, \
				separator, mean, separator, sd, separator, \
				cifactor * sd, separator, total);
		#else
		fprintf(ofp, "%c%ld%c%Lf%c%Lf%c%Lf%c%ld\n", separator, target_level, \
				separator, mean, separator, sd, separator, \
				cifactor * sd, separator, total);
		#endif
	} else
		fprintf(ofp, "%c%ld%c-1%c0%c0%c0\n", separator, target_level, \
				separator, separator, separator, separator);

	// tidy up
	free(levels);
	free(values);
}

// find time taken to reach level from mean levels
void output_average_level_time_mean(void)
{
	register unsigned long i, j, k, ind, timepoint, entries, total, run;
	#ifdef WIN32
	double mean, sd;
	#else
	long double mean, sd;
	#endif
	bool fall = false;
	char dset_name[1024];
	unsigned int ****values;
	bool found;

	// if target_level negative find time taken to fall (as opposed to rise)
	// to level
	if(target_level <= 0) {
		fall = true;
		target_level *= -1;
	}

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// allocate memory for computing statistics
	if(!(values = (unsigned int ****) calloc(num_output_species, \
			sizeof(unsigned int ***)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_output_species; i++) {
		if(!(values[i] = (unsigned int ***) calloc(num_output_compartments, \
				sizeof(unsigned int **)))) {
			fprintf(stderr, "error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < num_output_compartments; j++) {
			if(!(values[i][j] = (unsigned int **) calloc(num_runs, \
					sizeof(unsigned int *)))) {
				fprintf(stderr, \
						"error: couldn't allocate memory for statistics\n");
				exit(EXIT_FAILURE);
			}
			for(k = 0; k < num_runs; k++)
				if(!(values[i][j][k] = (unsigned int *) \
						calloc(chunk_size, sizeof(unsigned int)))) {
					fprintf(stderr, \
							"error: couldn't allocate memory for statistics\n");
					exit(EXIT_FAILURE);
				}
		}
	}

	// prepare to read hdf5 dataset
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[3], offset[3];

	// create datatype
	dtype = createAmountDatatype();

	// create memspace
	dims[0] = chunk_size;
	mspace = createDataspace(1, dims);

	// create dataspace
	dims[0] = num_species;
	dims[1] = num_compartments;
	dims[2] = num_timepoints;
	dspace = createDataspace(3, dims);

	// calculate average levels
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[3] = {1, 1, entries};
	found = false;
	while(timepoint < end_timepoint && !found) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			count[2] = entries;
			dims[0] = entries;
			closeDataspace(mspace);
			mspace = createDataspace(1, dims);
		}

		// read in levels data
		offset[2] = timepoint;
		for(i = 0; i < num_output_compartments; i++) {
			offset[1] = output_compartments[i];
			for(j = 0; j < num_output_species; j++) {
				offset[0] = output_species[j];
				for(k = 0; k < num_runs; k++) {
					snprintf(dset_name, 1024, "/run%u/amounts", (unsigned int) k + 1);
					dset = openDataset(dset_name, dfile);
					if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, \
							NULL, count, NULL) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
							values[j][i][k]) < 0) {
						std::cerr << "error: couldn't read dataset\n";
						exit(EXIT_FAILURE);
					}
					closeDataset(dset);
				}
			}
		}

		// calculate statistics
		ind = 0;
		while(ind < entries) {
			mean = 0.0;
			for(i = 0; i < num_output_compartments; i++) {
				for(j = 0; j < num_output_species; j++) {
					for(k = 0; k < num_runs; k++)
						#ifdef WIN32
						mean += (double) values[j][i][k][ind];
						#else
						mean += (long double) values[j][i][k][ind];
						#endif
				}
			}
			mean /= num_runs;
			if(fall) {
				#ifdef WIN32
				if(mean <= (double) target_level) {
				#else
				if(mean <= (long double) target_level) {
				#endif
					found = true;
					break;
				}
			} else {
				#ifdef WIN32
				if(mean >= (double) target_level) {
				#else
				if(mean >= (long double) target_level) {
				#endif
					found = true;
					break;
				}
			}
			if(found)
				break;
			ind++;
			timepoint++;
		}
	}

	// calculate statistics
	if(output_header)
		fprintf(ofp, "species%ccompartment%clevel%ctime\n" \
				, separator, separator, separator);
	if(use_species_indices)
		fprintf(ofp, "%ld", output_species[0]);
	else
		fprintf(ofp, "%s", species_buffer[output_species[0]].sn);
	if(use_compartment_indices)
		fprintf(ofp, "%c%ld", separator, output_compartments[0]);
	else
		fprintf(ofp, "%c[%ld,%ld]", separator, \
				compartment_buffer[output_compartments[0]].cx,
				compartment_buffer[output_compartments[0]].cy);

	if(found) {
		#ifdef WIN32
		fprintf(ofp, "%c%ld%c%f\n", separator, target_level, separator, \
				(double) timepoint * log_interval);
		#else
		fprintf(ofp, "%c%ld%c%Lf\n", separator, target_level, separator, \
				(long double) timepoint * log_interval);
		#endif
	} else
		fprintf(ofp, "%c%ld%c-1\n", separator, target_level, separator);

	// free memory
	for(i = 0; i < num_output_species; i++) {
		for(j = 0; j < num_output_compartments; j++) {
			for(k = 0; k < num_runs; k++)
				free(values[i][j][k]);
			free(values[i][j]);
		}
		free(values[i]);
	}
	free(values);
}

// output information on volumes
void output_compartment_volumes(void)
{
	register unsigned long i, ind, timepoint, entries;
	char dset_name[1024];

	// save data header
	if(output_header) {
		fprintf(ofp, "time");

		for(i = 0; i < num_output_compartments; i++)
/*			if(use_compartment_indices)*/
				fprintf(ofp, "%c%ld", separator, output_compartments[i]);
/*			else
				fprintf(ofp, "%c[%ld,%ld]", separator, \
						compartment_buffer[output_compartments[i]].cx, \
						compartment_buffer[output_compartments[i]].cy);*/
		fprintf(ofp, "\n");
	}

	// prepare to read hdf5 datasets
	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dspace, mspace;  // hdf5 dataspaces
	hid_t dtype;  // hdf5 datatype
	hsize_t dims[2];

	// open dataset
	snprintf(dset_name, 1024, "/run%u/%s", (unsigned int) output_run, "volumes");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createVolumeDatatype();

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// create data buffer
	dims[0] = num_compartments;
	dims[1] = chunk_size;
	volume_buffer = create2DArrayDouble(dims);

	// create memspace
	mspace = createDataspace(2, dims);

	// create dataspace
	dims[1] = num_timepoints;
	dspace = createDataspace(2, dims);

	// output levels
	timepoint = start_timepoint;
	entries = chunk_size;
	hsize_t count[2] = {num_compartments, chunk_size};
	hsize_t offset[2] = {0, start_timepoint};
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			free2DArrayDouble(volume_buffer);
			dims[1] = entries;
			volume_buffer = create2DArrayDouble(dims);
			closeDataspace(mspace);
			mspace = createDataspace(2, dims);
			count[1] = entries;
		}

		// read in data
		offset[1] = timepoint;
		if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, \
				count, NULL) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, \
				&volume_buffer[0][0]) < 0) {
			std::cerr << "error: couldn't read dataset\n";
			exit(EXIT_FAILURE);
		}

		// output data
		ind = 0;
		while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
				#ifdef WIN32
				fprintf(ofp, "%f", (double) timepoint * log_interval);
				#else
				fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
				#endif
				for(i = 0; i < num_output_compartments; i++) {
					fprintf(ofp, "%c%f", separator, \
							volume_buffer[output_compartments[i]][ind]);
				}
				fprintf(ofp, "\n");
			}
			ind++;
			timepoint++;
		}
	}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	free2DArrayDouble(volume_buffer);
}

// output average volumes
void output_average_compartment_volumes(void)
{
	register unsigned long i, j, k, entries, timepoint, ind;
	#ifdef WIN32
	register double mean, sd, total;
	#else
	register long double mean, sd, total;
	#endif
	char dset_name[1024];
	double ***vvalues;
	hid_t vdset;  // hdf5 dataset for volumes
	hid_t vdspace, vmspace;  // hdf5 dataspaces for volumes
	hid_t vdtype;  // hdf5 datatype for volumes
	hsize_t vdims[2], vcount[2], voffset[2];

	// reduce chunk size if greater than total number of timepoints
	if(num_timepoints_selected < chunk_size)
		chunk_size = num_timepoints_selected;

	// allocate memory for computing statistics
	if(!(vvalues = (double ***) calloc(num_output_compartments, \
			sizeof(double **)))) {
		fprintf(stderr, "error: couldn't allocate memory for statistics\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < num_output_compartments; i++) {
		if(!(vvalues[i] = (double **) calloc(num_runs, \
				sizeof(double *)))) {
			fprintf(stderr, \
					"error: couldn't allocate memory for statistics\n");
			exit(EXIT_FAILURE);
		}
		for(j = 0; j < num_runs; j++) {
			if(!(vvalues[i][j] = (double *) calloc(chunk_size, \
					sizeof(double)))) {
				fprintf(stderr, \
						"error: couldn't allocate memory for statistics\n");
				exit(EXIT_FAILURE);
			}
		}
	}

	// save data header
	if(output_header) {
		fprintf(ofp, "time");

		for(i = 0; i < num_output_compartments; i++) {
			string header_name = "";
			char header_index[64];
			if(use_compartment_indices)
				snprintf(header_index, 64, "%ld", \
						output_compartments[i]);
			else
				snprintf(header_index, 64, "[%ld,%ld]", \
						compartment_buffer[output_compartments[i]].cx, \
						compartment_buffer[output_compartments[i]].cy);
			header_name.append(header_index);
			if(upper_lower_bounds)
				fprintf(ofp, "%c%s_mean%c%s_sdl%c%s_sdu", separator, \
						header_name.c_str(), separator, \
						header_name.c_str(), separator, \
						header_name.c_str());
			else
				fprintf(ofp, "%c%s_mean%c%s_sd%c%s_ci", separator, \
						header_name.c_str(), separator, \
						header_name.c_str(), separator, \
						header_name.c_str());
		}
		fprintf(ofp, "\n");
	}

	// create datatype
	vdtype = createVolumeDatatype();

	// create memspace
	vdims[0] = chunk_size;
	vmspace = createDataspace(1, vdims);

	// create dataspace
	vdims[0] = num_compartments;
	vdims[1] = num_timepoints;
	vdspace = createDataspace(2, vdims);

	// calculate average levels
	timepoint = start_timepoint;
	entries = chunk_size;
	vcount[0] = 1;
	vcount[1] = chunk_size;
	while(timepoint < end_timepoint) {
		// adjust memspace size if remaining timepoints are less than chunk size
		if((end_timepoint - timepoint) < chunk_size) {
			entries = end_timepoint - timepoint;
			vcount[1] = entries;
			vdims[0] = entries;
			closeDataspace(vmspace);
			vmspace = createDataspace(1, vdims);
		}

		// read in volumes data
		voffset[1] = timepoint;
		for(i = 0; i < num_output_compartments; i++) {
			voffset[0] = output_compartments[i];
			for(k = 0; k < num_runs; k++) {
				snprintf(dset_name, 1024, "/run%u/volumes", (unsigned int) k + 1);
				vdset = openDataset(dset_name, dfile);
				if(H5Sselect_hyperslab(vdspace, H5S_SELECT_SET, voffset, \
						NULL, vcount, NULL) < 0) {
					std::cerr << "error: couldn't read dataset\n";
					exit(EXIT_FAILURE);
				}
				if(H5Dread(vdset, vdtype, vmspace, vdspace, H5P_DEFAULT, \
						vvalues[i][k]) < 0) {
					std::cerr << "error: couldn't read dataset\n";
					exit(EXIT_FAILURE);
				}
				closeDataset(vdset);
			}
		}

		// calculate statistics
		ind = 0;
		while(ind < entries) {
				if(fmod(timepoint * log_interval, output_interval) == 0.0 || \
							log_interval == output_interval) {
				#ifdef WIN32
				fprintf(ofp, "%f", (double) timepoint * log_interval);
				#else
				fprintf(ofp, "%Lf", (long double) timepoint * log_interval);
				#endif
				for(i = 0; i < num_output_compartments; i++) {
					mean = 0.0;
					for(k = 0; k < num_runs; k++)
						mean += vvalues[i][k][ind];
					mean /= num_runs;
					sd = 0.0;
					for(k = 0; k < num_runs; k++)
						sd += ((vvalues[i][k][ind] - mean) * \
								(vvalues[i][k][ind] - mean));
					sd = sqrt(sd / (num_runs - 1));
					if(upper_lower_bounds)
						#ifdef WIN32
						fprintf(ofp, "%c%f%c%f%c%f", separator, \
								mean, separator, mean - sd, separator, \
								mean + sd);
						#else
						fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
								mean, separator, mean - sd, separator, \
								mean + sd);
						#endif
					else
						#ifdef WIN32
						fprintf(ofp, "%c%f%c%f%c%f", separator, \
								mean, separator, sd, separator, \
								cifactor * sd);
						#else
						fprintf(ofp, "%c%Lf%c%Lf%c%Lf", separator, \
								mean, separator, sd, separator, \
								cifactor * sd);
						#endif
				}
				fprintf(ofp, "\n");
			}
			ind++;
			timepoint++;
		}
	}

	// free memory
	for(i = 0; i < num_output_compartments; i++) {
		for(j = 0; j < num_runs; j++)
			free(vvalues[i][j]);
		free(vvalues[i]);
	}
	free(vvalues);
}

// output information on dataset
void output_info(void)
{
	char *time, *date, *seconds, *str, *str1, *str2;
	unsigned long ulong, run;
	double dble, stime, etime;
	char run_group[64], timestamp[256];
	bool b;

 	readAttribute("mcss_version", str, dfile);
 	printf("%s: %s\n", "mcss version", str);
 	free(str);
 	readAttribute("model_input_file", str, dfile);
 	printf("%s: %s\n", "model input file", str);
 	free(str);
 	readAttribute("model_format", str, dfile);
 	printf("%s: %s\n", "model format", str);
 	free(str);
 	readAttribute("data_file", str, dfile);
 	printf("%s: %s\n", "data file", str);
 	free(str);
 	readAttribute("number_of_runs", ulong, dfile);
 	printf("%s: %lu\n", "number of runs", ulong);
 	readAttribute("simulation_algorithm", str, dfile);
 	printf("%s: %s\n", "simulation algorithm", str);
 	free(str);
 	readAttribute("simulation_algorithm_name", str, dfile);
 	printf("%s: %s\n", "simulation algorithm name", str);
 	free(str);
 	readAttribute("log_type", str, dfile);
 	printf("%s: %s\n", "log type", str);
 	free(str);
 	readAttribute("log_interval", dble, dfile);
 	printf("%s: %f\n", "log interval", dble);
 	readAttribute("max_time", dble, dfile);
 	printf("%s: %f\n", "max time", dble);
 	readAttribute("seed", ulong, dfile);
 	printf("%s: %lu\n", "seed", ulong);
 	readAttribute("duplicate_initial_amounts", b, dfile);
 	printf("%s: %d\n", "duplicate initial amounts", b);
 	readAttribute("log_degraded", b, dfile);
 	printf("%s: %d\n", "log degraded", b);
 	readAttribute("log_volumes", b, dfile);
 	printf("%s: %d\n", "log volumes", b);
 	readAttribute("log_propensities", b, dfile);
 	printf("%s: %d\n", "log propensities", b);
 	readAttribute("periodic_x", b, dfile);
 	printf("%s: %d\n", "periodic x", b);
 	readAttribute("periodic_y", b, dfile);
 	printf("%s: %d\n", "periodic y", b);
 	readAttribute("periodic_z", b, dfile);
 	printf("%s: %d\n", "periodic z", b);
 	readAttribute("number_of_species", ulong, dfile);
 	printf("%s: %lu\n", "number of species", ulong);
 	readAttribute("number_of_rule_templates", ulong, dfile);
 	printf("%s: %lu\n", "number of rule templates", ulong);
 	readAttribute("number_of_rules_in_templates", ulong, dfile);
 	printf("%s: %lu\n", "number of rules in templates", ulong);
 	readAttribute("total_number_of_rules", ulong, dfile);
 	printf("%s: %lu\n", "total number of rules", ulong);
 	readAttribute("lattice_x_dimension", ulong, dfile);
 	printf("%s: %lu\n", "lattice x dimension", ulong);
 	readAttribute("lattice_y_dimension", ulong, dfile);
 	printf("%s: %lu\n", "lattice y dimension", ulong);
 	readAttribute("lattice_z_dimension", ulong, dfile);
 	printf("%s: %lu\n", "lattice z dimension", ulong);
 	readAttribute("growth_type", str1, dfile);
 	printf("%s: %s\n", "growth type", str1);
 	readAttribute("keep_divisions", b, dfile);
 	printf("%s: %d\n", "keep divisions", b);
 	readAttribute("division_direction", str1, dfile);
 	printf("%s: %s\n", "division direction", str1);
 	readAttribute("simulation_start_time", str1, dfile);
 	printf("%s: %s %s\n", "simulation start time", \
 			timestamp_time(str1, timestamp), timestamp_date(str1, timestamp));
 	readAttribute("simulation_end_time", str2, dfile);
 	printf("%s: %s %s\n", "simulation end time", \
 			timestamp_time(str2, timestamp), timestamp_date(str2, timestamp));
 	printf("total simulation time: %s\n", \
 			convert_time(timestamp_diff(str1, str2), timestamp));
 	free(str1);
 	free(str2);
 	if(output_run) {
		snprintf(run_group, 64, "/run%u", (unsigned int) output_run);
	 	readAttribute("number_of_compartments", ulong, dfile, run_group);
 		printf("%s: %lu\n", "number of compartments", ulong);
	 	readAttribute("simulated_time", dble, dfile, run_group);
 		printf("%s: %s (%lu seconds)\n", "simulated time", \
 				convert_time(dble, timestamp), (unsigned long) dble);
	 	readAttribute("run_start_time", str1, dfile, run_group);
 		printf("%s: %s %s\n", "run start time", \
 				timestamp_time(str1, timestamp), timestamp_date(str1, \
				timestamp));
 		readAttribute("run_end_time", str2, dfile, run_group);
 		printf("%s: %s %s\n", "run end time", \
 				timestamp_time(str2, timestamp), timestamp_date(str2, \
				timestamp));
 		printf("total run time: %s\n", \
 				convert_time(timestamp_diff(str1, str2), timestamp));
	 	free(str1);
 		free(str2);
	 	readAttribute("preprocess_start_time", str1, dfile, run_group);
	 	readAttribute("preprocess_end_time", str2, dfile, run_group);
	 	printf("total preprocessing time: %s\n", \
 				convert_time(timestamp_diff(str1, str2), timestamp));
 		free(str1);
 		free(str2);
 		readAttribute("main_loop_start_time", str1, dfile, run_group);
 		readAttribute("main_loop_end_time", str2, dfile, run_group);
 		printf("main loop total time: %s\n", \
 				convert_time(timestamp_diff(str1, str2), timestamp));
 		readAttribute("total_reactions_simulated", ulong, dfile, run_group);
 		printf("%s: %lu\n", "total reactions simulated", ulong);
		if(ulong == 0)
			printf("%s: %d\n", "reactions per second", 0);
		else
			printf("%s: %lu\n", "reactions per second", \
					(unsigned long) ((double) ulong / \
					timestamp_diff(str1, str2), timestamp));
 		free(str1);
 		free(str2);
 	} else {
		#ifdef WIN32
 		double av_num_compartments = 0.0;
 		double av_simulated_time = 0.0;
 		double av_run_time = 0.0;
 		double av_total_reactions = 0.0;
 		double av_preprocess_time = 0.0;
 		double av_main_loop_time = 0.0;
 		double av_rps = 0.0;
		#else
 		long double av_num_compartments = 0.0;
 		long double av_simulated_time = 0.0;
 		long double av_run_time = 0.0;
 		long double av_total_reactions = 0.0;
 		long double av_preprocess_time = 0.0;
 		long double av_main_loop_time = 0.0;
 		long double av_rps = 0.0;
		#endif
 		for(run = 1; run <= num_runs; run++) {
			snprintf(run_group, 64, "/run%u", (unsigned int) run);
		 	readAttribute("number_of_compartments", ulong, dfile, run_group);
			#ifdef WIN32
		 	av_num_compartments += (double) ulong;
			#else
		 	av_num_compartments += (long double) ulong;
			#endif
		 	readAttribute("simulated_time", dble, dfile, run_group);
			#ifdef WIN32
		 	av_simulated_time += (double) dble;
			#else
		 	av_simulated_time += (long double) dble;
			#endif
		 	readAttribute("run_start_time", str1, dfile, run_group);
	 		readAttribute("run_end_time", str2, dfile, run_group);
			#ifdef WIN32
			av_run_time += (double) timestamp_diff(str1, str2);
			#else
			av_run_time += (long double) timestamp_diff(str1, str2);
			#endif
	 		readAttribute("total_reactions_simulated", ulong, dfile, run_group);
			#ifdef WIN32
	 		av_total_reactions += (double) ulong;
			#else
	 		av_total_reactions += (long double) ulong;
			#endif
		 	free(str1);
 			free(str2);
		 	readAttribute("preprocess_start_time", str1, dfile, run_group);
		 	readAttribute("preprocess_end_time", str2, dfile, run_group);
			av_preprocess_time += timestamp_diff(str1, str2);
	 		free(str1);
 			free(str2);
 			readAttribute("main_loop_start_time", str1, dfile, run_group);
 			readAttribute("main_loop_end_time", str2, dfile, run_group);
			av_main_loop_time += timestamp_diff(str1, str2);
			#ifdef WIN32
	 		av_rps += ((double) ulong / timestamp_diff(str1, str2));
			#else
	 		av_rps += ((long double) ulong / timestamp_diff(str1, str2));
			#endif
		 	free(str1);
 			free(str2);
 		}
		#ifdef WIN32
 		printf("%s: %ld\n", "average number of compartments", \
 				(unsigned long) ((double) av_num_compartments / num_runs));
 		printf("%s: %s (%lu seconds)\n", "average simulated time", \
 				convert_time((double) ((double) av_simulated_time / \
 				num_runs), timestamp), (unsigned long) ((double) av_simulated_time / \
 				num_runs));
 		printf("%s: %s\n", "average run time", \
 				convert_time((double) ((double) av_run_time / num_runs), timestamp));
 		printf("%s: %s\n", "average preprocessing time", \
 				convert_time((double) ((double) av_preprocess_time / \
 				num_runs), timestamp));
 		printf("%s: %s\n", "average main loop time", \
 				convert_time((double) ((double) av_main_loop_time / \
 				num_runs), timestamp));
 		printf("%s: %ld\n", "average total reactions simulated", \
 				(unsigned long) ((double) av_total_reactions / num_runs));
 		printf("%s: %ld\n", "average reactions per second", \
 				(unsigned long) ((double) av_rps / num_runs));
		#else
 		printf("%s: %ld\n", "average number of compartments", \
 				(unsigned long) ((long double) av_num_compartments / num_runs));
 		printf("%s: %s (%lu seconds)\n", "average simulated time", \
 				convert_time((double) ((long double) av_simulated_time / \
 				num_runs), timestamp), (unsigned long) ((long double) \
 				av_simulated_time / num_runs), timestamp);
 		printf("%s: %s\n", "average run time", \
 				convert_time((double) ((long double) av_run_time / num_runs), \
				timestamp));
 		printf("%s: %s\n", "average preprocessing time", \
 				convert_time((double) ((long double) av_preprocess_time / \
 				num_runs), timestamp));
 		printf("%s: %s\n", "average main loop time", \
 				convert_time((double) ((long double) av_main_loop_time / \
 				num_runs), timestamp));
 		printf("%s: %ld\n", "average total reactions simulated", \
 				(unsigned long) ((long double) av_total_reactions / num_runs));
 		printf("%s: %ld\n", "average reactions per second", \
 				(unsigned long) ((long double) av_rps / num_runs));
		#endif
 	}
}

// output information on compartments
void output_compartment_info(void)
{
	unsigned int i;

	for(i = 0; i < num_compartments; i++) {
		fprintf(ofp, "compartment index %ld:", compartment_buffer[i].ci);
		fprintf(ofp, " id=%s", compartment_buffer[i].cid);
		fprintf(ofp, " name=%s", compartment_buffer[i].cn);
		fprintf(ofp, " position=(%ld,%ld)", compartment_buffer[i].cx, \
				compartment_buffer[i].cy);
		fprintf(ofp, " template=%ld", compartment_buffer[i].ct);
		fprintf(ofp, " creation_time=%f", compartment_buffer[i].cst);
		fprintf(ofp, " destruction_time=%f", compartment_buffer[i].cet);
		fprintf(ofp, "\n");
	}
}

// output information on species
void output_species_info(void)
{
	unsigned int i;

	for(i = 0; i < num_species; i++) {
		fprintf(ofp, "species index %ld:", species_buffer[i].si);
		fprintf(ofp, " name=%s", species_buffer[i].sn);
		fprintf(ofp, "\n");
	}
}

// output propensities data
void output_propensities_info(void)
{
	register unsigned long i, j, buffer_size, buffer_index, entries;
	char run_group[64], dset_name[1024];

	// declare variables
	hid_t dset;  // hdf5 dataset
	hid_t dtype;  // hdf5 datatype

	// open dataset
	snprintf(run_group, 64, "/run%u", (unsigned int) (output_run ? output_run : 1));
	snprintf(dset_name, 1024, "%s/%s", run_group, "propensities");
	dset = openDataset(dset_name, dfile);

	// create datatype
	dtype = createPropensityDatatype(total_rules);

	// create data buffer
	buffer_size = chunk_size;
	propensity_buffer = createDataBuffer(buffer_size, dtype);

	// read and output data
	PropensityEntry propensity_entry;
	buffer_index = 0;
	while((entries = readDataset(propensity_buffer, dset, dtype, \
			buffer_index, buffer_size)))
		// access data in buffer
		for(i = 0; i < buffer_size; i++) {
			convertPropensityEntry(propensity_buffer, dtype, \
					propensity_entry, i);
			fprintf(ofp, "%f", *propensity_entry.t);
			for(j = 0; j < total_rules; j++)
				fprintf(ofp, " [%ld,%ld,%ld,%f]",propensity_entry.ri[j], \
						propensity_entry.rsi[j], propensity_entry.ci[j], \
						propensity_entry.rp[j]);
			fprintf(ofp, "\n");
		}

	// close dataset
	closeDataset(dset);

	// free data buffer memory
	freeDataBuffer(propensity_buffer);
}

// output information on rules
void output_rule_info(void)
{
	unsigned int i;

	for(i = 0; i < num_rules; i++)
		fprintf(ofp, "rule_index=%ld template_index=%ld rule_id=\"%s\" " \
				"rule_name=\"%s\" rule_vector=(%ld,%ld) rule_string=\"%s\"\n", \
				rule_buffer[i].ri, \
				rule_buffer[i].rsi, \
				rule_buffer[i].rid, \
				rule_buffer[i].rn, \
				rule_buffer[i].rx, \
				rule_buffer[i].ry, \
				rule_buffer[i].rstr \
				);
}

// output attribute value
void output_attribute_value(void)
{
	char *str, run_group[64];
	H5E_auto2_t old_func;
	void *old_client_data;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	if(attributeAsString(attribute, str, dfile)) {
		printf("%s\n", str);
		free(str);
	} else {
		snprintf(run_group, 64, "/run%u", (unsigned int) (output_run ? output_run : 1));
		if(attributeAsString(attribute, str, dfile, run_group)) {
			printf("%s\n", str);
			free(str);
		} else {
			cerr << "error: unable to find attribute \"" << attribute << "\"\n";
			exit(EXIT_FAILURE);
		}
	}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);
}

int main(int argc, char **argv)
{
	// parse command line
	parse_command_line(argc, argv);

	// read initial data
	initialise_input();

	// perform specified operation on dataset
	switch(mode) {
		case MODE_INFO:
			// output information on dataset
			output_info();
			break;

		case MODE_COMPARTMENT_INFO:
			// output information on compartments
			output_compartment_info();
			break;

		case MODE_SPECIES_INFO:
			// output information on species
			output_species_info();
			break;

		case MODE_RULES_INFO:
			// output information on rules
			output_rule_info();
			break;

		case MODE_VOLUMES:
			// output compartment volumes data
			if(average_runs)
				output_average_compartment_volumes();
			else
				output_compartment_volumes();
			break;

		case MODE_PROPENSITIES:
			// output propensities data
			output_propensities_info();
			break;

		case MODE_LEVELS:
			// output species levels
			if(average_runs)
				output_average_levels();
			else
				output_levels();
			break;

		case MODE_LEVEL_TIME:
			// find time taken to reach level
			if(average_runs)
				output_average_level_time();
			else
				output_level_time();
			break;

		case MODE_LEVEL_TIME_MEAN:
			// find time taken to reach level
			if(average_runs)
				output_average_level_time_mean();
			else
				output_level_time();
			break;

		case MODE_LEVELS_RATIO:
			// calculate ratio of species levels
			if(average_runs)
				output_average_levels_ratio();
			else
				output_levels_ratio();
			break;

		case MODE_ATTRIBUTE_INFO:
			// output attribute value
			output_attribute_value();
			break;

		case MODE_REACTION_FREQUENCY:
			// output reaction frequencies
			output_average_rule_frequencies();
			break;

		default:
			break;
	}

	fclose(ofp);

	return EXIT_SUCCESS;
}
