// tissue p-system simulator
// copyright 2008, 2009, 2010 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <mcss/mcss.h>

using namespace std;
using namespace Simulation;

#define SIMULATION_FINISHED	0
#define SIMULATION_CANCELLED	1

#ifndef WIN32
static pthread_mutex_t mcss_status_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t mcss_status = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t mcss_execute_running_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t timer_running_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t signal_handler_running_mutex = PTHREAD_MUTEX_INITIALIZER;
static bool mcss_execute_running = false;
static bool timer_running = false;
static bool signal_handler_running = false;

// timer thread data
struct TimerData {
	unsigned long max_time;
	pthread_t tid;
};

// mcss thread data
struct McssData {
	Psystem *psystem;
	double max_time;
	int simulation_status;
};

// signal handler thread
void *signal_handler(void *args)
{
	int sig;

	// set thread as running
	pthread_mutex_lock(&signal_handler_running_mutex);
	signal_handler_running = true;
	pthread_mutex_unlock(&signal_handler_running_mutex);

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// wait for signal
	sigset_t sig_mask;
	sigemptyset(&sig_mask);
	sigaddset(&sig_mask, SIGINT);
	sigwait(&sig_mask, &sig);

	pthread_mutex_lock(&mcss_status_mutex);

	// set thread as finished
	pthread_mutex_lock(&signal_handler_running_mutex);
	signal_handler_running = false;
	pthread_mutex_unlock(&signal_handler_running_mutex);

	pthread_cond_signal(&mcss_status);
	pthread_mutex_unlock(&mcss_status_mutex);

	pthread_exit(NULL);
}

// timer thread handler
void *timer(void *args)
{
	// set thread as running
	pthread_mutex_lock(&timer_running_mutex);
	timer_running = true;
	pthread_mutex_unlock(&timer_running_mutex);

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// get arguments
	struct TimerData *timer_data = (struct TimerData *) args;

	// sleep for specified runtime
	sleep_ptime(timer_data->max_time, timer_data->tid);

	pthread_mutex_lock(&mcss_status_mutex);

	// set thread as finished
	pthread_mutex_lock(&timer_running_mutex);
	timer_running = false;
	pthread_mutex_unlock(&timer_running_mutex);

	// woken up
	pthread_cond_signal(&mcss_status);
	pthread_mutex_unlock(&mcss_status_mutex);

	pthread_exit(NULL);
}

// mcss execute thread handler
void *mcss_execute(void *args)
{
	int ret;

	// set thread as running
	pthread_mutex_lock(&mcss_execute_running_mutex);
	mcss_execute_running = true;
	pthread_mutex_unlock(&mcss_execute_running_mutex);

	// set thread as cancelable
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	// set thread to cancel immediately
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	// get arguments
	struct McssData *mcss_data = (struct McssData *) args;

	// execute simulation
	mcss_data->psystem->execute(mcss_data->max_time);

	pthread_mutex_lock(&mcss_status_mutex);

	// finished ok
	mcss_data->simulation_status = SIMULATION_FINISHED;

	// set thread as finished
	pthread_mutex_lock(&mcss_execute_running_mutex);
	mcss_execute_running = false;
	pthread_mutex_unlock(&mcss_execute_running_mutex);

	pthread_cond_signal(&mcss_status);
	pthread_mutex_unlock(&mcss_status_mutex);

	pthread_exit(NULL);
}
#endif

int main(int argc, char **argv)
{
	#ifndef WIN32
	pthread_t tid_mcss, tid_timer, tid_signal;
	pthread_attr_t tattr;
	sigset_t sig_mask;
	struct TimerData timer_data;
	struct McssData mcss_data;
	#endif

	#ifdef HAVE_MPI
	// initialise mpi
	MPI_Init(&argc, &argv);
	atexit((void (*)()) MPI_Finalize);
	#endif

	// parse command line and read in parameters
	if(argc == 1)
		#ifdef HAVE_MPI
		fprintf(stderr, "mcss version %s (MPI)\n", VERSION);
		#else
		fprintf(stderr, "mcss version %s\n", VERSION);
		#endif
	SimulationParameters parameters;
	parameters.read(argc, argv);

	#ifdef HAVE_MPI
	// finish if more threads than runs and thread unneeded
	int tid;
	MPI_Comm_rank(MPI_COMM_WORLD, &tid);
	if(tid > parameters.runs)
		exit(EXIT_SUCCESS);
	#endif

	// create p-system from parameters specified in parameter file
	Psystem psystem(parameters);

	// only create p system - don't simulate
	if(parameters.just_psystem) {
		exit(EXIT_SUCCESS);
	}

	// dump model if not simulating
	if(parameters.dump) {
		char dump_file[PARAMETER_STR_MAX];
		char *ptr = strrchr(parameters.data_file, '.');
		*ptr = '\0';
		snprintf(dump_file, PARAMETER_STR_MAX, "%s.dump", \
			parameters.data_file);
		psystem.dump(dump_file);
		exit(EXIT_SUCCESS);
	}

	#ifndef WIN32
	// block all signals
	sigfillset(&sig_mask);
	pthread_sigmask(SIG_SETMASK, &sig_mask, NULL);

	// run simulation in thread
	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_JOINABLE);
	mcss_data.psystem = &psystem;
	mcss_data.max_time = parameters.max_time;
	mcss_data.simulation_status = SIMULATION_CANCELLED;
	if(pthread_create(&tid_mcss, &tattr, mcss_execute, (void *) \
				&mcss_data)) {
		fprintf(stderr, "error: unable to create mcss_execute thread\n");
		exit(EXIT_FAILURE);
	}

	// start timer thread if max_runtime specified
	if(parameters.max_runtime != 0.0) {
		timer_data.max_time = (unsigned int) parameters.max_runtime;
		timer_data.tid = tid_mcss;
		if(pthread_create(&tid_timer, &tattr, timer, (void *) &timer_data)) {
			fprintf(stderr, "error: unable to create timer thread\n");
			exit(EXIT_FAILURE);
		}
	}

	// start signal handler thread
	if(pthread_create(&tid_signal, &tattr, signal_handler, (void *) NULL)) {
		fprintf(stderr, "error: unable to create signal handler thread\n");
		exit(EXIT_FAILURE);
	}

	// tidy up
	pthread_attr_destroy(&tattr);

	// wait for either mcss or timer to finish
	pthread_mutex_lock(&mcss_status_mutex);
	pthread_cond_wait(&mcss_status, &mcss_status_mutex);

	switch(mcss_data.simulation_status) {
		case SIMULATION_FINISHED:
			// mcss execute thread exited first
			// cancel timer thread
			if(parameters.max_runtime != 0.0) {
				pthread_mutex_lock(&timer_running_mutex);
				if(timer_running)
					pthread_cancel(tid_timer);  // cancel thread
				pthread_mutex_unlock(&timer_running_mutex);
			}

			// cancel signal handler thread
			pthread_mutex_lock(&signal_handler_running_mutex);
			if(signal_handler_running)
				pthread_cancel(tid_signal);  // cancel thread
			pthread_mutex_unlock(&signal_handler_running_mutex);
			break;

		case SIMULATION_CANCELLED:
			// cancel timer thread
			if(parameters.max_runtime != 0.0) {
				pthread_mutex_lock(&timer_running_mutex);
				if(timer_running)
					pthread_cancel(tid_timer);  // cancel thread
				pthread_mutex_unlock(&timer_running_mutex);
			}

			// cancel signal handler thread
			pthread_mutex_lock(&signal_handler_running_mutex);
			if(signal_handler_running)
				pthread_cancel(tid_signal);  // cancel thread
			pthread_mutex_unlock(&signal_handler_running_mutex);

			// cancel mcss execute thread
			pthread_mutex_lock(&mcss_execute_running_mutex);
			if(mcss_execute_running)
				pthread_kill(tid_mcss, SIGUSR2);  // terminate simuation
			pthread_mutex_unlock(&mcss_execute_running_mutex);
			break;

		default:
			cerr << "error: unknown thread state\n";
			exit(EXIT_FAILURE);
			break;
	}
	pthread_mutex_unlock(&mcss_status_mutex);

	// wait for all threads to finish
	pthread_join(tid_mcss, NULL);
	pthread_join(tid_signal, NULL);
	if(parameters.max_runtime != 0.0)
		pthread_join(tid_timer, NULL);

	// tidy up
	pthread_mutex_destroy(&mcss_status_mutex);
	pthread_cond_destroy(&mcss_status);
	pthread_mutex_destroy(&mcss_execute_running_mutex);
	pthread_mutex_destroy(&timer_running_mutex);
	pthread_mutex_destroy(&signal_handler_running_mutex);
	#endif

	#ifdef WIN32
	// run simulation
	if(parameters.max_runtime != 0.0)
		fprintf(stderr, "warning: max_runtime parameter not supported " \
				"under windows\n");
	psystem.execute(parameters.max_time);
	#endif

	exit(EXIT_SUCCESS);
}
